/*
 *  Copyright (c) Texas Instruments Incorporated 2018-21
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_app.c
 *
 * \brief    Common CPSW utility.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <ti/csl/soc.h>

#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/include/cpsw_cfg.h>

#include <ti/drv/udma/udma.h>

#include "j7_cpsw.h"
#include "cpsw_app.h"
#include "cpsw_appmem.h"
#include "cpsw_lld_if.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */
static void CpswApp_printStatsNonZero(const char *pcString,
                                           uint64_t statVal);
static void CpswApp_printStatsWithIdxNonZero(const char *pcString,
                                                  uint32_t idx,
                                                  uint64_t statVal);
static void CpswApp_unlockMmr(volatile uint32_t *kick0,
                              volatile uint32_t *kick1);
static void CpswApp_lockMmr(volatile uint32_t *kick0,
                            volatile uint32_t *kick1);
static int32_t CpswApp_allocTxCh(Cpsw_Handle hCpsw,
                                 uint32_t coreKey,
                                 uint32_t coreId,
                                 uint32_t *txPSILThreadId);
static void CpswApp_reduceCoreMacAllocation(CpswRm_ResourcePrms *rscPrms,
                                            uint32_t *pReduceCount,
                                            uint32_t coreMinCount,
                                            bool skipCore,
                                            uint32_t skipCoreId);
static void CpswApp_updateMacResourcePartition(CpswRm_ResourcePrms *rscPrms,
                                               uint32_t availMacCount,
                                               uint32_t selfCoreId);
static void CpswApp_delAddrEntry(Cpsw_Handle hCpsw,
                                 uint32_t coreId,
                                 uint8_t *macAddr);
static void CpswApp_printDmaDescStats(CpswDma_DmaDescStats *descstats);
static void CpswApp_printCbStats(CpswDma_CbFxnStats *cbStats);
static void CpswApp_printRxFlowStats(CpswDma_RxFlowStats *stats);
static void CpswApp_printTxChStats(CpswDma_TxChStats *stats);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* UDMA driver objects */
static struct Udma_DrvObj udmaDrvObj[CPSW_COUNT];

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

uint32_t CpswApp_getNavSSInstanceId(Cpsw_Type cpswType)
{
    uint32_t instId;

#if defined(BUILD_MCU1_0) || defined(BUILD_MCU1_1)
    instId = UDMA_INST_ID_MCU_0;
#else
    instId = UDMA_INST_ID_MAIN_0;
#endif
    return instId;
}

Udma_DrvHandle CpswApp_udmaOpen(Cpsw_Type cpswType,
                                Udma_InitPrms *pInitPrms)
{
    Udma_InitPrms initPrms;
    Udma_DrvHandle hUdmaDrv;
    int32_t retVal;
    uint32_t instId;

    hUdmaDrv = &udmaDrvObj[cpswType];
    memset(hUdmaDrv, 0U, sizeof(*hUdmaDrv));

    /* Use default params if not passed by caller */
    if (NULL == pInitPrms)
    {
        instId = CpswApp_getNavSSInstanceId(cpswType);
        /* Initialize the UDMA driver based on NAVSS instance */
        UdmaInitPrms_init(instId, &initPrms);
        initPrms.printFxn = (Udma_PrintFxn) & CpswApp_print;
        initPrms.virtToPhyFxn = &CpswApp_udmaVirtToPhyFxn;
        initPrms.phyToVirtFxn = &CpswApp_udmaPhyToVirtFxn;

        //Note: Over-ride the osal api to disableAllIntr/restoreAllIntr
        initPrms.osalPrms.disableAllIntr = &CpswApp_disableAllIntr;
        initPrms.osalPrms.restoreAllIntr = &CpswApp_restoreAllIntr;
        //Note: Over-ride the osal api to registerIntr/restoreAllIntr
        initPrms.osalPrms.registerIntr = &CpswApp_registerIntr;
        initPrms.osalPrms.unRegisterIntr = &CpswApp_unRegisterIntr;

        retVal            = Udma_init(hUdmaDrv, &initPrms);
    }
    else
    {
        retVal = Udma_init(hUdmaDrv, pInitPrms);
    }

    /* assert if UDMA failed to open */
    CpswApp_assert(UDMA_SOK == retVal);

    return hUdmaDrv;
}

void CpswApp_udmaclose(Udma_DrvHandle hUdmaDrv)
{
    CpswApp_assert(NULL != hUdmaDrv);
    /* App should make sure that all Rx & Tx channels & are closed */
    /* Deinitialize the UDMA driver */
    Udma_deinit(hUdmaDrv);
}

void CpswApp_print(const char *pcString, ...)
{
    char printBuffer[CPSW_CFG_PRINT_BUF_LEN];
    va_list arguments;

    if (CPSW_CFG_PRINT_BUF_LEN < strlen(pcString))
    {
        CpswApp_assert(false);
    }

    /* Start the varargs processing */
    va_start(arguments, pcString);
    vsnprintf(printBuffer, sizeof(printBuffer), pcString, arguments);

    slogf(_SLOGC_NETWORK, _SLOG_INFO, printBuffer);

    /* End the varargs processing */
    va_end(arguments);
}

static void CpswApp_printStatsNonZero(const char *pcString,
                                      uint64_t statVal)
{
    if (0U != statVal)
    {
        CpswApp_print(pcString, statVal);
    }
}

static void CpswApp_printStatsWithIdxNonZero(const char *pcString,
                                             uint32_t idx,
                                             uint64_t statVal)
{
    if (0U != statVal)
    {
        CpswApp_print(pcString, idx, statVal);
    }
}

void CpswApp_printMacAddr(uint8_t macAddr[])
{
    CpswApp_print("%02x:%02x:%02x:%02x:%02x:%02x",
                       macAddr[0] & 0xFF,
                       macAddr[1] & 0xFF,
                       macAddr[2] & 0xFF,
                       macAddr[3] & 0xFF,
                       macAddr[4] & 0xFF,
                       macAddr[5] & 0xFF);
}

void CpswApp_printFrame(EthFrame *frame,
                        uint32_t len)
{
    uint8_t *payload;
    uint32_t i;

    CpswApp_print("Dst addr : ");
    CpswApp_printMacAddr(&frame->hdr.dstMac[0]);

    CpswApp_print("Src addr : ");
    CpswApp_printMacAddr(&frame->hdr.srcMac[0]);

    if (frame->hdr.etherType == htons(ETHERTYPE_VLAN_TAG))
    {
        EthVlanFrame *vlanFrame = (EthVlanFrame *)frame;

        CpswApp_print("TPID     : 0x%04x", ntohs(vlanFrame->hdr.tpid) & 0xFFFFU);
        CpswApp_print("Priority : %d", (ntohs(vlanFrame->hdr.tci) & 0xFFFFU) >> 13);
        CpswApp_print("VLAN Id  : %d", ntohs(vlanFrame->hdr.tci) & 0xFFFU);
        CpswApp_print("EtherType: 0x%04x", ntohs(vlanFrame->hdr.etherType) & 0xFFFFU);
        payload = vlanFrame->payload;
        len    -= ETH_VLAN_TAG_LEN;
    }
    else
    {
        CpswApp_print("EtherType: 0x%04x", ntohs(frame->hdr.etherType) & 0xFFFFU);
        payload = frame->payload;
    }

    CpswApp_print("Payload  : ");
    for (i = 0; i < len; i++)
    {
        CpswApp_print("0x%02x ", payload[i]);
        if (i && (((i + 1) % OCTETS_PER_ROW) == 0))
        {
            CpswApp_print("           ");
        }
    }

    if (len && ((len % OCTETS_PER_ROW) != 0))
    {
        CpswApp_print("");
    }

    CpswApp_print("");
}

void CpswApp_printHostPortStats2G(CpswStats_HostPort_2g *st)
{
    CpswApp_printStatsNonZero("  rxGoodFrames            = %llu", st->rxGoodFrames);
    CpswApp_printStatsNonZero("  rxBcastFrames           = %llu", st->rxBcastFrames);
    CpswApp_printStatsNonZero("  rxMcastFrames           = %llu", st->rxMcastFrames);
    CpswApp_printStatsNonZero("  rxCrcErrors             = %llu", st->rxCrcErrors);
    CpswApp_printStatsNonZero("  rxOversizedFrames       = %llu",
                                   st->rxOversizedFrames);
    CpswApp_printStatsNonZero("  rxUndersizedFrames      = %llu",
                                   st->rxUndersizedFrames);
    CpswApp_printStatsNonZero("  rxFragments             = %llu", st->rxFragments);
    CpswApp_printStatsNonZero("  aleDrop                 = %llu", st->aleDrop);
    CpswApp_printStatsNonZero("  aleOverrunDrop          = %llu", st->aleOverrunDrop);
    CpswApp_printStatsNonZero("  rxOctets                = %llu", st->rxOctets);
    CpswApp_printStatsNonZero("  txGoodFrames            = %llu", st->txGoodFrames);
    CpswApp_printStatsNonZero("  txBcastFrames           = %llu", st->txBcastFrames);
    CpswApp_printStatsNonZero("  txMcastFrames           = %llu", st->txMcastFrames);
    CpswApp_printStatsNonZero("  txOctets                = %llu", st->txOctets);
    CpswApp_printStatsNonZero("  octetsFrames64          = %llu", st->octetsFrames64);
    CpswApp_printStatsNonZero("  octetsFrames65to127     = %llu",
                                   st->octetsFrames65to127);
    CpswApp_printStatsNonZero("  octetsFrames128to255    = %llu",
                                   st->octetsFrames128to255);
    CpswApp_printStatsNonZero("  octetsFrames256to511    = %llu",
                                   st->octetsFrames256to511);
    CpswApp_printStatsNonZero("  octetsFrames512to1023   = %llu",
                                   st->octetsFrames512to1023);
    CpswApp_printStatsNonZero("  octetsFrames1024        = %llu",
                                   st->octetsFrames1024);
    CpswApp_printStatsNonZero("  netOctets               = %llu", st->netOctets);
    CpswApp_printStatsNonZero("  rxBottomOfFifoDrop      = %llu",
                                   st->rxBottomOfFifoDrop);
    CpswApp_printStatsNonZero("  portMaskDrop            = %llu", st->portMaskDrop);
    CpswApp_printStatsNonZero("  rxTopOfFifoDrop         = %llu",
                                   st->rxTopOfFifoDrop);
    CpswApp_printStatsNonZero("  aleRateLimitDrop        = %llu",
                                   st->aleRateLimitDrop);
    CpswApp_printStatsNonZero("  aleVidIngressDrop       = %llu",
                                   st->aleVidIngressDrop);
    CpswApp_printStatsNonZero("  aleDAEqSADrop           = %llu", st->aleDAEqSADrop);
    CpswApp_printStatsNonZero("  aleBlockDrop            = %llu", st->aleBlockDrop);
    CpswApp_printStatsNonZero("  aleSecureDrop           = %llu", st->aleSecureDrop);
    CpswApp_printStatsNonZero("  aleAuthDrop             = %llu", st->aleAuthDrop);
    CpswApp_printStatsNonZero("  aleUnknownUcast         = %llu",
                                   st->aleUnknownUcast);
    CpswApp_printStatsNonZero("  aleUnknownUcastBcnt     = %llu",
                                   st->aleUnknownUcastBcnt);
    CpswApp_printStatsNonZero("  aleUnknownMcast         = %llu",
                                   st->aleUnknownMcast);
    CpswApp_printStatsNonZero("  aleUnknownMcastBcnt     = %llu",
                                   st->aleUnknownMcastBcnt);
    CpswApp_printStatsNonZero("  aleUnknownBcast         = %llu",
                                   st->aleUnknownBcast);
    CpswApp_printStatsNonZero("  aleUnknownBcastBcnt     = %llu",
                                   st->aleUnknownBcastBcnt);
    CpswApp_printStatsNonZero("  alePolicyMatch          = %llu", st->alePolicyMatch);
    CpswApp_printStatsNonZero("  alePolicyMatchRed       = %llu",
                                   st->alePolicyMatchRed);
    CpswApp_printStatsNonZero("  alePolicyMatchYellow    = %llu",
                                   st->alePolicyMatchYellow);
    CpswApp_printStatsNonZero("  txMemProtectError       = %llu",
                                   st->txMemProtectError);
}

void CpswApp_printMacPortStats2G(CpswStats_MacPort_2g *st)
{
    uint32_t i;

    CpswApp_printStatsNonZero("  rxGoodFrames            = %llu", st->rxGoodFrames);
    CpswApp_printStatsNonZero("  rxBcastFrames           = %llu", st->rxBcastFrames);
    CpswApp_printStatsNonZero("  rxMcastFrames           = %llu", st->rxMcastFrames);
    CpswApp_printStatsNonZero("  rxPauseFrames           = %llu", st->rxPauseFrames);
    CpswApp_printStatsNonZero("  rxCrcErrors             = %llu", st->rxCrcErrors);
    CpswApp_printStatsNonZero("  rxAlignCodeErrors       = %llu",
                                   st->rxAlignCodeErrors);
    CpswApp_printStatsNonZero("  rxOversizedFrames       = %llu",
                                   st->rxOversizedFrames);
    CpswApp_printStatsNonZero("  rxJabberFrames          = %llu", st->rxJabberFrames);
    CpswApp_printStatsNonZero("  rxUndersizedFrames      = %llu",
                                   st->rxUndersizedFrames);
    CpswApp_printStatsNonZero("  rxFragments             = %llu", st->rxFragments);
    CpswApp_printStatsNonZero("  aleDrop                 = %llu", st->aleDrop);
    CpswApp_printStatsNonZero("  aleOverrunDrop          = %llu", st->aleOverrunDrop);
    CpswApp_printStatsNonZero("  rxOctets                = %llu", st->rxOctets);
    CpswApp_printStatsNonZero("  txGoodFrames            = %llu", st->txGoodFrames);
    CpswApp_printStatsNonZero("  txBcastFrames           = %llu", st->txBcastFrames);
    CpswApp_printStatsNonZero("  txMcastFrames           = %llu", st->txMcastFrames);
    CpswApp_printStatsNonZero("  txPauseFrames           = %llu", st->txPauseFrames);
    CpswApp_printStatsNonZero("  txDeferredFrames        = %llu",
                                   st->txDeferredFrames);
    CpswApp_printStatsNonZero("  txCollisionFrames       = %llu",
                                   st->txCollisionFrames);
    CpswApp_printStatsNonZero("  txSingleCollFrames      = %llu",
                                   st->txSingleCollFrames);
    CpswApp_printStatsNonZero("  txMultipleCollFrames    = %llu",
                                   st->txMultipleCollFrames);
    CpswApp_printStatsNonZero("  txExcessiveCollFrames   = %llu",
                                   st->txExcessiveCollFrames);
    CpswApp_printStatsNonZero("  txLateCollFrames        = %llu",
                                   st->txLateCollFrames);
    CpswApp_printStatsNonZero("  rxIPGError              = %llu", st->rxIPGError);
    CpswApp_printStatsNonZero("  txCarrierSenseErrors    = %llu",
                                   st->txCarrierSenseErrors);
    CpswApp_printStatsNonZero("  txOctets                = %llu", st->txOctets);
    CpswApp_printStatsNonZero("  octetsFrames64          = %llu", st->octetsFrames64);
    CpswApp_printStatsNonZero("  octetsFrames65to127     = %llu",
                                   st->octetsFrames65to127);
    CpswApp_printStatsNonZero("  octetsFrames128to255    = %llu",
                                   st->octetsFrames128to255);
    CpswApp_printStatsNonZero("  octetsFrames256to511    = %llu",
                                   st->octetsFrames256to511);
    CpswApp_printStatsNonZero("  octetsFrames512to1023   = %llu",
                                   st->octetsFrames512to1023);
    CpswApp_printStatsNonZero("  octetsFrames1024        = %llu",
                                   st->octetsFrames1024);
    CpswApp_printStatsNonZero("  netOctets               = %llu", st->netOctets);
    CpswApp_printStatsNonZero("  rxBottomOfFifoDrop      = %llu",
                                   st->rxBottomOfFifoDrop);
    CpswApp_printStatsNonZero("  portMaskDrop            = %llu", st->portMaskDrop);
    CpswApp_printStatsNonZero("  rxTopOfFifoDrop         = %llu",
                                   st->rxTopOfFifoDrop);
    CpswApp_printStatsNonZero("  aleRateLimitDrop        = %llu",
                                   st->aleRateLimitDrop);
    CpswApp_printStatsNonZero("  aleVidIngressDrop       = %llu",
                                   st->aleVidIngressDrop);
    CpswApp_printStatsNonZero("  aleDAEqSADrop           = %llu", st->aleDAEqSADrop);
    CpswApp_printStatsNonZero("  aleBlockDrop            = %llu", st->aleBlockDrop);
    CpswApp_printStatsNonZero("  aleSecureDrop           = %llu", st->aleSecureDrop);
    CpswApp_printStatsNonZero("  aleAuthDrop             = %llu", st->aleAuthDrop);
    CpswApp_printStatsNonZero("  aleUnknownUcast         = %llu",
                                   st->aleUnknownUcast);
    CpswApp_printStatsNonZero("  aleUnknownUcastBcnt     = %llu",
                                   st->aleUnknownUcastBcnt);
    CpswApp_printStatsNonZero("  aleUnknownMcast         = %llu",
                                   st->aleUnknownMcast);
    CpswApp_printStatsNonZero("  aleUnknownMcastBcnt     = %llu",
                                   st->aleUnknownMcastBcnt);
    CpswApp_printStatsNonZero("  aleUnknownBcast         = %llu",
                                   st->aleUnknownBcast);
    CpswApp_printStatsNonZero("  aleUnknownBcastBcnt     = %llu",
                                   st->aleUnknownBcastBcnt);
    CpswApp_printStatsNonZero("  alePolicyMatch          = %llu", st->alePolicyMatch);
    CpswApp_printStatsNonZero("  alePolicyMatchRed       = %llu",
                                   st->alePolicyMatchRed);
    CpswApp_printStatsNonZero("  alePolicyMatchYellow    = %llu",
                                   st->alePolicyMatchYellow);
    CpswApp_printStatsNonZero("  txMemProtectError       = %llu",
                                   st->txMemProtectError);

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPri); i++)
    {
        CpswApp_printStatsWithIdxNonZero("  txPri[%u]                = %llu", i, st->txPri[i]);
    }

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPriBcnt); i++)
    {
        CpswApp_printStatsWithIdxNonZero("  txPriBcnt[%u]            = %llu", i, st->txPriBcnt[i]);
    }

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPriDrop); i++)
    {
        CpswApp_printStatsWithIdxNonZero("  txPriDrop[%u]            = %llu", i, st->txPriDrop[i]);
    }

    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(st->txPriDropBcnt); i++)
    {
        CpswApp_printStatsWithIdxNonZero("  txPriDropBcnt[%u]        = %llu", i, st->txPriDropBcnt[i]);
    }
}


void CpswApp_submitRxPackets(CpswDma_RxFlowHandle hRxFlow,
                             CpswDma_PktInfoQ *pSubmitQ)
{
    int32_t retVal;
    CpswDma_PktInfoQ unUsedQ;

    CpswUtils_initQ(&unUsedQ);
    retVal = CpswDma_submitRxPackets(hRxFlow, pSubmitQ, &unUsedQ);
    if (CPSW_SOK != retVal)
    {
        CpswApp_print("%s:%d Error: failed to submit pkts: %d",
            __FUNCTION__, __LINE__, retVal);
    }

    if (CpswUtils_getQCount(&unUsedQ))
    {
        /* Copy unused packets to back to readyQ */
        CpswUtils_appendQ(pSubmitQ, &unUsedQ);
    }
}

int32_t CpswApp_submitTxPackets(CpswDma_TxChHandle hTxCh,
                                CpswDma_PktInfoQ *pSubmitQ)
{
    int32_t retVal;
    CpswDma_PktInfoQ unUsedQ;

    CpswUtils_initQ(&unUsedQ);
    retVal = CpswDma_submitTxReadyPackets(hTxCh, pSubmitQ, &unUsedQ);
    if (CPSW_SOK != retVal)
    {
        CpswApp_print("%s:%d Error: failed to submit pkts: %d",
            __FUNCTION__, __LINE__, retVal);
    }

    if (CpswUtils_getQCount(&unUsedQ))
    {
        CpswUtils_appendQ(pSubmitQ, &unUsedQ);
    }

    return retVal;
}

void CpswApp_validatePacketState(CpswDma_PktInfoQ *pQueue,
                                 uint32_t expectedState,
                                 uint32_t newState)
{
    uint32_t i;
    CpswDma_PktInfo *pktInfo = (CpswDma_PktInfo *)pQueue->pHead;

    for (i = 0; i < CpswUtils_getQCount(pQueue); i++)
    {
        CpswUtils_checkPktState(&pktInfo->pktState,
                                CPSW_PKTSTATE_MODULE_APP,
                                expectedState,
                                newState);
        pktInfo = (CpswDma_PktInfo *)pktInfo->node.pNext;
    }
}

uint64_t CpswApp_virtToPhyFxn(const void *virtAddr,
                              void *appData)
{
    int ret;
    off64_t    phyAddr;
    uint32_t   length;

    CpswApp_assert(virtAddr != NULL);

    if(appData != NULL_PTR) {
        length = (uint32_t) *((uint32_t *) appData);
    }
    else {
        length = 4;
    }

    /* Get destination physical address */
    ret = mem_offset64(virtAddr, NOFD, length, &phyAddr, NULL);
    if (ret) {
        CpswApp_print("%s:%d Error: failed mem_offset", __FUNCTION__, __LINE__);
    }
    //CpswApp_print("%s: 0x%lx -> 0x%lx ",__FUNCTION__, virtAddr, phyAddr);

    return (uint64_t ) phyAddr;
}

void *CpswApp_phyToVirtFxn(uint64_t phyAddr,
                           void *appData)
{
    uint64_t *temp = 0;
    uint32_t length = 0;

    CpswApp_assert(phyAddr != 0);

    if(appData != NULL_PTR) {
        length = (uint32_t) *((uint32_t *) appData);
    }
    else {
        length = 8192; // Default to 8K
    }

    temp  = mmap_device_memory(0, length, PROT_READ|PROT_WRITE, 0, phyAddr);
    if((temp == MAP_FAILED))
    {
        CpswApp_print("%s:%d Error: mmmap_device_memory failed",
            __FUNCTION__, __LINE__);
    }
    //CpswApp_print("%s: 0x%lx -> 0x%lx size -> %d",__FUNCTION__, phyAddr, temp, length);

    return ((void *) temp);
}

uint64_t CpswApp_udmaVirtToPhyFxn(const void *virtAddr,
                                  uint32_t chNum,
                                  void *appData)
{
    return CpswApp_virtToPhyFxn(virtAddr, appData);
}

void *CpswApp_udmaPhyToVirtFxn(uint64_t phyAddr,
                               uint32_t chNum,
                               void *appData)
{
    return CpswApp_phyToVirtFxn(phyAddr, appData);
}

static void CpswApp_unlockMmr(volatile uint32_t *kick0,
                              volatile uint32_t *kick1)
{
    uint32_t kick0Val;

    /* if kick lock registers are locked*/
    if ((CSL_REG32_RD(kick0) & 0x1U) == 0U)
    {
        /* unlock the partition by writing the unlock values to the kick lock registers*/
        CSL_REG32_WR(kick0, MMR_KICK0_UNLOCK_VAL);
        CSL_REG32_WR(kick1, MMR_KICK1_UNLOCK_VAL);
    }

    kick0Val = CSL_REG32_FEXT(kick0, MMR_KICK_UNLOCKED);

    /* check to see if unlock bit in lock0 register is set */
    if (1U != kick0Val)
    {
        CpswApp_assert(false);
    }
}

static void CpswApp_lockMmr(volatile uint32_t *kick0,
                            volatile uint32_t *kick1)
{
    uint32_t kick0Val;

    /* check to see if either of the kick registers are unlocked.*/
    if ((CSL_REG32_RD(kick0) & 0x1) == 1U)
    {
        /* write the kick lock value to the kick lock registers to lock the partition*/
        CSL_REG32_WR(kick0, MMR_KICK_LOCK_VAL);
        CSL_REG32_WR(kick1, MMR_KICK_LOCK_VAL);
    }

    /* check to see if unlock bit in lock0 register is set */
    kick0Val = CSL_REG32_FEXT(kick0, MMR_KICK_UNLOCKED);

    if (0U != kick0Val)
    {
        CpswApp_assert(false);
    }
}

void CpswApp_mcuMmrCtrl(CpswApp_CtrlMmrType mmrNum,
                        CpswApp_MmrLockType lock)
{
#if defined(SOC_AM65XX) || defined(SOC_J721E)
    CSL_mcu_ctrl_mmr_cfg0Regs *regs = NULL;
    volatile uint32_t *kick0 = NULL, *kick1 = NULL;

    regs = (CSL_mcu_ctrl_mmr_cfg0Regs *)mmap_device_memory(0,
          sizeof(CSL_mcu_ctrl_mmr_cfg0Regs), PROT_READ|PROT_WRITE|PROT_NOCACHE, 0,
                (uint64_t)CSL_MCU_CTRL_MMR0_CFG0_BASE);

    CpswApp_assert(regs != NULL);

    switch (mmrNum)
    {
        case CPSWAPP_MMR_LOCK0:
            kick0 = &regs->LOCK0_KICK0;
            kick1 = &regs->LOCK0_KICK1;
            break;

        case CPSWAPP_MMR_LOCK1:
            kick0 = &regs->LOCK1_KICK0;
            kick1 = &regs->LOCK1_KICK1;
            break;

        case CPSWAPP_MMR_LOCK2:
            kick0 = &regs->LOCK2_KICK0;
            kick1 = &regs->LOCK2_KICK1;
            break;

        case CPSWAPP_MMR_LOCK3:
            kick0 = &regs->LOCK3_KICK0;
            kick1 = &regs->LOCK3_KICK1;
            break;

        case CPSWAPP_MMR_LOCK4:
            kick0 = &regs->LOCK4_KICK0;
            kick1 = &regs->LOCK4_KICK1;
            break;

        default:
            CpswApp_assert(false);
            break;
    }

    if ((NULL != kick0) && (NULL != kick1))
    {
        if (CPSWAPP_UNLOCK_MMR == lock)
        {
            CpswApp_unlockMmr(kick0, kick1);
        }
        else
        {
            CpswApp_lockMmr(kick0, kick1);
        }
    }

    munmap_device_memory((void *)regs,sizeof(CSL_mcu_ctrl_mmr_cfg0Regs));
#else
#error "Unsupported platform"
#endif
}

void CpswApp_mainMmrCtrl(CpswApp_CtrlMmrType mmrNum,
                         CpswApp_MmrLockType lock)
{
#if defined(SOC_AM65XX) || defined(SOC_J721E)
    CSL_main_ctrl_mmr_cfg0Regs *regs = NULL;
    volatile uint32_t *kick0 = NULL, *kick1 = NULL;

    regs = (CSL_main_ctrl_mmr_cfg0Regs *)mmap_device_memory(0,
            sizeof(CSL_main_ctrl_mmr_cfg0Regs), PROT_READ|PROT_WRITE|PROT_NOCACHE, 0,
            (uint64_t)CSL_CTRL_MMR0_CFG0_BASE);

    CpswApp_assert(regs != NULL);

    switch (mmrNum)
    {
        case CPSWAPP_MMR_LOCK0:
            kick0 = &regs->LOCK0_KICK0;
            kick1 = &regs->LOCK0_KICK1;
            break;

        case CPSWAPP_MMR_LOCK1:
            kick0 = &regs->LOCK1_KICK0;
            kick1 = &regs->LOCK1_KICK1;
            break;

        case CPSWAPP_MMR_LOCK2:
            kick0 = &regs->LOCK2_KICK0;
            kick1 = &regs->LOCK2_KICK1;
            break;

        case CPSWAPP_MMR_LOCK3:
            kick0 = &regs->LOCK3_KICK0;
            kick1 = &regs->LOCK3_KICK1;
            break;

        case CPSWAPP_MMR_LOCK4:
            kick0 = &regs->LOCK4_KICK0;
            kick1 = &regs->LOCK4_KICK1;
            break;

#if defined(SOC_J721E)
        case CPSWAPP_MMR_LOCK5:
            kick0 = &regs->LOCK5_KICK0;
            kick1 = &regs->LOCK5_KICK1;
            break;
#endif
        case CPSWAPP_MMR_LOCK6:
            kick0 = &regs->LOCK6_KICK0;
            kick1 = &regs->LOCK6_KICK1;
            break;

        case CPSWAPP_MMR_LOCK7:
            kick0 = &regs->LOCK7_KICK0;
            kick1 = &regs->LOCK7_KICK1;
            break;

        default:
            CpswApp_assert(false);
            break;
    }

    if ((NULL != kick0) && (NULL != kick1))
    {
        if (CPSWAPP_UNLOCK_MMR == lock)
        {
            CpswApp_unlockMmr(kick0, kick1);
        }
        else
        {
            CpswApp_lockMmr(kick0, kick1);
        }
    }

    munmap_device_memory((void *)regs,sizeof(CSL_main_ctrl_mmr_cfg0Regs));
#else
#error "Unsupported platform"
#endif
}

void CpswApp_selectCptsClock(Cpsw_Type cpswType,
                             CpswApp_CptsClkSelMux clkSelMux)
{
#if defined(SOC_AM65XX) || defined(SOC_J721E)
    uint32_t muxVal;

    muxVal = (uint32_t) clkSelMux;
    switch (cpswType)
    {
        case CPSW_2G:
        {
            CSL_mcu_ctrl_mmr_cfg0Regs *mcuRegs;
            mcuRegs = (CSL_mcu_ctrl_mmr_cfg0Regs *)mmap_device_memory(0,
                sizeof(CSL_mcu_ctrl_mmr_cfg0Regs), PROT_READ|PROT_WRITE|PROT_NOCACHE, 0,
                (uint64_t)CSL_MCU_CTRL_MMR0_CFG0_BASE);

            CpswApp_assert(mcuRegs != NULL);

            CpswApp_mcuMmrCtrl(CPSWAPP_MMR_LOCK1, CPSWAPP_UNLOCK_MMR);
            CSL_REG32_FINS(&mcuRegs->MCU_ENET_CLKSEL,
                           MCU_CTRL_MMR_CFG0_MCU_ENET_CLKSEL_CPTS_CLKSEL,
                           muxVal);
            CpswApp_mcuMmrCtrl(CPSWAPP_MMR_LOCK1, CPSWAPP_LOCK_MMR);

            CpswApp_print("Success - write to MCU_ENET_CLKSEL - 0x%x",
                CSL_REG32_RD(&mcuRegs->MCU_ENET_CLKSEL));

            munmap_device_memory((void *)mcuRegs,sizeof(CSL_mcu_ctrl_mmr_cfg0Regs));
        }
        break;

#if defined(SOC_J721E)
        case CPSW_9G:
        {
            CSL_main_ctrl_mmr_cfg0Regs *mainRegs;
            mainRegs = (CSL_main_ctrl_mmr_cfg0Regs *)mmap_device_memory(0,
                sizeof(CSL_main_ctrl_mmr_cfg0Regs), PROT_READ|PROT_WRITE|PROT_NOCACHE, 0,
                (uint64_t)CSL_CTRL_MMR0_CFG0_BASE);

            CpswApp_assert(mainRegs != NULL);

            CpswApp_mainMmrCtrl(CPSWAPP_MMR_LOCK1, CPSWAPP_UNLOCK_MMR);
            CSL_REG32_FINS(&mainRegs->CPSW_CLKSEL,
                           MAIN_CTRL_MMR_CFG0_CPSW_CLKSEL_CPTS_CLKSEL,
                           muxVal);
            CpswApp_mainMmrCtrl(CPSWAPP_MMR_LOCK1, CPSWAPP_LOCK_MMR);

            CpswApp_print("Success - write to CPSW_CLKSEL - 0x%x",
                CSL_REG32_RD(&mainRegs->CPSW_CLKSEL));

            munmap_device_memory((void *)mainRegs,sizeof(CSL_main_ctrl_mmr_cfg0Regs));
        }
        break;
#endif
        default:
            CpswApp_assert(false);
            break;
    }

#else
#error "Unsupported platform"
#endif
}

int32_t CpswApp_allocRxFlow(Cpsw_Handle hCpsw,
                            uint32_t coreKey,
                            uint32_t coreId,
                            uint32_t *rxFlowStartIdx,
                            uint32_t *flowIdx)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    CpswRm_AllocRxFlowOutArgs allocRxFlowOutArgs;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &coreKey, &allocRxFlowOutArgs);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_RM_IOCTL_ALLOC_RX_FLOW,
                        &prms);

    if (status == CPSW_SOK)
    {
        *rxFlowStartIdx = allocRxFlowOutArgs.startIdx;
        *flowIdx        = allocRxFlowOutArgs.flowIdx;
    }
    else
    {
        CpswApp_print("%s:%d Error: IOCTL_ALLOC_RX_FLOW failed : %d",
            __FUNCTION__, __LINE__, status);
    }

    return status;
}

static int32_t CpswApp_allocTxCh(Cpsw_Handle hCpsw,
                                 uint32_t coreKey,
                                 uint32_t coreId,
                                 uint32_t *txPSILThreadId)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;

    /* Allocate Tx Ch */
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &coreKey, txPSILThreadId);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_RM_IOCTL_ALLOC_TX_CH_PEERID,
                        &prms);
    if (status != CPSW_SOK)
    {
        *txPSILThreadId = CPSW_RM_TXCHNUM_INVALID;
        CpswApp_print("%s:%d Error: IOCTL_ALLOC_TX_CH_PEERID failed: %d",
            __FUNCTION__, __LINE__, status);
    }

    return status;
}

int32_t CpswApp_allocMac(Cpsw_Handle hCpsw,
                         uint32_t coreKey,
                         uint32_t coreId,
                         uint8_t *macAddress)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    CpswRm_AllocMacAddrOutArgs allocMacOutArgs;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &coreKey, &allocMacOutArgs);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_RM_IOCTL_ALLOC_MAC_ADDR,
                        &prms);

    if (status == CPSW_SOK)
    {
        memcpy(macAddress, allocMacOutArgs.macAddr, sizeof(allocMacOutArgs.macAddr));
    }
    else
    {
        CpswApp_print("%s:%d Error: IOCTL_ALLOC_MAC_ADDR failed : %d",
            __FUNCTION__, __LINE__, status);
    }

    return status;
}

int32_t CpswApp_freeRxFlow(Cpsw_Handle hCpsw,
                           uint32_t coreKey,
                           uint32_t coreId,
                           uint32_t rxFlowIdx)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    CpswRm_FreeRxFlowInArgs freeRxFlowInArgs;

    /*Free Rx Flow*/
    freeRxFlowInArgs.coreKey = coreKey;
    freeRxFlowInArgs.flowIdx = rxFlowIdx;

    CPSW_IOCTL_SET_IN_ARGS(&prms, &freeRxFlowInArgs);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_RM_IOCTL_FREE_RX_FLOW,
                        &prms);

    return status;
}

int32_t CpswApp_freeTxCh(Cpsw_Handle hCpsw,
                         uint32_t coreKey,
                         uint32_t coreId,
                         uint32_t txChNum)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    CpswRm_FreeTxChInArgs freeTxChInArgs;

    /* Release Tx Ch */
    freeTxChInArgs.coreKey = coreKey;
    freeTxChInArgs.txChNum = txChNum;

    CPSW_IOCTL_SET_IN_ARGS(&prms, &freeTxChInArgs);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_RM_IOCTL_FREE_TX_CH_PEERID,
                        &prms);

    return status;
}

int32_t CpswApp_freeMac(Cpsw_Handle hCpsw,
                        uint32_t coreKey,
                        uint32_t coreId,
                        uint8_t *macAddress)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    CpswRm_FreeMacAddrInArgs freeMacInArgs;

    freeMacInArgs.coreKey = coreKey;
    CPSW_UTILS_ARRAY_COPY(freeMacInArgs.macAddr, macAddress);

    CPSW_IOCTL_SET_IN_ARGS(&prms, &freeMacInArgs);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_RM_IOCTL_FREE_MAC_ADDR,
                        &prms);

    if (status != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: IOCTL_FREE_MAC_ADDR failed : %d",
            __FUNCTION__, __LINE__, status);
    }

    return status;
}

int32_t CpswApp_registerDefaultRxFlow(Cpsw_Handle hCpsw,
                                      uint32_t coreKey,
                                      uint32_t coreId,
                                      uint32_t rxFlowStartIdx,
                                      uint32_t rxFlowIdx)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    Cpsw_DefaultFlowInfo defaultFlow;

    defaultFlow.coreKey  = coreKey;
    defaultFlow.startIdx = rxFlowStartIdx;
    defaultFlow.flowIdx  = rxFlowIdx;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &defaultFlow);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_IOCTL_REGISTER_RX_DEFAULT_FLOW,
                        &prms);

    if (status != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: REGISTER_RX_DEFAULT_FLOW failed : %d",
            __FUNCTION__, __LINE__, status);
    }

    return status;
}

int32_t CpswApp_unregisterDefaultRxFlow(Cpsw_Handle hCpsw,
                                        uint32_t coreKey,
                                        uint32_t coreId,
                                        uint32_t rxFlowStartIdx,
                                        uint32_t rxFlowIdx)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    Cpsw_DefaultFlowInfo defaultFlow;

    defaultFlow.coreKey  = coreKey;
    defaultFlow.startIdx = rxFlowStartIdx;
    defaultFlow.flowIdx  = rxFlowIdx;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &defaultFlow);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_IOCTL_UNREGISTER_RX_DEFAULT_FLOW,
                        &prms);

    if (status != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: UNREGISTER_RX_DEFAULT_FLOW failed : %d",
            __FUNCTION__, __LINE__, status);
    }

    return status;
}

int32_t CpswApp_registerDstMacRxFlow(Cpsw_Handle hCpsw,
                                     uint32_t coreKey,
                                     uint32_t coreId,
                                     uint32_t rxFlowStartIdx,
                                     uint32_t rxFlowIdx,
                                     uint8_t macAddress[CPSW_MAC_ADDR_LEN])
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    Cpsw_MacDstFlowInfo macDstFlow;

    macDstFlow.coreKey  = coreKey;
    macDstFlow.startIdx = rxFlowStartIdx;
    macDstFlow.flowIdx  = rxFlowIdx;
    CPSW_UTILS_ARRAY_COPY(macDstFlow.macAddress, macAddress);
    CPSW_IOCTL_SET_IN_ARGS(&prms, &macDstFlow);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_IOCTL_REGISTER_DSTMAC_RX_FLOW,
                        &prms);

    if (status != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: REGISTER_DSTMAC_RX_FLOW failed : %d",
            __FUNCTION__, __LINE__, status);
    }

    return status;
}

int32_t CpswApp_unregisterDstMacRxFlow(Cpsw_Handle hCpsw,
                                       uint32_t coreKey,
                                       uint32_t coreId,
                                       uint32_t rxFlowStartIdx,
                                       uint32_t rxFlowIdx,
                                       uint8_t macAddress[CPSW_MAC_ADDR_LEN])
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    Cpsw_MacDstFlowInfo macDstFlow;

    macDstFlow.coreKey  = coreKey;
    macDstFlow.startIdx = rxFlowStartIdx;
    macDstFlow.flowIdx  = rxFlowIdx;
    CPSW_UTILS_ARRAY_COPY(macDstFlow.macAddress, macAddress);
    CPSW_IOCTL_SET_IN_ARGS(&prms, &macDstFlow);
    status = Cpsw_ioctl(hCpsw,
                        coreId,
                        CPSW_IOCTL_UNREGISTER_DSTMAC_RX_FLOW,
                        &prms);

    if (status != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: UNREGISTER_DSTMAC_RX_FLOW failed : %d",
            __FUNCTION__, __LINE__, status);
    }

    return status;
}

bool CpswApp_isPortLinkUp(Cpsw_Handle hCpsw,
                          uint32_t coreId,
                          Cpsw_MacPort portNum)
{
    Cpsw_IoctlPrms prms;
    Cpsw_GenericPortLinkInArgs portLinkInArgs;
    bool linked = true;
    int32_t status;

    portLinkInArgs.portNum = portNum;
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &portLinkInArgs, &linked);

    status = Cpsw_ioctl(hCpsw, coreId, CPSW_IOCTL_IS_PORT_LINK_UP, &prms);
    if (status != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: failed to get port %u's link status: %d",
              __FUNCTION__, __LINE__, portNum, status);
        linked = false;
    }

    return linked;
}

static void CpswApp_reduceCoreMacAllocation(CpswRm_ResourcePrms *rscPrms,
                                            uint32_t *pReduceCount,
                                            uint32_t coreMinCount,
                                            bool skipCore,
                                            uint32_t skipCoreId)
{
    uint32_t i;

    for (i = 0; (i < rscPrms->numCores) && (*pReduceCount > 0); i++)
    {
        if ((rscPrms->coreDmaResourceInfo[i].numMacAddress > coreMinCount)
            &&
            ((skipCore == false) || (skipCoreId != rscPrms->coreDmaResourceInfo[i].coreId)))
        {
            uint32_t coreMacAddrReducedCount = (rscPrms->coreDmaResourceInfo[i].numMacAddress - coreMinCount);

            if (*pReduceCount >= coreMacAddrReducedCount)
            {
                *pReduceCount -= coreMacAddrReducedCount;
            }
            else
            {
                coreMacAddrReducedCount -= *pReduceCount;
                *pReduceCount            = 0;
            }

            CpswApp_print("CpswApp_reduceCoreMacAllocation: "
                           "Reduced Mac Address Allocation for CoreId:%u From %u To %u ",
                           rscPrms->coreDmaResourceInfo[i].coreId,
                           rscPrms->coreDmaResourceInfo[i].numMacAddress,
                           (rscPrms->coreDmaResourceInfo[i].numMacAddress - coreMacAddrReducedCount));
            rscPrms->coreDmaResourceInfo[i].numMacAddress -= coreMacAddrReducedCount;
        }
    }
}

static void CpswApp_updateMacResourcePartition(CpswRm_ResourcePrms *rscPrms,
                                               uint32_t availMacCount,
                                               uint32_t selfCoreId)
{
    uint32_t totalRscPatitionMacCount;
    uint32_t i;

    totalRscPatitionMacCount = 0;
    for (i = 0; i < rscPrms->numCores; i++)
    {
        totalRscPatitionMacCount += rscPrms->coreDmaResourceInfo[i].numMacAddress;
    }

    if (totalRscPatitionMacCount > availMacCount)
    {
        uint32_t reduceCount = totalRscPatitionMacCount - availMacCount;

        /* First reduce mac count for cores with more than one mac address allocation */
        CpswApp_reduceCoreMacAllocation(rscPrms, &reduceCount, 1, false, selfCoreId);
        if (reduceCount)
        {
            /* Next reduce mac address for core other than self core to 0 */
            CpswApp_reduceCoreMacAllocation(rscPrms, &reduceCount, 0, true, selfCoreId);
        }

        /* Finally reduce self core also to 0 */
        if (reduceCount)
        {
            /* Next reduce mac address for core other than self core to 0 */
            CpswApp_reduceCoreMacAllocation(rscPrms, &reduceCount, 0, false, selfCoreId);
        }

        CpswApp_assert(reduceCount == 0);
    }
}

void CpswApp_initResourceConfig(Cpsw_Type cpswType,
                                uint32_t selfCoreId,
                                CpswRm_ResourceConfig *rscCfg,
                                uint8_t *currentMacAddr)
{
    const CpswRm_ResourcePrms *rscPrms         = CpswAppRm_getResourcePartitionInfo(cpswType);
    const CpswRm_IoctlPermissionTable *ioPerms = CpswAppRm_getIoctlPermissionInfo(cpswType);

    CpswApp_assert(rscPrms != NULL);
    rscCfg->resourcePartitionInfo = *rscPrms;

    CpswApp_assert(ioPerms != NULL);
    rscCfg->ioctlPermissionInfo = *ioPerms;

#if defined (CPSW2G)
    CpswApp_assert(cpswType == CPSW_2G);
#elif defined (CPSW9G)
    CpswApp_assert(cpswType == CPSW_9G);
#else
    #error "Unsupported!"
#endif

#if 0
    CpswSoc_getEFusedMacAddr(rscCfg->macList.macAddress[0]);
#else
    // for qnx, we would need to use the current mac address passed by devnp
    memcpy(rscCfg->macList.macAddress[0], currentMacAddr, ETH_MAC_ADDR_LEN);
#endif
    rscCfg->macList.numMacAddress = 1;

    CpswApp_updateMacResourcePartition(&rscCfg->resourcePartitionInfo,
                                       rscCfg->macList.numMacAddress,
                                       selfCoreId);
    rscCfg->selfCoreId = selfCoreId;
}

void CpswApp_setCommonRxFlowPrms(CpswDma_OpenRxFlowPrms *pRxFlowPrms, uint32_t numRxPkts)
{

    pRxFlowPrms->numRxPkts           = numRxPkts;
    pRxFlowPrms->disableCacheOpsFlag = false;
    pRxFlowPrms->rxFlowMtu           = MAX_ETH_PACKET_SIZE;

    pRxFlowPrms->ringMemAllocFxn = &CpswAppMem_allocRingMemFxn;
    pRxFlowPrms->ringMemFreeFxn  = &CpswAppMem_freeRingMemFxn;

    pRxFlowPrms->dmaDescAllocFxn = &CpswAppMem_allocDmaDescFxn;
    pRxFlowPrms->dmaDescFreeFxn  = &CpswAppMem_freeDmaDescFxn;
}

void CpswApp_addHostPortEntry(Cpsw_Handle hCpsw,
                              uint32_t coreId,
                              uint8_t *macAddr)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setUcastOutArgs;
    CpswAle_SetUcastEntryInArgs setUcastInArgs;

    memset(&setUcastInArgs, 0, sizeof(setUcastInArgs));
    memcpy(&setUcastInArgs.addr.addr[0U], macAddr, sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = 0U;
    setUcastInArgs.info.portNum = CPSW_ALE_HOST_PORT_NUM;
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = true;
    setUcastInArgs.info.super   = false;
    setUcastInArgs.info.ageable = false;
    setUcastInArgs.info.trunk   = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Cpsw_ioctl(hCpsw, coreId, CPSW_ALE_IOCTL_ADD_UNICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: failed CPSW_ALE_IOCTL_ADD_UNICAST: %d",
             __FUNCTION__, __LINE__, status);
    }

    CpswApp_assert(status == CPSW_SOK);
}

static void CpswApp_delAddrEntry(Cpsw_Handle hCpsw,
                                 uint32_t coreId,
                                 uint8_t *macAddr)
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_MacAddrInfo addrInfo;

    memset(&addrInfo, 0, sizeof(addrInfo));
    memcpy(&addrInfo.addr[0U], macAddr, sizeof(addrInfo.addr));
    addrInfo.vlanId = 0U;

    CPSW_IOCTL_SET_IN_ARGS(&prms, &addrInfo);

    status = Cpsw_ioctl(hCpsw, coreId, CPSW_ALE_IOCTL_REMOVE_ADDR, &prms);
    if (status != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: failed CPSW_ALE_IOCTL_REMOVE_ADDR: %d",
             __FUNCTION__, __LINE__, status);
    }

    CpswApp_assert(status == CPSW_SOK);
}

void CpswApp_setCommonTxChPrms(CpswDma_OpenTxChPrms *pTxChPrms, uint32_t numTxPkts)
{

    pTxChPrms->numTxPkts           = numTxPkts;
    pTxChPrms->disableCacheOpsFlag = false;

    pTxChPrms->ringMemAllocFxn = &CpswAppMem_allocRingMemFxn;
    pTxChPrms->ringMemFreeFxn  = &CpswAppMem_freeRingMemFxn;

    pTxChPrms->dmaDescAllocFxn = &CpswAppMem_allocDmaDescFxn;
    pTxChPrms->dmaDescFreeFxn  = &CpswAppMem_freeDmaDescFxn;
}
#if defined (CPSW2G)
void CpswApp_openTxCh(Cpsw_Handle hCpsw,
                      uint32_t coreKey,
                      uint32_t coreId,
                      uint32_t *pTxChNum,
                      CpswDma_TxChHandle *pTxChHandle,
                      CpswDma_OpenTxChPrms *pCpswTxChCfg)
{
    int32_t status;

    status = CpswApp_allocTxCh(hCpsw,
                                    coreKey,
                                    coreId,
                                    pTxChNum);
    CpswApp_assert(CPSW_SOK == status);

    pCpswTxChCfg->chNum = *pTxChNum;

    *pTxChHandle = CpswDma_openTxCh(pCpswTxChCfg);
    CpswApp_assert(NULL != *pTxChHandle);
}

void CpswApp_closeTxCh(Cpsw_Handle hCpsw,
                       uint32_t coreKey,
                       uint32_t coreId,
                       CpswDma_PktInfoQ *pFqPktInfoQ,
                       CpswDma_PktInfoQ *pCqPktInfoQ,
                       CpswDma_TxChHandle hTxChHandle,
                       uint32_t txChNum)
{
    int32_t status;

    CpswUtils_initQ(pFqPktInfoQ);
    CpswUtils_initQ(pCqPktInfoQ);

    // Note: No need to call disable as we are not enabling TX Event
    //CpswDma_disableTxEvent(hTxChHandle);

    status = CpswDma_closeTxCh(hTxChHandle, pFqPktInfoQ, pCqPktInfoQ);
    CpswApp_assert(CPSW_SOK == status);

    status = CpswApp_freeTxCh(hCpsw,
                               coreKey,
                               coreId,
                               txChNum);
    CpswApp_assert(CPSW_SOK == status);
}

void CpswApp_openRxFlow(Cpsw_Handle hCpsw,
                        uint32_t coreKey,
                        uint32_t coreId,
                        bool useDefaultFlow,
                        uint32_t *pRxFlowStartIdx,
                        uint32_t *pRxFlowIdx,
                        uint8_t macAddr[CPSW_MAC_ADDR_LEN],
                        CpswDma_RxFlowHandle *pRxFlowHandle,
                        CpswDma_OpenRxFlowPrms *pRxFlowPrms)
{
    int32_t status = CPSW_SOK;

    status = CpswApp_allocRxFlow(hCpsw,
                                  coreKey,
                                  coreId,
                                  pRxFlowStartIdx,
                                  pRxFlowIdx);
    CpswApp_assert(status == CPSW_SOK);

    pRxFlowPrms->startIdx = *pRxFlowStartIdx;
    pRxFlowPrms->flowIdx  = *pRxFlowIdx;

    *pRxFlowHandle = CpswDma_openRxFlow(pRxFlowPrms);
    CpswApp_assert(*pRxFlowHandle != NULL);

    status = CpswApp_allocMac(hCpsw, coreKey, coreId, macAddr);
    CpswApp_assert(status == CPSW_SOK);

    CpswApp_addHostPortEntry(hCpsw, coreId, macAddr);
    if (useDefaultFlow)
    {
        status = CpswApp_registerDefaultRxFlow(hCpsw,
                                                coreKey,
                                                coreId,
                                                *pRxFlowStartIdx,
                                                *pRxFlowIdx);
    }
    else
    {
        status = CpswApp_registerDstMacRxFlow(hCpsw,
                                               coreKey,
                                               coreId,
                                               *pRxFlowStartIdx,
                                               *pRxFlowIdx,
                                               macAddr);
    }

    CpswApp_assert(status == CPSW_SOK);
}

void CpswApp_closeRxFlow(Cpsw_Handle hCpsw,
                         uint32_t coreKey,
                         uint32_t coreId,
                         bool useDefaultFlow,
                         CpswDma_PktInfoQ *pFqPktInfoQ,
                         CpswDma_PktInfoQ *pCqPktInfoQ,
                         uint32_t rxFlowStartIdx,
                         uint32_t rxFlowIdx,
                         uint8_t macAddr[CPSW_MAC_ADDR_LEN],
                         CpswDma_RxFlowHandle hRxFlow)
{
    int32_t status = CPSW_SOK;

    CpswUtils_initQ(pFqPktInfoQ);
    CpswUtils_initQ(pCqPktInfoQ);

    CpswDma_disableRxEvent(hRxFlow);

    if (useDefaultFlow)
    {
        status = CpswApp_unregisterDefaultRxFlow(hCpsw,
                                                 coreKey,
                                                 coreId,
                                                 rxFlowStartIdx,
                                                 rxFlowIdx);
    }
    else
    {
        status = CpswApp_unregisterDstMacRxFlow(hCpsw,
                                                coreKey,
                                                coreId,
                                                rxFlowStartIdx,
                                                rxFlowIdx,
                                                macAddr);
    }

    CpswApp_assert(status == CPSW_SOK);
    CpswApp_delAddrEntry(hCpsw, coreId, macAddr);
    status = CpswDma_closeRxFlow(hRxFlow,
                                 pFqPktInfoQ,
                                 pCqPktInfoQ);
    CpswApp_assert(status == CPSW_SOK);

    status = CpswApp_freeMac(hCpsw,
                                  coreKey,
                                  coreId,
                                  macAddr);
    CpswApp_assert(status == CPSW_SOK);

    status = CpswApp_freeRxFlow(hCpsw,
                                     coreKey,
                                     coreId,
                                     rxFlowIdx);
    CpswApp_assert(status == CPSW_SOK);
}
#elif defined (CPSW9G)
void CpswApp_openTxCh(CpswApp_Obj *obj,
                      CpswDma_OpenTxChPrms *pCpswTxChCfg)
{
    CpswProxy_allocTxCh(obj->hCpswProxy,
                        obj->hCpsw,
                        obj->coreKey,
                        &obj->txChNum);
    pCpswTxChCfg->chNum = obj->txChNum;

    obj->hTxCh = CpswDma_openTxCh(pCpswTxChCfg);
    CpswApp_assert(NULL != obj->hTxCh);
}

void CpswApp_closeTxCh(CpswApp_Obj *obj,
                       CpswDma_PktInfoQ *pFqPktInfoQ,
                       CpswDma_PktInfoQ *pCqPktInfoQ)
{
    int32_t status;

    CpswUtils_initQ(pFqPktInfoQ);
    CpswUtils_initQ(pCqPktInfoQ);

    // Note: No need to call disable as we are not enabling TX Event
    //CpswDma_disableTxEvent(obj->hTxCh);

    status = CpswDma_closeTxCh(obj->hTxCh, pFqPktInfoQ, pCqPktInfoQ);
    CpswApp_assert(CPSW_SOK == status);

    CpswProxy_freeTxCh(obj->hCpswProxy,
                       obj->hCpsw,
                       obj->coreKey,
                       obj->txChNum);
}

void CpswApp_openRxFlow(CpswApp_Obj *obj,
                        bool useDefaultFlow,
                        CpswDma_OpenRxFlowPrms *pRxFlowPrms)
{
    CpswProxy_allocRxFlow(obj->hCpswProxy,
                          obj->hCpsw,
                          obj->coreKey,
                          &obj->rxStartFlowIdx,
                          &obj->rxFlowIdx);

    pRxFlowPrms->startIdx = obj->rxStartFlowIdx;
    pRxFlowPrms->flowIdx  = obj->rxFlowIdx;

    obj->hRxFlow = CpswDma_openRxFlow(pRxFlowPrms);
    CpswApp_assert(obj->hRxFlow != NULL);

    CpswProxy_addHostPortEntry(obj->hCpswProxy, obj->hCpsw, obj->coreKey, obj->hostMacAddr);
    if (useDefaultFlow)
    {
        CpswProxy_registerDefaultRxFlow(obj->hCpswProxy,
                                        obj->hCpsw,
                                        obj->coreKey,
                                        obj->rxStartFlowIdx,
                                        obj->rxFlowIdx);
    }
    else
    {
        CpswProxy_registerDstMacRxFlow(obj->hCpswProxy,
                                       obj->hCpsw,
                                       obj->coreKey,
                                       obj->rxStartFlowIdx,
                                       obj->rxFlowIdx,
                                       obj->hostMacAddr);
    }
}

void CpswApp_closeRxFlow(CpswApp_Obj *obj,
                         bool useDefaultFlow,
                         CpswDma_PktInfoQ *pFqPktInfoQ,
                         CpswDma_PktInfoQ *pCqPktInfoQ)
{
    int32_t status = CPSW_SOK;

    CpswUtils_initQ(pFqPktInfoQ);
    CpswUtils_initQ(pCqPktInfoQ);

    CpswDma_disableRxEvent(obj->hRxFlow);

    if (useDefaultFlow)
    {
        CpswProxy_unregisterDefaultRxFlow(obj->hCpswProxy,
                                          obj->hCpsw,
                                          obj->coreKey,
                                          obj->rxStartFlowIdx,
                                          obj->rxFlowIdx);
    }
    else
    {
        CpswProxy_unregisterDstMacRxFlow(obj->hCpswProxy,
                                         obj->hCpsw,
                                         obj->coreKey,
                                         obj->rxStartFlowIdx,
                                         obj->rxFlowIdx,
                                         obj->hostMacAddr);
    }

    CpswProxy_delAddrEntry(obj->hCpswProxy, obj->hCpsw, obj->coreKey, obj->hostMacAddr);

    status = CpswDma_closeRxFlow(obj->hRxFlow,
                                 pFqPktInfoQ,
                                 pCqPktInfoQ);
    CpswApp_assert(status == CPSW_SOK);

    CpswProxy_freeMac(obj->hCpswProxy,
                      obj->hCpsw,
                      obj->coreKey,
                      obj->hostMacAddr);
    CpswProxy_freeRxFlow(obj->hCpswProxy,
                         obj->hCpsw,
                         obj->coreKey,
                         obj->rxFlowIdx);
}
#else
    error "Unsupported!"
#endif


static void CpswApp_printDmaDescStats(CpswDma_DmaDescStats *descstats)
{
    // TODO when DMA stats are enabled.
}

static void CpswApp_printCbStats(CpswDma_CbFxnStats *cbStats)
{
    uint32_t i;

    CpswApp_printStatsNonZero("Data Notify Count          = %llu", cbStats->dataNotifyCnt);
    CpswApp_printStatsNonZero("Zero Notify Count          = %llu", cbStats->zeroNotifyCnt);
    CpswApp_printStatsNonZero("Total Packets Count        = %llu", cbStats->totalPktCnt);
    CpswApp_printStatsNonZero("Total Cycles Count         = %llu", cbStats->totalCycleCnt);
    CpswApp_printStatsNonZero("Packets per Notify Max     = %llu", cbStats->pktsPerNotifyMax);
    for (i = 0U; i < CPSW_DMA_STATS_HISTORY_CNT; i++)
    {
        CpswApp_printStatsWithIdxNonZero("Packets per Notify[%d] = %llu", i, cbStats->pktsPerNotify[i]);
    }

    CpswApp_printStatsNonZero("Cycles per Notify Max      = %llu", cbStats->cycleCntPerNotifyMax);
    for (i = 0U; i < CPSW_DMA_STATS_HISTORY_CNT; i++)
    {
        CpswApp_printStatsWithIdxNonZero("Cycles per Notify[%d]  = %llu", i, cbStats->cycleCntPerNotify[i]);
    }

    CpswApp_printStatsNonZero("Cycles per Packet Max      = %llu", cbStats->cycleCntPerPktMax);
    for (i = 0U; i < CPSW_DMA_STATS_HISTORY_CNT; i++)
    {
        CpswApp_printStatsWithIdxNonZero("Cycles per Packet[%d]  = %llu", i, cbStats->cycleCntPerPkt[i]);
    }

    for (i = 0U; i < CPSW_DMA_STATS_HISTORY_CNT; i++)
    {
        CpswApp_printStatsWithIdxNonZero("Ready Desc Q Count[%d] = %llu", i, cbStats->readyDmaDescQCnt[i]);
    }
}

static void CpswApp_printRxFlowStats(CpswDma_RxFlowStats *stats)
{
    CpswApp_print(" RX Flow Statistics");
    CpswApp_print("-----------------------------------------");
    CpswApp_printCbStats(&stats->submitPktStats);
    CpswApp_printCbStats(&stats->retrievePktStats);
    CpswApp_printDmaDescStats(&stats->dmaDescStats);

    CpswApp_printStatsNonZero("RX Submit Packet EnQ count           = %llu", stats->rxSubmitPktEnq);
    CpswApp_printStatsNonZero("RX Submit Packet Underflow           = %llu", stats->rxSubmitPktUnderFlowCnt);
    CpswApp_printStatsNonZero("RX Submit Packet DeQ count           = %llu", stats->rxRetrievePktDq);
}

static void CpswApp_printTxChStats(CpswDma_TxChStats *stats)
{
    CpswApp_print(" RX Flow Statistics");
    CpswApp_print("-----------------------------------------");
    CpswApp_printCbStats(&stats->submitPktStats);
    CpswApp_printCbStats(&stats->retrievePktStats);
    CpswApp_printDmaDescStats(&stats->dmaDescStats);

    CpswApp_printStatsNonZero("TX Submit Packet EnQ count           = %llu", stats->txSubmitPktEnq);
    CpswApp_printStatsNonZero("TX Submit Packet Underflow           = %llu", stats->txSubmitPktOverFlowCnt);
    CpswApp_printStatsNonZero("TX Submit Packet DeQ count           = %llu", stats->txRetrievePktDq);
}

int32_t CpswApp_showRxFlowStats(CpswDma_RxFlowHandle hRxFlow)
{
    int32_t retVal = CPSW_SOK;

    CpswDma_RxFlowStats rxFlowStats;

    retVal = CpswDma_getRxFlowStats(hRxFlow, &rxFlowStats);
    if (CPSW_ENOTSUPPORTED != retVal)
    {
        CpswApp_printRxFlowStats(&rxFlowStats);
        CpswDma_resetRxFlowStats(hRxFlow);
    }

    return retVal;
}

int32_t CpswApp_showTxChStats(CpswDma_TxChHandle hTxCh)
{
    int32_t retVal = CPSW_SOK;

    CpswDma_TxChStats txChStats;

    retVal = CpswDma_getTxChStats(hTxCh, &txChStats);
    if (CPSW_ENOTSUPPORTED != retVal)
    {
        CpswApp_printTxChStats(&txChStats);
        CpswDma_resetTxChStats(hTxCh);
    }

    return retVal;
}

void CpswApp_setNoPhyCfgRgmii(CpswMacPort_Interface *interface,
                              CpswPhy_Config *phyConfig)
{
    phyConfig->phyAddr      = CPSW_PHY_INVALID_PHYADDR;
    interface->layerType    = CPSW_MAC_LAYER_GMII;
    interface->sublayerType = CPSW_MAC_SUBLAYER_REDUCED;
    interface->variantType  = CPSW_MAC_VARIANT_FORCED;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/**
 * \brief  Configures the CPSW2G Subsytem for RGMII mode
 *
 * \param  portNum [IN]    EMAC port number
 * \param  mode    [IN]    Mode selection for the specified port number
 *                         00 - GMII
 *                         01 - RMII
 *                         10 - RGMII
 *                         11 - SGMII
 *
 * \return  CPSW_SOK in case of success or appropriate error code
 */
int32_t CpswAppBoard_cpsw2gMacModeConfig(uint32_t portNum, uint8_t mode)
{
    uint32_t status;
    uintptr_t ethModeCtrl;
    uint32_t regData;

    ethModeCtrl = (uintptr_t)mmap_device_memory(0,
        sizeof(uintptr_t), PROT_READ|PROT_WRITE|PROT_NOCACHE, 0,
        (uint64_t)CSL_MCU_CTRL_MMR0_CFG0_BASE + CSL_MCU_CTRL_MMR_CFG0_MCU_ENET_CTRL);

    regData = CSL_REG32_RD(ethModeCtrl);
    regData = mode;
    if (RGMII == mode)
    {
        regData |= (RGMII_ID_DISABLE_MASK);
    }

    CSL_REG32_WR(ethModeCtrl , regData);
    status = CSL_REG32_RD(ethModeCtrl);
    if (status != regData)
    {
        CpswApp_print("%s:%d Error: Failed to write MCU_ENET_CTRL",
            __FUNCTION__, __LINE__);
        status =  CPSW_EFAIL;
    }
    else
    {
        CpswApp_print("Success - write to MCU_ENET_CTRL - 0x%x", status);
        status = CPSW_SOK;
    }

    munmap_device_memory((void *)ethModeCtrl,sizeof(uintptr_t));
    return status;
}

uint32_t CpswAppBoard_getPhyAddr(Cpsw_Type cpswType,
                                 Cpsw_MacPort portNum)
{
    uint32_t phyAddr = CPSW_PHY_INVALID_PHYADDR;

    switch (cpswType)
    {
        case CPSW_2G:
            phyAddr = 0x00U;
            break;

        case CPSW_9G:
            switch (portNum)
            {
                case CPSW_MAC_PORT_0:
                    phyAddr = 0x0CU;
                    break;

                case CPSW_MAC_PORT_1:
                    phyAddr = 0x0FU;
                    break;

                case CPSW_MAC_PORT_2:
                    phyAddr = 0x00U;
                    break;

                case CPSW_MAC_PORT_3:
                    phyAddr = 0x03U;
                    break;

                case CPSW_MAC_PORT_7:
                    phyAddr = 0x17U;
                    break;

                default:
                    CpswApp_assert(false);
                    break;
            }

            break;

        default:
            CpswApp_assert(false);
    }

    return phyAddr;
}


void CpswAppBoard_setPhyConfig(Cpsw_Type cpswType,
                               Cpsw_MacPort portNum,
                               CpswMacPort_Interface *interface,
                               CpswPhy_Config *phyConfig)
{
    Dp83867_Config extendedConfig;

    Dp83867_initParams(&extendedConfig);
    Cpsw_initPhyParams(phyConfig);
    phyConfig->phyAddr = CpswAppBoard_getPhyAddr(cpswType, portNum);

    switch (cpswType)
    {
        case CPSW_2G:
            CpswApp_assert(portNum == CPSW_MAC_PORT_0);

            /* DP83867 specific configuration */
            extendedConfig.txFifoDepth = 4U;

            extendedConfig.txClkShiftEnable     = true;
            extendedConfig.rxClkShiftEnable     = true;
            extendedConfig.impedanceInMilliOhms = 35000; /* 35 ohms */
            extendedConfig.gpio0Mode            = DP83867_GPIO0_LED3;
            extendedConfig.ledMode[1]           = DP83867_LED_LINKED_100BTX;
            extendedConfig.ledMode[2]           = DP83867_LED_RXTXACT;
            extendedConfig.ledMode[3]           = DP83867_LED_LINKED_1000BT;

            /* RGMII interface type */
            interface->layerType    = CPSW_MAC_LAYER_GMII;
            interface->sublayerType = CPSW_MAC_SUBLAYER_REDUCED;
            interface->variantType  = CPSW_MAC_VARIANT_NONE;
            break;

        case CPSW_9G:
            /* DP83867 specific configuration */
            extendedConfig.txDelayInPs          = 2750U; /* 2.75 ns */
            extendedConfig.rxDelayInPs          = 2500U; /* 2.50 ns */
            extendedConfig.txFifoDepth          = 4U;
            extendedConfig.txClkShiftEnable     = true;
            extendedConfig.rxClkShiftEnable     = true;
            extendedConfig.impedanceInMilliOhms = 35000; /* 35 ohms */
            extendedConfig.gpio0Mode            = DP83867_GPIO0_LED3;
            extendedConfig.ledMode[1]           = DP83867_LED_LINKED_100BTX;
            extendedConfig.ledMode[2]           = DP83867_LED_RXTXACT;
            extendedConfig.ledMode[3]           = DP83867_LED_LINKED_1000BT;

            /* RGMII interface type */
            interface->layerType    = CPSW_MAC_LAYER_GMII;
            interface->sublayerType = CPSW_MAC_SUBLAYER_REDUCED;
            interface->variantType  = CPSW_MAC_VARIANT_NONE;
            break;

        default:
            break;
    }

    Cpsw_setPhyExtendedParams(phyConfig,
                              &extendedConfig,
                              sizeof(extendedConfig));
}
