/*
 *  Copyright (c) Texas Instruments Incorporated 2018-21
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *  \file     cpsw_lld_if.c
 *
 *  \brief    CPSW interface private source file.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <hw/inout.h>
#include <stdint.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/mman.h>
#include <hw/nicinfo.h>
#include <io-pkt/iopkt_driver.h>
#include <sys/io-pkt.h>

#include <ti/osal/osal.h>
#include <ti/drv/uart/UART_stdio.h>
#include <ti/drv/udma/udma.h>

#include <ti/drv/cpsw/cpsw.h>

#if defined (CPSW9G)
#include <ti/drv/ipc/ipc.h>
#include <ti/csl/cslr_gtc.h>
#endif /* defined (CPSW9G)*/

#include "j7_cpsw.h"
#include "cpsw_app.h"
#include "cpsw_appmem.h"
#include "cpsw_lld_if.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#if defined (CPSW9G)
  #define IPC_RPMESSAGE_OBJ_SIZE  256

  #define RPMSG_DATA_SIZE         (256*512 + IPC_RPMESSAGE_OBJ_SIZE)
  #define VQ_BUF_SIZE             2048
  #define REMOTE_DEVICE_ENDPT     (26)

  #define VRING_BASE_ADDRESS      0xAA000000

  #define CORE_IN_TEST            9

  #define CPSW_REMOTE_APP_GTC_PUSHEVT_BIT_SEL   (28U)
  #define CPSW_REMOTE_APP_CPTS_HW_PUSH_NUM      (2U)
#endif /* defined (CPSW9G)*/

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* CPSW configuration */
CpswApp_Obj gCpswAppObj;

/* SMMU enable flag.. referred by lld driver */
uint32_t gSMMU = 0;
uint32_t gSMMU_virtid = 0;

/* Cache Op flag */
uint32_t gCache_ops = 0;

/* Cache Op stucture */
extern struct cache_ctrl cachectl;

/* devnp driver object */
extern struct cpsw_dev *g_cpsw;

#if defined (CPSW9G)
uint8_t  g_taskStackBuf[IPC_TASK_STACKSIZE];

uint8_t  gCntrlBuf[RPMSG_DATA_SIZE] __attribute__ ((aligned (8)));
uint8_t  sysVqBuf[VQ_BUF_SIZE]  __attribute__ ((aligned (8)));
uint8_t *pCntrlBuf = gCntrlBuf;
uint8_t *pTaskBuf = g_taskStackBuf;
uint8_t *pSysVqBuf = sysVqBuf;

uint32_t selfProcId = IPC_MPU1_0;
uint32_t remoteProc[] =
{
    IPC_MCU1_0, IPC_MCU1_1, IPC_MCU2_0, IPC_MCU2_1, IPC_MCU3_0, IPC_MCU3_1, IPC_C66X_1, IPC_C66X_2, IPC_C7X_1
};

uint32_t *pRemoteProcArray = remoteProc;
uint32_t  gNumRemoteProc = sizeof(remoteProc)/sizeof(uint32_t);
uint32_t remoteProcNum[IPC_MAX_PROCS];
uint32_t rpmsgDataSize = RPMSG_DATA_SIZE;
#endif /* defined (CPSW9G) */

CpswDma_PktInfo gPktInfoMem[CPSW_MEM_TX_PKTS_MAX + CPSW_MEM_RX_PKTS_MAX];

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */
#if defined (CPSW9G)
void appLogPrintf(const char *format, ...)
{
    char printBuffer[256];
    va_list arguments;

    if (256 < strlen(format))
    {
        assert(false);
    }

    /* Start the varargs processing */
    va_start(arguments, format);
    vsnprintf(printBuffer, sizeof(printBuffer), format, arguments);

    slogf(_SLOGC_NETWORK, _SLOG_INFO, printBuffer);

    /* End the varargs processing */
    va_end(arguments);
}

static void printDevInfo(struct rpmsg_kdrv_ethswitch_device_data *ethDevData)
{
    char *tf[] = {"false", "true"};

    CpswApp_print("ETHFW Version:%2d.%2d.%2d\n",
                  ethDevData->fw_ver.major,
                  ethDevData->fw_ver.minor,
                  ethDevData->fw_ver.rev);
    CpswApp_print("ETHFW Build Date (YYYY/MMM/DD):%c%c%c%c/%c%c%c/%c%c\n",
                  ethDevData->fw_ver.year[0], ethDevData->fw_ver.year[1], ethDevData->fw_ver.year[2], ethDevData->fw_ver.year[3],
                  ethDevData->fw_ver.month[0], ethDevData->fw_ver.month[1], ethDevData->fw_ver.month[2],
                  ethDevData->fw_ver.date[0], ethDevData->fw_ver.date[1]);
    CpswApp_print("ETHFW Commit SHA:%c%c%c%c%c%c%c%c\n",
                  ethDevData->fw_ver.commit_hash[0],
                  ethDevData->fw_ver.commit_hash[1],
                  ethDevData->fw_ver.commit_hash[2],
                  ethDevData->fw_ver.commit_hash[3],
                  ethDevData->fw_ver.commit_hash[4],
                  ethDevData->fw_ver.commit_hash[5],
                  ethDevData->fw_ver.commit_hash[6],
                  ethDevData->fw_ver.commit_hash[7]);
    CpswApp_print("ETHFW PermissionFlag:0x%x, UART Connected:%s,UART Id:%d",
                  ethDevData->permission_flags,
                  tf[ethDevData->uart_connected],
                  ethDevData->uart_id);
}
static CpswProxy_Handle CpswRemoteApp_initCpswProxy(void)
{
    CpswProxy_Config proxyConfig;
    CpswProxy_Handle hProxy;

    strncpy(proxyConfig.device_name, ETHREMOTEDEVICE_DEVICE_NAME_MPU_1_0, (sizeof(proxyConfig.device_name) - 1));
    proxyConfig.device_name[(sizeof(proxyConfig.device_name) - 1)] = 0;
    proxyConfig.deviceDataNotifyCb = &printDevInfo;
    proxyConfig.masterCoreId = IPC_MCU2_0;
    proxyConfig.rpmsgEndPointId = REMOTE_DEVICE_ENDPT;

    hProxy = CpswProxy_init(&proxyConfig);

    return hProxy;
}

static uint64_t CpswApp_getLocalTime(uint64_t baseAddress)
{
    uint64_t gtcTimeLo = 0UL, gtcTimeHi = 0UL;
    uint64_t gtcTime = 0UL;

    gtcTimeLo = (uint64_t) CSL_REG32_RD(baseAddress + CSL_GTC_CFG1_CNTCV_LO);
    gtcTimeHi = (uint64_t) CSL_REG32_RD(baseAddress + CSL_GTC_CFG1_CNTCV_HI);
    gtcTime = (gtcTimeHi << 32U) | gtcTimeLo;

    return gtcTime;
}

static uint64_t CpswApp_getSynchronizedTime(CpswApp_SyncTimerObj *hSyncTimerObj)
{
    uint64_t gtcTime = 0UL, synchronizedTime = 0UL;

    /* Get GTC time */
    gtcTime = CpswApp_getLocalTime((uint64_t) hSyncTimerObj->GTC0_GTC_CFG1_Base);

    /* Compute synchronized time from GTC time */
    synchronizedTime = (uint64_t)((hSyncTimerObj->rate * (double)gtcTime) + hSyncTimerObj->offset);

    return synchronizedTime;
}

static void CpswApp_calcSyncTimeParams(CpswCpts_HwPush hwPushNum, uint64_t syncTime)
{
    uint64_t gtcTime = 0U;
    CpswApp_SyncTimerObj *hSyncTimerObj = (CpswApp_SyncTimerObj*)&gCpswAppObj.syncTimerObj;

    if (hwPushNum == CPSW_REMOTE_APP_CPTS_HW_PUSH_NUM)
    {
        if(hSyncTimerObj->prevLocalTime == 0U)
        {
            /* Set GTC time */
            gtcTime = CpswApp_getLocalTime((uint64_t) gCpswAppObj.syncTimerObj.GTC0_GTC_CFG1_Base) &
                                            (0xFFFFFFFFFFFFFFFFUL << CPSW_REMOTE_APP_GTC_PUSHEVT_BIT_SEL);
        }
        else
        {
            /* Increment GTC time used for computation based on selected bit for event */
            gtcTime = hSyncTimerObj->prevLocalTime + (1UL << (CPSW_REMOTE_APP_GTC_PUSHEVT_BIT_SEL+1));
        }

        if ((hSyncTimerObj->prevLocalTime != 0U) &&
            (hSyncTimerObj->prevCptsTime != 0U))
        {
            /* Logic:
             *  T1, T2 - Previous & Current CPTS time
             *  t1, t2 - Previous & Current local GTC time
             *  rate = (T2-T1) / (t2-t1)
             *  offset = T2 - rate*t2
             */
            hSyncTimerObj->rate = (double)((double)(syncTime - hSyncTimerObj->prevCptsTime) /
                                          (double)(gtcTime - hSyncTimerObj->prevLocalTime));
            hSyncTimerObj->offset = (double) syncTime - (hSyncTimerObj->rate * (double) gtcTime);
        }

        hSyncTimerObj->prevLocalTime = gtcTime;
        hSyncTimerObj->prevCptsTime = syncTime;
    }
}

#endif /* defined (CPSW9G) */

void CpswApp_InitObj()
{
#if defined (CPSW2G)

    memset(&gCpswAppObj, 0, sizeof(CpswApp_Obj));

    gCpswAppObj.cpswType = CPSW_2G;
    gCpswAppObj.portNum = CPSW_MAC_PORT_0;
    gCpswAppObj.coreId = IPC_MPU1_0;
    gCpswAppObj.preferredUdmaChannel = UDMA_DMA_CH_ANY;
    gCpswAppObj.numTxPkts = CPSW_MEM_TX_PKTS_DEFAULT;
    gCpswAppObj.numRxPkts = CPSW_MEM_RX_PKTS_DEFAULT;
    gCache_ops = 0;
    gSMMU = 0;

    CpswAppBoard_cpsw2gMacModeConfig(gCpswAppObj.portNum, RGMII);

    CpswApp_selectCptsClock(gCpswAppObj.cpswType, CPSWAPP_CPTS_CLKSEL_MAIN_SYSCLK0);

#elif defined (CPSW9G)

    uint32_t         numProc = gNumRemoteProc;
    Ipc_VirtIoParams vqParam;

    memset(&gCpswAppObj, 0, sizeof(CpswApp_Obj));

    gCpswAppObj.portNum = CPSW_MAC_PORT_0;
    gCpswAppObj.coreId = IPC_MPU1_0;
    gCpswAppObj.cpswType = CPSW_9G;
    gCpswAppObj.preferredUdmaChannel = UDMA_DMA_CH_ANY;
    gCpswAppObj.numTxPkts = CPSW_MEM_TX_PKTS_DEFAULT;
    gCpswAppObj.numRxPkts = CPSW_MEM_RX_PKTS_DEFAULT;
    gCache_ops= 0;
    gSMMU = 0;

    /* Step1 : Initialize the multiproc */
    Ipc_mpSetConfig(selfProcId, numProc, pRemoteProcArray);

    CpswApp_print("CPSW9G Client (core : %s) .....\r\n",
         Ipc_mpGetSelfName());

    Ipc_init(NULL);

    /* Step2 : Initialize Virtio */
    vqParam.vqObjBaseAddr = (void*)pSysVqBuf;
    vqParam.vqBufSize     = numProc * Ipc_getVqObjMemoryRequiredPerCore();
    vqParam.vringBaseAddr = (void*)VRING_BASE_ADDRESS;
    vqParam.vringBufSize  = IPC_VRING_BUFFER_SIZE;
    vqParam.timeoutCnt    = 100;  /* Wait for counts */
    Ipc_initVirtIO(&vqParam);

    /* Step 3: Initialize RPMessage */
    RPMessage_Params cntrlParam;

    /* Initialize the param */
    RPMessageParams_init(&cntrlParam);

    /* Set memory for HeapMemory for control task */
    cntrlParam.buf         = pCntrlBuf;
    cntrlParam.bufSize     = rpmsgDataSize;
    cntrlParam.stackBuffer = pTaskBuf;
    cntrlParam.stackSize   = IPC_TASK_STACKSIZE;
    RPMessage_init(&cntrlParam);

    /* Step 5: Start Cpsw Proxy */
    gCpswAppObj.hCpswProxy = CpswRemoteApp_initCpswProxy();
    gCpswAppObj.cpswType = CPSW_9G;
    gCpswAppObj.portNum = CPSW_MAC_PORT_0;
    gCpswAppObj.coreId = IPC_MPU1_0;

    CpswApp_print("CpswProxy_start .....\r\n");
    CpswProxy_start(gCpswAppObj.hCpswProxy);

    CpswProxy_attach(gCpswAppObj.hCpswProxy,
                   gCpswAppObj.cpswType,
                   &gCpswAppObj.hCpsw,
                   &gCpswAppObj.coreKey,
                   &gCpswAppObj.rxMtu,
                   gCpswAppObj.txMtu);
    CpswProxy_allocMac(gCpswAppObj.hCpswProxy,
                       gCpswAppObj.hCpsw,
                       gCpswAppObj.coreKey,
                       gCpswAppObj.hostMacAddr);
#else
    #error "Unsupported!"
#endif

    return;
}

void CpswApp_SetPerferredDmaChNum(uint32_t preferred_udma_channel)
{
    gCpswAppObj.preferredUdmaChannel = preferred_udma_channel;
}

void CpswApp_EnableSmmu(uint32_t virt_id)
{
    CpswApp_print("%s:%d SMMU", __FUNCTION__, __LINE__);
    gSMMU = 1;
    gSMMU_virtid = virt_id;
}

void CpswApp_EnableCacheOps()
{
    CpswApp_print("%s:%d SMMU Cache Ops enabled", __FUNCTION__, __LINE__);
    gCache_ops = 1;
}

void CpswApp_SetDescriptorCount(uint32_t tx_descriptors_count,
                                uint32_t rx_descriptors_count)
{
    gCpswAppObj.numTxPkts = tx_descriptors_count;
    gCpswAppObj.numRxPkts = rx_descriptors_count;
}

int32_t CpswApp_GetMacAddr(uint8_t *macAddr)
{
#if defined (CPSW2G)
    CpswSoc_getEFusedMacAddr(macAddr);
#elif defined (CPSW9G)
    memcpy(macAddr, gCpswAppObj.hostMacAddr, ETH_MAC_ADDR_LEN);
    CpswApp_print("MAC Address: %x:%x:%x:%x:%x:%x\n", macAddr[0],macAddr[1],
                                               macAddr[2],macAddr[3],
                                               macAddr[4],macAddr[5]);
#else
    #error "Unsupported!"
#endif
    return CPSW_SOK;
}

static void CpswCptsEvent_setDefaultPortEventPrms(
    CpswMacPort_TsEventConfig *tsPortEventCfg)
{
    tsPortEventCfg->commonPortIpConfig.enableTTLNonzero = true;
    tsPortEventCfg->commonPortIpConfig.enableTsIp107    = true;
    tsPortEventCfg->commonPortIpConfig.enableTsIp129    = true;
    tsPortEventCfg->commonPortIpConfig.enableTsIp130    = true;
    tsPortEventCfg->commonPortIpConfig.enableTsIp131    = true;
    tsPortEventCfg->commonPortIpConfig.enableTsIp132    = true;
    tsPortEventCfg->commonPortIpConfig.enableTsPort319  = true;
    tsPortEventCfg->commonPortIpConfig.enableTsPort320  = true;
    tsPortEventCfg->commonPortIpConfig.enableUnicast    = false;
    tsPortEventCfg->domainOffset                        = 4U;
    tsPortEventCfg->enableLType2                        = false;
    tsPortEventCfg->enableRxAnnexD                      = true;
    tsPortEventCfg->enableRxAnnexE                      = true;
    tsPortEventCfg->enableRxAnnexF                      = true;
    tsPortEventCfg->enableTxAnnexD                      = true;
    tsPortEventCfg->enableTxAnnexE                      = true;
    tsPortEventCfg->enableTxAnnexF                      = true;
    tsPortEventCfg->enableTxHostTs                      = true;
    tsPortEventCfg->mcastType                           = 0U;
    tsPortEventCfg->messageType                         = 0xFFFFU;
    tsPortEventCfg->rxVlanType                          = CPSW_MACPORT_VLAN_TYPE_SINGLE_TAG;
    tsPortEventCfg->seqIdOffset                         = 30U;
    tsPortEventCfg->txVlanType                          = CPSW_MACPORT_VLAN_TYPE_SINGLE_TAG;
    tsPortEventCfg->vlanLType1                          = 0x8100;
    tsPortEventCfg->vlanLType2                          = 0U;
}

int32_t CpswApp_PtpInit()
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    CpswCpts_RegisterStackInArgs registerStackInArgs;
    CpswMacPort_EnableTsEventInArgs enableTsEventInArgs;

    CpswApp_print("%s:%d PTP", __FUNCTION__, __LINE__);

    /* CPTS */
    if (status == CPSW_SOK)
    {
        registerStackInArgs.pEventNotifyCb = CpswApp_EventNotifyCallback;
        registerStackInArgs.hEventNotifyCbArg = &gCpswAppObj;

        CPSW_IOCTL_SET_IN_ARGS(&prms, &registerStackInArgs);
        status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                    gCpswAppObj.coreId,
                    CPSW_CPTS_IOCTL_REGISTER_STACK,
                    &prms);
        if (status != CPSW_SOK)
        {
            CpswApp_print("%s:%d Error: failed CPSW_CPTS_IOCTL_REGISTER_STACK: %d",
                __FUNCTION__, __LINE__, status);
        }
        else {
            CpswApp_print("%s:%d CPSW_CPTS_IOCTL_REGISTER_STACK is sucessful",
                __FUNCTION__, __LINE__);
        }
    }

    if (status == CPSW_SOK)
    {
        /* Enable PTP over Bare Ethernet - Annex F (IEEE802.3) */
        enableTsEventInArgs.portNum = CPSW_MAC_PORT_FIRST;
        CpswCptsEvent_setDefaultPortEventPrms(&enableTsEventInArgs.portCfg);

        CPSW_IOCTL_SET_IN_ARGS(&prms, &enableTsEventInArgs);
        status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                            gCpswAppObj.coreId,
                            CPSW_MACPORT_IOCTL_ENABLE_CPTS_EVENT,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswApp_print("%s:%d Error: failed CPSW_MACPORT_IOCTL_ENABLE_CPTS_EVENT: %d",
                __FUNCTION__, __LINE__, status);
        }
    }

    return status;
}

void CpswApp_EventNotifyCallback(void *hEventNotifyCallbackArg,
                                 CpswCpts_Event *eventInfo)
{
    /* Do Nothing */
    if (g_cpsw->cfg.verbose & DEBUG_PTP_APP) {
        CpswApp_print("Entered cpts Callback eventtype=%d, ts=%lld, msgType=%d, seqId=%d ",
            eventInfo->eventType, eventInfo->tsVal, eventInfo->msgType, eventInfo->seqId);
    }
}

int32_t CpswApp_GetTime(uint64_t *ts)
{
    int32_t status = CPSW_SOK;
#if defined (CPSW2G)
    Cpsw_IoctlPrms prms;

    if (gCpswAppObj.hCpsw == NULL)
    {
        return CPSW_ENOTFOUND;
    }

    if (ts == NULL)
    {
        return CPSW_EBADARGS;
    }

    *ts = 0;
    CPSW_IOCTL_SET_OUT_ARGS(&prms, ts);
    status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                        gCpswAppObj.coreId,
                        CPSW_CPTS_IOCTL_PUSH_TIMESTAMP_EVENT,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: failed CPSW_CPTS_IOCTL_PUSH_TIMESTAMP_EVENT: %d",
            __FUNCTION__, __LINE__, status);
    }
    else {
        if (g_cpsw->cfg.verbose & DEBUG_PTP_APP) {
            CpswApp_print("%s:%d The timestamp is: %lld", __FUNCTION__, __LINE__, *ts);
        }
    }
#elif defined (CPSW9G)
    if (ts == NULL)
    {
        return CPSW_EBADARGS;
    }

    *ts = CpswApp_getSynchronizedTime(&gCpswAppObj.syncTimerObj);
#endif
    return status;
}

int32_t CpswApp_SetTime(uint64_t ts)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;

    if (gCpswAppObj.hCpsw == NULL)
    {
        return CPSW_ENOTFOUND;
    }

    if (g_cpsw->cfg.verbose & DEBUG_PTP_APP) {
        CpswApp_print("%s:%d set timestamp is: %lld", __FUNCTION__, __LINE__, ts);
    }

    CPSW_IOCTL_SET_IN_ARGS(&prms, &ts);
    status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                        gCpswAppObj.coreId,
                        CPSW_CPTS_IOCTL_LOAD_TIMESTAMP,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: failed CPSW_CPTS_IOCTL_LOAD_TIMESTAMP: %d",
            __FUNCTION__, __LINE__, status);
    }

    return status;
}

int32_t CpswApp_GetTimestamp(uint8_t msgType,
                             uint16_t sequenceId,
                             uint8_t tx,
                             uint64_t *ts)
{
    int32_t status = CPSW_SOK;
    CpswCpts_LookUpEventInArgs lookUpEventInArgs;
    CpswCpts_Event lookUpEventOutArgs;
    Cpsw_IoctlPrms prms;

    if (gCpswAppObj.hCpsw == NULL)
    {
        return CPSW_ENOTFOUND;
    }

    if (ts == NULL)
    {
        return CPSW_EBADARGS;
    }

    *ts = 0;
    if (tx)
    {
        lookUpEventInArgs.eventType = CPSW_CPTS_EVENTTYPE_ETH_TRANSMIT;
    }
    else
    {
        lookUpEventInArgs.eventType = CPSW_CPTS_EVENTTYPE_ETH_RECEIVE;
    }
    lookUpEventInArgs.msgType = msgType;
    lookUpEventInArgs.seqId = sequenceId;
    lookUpEventInArgs.portNum = gCpswAppObj.portNum;

    if (g_cpsw->cfg.verbose & DEBUG_PTP_APP) {
        CpswApp_print("%s:%d PTP_GET_TX_TIMESTAMP/PTP_GET_RX_TIMESTAMP",
                __FUNCTION__, __LINE__);
        CpswApp_print("%s:%d msg_type:%d, sequence_id:%d, tx:%d",
                __FUNCTION__, __LINE__, msgType, sequenceId, tx);
    }
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &lookUpEventInArgs, &lookUpEventOutArgs);

    status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                        gCpswAppObj.coreId,
                        CPSW_CPTS_IOCTL_LOOKUP_EVENT,
                        &prms);
    if (status != CPSW_SOK)
    {
        if (g_cpsw->cfg.verbose & DEBUG_PTP_APP) {
            CpswApp_print("%s:%d Error: failed, no retries, CPSW_CPTS_IOCTL_LOOKUP_EVENT: %d",
                __FUNCTION__, __LINE__, status);
        }
    }
    else
    {
        *ts = lookUpEventOutArgs.tsVal;
        if (g_cpsw->cfg.verbose & DEBUG_PTP_APP) {
            CpswApp_print("%s:%d The CPSW_CPTS_IOCTL_LOOKUP_EVENT timestamp is: %lld",
                __FUNCTION__, __LINE__, *ts);
        }
    }

    return status;
}

int32_t CpswApp_SetCompensation(int32_t nudge)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;

    if (gCpswAppObj.hCpsw == NULL)
    {
        return CPSW_ENOTFOUND;
    }

    if (g_cpsw->cfg.verbose & DEBUG_PTP_APP) {
        CpswApp_print("%s:%d nudge is: %d", __FUNCTION__, __LINE__, nudge);
    }

    if (nudge > 127 || nudge < -128)
    {
        uint64_t ts = 0;
        CpswApp_GetTime(&ts);
#if 0
        ts += nudge;
#else
        ts -= nudge;
#endif
        CpswApp_SetTime(ts);
        return CPSW_SOK;
    }

    CPSW_IOCTL_SET_IN_ARGS(&prms, &nudge);
    status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                        gCpswAppObj.coreId,
                        CPSW_CPTS_IOCTL_SET_TS_NUDGE,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: failed CPSW_CPTS_IOCTL_SET_TS_NUDGE: %d",
            __FUNCTION__, __LINE__, status);
    }

    return status;
}


int32_t CpswApp_Init(int mac_to_mac, int speed, uint8_t *currentMacAddr)
{
    int32_t status = CPSW_SOK;
#if defined (CPSW2G)
    Cpsw_IoctlPrms prms;
    CpswAle_SetPortStateInArgs setPortStateInArgs;

#if 0
    CpswAppUtils_enableClocks(gCpswAppObj.cpswType,
                              MAC_CONN_TYPE_RGMII_FORCE_1000_FULL);
#endif
#endif

    /* Open the CPSW */
    status = CpswApp_openCpsw(mac_to_mac, speed, currentMacAddr);
    if (status != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: Failed to open CPSW: %d",
            __FUNCTION__, __LINE__, status);
    }

    /* Open CPSW DMA driver */
    if (status == CPSW_SOK)
    {
        status = CpswApp_openDma();
        if (status != CPSW_SOK)
        {
            CpswApp_print("%s:%d Error: Failed to open CPSW DMA: %d",
                __FUNCTION__, __LINE__, status);
        }
    }

#if defined (CPSW2G)
    /* Enable host port */
    if (status == CPSW_SOK)
    {
        setPortStateInArgs.portNum   = CPSW_ALE_HOST_PORT_NUM;
        setPortStateInArgs.portState = CPSW_ALE_PORTSTATE_FORWARD;
        CPSW_IOCTL_SET_IN_ARGS(&prms, &setPortStateInArgs);
        prms.outArgs = NULL;
        status       = Cpsw_ioctl(gCpswAppObj.hCpsw,
                                  gCpswAppObj.coreId,
                                  CPSW_ALE_IOCTL_SET_PORT_STATE,
                                  &prms);
        if (status != CPSW_SOK)
        {
            CpswApp_print( "%s:%d Error: Failed CPSW_ALE_IOCTL_SET_PORT_STATE: %d",
                __FUNCTION__, __LINE__, status);
        }

        if (status == CPSW_SOK)
        {
            CPSW_IOCTL_SET_NO_ARGS(&prms);
            status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                                gCpswAppObj.coreId,
                                CPSW_HOSTPORT_IOCTL_ENABLE,
                                &prms);
            if (status != CPSW_SOK)
            {
                CpswApp_print("%s:%d Error: Failed to enable host port: %d",
                    __FUNCTION__, __LINE__, status);
            }
        }
    }
#endif
    return status;
}

int32_t CpswApp_InitPhy()
{
#if defined (CPSW2G)
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    int8_t         i;
    bool           alive;

    /* Show alive PHYs */
    if (status == CPSW_SOK)
    {
        for (i = 0U; i < CPSW_MDIO_MAX_PHY_CNT; i++)
        {
            CPSW_IOCTL_SET_INOUT_ARGS(&prms, &i, &alive);
            status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                                gCpswAppObj.coreId,
                                CPSW_MDIO_IOCTL_GET_ALIVE_STATUS,
                                &prms);
            if (status == CPSW_SOK)
            {
                if (alive == true)
                {
                    CpswApp_print("PHY %d is alive", i);
                }
            }
            else
            {
                CpswApp_print("%s:%d Error: Failed to get PHY %d alive status: %d",
                    __FUNCTION__, __LINE__, i, status);
            }
        }
    }
    return status;
#elif defined (CPSW9G)
    return CPSW_SOK;
#endif
}


int32_t CpswApp_CheckPortLinkUp(uint32_t timeout_in_sec)
{
    bool linked = false;
    int32_t timeout_cnt = (timeout_in_sec * 1000 / 100);

    do {
#if defined (CPSW2G)
        linked = CpswApp_isPortLinkUp(gCpswAppObj.hCpsw,
                                    gCpswAppObj.coreId,
                                    gCpswAppObj.portNum);
#elif defined (CPSW9G)
        linked = CpswProxy_isPhyLinked(gCpswAppObj.hCpswProxy,
                                    gCpswAppObj.hCpsw,
                                    gCpswAppObj.coreKey,
                                    gCpswAppObj.portNum);
#endif
        if ((linked) || (!timeout_cnt)) {
            break;
        }
        CpswApp_wait(100);
        Cpsw_periodicTick(gCpswAppObj.hCpsw);
        timeout_cnt--;
    } while (timeout_cnt);

    if (!linked) {
        return CPSW_EFAIL;
    }
    return CPSW_SOK;
}


int32_t CpswApp_Close()
{
#if defined (CPSW2G)
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;

    /* Retreive any pending packets */
    CpswApp_retrieveFreeTxPkts();

    /* Print CPSW statistics of all ports */
    if (status == CPSW_SOK)
    {
        CpswApp_showStats();

        CpswApp_print("rxReadyQ count = %d", CpswUtils_getQCount(&gCpswAppObj.rxReadyQ));
        CpswApp_print("freePktInfoQ count = %d", CpswUtils_getQCount(&gCpswAppObj.freePktInfoQ));
    }

    /* Disable host port */
    if (status == CPSW_SOK)
    {
        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                            gCpswAppObj.coreId,
                            CPSW_HOSTPORT_IOCTL_DISABLE,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswApp_print("%s:%d Error: Failed to disable host port: %d",
                __FUNCTION__, __LINE__, status);
        }
    }

    /* De-register ptp */
    if (status == CPSW_SOK)
    {
        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                            gCpswAppObj.coreId,
                            CPSW_CPTS_IOCTL_UNREGISTER_STACK,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswApp_print("%s:%d Error: Failed to de-register ptp: %d",
                __FUNCTION__, __LINE__, status);
        }
    }

    /* Print DMA statistics */
    if (status == CPSW_SOK)
    {
        /* Print RX flow DMA statistics */
        CpswApp_showRxFlowStats(gCpswAppObj.hRxFlow);
        /* Print TX channel DMA statistics */
        CpswApp_showTxChStats(gCpswAppObj.hTxCh);
    }

    /* Retreive any pending packets */
    CpswApp_retrieveFreeTxPkts();

    /* Close CPSW DMA driver */
    CpswApp_closeDma();
    /* Close CPSW */
    CpswApp_closeCpsw();
    /* Close UDMA driver */
    CpswApp_udmaclose(gCpswAppObj.hUdmaDrv);

    pthread_mutex_destroy(&gCpswAppObj.cpswMutex);

    gCpswAppObj.hCpsw = NULL;
    CpswApp_print("cpsw deinit done");

#if 0
    CpswAppUtils_disableClocks(gCpswAppObj.cpswType);
#endif

    return status;
#elif defined (CPSW9G)
    return CPSW_SOK;
#endif
}

void CpswApp_initAleConfig(CpswAle_Config *aleConfig)
{

    // Default ALE config
    aleConfig->modeFlags = (CPSW_ALE_CONFIG_MASK_ALE_MODULE_ENABLE);
    aleConfig->agingConfig.enableAutoAging           = TRUE;
    aleConfig->agingConfig.agingPeriodInMs           = 1000;
    aleConfig->nwSecCfg.enableVid0Mode               = TRUE;
    aleConfig->vlanConfig.aleVlanAwareMode           = FALSE;
    aleConfig->vlanConfig.cpswVlanAwareMode          = FALSE;
    aleConfig->vlanConfig.unknownUnregMcastFloodMask = CPSW_ALE_ALL_PORTS_MASK;
    aleConfig->vlanConfig.unknownRegMcastFloodMask   = CPSW_ALE_ALL_PORTS_MASK;
    aleConfig->vlanConfig.unknownVlanMemberListMask  = CPSW_ALE_ALL_PORTS_MASK;

    if (g_cpsw->promiscious)
    {
        aleConfig->modeFlags = (CPSW_ALE_CONFIG_MASK_ALE_BYPASS_ENABLE);
    }
}

uintptr_t CpswApp_disableAllIntr(void)
{
    pthread_mutex_lock(&gCpswAppObj.cpswMutex);
    return 0;
}

void CpswApp_restoreAllIntr(uintptr_t key)
{
    pthread_mutex_unlock(&gCpswAppObj.cpswMutex);
}

///////////////////////////////////////////////////////////////////////////////
/* Thread level ISR handler */
void *CpswApp_isr_thread (void *arg)
{
    CpswApp_hwi_info *hwi = (CpswApp_hwi_info *)arg;
    int             rcvid;
    struct _pulse   pulse;
    uintptr_t       interruptNum;

    while (1) {
        rcvid = MsgReceivePulse(hwi->chid, &pulse, sizeof(struct _pulse), NULL);
        if (rcvid != -1) {
            switch (pulse.code) {
                case _PULSE_CODE_MINAVAIL:
                    //CpswApp_print("%s: Interrupt %d received\n",__func__, pulse.value.sival_int);
                    interruptNum = pulse.value.sival_int;

                    /* Call the callback function */
                    hwi->isrFxn(hwi->arg);

                    HwiP_enableInterrupt(interruptNum);
                    break;
                default:
                    CpswApp_print("%s: Rx Unknown pulse %d received\n",__func__, pulse.code);
                    break;
            }
        }
        else {
            CpswApp_print("%s: MsgReceivePulse failed\n",__func__);
            delay(20);
        }
    }
    return NULL;
}
///////////////////////////////////////////////////////////////////////////////

void *CpswApp_registerIntr(CpswOsal_IsrFxn isrFxn,
                           uint32_t coreIntrNum,
                           uint32_t intrPriority,
                           void *arg)
{
    pthread_t      tid;
    pthread_attr_t thread_attr;
    struct sched_param  param;
    CpswApp_hwi_info  *hwi = NULL;
    char threadName[128];

    if (gCpswAppObj.currIntrCount > CPSWAPP_MAX_INTR_COUNT) {
        CpswApp_print("%s: MAXed out on the hwi structure\n", __FUNCTION__);
        CpswApp_assert(FALSE);
    }
    hwi = &gCpswAppObj.hwi[gCpswAppObj.currIntrCount];
    gCpswAppObj.currIntrCount++;

    hwi->chid = ChannelCreate( _NTO_CHF_DISCONNECT | _NTO_CHF_UNBLOCK);

    if (coreIntrNum == 890) {
        // ptp event interrupt
        intrPriority = 22;
    }
    else {
        intrPriority = 21;
    }

    pthread_attr_init(&thread_attr);
    pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_DETACHED);
    param.sched_priority = intrPriority;
    pthread_attr_setschedparam(&thread_attr, &param);

    if (pthread_create(&tid, &thread_attr, (void *)CpswApp_isr_thread, (void *)hwi) != EOK) {
        CpswApp_print("%s: Unable to create isr thread\n", __FUNCTION__);
        CpswApp_assert(FALSE);
    }
    sprintf(threadName, "IntrThread_%d",coreIntrNum);
    pthread_setname_np(tid, threadName);

    /* Store pointer to the hwi structure */
    hwi->isrFxn = isrFxn;
    hwi->coreIntrNum = coreIntrNum;
    hwi->intrPriority = intrPriority;
    hwi->arg = (uintptr_t)arg;

    /* Init the pulse for interrupt event */
    hwi->isr_event.sigev_notify = SIGEV_PULSE;
    hwi->isr_event.sigev_code = _PULSE_CODE_MINAVAIL;
    hwi->isr_event.sigev_coid = ConnectAttach(0, 0, hwi->chid, _NTO_SIDE_CHANNEL, 0);
    hwi->isr_event.sigev_priority = intrPriority;     /* service interrupts at a higher priority then client requests */
    hwi->isr_event.sigev_value.sival_int = coreIntrNum;

    /*
     * Attach interrupt handler (thread level)
     *
     * _NTO_INTR_FLAGS_TRK_MSK - Track mask/unmask
     * _NTO_INTR_FLAGS_NO_UNMASK - Start with interrupt masked
     */
    hwi->evtId = InterruptAttachEvent (coreIntrNum, &hwi->isr_event,  0 /*_NTO_INTR_FLAGS_NO_UNMASK*/);
    if(hwi->evtId == -1)
    {
        CpswApp_print("%s: InterruptAttachEvent failed\n",__FUNCTION__);
    }
    else
    {
        CpswApp_print("%s: InterruptAttachEvent succeed irq/%d coid/%d event/%p\n",__FUNCTION__, 
            coreIntrNum, hwi->isr_event.sigev_coid, hwi->evtId);
    }

    return ((void *)hwi);
}

void CpswApp_unRegisterIntr(void *hwiHandle)
{
    // do nothing
    return;
}

int32_t CpswApp_openCpsw(int mac_to_mac, int speed, uint8_t *currentMacAddr)
{
    Cpsw_Config cpswCfg;
    CpswOsal_Prms osalPrms;
    CpswUtils_Prms utilsPrms;
    int32_t status = CPSW_SOK;
#if defined (CPSW2G)
    Cpsw_IoctlPrms prms;
    uint32_t coreId;
    Cpsw_AttachCoreOutArgs attachCoreOutArgs;
#endif

    /* Set configuration parameters */
#if defined (CPSW9G)
    /* TODO: May be ignored for 9G */
#endif
    Cpsw_initParams(&cpswCfg);
    cpswCfg.vlanConfig.vlanAware          = false;
    cpswCfg.hostPortConfig.removeCrc      = true;
    cpswCfg.hostPortConfig.padShortPacket = true;
    cpswCfg.hostPortConfig.passCrcErrors  = true;
    CpswApp_initResourceConfig(gCpswAppObj.cpswType,
                               gCpswAppObj.coreId,
                               &cpswCfg.resourceConfig,
                               currentMacAddr);

    CpswApp_initAleConfig(&cpswCfg.aleConfig);

    /* Setting cpts refrence clock frequency */
    cpswCfg.cptsConfig.cptsRftClkFreq = CPSW_CPTS_RFTCLK_FREQ_500MHZ;

    if (gCpswAppObj.cpswType == CPSW_9G)
    {
        CpswApp_print("CPSW_9G on MAIN NAVSS");
    }
    else if (gCpswAppObj.cpswType == CPSW_2G)
    {
        CpswApp_print("CPSW_2G on MCU NAVSS");
    }

    // clear hwi strucure
    gCpswAppObj.currIntrCount = 0;
    memset(&gCpswAppObj.hwi, 0, sizeof(CpswApp_hwi_info) * CPSWAPP_MAX_INTR_COUNT);

    // set the mutex with resursive attribute set
    pthread_mutexattr_t mattr;
    pthread_mutexattr_init(&mattr);
    pthread_mutexattr_setrecursive(&mattr, PTHREAD_RECURSIVE_ENABLE);
    pthread_mutex_init(&gCpswAppObj.cpswMutex, &mattr);
    pthread_mutexattr_destroy (&mattr);

    cpswCfg.dmaConfig.rxChInitPrms.dmaPriority = UDMA_DEFAULT_RX_CH_DMA_PRIORITY;
    // Note: Using preferred channel
    cpswCfg.dmaConfig.rxChInitPrms.preferredChNum = gCpswAppObj.preferredUdmaChannel;

    /* App should open UDMA first as UDMA handle is needed to initialize
     * CPSW RX channel */
    gCpswAppObj.hUdmaDrv = CpswApp_udmaOpen(gCpswAppObj.cpswType, NULL);
    CpswApp_assert(NULL != gCpswAppObj.hUdmaDrv);

    cpswCfg.dmaConfig.hUdmaDrv = gCpswAppObj.hUdmaDrv;

    /* Initialize CPSW driver with default OSAL, utils */
    utilsPrms.printFxn     = (Cpsw_PrintFxnCb)CpswApp_print;
    utilsPrms.traceFxn     = (Cpsw_PrintFxnCb)CpswApp_print;
    utilsPrms.phyToVirtFxn = &CpswApp_phyToVirtFxn;
    utilsPrms.virtToPhyFxn = &CpswApp_virtToPhyFxn;

    Cpsw_initOsalPrms(&osalPrms);
    //Note: Over-ride the osal api to disableAllIntr/restoreAllIntr
    osalPrms.disableAllIntr = &CpswApp_disableAllIntr;
    osalPrms.restoreAllIntr = &CpswApp_restoreAllIntr;
    //Note: Over-ride the osal api to registerIntr/restoreAllIntr
    osalPrms.registerIntr = &CpswApp_registerIntr;
    osalPrms.unRegisterIntr = &CpswApp_unRegisterIntr;

    Cpsw_init(gCpswAppObj.cpswType, &osalPrms, &utilsPrms);

#if defined (CPSW2G)
    /* Open the CPSW driver */
    gCpswAppObj.hCpsw = Cpsw_open(gCpswAppObj.cpswType, &cpswCfg);
    if (gCpswAppObj.hCpsw == NULL)
    {
        CpswApp_print("%s:%d Error: Failed to open: %d",
            __FUNCTION__, __LINE__, status);
        status = CPSW_EFAIL;
    }

    if (status == CPSW_SOK)
    {
        /* Attach the core with RM */
        coreId = gCpswAppObj.coreId;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &coreId, &attachCoreOutArgs);
        status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                            gCpswAppObj.coreId,
                            CPSW_IOCTL_ATTACH_CORE,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswApp_print("%s:%d Error: failed CPSW_IOCTL_ATTACH_CORE: %d",
                __FUNCTION__, __LINE__, status);
        }
        else
        {
            gCpswAppObj.coreKey = attachCoreOutArgs.coreKey;
        }
    }
#elif defined (CPSW9G)
    /* CPSW9G proxy driver open & attach in cpsw_attach() */
    /* Open UDMA for CPSW NAVSS instance type */
    gCpswAppObj.dmaDataPathConfig.hUdmaDrv = gCpswAppObj.hUdmaDrv;
    gCpswAppObj.dma = CpswDma_initDataPath(gCpswAppObj.cpswType, &gCpswAppObj.dmaDataPathConfig);
    CpswApp_assert(NULL != gCpswAppObj.dma);
#endif

    if (status == CPSW_SOK)
    {
        /* appmem open should happen after Cpsw is opened as it uses CpswUtils_Q
         * functions */
        status = CpswAppMem_init();
        CpswApp_assert(CPSW_SOK == status);
    }

#if defined (CPSW2G)
    if (status == CPSW_SOK)
    {
        Cpsw_OpenPortLinkInArgs linkArgs =
        {
            .portNum = gCpswAppObj.portNum,
        };
        CpswMacPort_Config *macConfig      = &linkArgs.macConfig;
        CpswMacPort_LinkConfig *linkConfig = &linkArgs.linkConfig;
        CpswPhy_Config *phyConfig          = &linkArgs.phyConfig;
        CpswMacPort_Interface *interface   = &linkArgs.interface;

        Cpsw_initMacPortParams(macConfig);

        if (mac_to_mac == 1) {
            /* mac-to-mac configuration -> no phy */
            CpswApp_setNoPhyCfgRgmii(interface, phyConfig);
            if (speed == 100) {
                linkConfig->speed     = CPSW_SPEED_100MBIT;
            }
            else if (speed == 1000) {
                linkConfig->speed     = CPSW_SPEED_1GBIT;
            }
            linkConfig->duplexity = CPSW_DUPLEX_FULL;
        }
        else
        {
            CpswAppBoard_setPhyConfig(gCpswAppObj.cpswType,
                                       linkArgs.portNum,
                                       interface,
                                       phyConfig);

            interface->variantType  = CPSW_MAC_VARIANT_FORCED;
            linkConfig->speed     = CPSW_SPEED_AUTO;
            linkConfig->duplexity = CPSW_DUPLEX_AUTO;
        }

        CPSW_IOCTL_SET_IN_ARGS(&prms, &linkArgs);
        status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                            gCpswAppObj.coreId,
                            CPSW_IOCTL_OPEN_PORT_LINK,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswApp_print("%s:%d Error: failed to open MAC port: %d",
                __FUNCTION__, __LINE__, status);
        }
    }
#endif

    return status;
}

void CpswApp_closeCpsw(void)
{
    Cpsw_IoctlPrms prms;
    uint32_t coreKey;
    Cpsw_GenericPortLinkInArgs args =
    {
        .portNum = gCpswAppObj.portNum,
    };
    int32_t status;

    coreKey = gCpswAppObj.coreKey;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &coreKey);

    status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                        gCpswAppObj.coreId,
                        CPSW_IOCTL_DETACH_CORE,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: failed CPSW_IOCTL_DETACH_CORE: %d",
            __FUNCTION__, __LINE__, status);
    }

    CPSW_IOCTL_SET_IN_ARGS(&prms, &args);
    status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                        gCpswAppObj.coreId,
                        CPSW_IOCTL_CLOSE_PORT_LINK,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: failed to close MAC port: %d",
            __FUNCTION__, __LINE__, status);
    }

    Cpsw_close(gCpswAppObj.hCpsw);

    Cpsw_deinit(gCpswAppObj.cpswType);
}

void CpswApp_showStats(void)
{
    Cpsw_IoctlPrms prms;
    CpswStats_GenericMacPortInArgs inArgs;
    CpswStats_PortStats portStats;
    int32_t status = CPSW_SOK;

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &portStats);
    status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                        gCpswAppObj.coreId,
                        CPSW_STATS_IOCTL_GET_HOSTPORT_STATS,
                        &prms);
    if (status == CPSW_SOK)
    {
        CpswApp_print(" Port 0 Statistics");
        CpswApp_print("-----------------------------------------");
        CpswApp_printHostPortStats2G((CpswStats_HostPort_2g *)&portStats);
        CpswApp_print("");
    }
    else
    {
        CpswApp_print("%s:%d Error: failed to get host stats: %d",
            __FUNCTION__, __LINE__, status);
    }

    if (status == CPSW_SOK)
    {
        inArgs.portNum = gCpswAppObj.portNum;
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &portStats);
        status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                            gCpswAppObj.coreId,
                            CPSW_STATS_IOCTL_GET_MACPORT_STATS,
                            &prms);
        if (status == CPSW_SOK)
        {
            CpswApp_print(" Port 1 Statistics");
            CpswApp_print("-----------------------------------------");
            CpswApp_printMacPortStats2G((CpswStats_MacPort_2g *)&portStats);
            CpswApp_print("");
        }
        else
        {
            CpswApp_print("%s:%d Error: failed to get MAC stats: %d",
                __FUNCTION__, __LINE__, status);
        }
    }
    if (status == CPSW_SOK)
    {
        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                            gCpswAppObj.coreId,
                            CPSW_ALE_IOCTL_DUMP_TABLE,
                            &prms);
        CpswApp_assert(status == CPSW_SOK);
    }
}

extern void SetRxPulse();
void CpswApp_rxIsrFxn(void *appData)
{
    SetRxPulse();
}

void CpswApp_txIsrFxn(void *appData)
{
    // nothing to do here
}

int32_t CpswApp_initRxReadyPktQ(void)
{
    CpswDma_PktInfoQ rxFreeQ;
    CpswDma_PktInfo *pPktInfo;
    int32_t status = CPSW_SOK;
    int i;
    struct mbuf *m;
    struct nw_work_thread   *wtp = WTP;

    CpswUtils_initQ(&rxFreeQ);
    CpswUtils_initQ(&gCpswAppObj.rxReadyQ);

    for (i = 0U; i < gCpswAppObj.numRxPkts; i++)
    {
        m = m_getcl_wtp (M_DONTWAIT, MT_DATA, M_PKTHDR, wtp);
        if (m == NULL) {
            CpswApp_print("%s:%d Error: m_getcl_wtp failed ",
                __FUNCTION__, __LINE__);
            status = CPSW_EALLOC;
            return status;
        }

        pPktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&gCpswAppObj.freePktInfoQ);
        CpswApp_assert(pPktInfo != NULL);
        pPktInfo->bufPtr     = (uint8_t *)m->m_data;
        pPktInfo->bufPtrPhys = (uint8_t *)pool_phys(m->m_data, m->m_ext.ext_page);
        pPktInfo->userBufLen = MAX_ETH_PACKET_SIZE;
        pPktInfo->orgBufLen  = MAX_ETH_PACKET_SIZE;
        if (gCache_ops)
        {
            CACHE_INVAL(&cachectl, (void *)pPktInfo->bufPtr, (uint64_t)pPktInfo->bufPtrPhys, m->m_ext.ext_size);
        }
        pPktInfo->appPriv    = m;
        CPSW_UTILS_SET_PKT_APP_STATE(&pPktInfo->pktState, CPSW_PKTSTATE_APP_WITH_FREEQ);
        CpswUtils_enQ(&rxFreeQ, &pPktInfo->node);
    }

    /* Retrieve any CPSW packets which are ready */
    status = CpswDma_retrieveRxPackets(gCpswAppObj.hRxFlow, &gCpswAppObj.rxReadyQ);
    CpswApp_assert(status == CPSW_SOK);
    /* There should not be any packet with DMA during init */
    CpswApp_assert(CpswUtils_getQCount(&gCpswAppObj.rxReadyQ) == 0U);

#if (1U == CPSW_CFG_DEV_ERROR)
    CpswApp_validatePacketState(&rxFreeQ,
                                 CPSW_PKTSTATE_APP_WITH_FREEQ,
                                 CPSW_PKTSTATE_APP_WITH_DRIVER);
#endif

    CpswApp_submitRxPackets(gCpswAppObj.hRxFlow,
                            &rxFreeQ);

    /* Assert here as during init no. of DMA descriptors should be equal to
     * no. of free Ethernet buffers available with app */

    CpswApp_assert(0U == CpswUtils_getQCount(&rxFreeQ));

    return status;
}

void CpswApp_initPktInfoQ()
{
    uint32_t i;
    CpswDma_PktInfo *pPktInfo;

    /* Initialize all queues */
    CpswUtils_initQ(&gCpswAppObj.freePktInfoQ);

    memset(&gPktInfoMem, 0U,
        sizeof(CpswDma_PktInfo) * (CPSW_MEM_TX_PKTS_MAX + CPSW_MEM_RX_PKTS_MAX));

    /* Initialize TX & RX EthPkts and queue them to freePktInfoQ */
    for (i = 0U; i < (gCpswAppObj.numTxPkts + gCpswAppObj.numRxPkts); i++)
    {
        pPktInfo = &gPktInfoMem[i];
        CpswDma_pktInfoInit(pPktInfo);
        CPSW_UTILS_SET_PKT_APP_STATE(&pPktInfo->pktState, CPSW_PKTSTATE_APP_WITH_FREEQ);
        CpswUtils_enQ(&gCpswAppObj.freePktInfoQ, &pPktInfo->node);
    }

    CpswApp_print("initQs() freePktInfoQ initialized with %d pkts",
                       CpswUtils_getQCount(&gCpswAppObj.freePktInfoQ));

}

int32_t CpswApp_openDma(void)
{
    int32_t status = CPSW_SOK;
    CpswDma_OpenRxFlowPrms cpswRxFlowCfg;
    CpswDma_OpenTxChPrms cpswTxChCfg;

    CpswApp_initPktInfoQ();

    /* Open the CPSW TX channel  */
    if (status == CPSW_SOK)
    {
        CpswDma_initTxChParams(&cpswTxChCfg);

        cpswTxChCfg.hUdmaDrv = gCpswAppObj.hUdmaDrv;
        cpswTxChCfg.hCbArg   = &gCpswAppObj;
        // Note: we do not register callback for TX channel
        cpswTxChCfg.notifyCb = NULL;
        // Note: Using preferred channel
        cpswTxChCfg.preferredChNum = gCpswAppObj.preferredUdmaChannel;

        CpswApp_setCommonTxChPrms(&cpswTxChCfg, gCpswAppObj.numTxPkts);

#if defined (CPSW2G)
        CpswApp_openTxCh(gCpswAppObj.hCpsw,
                              gCpswAppObj.coreKey,
                              gCpswAppObj.coreId,
                              &gCpswAppObj.txChNum,
                              &gCpswAppObj.hTxCh,
                              &cpswTxChCfg);
#elif defined (CPSW9G)
        CpswApp_openTxCh(&gCpswAppObj,
                         &cpswTxChCfg);
#endif
        // Note: We do not call EnetDma_enableTxEvent call for
        // registering the TX event

        if (NULL == gCpswAppObj.hTxCh)
        {
#if defined (CPSW2G)
            /* Free the Ch Num if open Tx Ch failed */
            CpswApp_freeTxCh(gCpswAppObj.hCpsw,
                             gCpswAppObj.coreKey,
                             gCpswAppObj.coreId,
                             gCpswAppObj.txChNum);
#elif defined (CPSW9G)
            CpswProxy_freeTxCh(gCpswAppObj.hCpswProxy,
                               gCpswAppObj.hCpsw,
                               gCpswAppObj.coreKey,
                               gCpswAppObj.txChNum);
#endif
            CpswApp_print("%s:%d Error: CpswDma_openTxCh failed to open: %d",
                __FUNCTION__, __LINE__, status);
            status = CPSW_EFAIL;
        }
    }

    /* Open the CPSW RX flow  */
    if (status == CPSW_SOK)
    {
#if defined (CPSW2G)
        CpswDma_initRxFlowParams(&cpswRxFlowCfg);
        cpswRxFlowCfg.notifyCb = CpswApp_rxIsrFxn;
        cpswRxFlowCfg.hUdmaDrv = gCpswAppObj.hUdmaDrv;
        cpswRxFlowCfg.hCbArg   = &gCpswAppObj;

        CpswApp_setCommonRxFlowPrms(&cpswRxFlowCfg, gCpswAppObj.numRxPkts);

        CpswApp_openRxFlow(gCpswAppObj.hCpsw,
                           gCpswAppObj.coreKey,
                           gCpswAppObj.coreId,
                           true,
                           &gCpswAppObj.rxStartFlowIdx,
                           &gCpswAppObj.rxFlowIdx,
                           &gCpswAppObj.hostMacAddr[0U],
                           &gCpswAppObj.hRxFlow,
                           &cpswRxFlowCfg);
#elif defined (CPSW9G)
        CpswDma_initRxFlowParams(&cpswRxFlowCfg);
        CpswApp_setCommonRxFlowPrms(&cpswRxFlowCfg, gCpswAppObj.numRxPkts);
        cpswRxFlowCfg.notifyCb = CpswApp_rxIsrFxn;
        cpswRxFlowCfg.hUdmaDrv = gCpswAppObj.hUdmaDrv;
        cpswRxFlowCfg.hCbArg   = &gCpswAppObj;
        cpswRxFlowCfg.rxFlowMtu = gCpswAppObj.rxMtu;

        CpswApp_openRxFlow(&gCpswAppObj,
                           false,
                           &cpswRxFlowCfg);
#endif

        if (NULL == gCpswAppObj.hRxFlow)
        {
#if defined (CPSW2G)
            CpswApp_freeRxFlow(gCpswAppObj.hCpsw,
                               gCpswAppObj.coreKey,
                               gCpswAppObj.coreId,
                               gCpswAppObj.rxFlowIdx);
#elif defined (CPSW9G)
            CpswProxy_freeRxFlow(gCpswAppObj.hCpswProxy,
                                 gCpswAppObj.hCpsw,
                                 gCpswAppObj.coreKey,
                                 gCpswAppObj.rxFlowIdx);
#endif
            CpswApp_print("%s:%d Error: CpswDma_openRxFlow failed to open: %d",
                __FUNCTION__, __LINE__, status);
            CpswApp_assert(NULL != gCpswAppObj.hRxFlow);
        }
        else
        {
            CpswApp_print("Host MAC address: ");
            CpswApp_printMacAddr(gCpswAppObj.hostMacAddr);
#if defined (CPSW2G)
            /* For MAC loopback disable secure flag for the host port entry */
            CpswApp_changeHostAleEntry(&gCpswAppObj.hostMacAddr[0U]);
#elif defined (CPSW9G)
            /* TODO: Is this needed for 9G */
#endif

            /* Submit all ready RX buffers to DMA.*/
            status = CpswApp_initRxReadyPktQ();
        }
    }

    return status;
}

void CpswApp_freePktInfoQ(CpswDma_PktInfoQ *pPktInfoQ)
{
    CpswDma_PktInfo *pktInfo;
    uint32_t pktCnt, i;
    struct mbuf *m = NULL;

    pktCnt = CpswUtils_getQCount(pPktInfoQ);
    /* Free all retrieved packets from DMA */
    for (i = 0U; i < pktCnt; i++)
    {
        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(pPktInfoQ);
        if (pktInfo->appPriv != NULL) {
            m = pktInfo->appPriv;
            m_freem (m);
            pktInfo->appPriv = NULL;
        }
    }
}

void CpswApp_closeDma(void)
{
    CpswDma_PktInfoQ fqPktInfoQ;
    CpswDma_PktInfoQ cqPktInfoQ;

    CpswUtils_initQ(&fqPktInfoQ);
    CpswUtils_initQ(&cqPktInfoQ);

    /* There should not be any ready packet */
    CpswApp_assert(0U == CpswUtils_getQCount(&gCpswAppObj.rxReadyQ));

#if defined (CPSW2G)
    /* Close RX Flow */
    CpswApp_closeRxFlow(gCpswAppObj.hCpsw,
                        gCpswAppObj.coreKey,
                        gCpswAppObj.coreId,
                        true,
                        &fqPktInfoQ,
                        &cqPktInfoQ,
                        gCpswAppObj.rxStartFlowIdx,
                        gCpswAppObj.rxFlowIdx,
                        gCpswAppObj.hostMacAddr,
                        gCpswAppObj.hRxFlow);
#elif defined (CPSW9G)
    /* Close RX Flow */
    CpswApp_closeRxFlow(&gCpswAppObj,
                        true,
                        &fqPktInfoQ,
                        &cqPktInfoQ);
#endif

    CpswApp_freePktInfoQ(&fqPktInfoQ);
    CpswApp_freePktInfoQ(&cqPktInfoQ);
    /* Close TX channel */
    CpswUtils_initQ(&fqPktInfoQ);
    CpswUtils_initQ(&cqPktInfoQ);

#if defined (CPSW2G)
    CpswApp_closeTxCh(gCpswAppObj.hCpsw,
                      gCpswAppObj.coreKey,
                      gCpswAppObj.coreId,
                      &fqPktInfoQ,
                      &cqPktInfoQ,
                      gCpswAppObj.hTxCh,
                      gCpswAppObj.txChNum);
#elif defined (CPSW9G)
    CpswApp_closeTxCh(&gCpswAppObj,
                      &fqPktInfoQ,
                      &cqPktInfoQ);
#endif
    CpswApp_freePktInfoQ(&fqPktInfoQ);
    CpswApp_freePktInfoQ(&cqPktInfoQ);
    CpswApp_freePktInfoQ(&gCpswAppObj.rxReadyQ);
    CpswApp_freePktInfoQ(&gCpswAppObj.freePktInfoQ);

    CpswAppMem_deInit();
}

void CpswApp_changeHostAleEntry(uint8_t macAddr[])
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setUcastOutArgs;
    CpswAle_SetUcastEntryInArgs setUcastInArgs;

    memcpy(&setUcastInArgs.addr.addr[0U], macAddr, sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = 0U;
    setUcastInArgs.info.portNum = 0U;
    setUcastInArgs.info.trunk   = false;
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = false;
    setUcastInArgs.info.super   = 0U;
    setUcastInArgs.info.ageable = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                        gCpswAppObj.coreId,
                        CPSW_ALE_IOCTL_ADD_UNICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: failed CPSW_ALE_IOCTL_SET_PORT_STATE: %d",
            __FUNCTION__, __LINE__, status);
    }
}

uint32_t CpswApp_retrieveFreeTxPkts(void)
{
    CpswDma_PktInfoQ txFreeQ;
    CpswDma_PktInfo *pktInfo;
    int32_t status;
    uint32_t txFreeQCnt = 0U;
    struct mbuf *m = NULL;

    CpswUtils_initQ(&txFreeQ);

    /* Retrieve any CPSW packets that may be free now */
    status = CpswDma_retrieveTxDonePackets(gCpswAppObj.hTxCh, &txFreeQ);
    if (status == CPSW_SOK)
    {
        txFreeQCnt = CpswUtils_getQCount(&txFreeQ);

        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&txFreeQ);
        while (NULL != pktInfo)
        {
            // start
            m = pktInfo->appPriv;
            CpswApp_assert(m != NULL);
            m_freem (m);
            pktInfo->appPriv = NULL;

#if (1U == CPSW_CFG_DEV_ERROR)
            CpswUtils_checkPktState(&pktInfo->pktState,
                                    CPSW_PKTSTATE_MODULE_APP,
                                    CPSW_PKTSTATE_APP_WITH_DRIVER,
                                    CPSW_PKTSTATE_APP_WITH_FREEQ);
#endif

            CpswUtils_enQ(&gCpswAppObj.freePktInfoQ, &pktInfo->node);
            pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&txFreeQ);
        }
    }
    else
    {
        CpswApp_print("%s:%d Error: failed to retrieve pkts: %d",
            __FUNCTION__, __LINE__, status);
    }

    return txFreeQCnt;
}

uint32_t CpswApp_receivePkts(void)
{
    int32_t status;
    uint32_t rxReadyCnt = 0U;

    CpswUtils_initQ(&gCpswAppObj.rxReadyQ);

    /* Retrieve any CPSW packets which are ready */
    status = CpswDma_retrieveRxPackets(gCpswAppObj.hRxFlow, &gCpswAppObj.rxReadyQ);
    if (status == CPSW_SOK)
    {
        rxReadyCnt = CpswUtils_getQCount(&gCpswAppObj.rxReadyQ);
    }
    else
    {
        CpswApp_print("%s:%d Error: failed to retrieve pkts: %d",
            __FUNCTION__, __LINE__, status);
    }

    return rxReadyCnt;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

void CpswApp_Tick()
{
    Cpsw_periodicTick(gCpswAppObj.hCpsw);
}

void CpswApp_GetStats(nic_stats_t *stats)
{
#if defined (CPSW2G)
    Cpsw_IoctlPrms prms;
    CpswStats_GenericMacPortInArgs inArgs;
    CpswStats_PortStats portStats;
    int32_t status = CPSW_SOK;
    nic_ethernet_stats_t    *estats = &stats->un.estats;
    CpswStats_MacPort_2g *st = (CpswStats_MacPort_2g *)&portStats;
    CpswPhy_GenericInArgs inArgsPhy;

    if (gCpswAppObj.hCpsw == NULL)
    {
        return;
    }

    inArgs.portNum = gCpswAppObj.portNum;
    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &portStats);
    status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                        gCpswAppObj.coreId,
                        CPSW_STATS_IOCTL_GET_MACPORT_STATS,
                        &prms);
    if (status == CPSW_SOK)
    {
        stats->rxed_ok = st->rxGoodFrames;
        stats->rxed_broadcast = st->rxBcastFrames;
        stats->rxed_multicast = st->rxMcastFrames;
        stats->octets_rxed_ok = st->rxOctets;

        stats->txed_ok = st->txGoodFrames;
        stats->txed_broadcast = st->txBcastFrames;
        stats->txed_multicast = st->txMcastFrames;
        stats->octets_txed_ok = st->txOctets;

        estats->align_errors = st->rxAlignCodeErrors;
        estats->single_collisions = st->txSingleCollFrames;
        estats->multi_collisions = st->txMultipleCollFrames;
        estats->fcs_errors = st->rxCrcErrors;
        estats->tx_deferred = st->txDeferredFrames;
        estats->late_collisions = st->txLateCollFrames;
        estats->xcoll_aborted = st->txExcessiveCollFrames;
        estats->no_carrier = st->txCarrierSenseErrors;
        estats->oversized_packets = st->rxOversizedFrames;
        estats->jabber_detected = st->rxJabberFrames;
        estats->short_packets = st->rxUndersizedFrames;
        estats->total_collision_frames = st->txCollisionFrames;
    }
    else
    {
        CpswApp_print("%s:%d Error: failed to get MAC stats: %d",
            __FUNCTION__, __LINE__, status);
    }
    CpswApp_showStats();

    CpswApp_print("rxReadyQ count = %d", CpswUtils_getQCount(&gCpswAppObj.rxReadyQ));
    CpswApp_print("freePktInfoQ count = %d", CpswUtils_getQCount(&gCpswAppObj.freePktInfoQ));

    /* Print CPSW regs */
    CPSW_IOCTL_SET_NO_ARGS(&prms);
    status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                        gCpswAppObj.coreId,
                        CPSW_IOCTL_PRINT_REGISTERS,
                        &prms);

    /* Print CPSW PHY regs */
    CPSW_IOCTL_SET_IN_ARGS(&prms, &inArgsPhy);
    inArgsPhy.portNum = CPSW_MAC_PORT_0;
    status = Cpsw_ioctl(gCpswAppObj.hCpsw,
                        gCpswAppObj.coreId,
                        CPSW_PHY_IOCTL_PRINT_REGS,
                        &prms);
#elif defined (CPSW9G)
    /* TODO: Add CPSW9G Proxy support for statistics */
#endif
}


int32_t CpswApp_SendPkt(struct mbuf* m)
{
    int32_t retVal;
    CpswDma_PktInfo *pktInfo;
    CpswDma_PktInfoQ txSubmitQ;

    CpswApp_assert(m != NULL);

    /* Transmit a single packet */
    CpswUtils_initQ(&txSubmitQ);

    /* Dequeue one free TX Eth packet */
    pktInfo = (CpswDma_PktInfo *) CpswUtils_deQ(&gCpswAppObj.freePktInfoQ);
    if (pktInfo == NULL)
    {
        return CPSW_EFAIL;
    }

    // start
    if (!CPSW_UTILS_IS_ALIGNED(m->m_data, UDMA_CACHELINE_ALIGNMENT))
    {
        CpswApp_print("%s:%d: Error: Send m buffer addr=0x%lx is not aligned",
            __FUNCTION__, __LINE__, m->m_data);
    }

    pktInfo->bufPtr = (uint8_t *)m->m_data;
    pktInfo->bufPtrPhys = (uint8_t *)pool_phys(m->m_data, m->m_ext.ext_page);
    pktInfo->orgBufLen = m->m_ext.ext_size;
    pktInfo->userBufLen = m->m_len;
    pktInfo->txPortNum = CPSW_MAC_PORT_INV;
    pktInfo->appPriv = m;
    if (gCache_ops)
    {
        CACHE_FLUSH(&cachectl, (void *)pktInfo->bufPtr, (uint64_t)pktInfo->bufPtrPhys, pktInfo->userBufLen);
    }
    // end

#if (1U == CPSW_CFG_DEV_ERROR)
    CpswUtils_checkPktState(&pktInfo->pktState,
                            CPSW_PKTSTATE_MODULE_APP,
                            CPSW_PKTSTATE_APP_WITH_FREEQ,
                            CPSW_PKTSTATE_APP_WITH_DRIVER);
#endif

    /* Enqueue the packet for later transmission */
    CpswUtils_enQ(&txSubmitQ, &pktInfo->node);

    retVal = CpswApp_submitTxPackets(gCpswAppObj.hTxCh,
                                          &txSubmitQ);
    if (retVal != CPSW_SOK)
    {
        CpswApp_print("%s:%d Error: failed to send packet.. ret-%d",
            __FUNCTION__, __LINE__, retVal);
    }

    return retVal;
}

extern struct mbuf * cpsw_process_rx(void *arg, struct mbuf *);

int32_t CpswApp_GetRx(void *arg)
{
    int32_t retVal = CPSW_SOK;
    CpswDma_PktInfoQ rxFreeQ;
    CpswDma_PktInfo *pktInfo;
    uint32_t rxReadyCnt;
    struct mbuf *m, *mNew;

    /* Get the packets received so far */
    rxReadyCnt = CpswApp_receivePkts();
    if (rxReadyCnt > 0U)
    {
        CpswUtils_initQ(&rxFreeQ);

        /* Consume the received packets and release them */
        pktInfo = (CpswDma_PktInfo *) CpswUtils_deQ( &gCpswAppObj.rxReadyQ);
        while (NULL != pktInfo)
        {
#if (1U == CPSW_CFG_DEV_ERROR)
            CpswUtils_checkPktState(&pktInfo->pktState,
                                    CPSW_PKTSTATE_MODULE_APP,
                                    CPSW_PKTSTATE_APP_WITH_DRIVER,
                                    CPSW_PKTSTATE_APP_WITH_READYQ);
#endif
            m = pktInfo->appPriv;
            m->m_pkthdr.len = m->m_len = pktInfo->userBufLen;
            if (gCache_ops)
            {
                CACHE_INVAL(&cachectl, (void *)pktInfo->bufPtr, (uint64_t)pktInfo->bufPtrPhys, m->m_ext.ext_size);
            }
            /* Get a packet/buffer to replace the one that was filled */
            /* Consume the packet by sending to upper layer */
            mNew = cpsw_process_rx(arg, m);
            if (mNew == NULL)
            {
                CpswApp_print("%s:%d: Error: cpsw_process_rx returned NULL",
                    __FUNCTION__, __LINE__);
                CpswApp_assert(false);
                //TODO - instead of assert, push pktInfo to freePktInfoQ and populate it
                // later when m bufs gets available.
            }

            CpswApp_assert(CPSW_UTILS_IS_ALIGNED(mNew->m_data, UDMA_CACHELINE_ALIGNMENT));

            pktInfo->bufPtr     = (uint8_t *)mNew->m_data;
            pktInfo->bufPtrPhys = (uint8_t *)pool_phys(mNew->m_data, mNew->m_ext.ext_page);
            pktInfo->orgBufLen  = MAX_ETH_PACKET_SIZE;
            pktInfo->userBufLen = MAX_ETH_PACKET_SIZE;
            if (gCache_ops)
            {
                CACHE_INVAL(&cachectl, (void *)pktInfo->bufPtr, (uint64_t)pktInfo->bufPtrPhys, m->m_ext.ext_size);
            }
            pktInfo->appPriv    = mNew;

#if (1U == CPSW_CFG_DEV_ERROR)
            CpswUtils_checkPktState(&pktInfo->pktState,
                                    CPSW_PKTSTATE_MODULE_APP,
                                    CPSW_PKTSTATE_APP_WITH_READYQ,
                                    CPSW_PKTSTATE_APP_WITH_FREEQ);
#endif

            CpswUtils_enQ(&rxFreeQ, &pktInfo->node);
            pktInfo = (CpswDma_PktInfo *) CpswUtils_deQ( &gCpswAppObj.rxReadyQ);
        }

        /*Submit now processed buffers */
        {
#if (1U == CPSW_CFG_DEV_ERROR)
            CpswApp_validatePacketState(&rxFreeQ,
                                             CPSW_PKTSTATE_APP_WITH_FREEQ,
                                             CPSW_PKTSTATE_APP_WITH_DRIVER);
#endif

            CpswApp_submitRxPackets(gCpswAppObj.hRxFlow,
                                    &rxFreeQ);
        }
    }

    return retVal;
}


uint32_t CpswApp_GetTxFreeQCnt() {
    return CpswUtils_getQCount(&gCpswAppObj.freePktInfoQ);
}

#if defined (CPSW9G)
void CpswApp_RegisterIPv4Address(uint8_t *ipv4Addr)
{
    CpswProxy_registerIPV4Addr(gCpswAppObj.hCpswProxy,
                               gCpswAppObj.hCpsw,
                               gCpswAppObj.coreKey,
                               gCpswAppObj.hostMacAddr,
                               ipv4Addr);
}

void CpswApp_InitSyncTimer(void)
{
    int32_t status;
    uint64_t baseaddr;

    memset(&gCpswAppObj.syncTimerObj, 0, sizeof(CpswApp_SyncTimerObj));

    gCpswAppObj.syncTimerObj.GTC0_GTC_CFG0_Base = (uint32_t *) mmap_device_memory(0,
            CSL_GTC0_GTC_CFG0_SIZE, PROT_READ|PROT_WRITE|PROT_NOCACHE, 0, CSL_GTC0_GTC_CFG0_BASE);
    CpswApp_assert(MAP_FAILED != gCpswAppObj.syncTimerObj.GTC0_GTC_CFG0_Base);

    gCpswAppObj.syncTimerObj.GTC0_GTC_CFG1_Base = (uint32_t *) mmap_device_memory(0,
            CSL_GTC0_GTC_CFG1_SIZE, PROT_READ|PROT_WRITE|PROT_NOCACHE, 0, CSL_GTC0_GTC_CFG1_BASE);
    CpswApp_assert(MAP_FAILED != gCpswAppObj.syncTimerObj.GTC0_GTC_CFG1_Base);

    gCpswAppObj.syncTimerObj.GTC0_GTC_CFG2_Base = (uint32_t *) mmap_device_memory(0,
            CSL_GTC0_GTC_CFG2_SIZE, PROT_READ|PROT_WRITE|PROT_NOCACHE, 0, CSL_GTC0_GTC_CFG2_BASE);
    CpswApp_assert(MAP_FAILED != gCpswAppObj.syncTimerObj.GTC0_GTC_CFG2_Base);

    gCpswAppObj.syncTimerObj.GTC0_GTC_CFG3_Base = (uint32_t *) mmap_device_memory(0,
            CSL_GTC0_GTC_CFG3_SIZE, PROT_READ|PROT_WRITE|PROT_NOCACHE, 0, CSL_GTC0_GTC_CFG3_BASE);
    CpswApp_assert(MAP_FAILED != gCpswAppObj.syncTimerObj.GTC0_GTC_CFG3_Base);

    CpswApp_print("CpswApp_InitSyncTimer(): Set GTC0 Push event from BIT %d\n", CPSW_REMOTE_APP_GTC_PUSHEVT_BIT_SEL);
    baseaddr = (uint64_t) gCpswAppObj.syncTimerObj.GTC0_GTC_CFG0_Base;
    baseaddr += CSL_GTC_CFG0_PUSHEVT;
    /* Configure GTC push event */
    CSL_REG32_WR(baseaddr, CPSW_REMOTE_APP_GTC_PUSHEVT_BIT_SEL);

    /* Register callback */
    status = CpswProxy_registerHwPushNotifyCb(gCpswAppObj.hCpswProxy, CpswApp_calcSyncTimeParams);
    if (status == CPSW_EALREADY_OPEN)
    {
        CpswApp_print("CpswProxy_registerHwPushNotifyCb(): Callback is registered already\n");
    }
    else
    {
        CpswApp_print("CpswProxy_registerRemoteTimer() for GTC0 Push Event\n");
        /* Send request to Ethfw to configure TSR */
        CpswProxy_registerRemoteTimer(gCpswAppObj.hCpswProxy,
                                      gCpswAppObj.hCpsw,
                                      gCpswAppObj.coreKey,
                                      CSLR_TIMESYNC_INTRTR0_IN_GTC0_GTC_PUSH_EVENT_0,
                                      CPSW_REMOTE_APP_CPTS_HW_PUSH_NUM);
    }

    /* Enable Timer */
    baseaddr = (uint64_t) gCpswAppObj.syncTimerObj.GTC0_GTC_CFG1_Base;
    baseaddr += CSL_GTC_CFG1_CNTCR;
    if ((CSL_REG32_RD(baseaddr) & 0x1U) == 0U)
    {
        CpswApp_print("CpswApp_InitSyncTimer(): GTC0 is not enabled. Enable it now.\n");
        CSL_REG32_WR(baseaddr, 0x1U);
    }

    return;
}

#endif
/* end of file */
