/*
 * $QNXLicenseC:
 * Copyright 2019, QNX Software Systems.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You
 * may not reproduce, modify or distribute this software except in
 * compliance with the License. You may obtain a copy of the License
 * at: http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OF ANY KIND, either express or implied.
 *
 * This file may contain contributions from others, either as
 * contributors under the License or as licensors under other terms.
 * Please review this entire file for other proprietary rights or license
 * notices, as well as the QNX Development Suite License Guide at
 * http://licensing.qnx.com/license-guide/ for other information.
 * $
 */

/*
 * Modfications copyright (c) 2019-2021, Texas Instruments Incorporated
 *
 */

#include    <j7_cpsw.h>
#include    <net/ifdrvcom.h>
#include    <sys/sockio.h>
#include    <netdrvr/ptp.h>
#if 1
#include "cpsw_lld_if.h"
#endif

pthread_mutex_t ts_mutex = PTHREAD_MUTEX_INITIALIZER;

int cpsw_ptp_ioctl (struct cpsw_dev *cpsw, struct ifdrv *ifd)
{
    ptp_time_t      time;
    ptp_comp_t      comp;
    ptp_extts_t     ts;
    uint8_t         tx;
    uint64_t        time_ns;
    int32_t         nudge;
    int32_t         status;

    if (ifd != NULL) {
        switch(ifd->ifd_cmd) {

        case PTP_GET_TX_TIMESTAMP:
        case PTP_GET_RX_TIMESTAMP:
            if (ifd->ifd_len != sizeof(ts)) {
                return EINVAL;
            }

            if (ISSTACK) {
                if (copyin((((uint8_t *)ifd) + sizeof(*ifd)),
                    &ts, sizeof(ts))) {
                    return EINVAL;
                }
            } else {
                memcpy(&ts, (((uint8_t *)ifd) + sizeof(*ifd)), sizeof(ts));
            }

            if (ifd->ifd_cmd == PTP_GET_TX_TIMESTAMP) {
                tx = 1;
            } else {
                tx = 0;
            }
            if (cpsw->cfg.verbose & DEBUG_PTP) {
                slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d PTP_GET_TX_TIMESTAMP/PTP_GET_RX_TIMESTAMP",
                        __FUNCTION__, __LINE__);
                slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d msg_type:%d, sequence_id:%d, tx:%d",
                        __FUNCTION__, __LINE__, ts.msg_type, ts.sequence_id, tx);
            }
            status = CpswApp_GetTimestamp(ts.msg_type, ts.sequence_id, tx, &time_ns);
            if (status != EOK)
            {
                ts.ts.sec = 0;
                ts.ts.nsec = 0;
            }
            else
            {
                ts.ts.sec = time_ns / (1000LL * 1000LL * 1000LL);
                ts.ts.nsec = time_ns % (1000LL * 1000LL * 1000LL);
            }

            if (cpsw->cfg.verbose & DEBUG_PTP) {
                slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d time_ns:%lud, ts.sec:%d, ts.nsec:%d",
                        __FUNCTION__, __LINE__, time_ns, ts.ts.sec, ts.ts.nsec);
            }

            if (ISSTACK) {
                return (copyout(&ts, (((uint8_t *)ifd) + sizeof(*ifd)),
                    sizeof(ts)));
            } else {
                memcpy((((uint8_t *)ifd) + sizeof(*ifd)), &ts, sizeof(ts));
                return EOK;
            }
            break;

        case PTP_GET_TIME:
            if (ifd->ifd_len != sizeof(time)) {
                return EINVAL;
            }
            pthread_mutex_lock(&ts_mutex);
            time_ns = 0;
            status = CpswApp_GetTime(&time_ns);
            if (status != EOK)
            {
                time.sec = 0;
                time.nsec = 0;
            }
            else
            {
                time.sec = (int32_t) (time_ns / (1000LL * 1000LL * 1000LL));
                time.nsec = (int32_t) (time_ns % (1000LL * 1000LL * 1000LL));
            }
            pthread_mutex_unlock(&ts_mutex);

            if (cpsw->cfg.verbose & DEBUG_PTP) {
                slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d PTP_GET_TIME",
                        __FUNCTION__, __LINE__);
                slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d time_ns:%lud, ts.sec:%d, ts.nsec:%d",
                        __FUNCTION__, __LINE__, time_ns, time.sec, time.nsec);
            }

            if (ISSTACK) {
                return (copyout(&time, (((uint8_t *)ifd) + sizeof(*ifd)),
                    sizeof(time)));
            } else {
                memcpy((((uint8_t *)ifd) + sizeof(*ifd)), &time, sizeof(time));
                return EOK;
            }
            break;

        case PTP_SET_TIME:
            if (ifd->ifd_len != sizeof(time)) {
                return EINVAL;
            }
            if (ISSTACK) {
                if (copyin((((uint8_t *)ifd) + sizeof(*ifd)),
                    &time, sizeof(time))) {
                return EINVAL;
                }
            } else {
                memcpy(&time, (((uint8_t *)ifd) + sizeof(*ifd)), sizeof(time));
            }

            time_ns = (time.sec * 1000LL * 1000LL * 1000LL) + time.nsec;
            if (cpsw->cfg.verbose & DEBUG_PTP) {
                slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d PTP_SET_TIME",
                        __FUNCTION__, __LINE__);
                slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d time_ns:%lud, ts.sec:%d, ts.nsec:%d",
                        __FUNCTION__, __LINE__, time_ns, time.sec, time.nsec);
            }

            pthread_mutex_lock(&ts_mutex);
            status = CpswApp_SetTime(time_ns);
            pthread_mutex_unlock(&ts_mutex);
            return EOK;
            break;

        case PTP_SET_COMPENSATION:
            if (ifd->ifd_len != sizeof(comp)) {
                return EINVAL;
            }
            if (ISSTACK) {
                if (copyin((((uint8_t *)ifd) + sizeof(*ifd)),
                    &comp, sizeof(comp))) {
                return EINVAL;
                }
            } else {
                memcpy(&comp, (((uint8_t *)ifd) + sizeof(*ifd)), sizeof(comp));
            }

            if (comp.positive)
                nudge = comp.comp;
            else
                nudge = -(comp.comp);

            if (cpsw->cfg.verbose & DEBUG_PTP) {
                slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d PTP_SET_COMPENSATION",
                        __FUNCTION__, __LINE__);
                slogf(_SLOGC_NETWORK, _SLOG_INFO, "%s:%d positive:%d, comp:%d",
                        __FUNCTION__, __LINE__, comp.positive, comp.comp);
            }
            status = CpswApp_SetCompensation(nudge);
            return EOK;
            break;

        default:
            slogf(_SLOGC_NETWORK, _SLOG_ERROR, "Unknown PTP ioctl 0x%lx", ifd->ifd_cmd);
            break;
        }
    }
    return EINVAL;
}


/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/

int     cpsw_ioctl (struct ifnet * ifp, unsigned long cmd, caddr_t data)

{
    int                     error = 0;
    struct cpsw_dev         *cpsw = ifp->if_softc;
    struct drvcom_config    *dcfgp;
    struct drvcom_stats     *dstp;
    struct ifdrv_com        *ifdc;
    struct ifdrv            *ifd;

    if (cpsw->cfg.verbose & DEBUG_TRACE) {
        slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d Came here -->cmd-0x%x", __FUNCTION__, __LINE__, (unsigned)cmd);
    }

    switch (cmd) {
        case    SIOCGDRVCOM:
            ifdc = (struct ifdrv_com *)data;
            switch (ifdc->ifdc_cmd) {
                case    DRVCOM_CONFIG:
                    dcfgp = (struct drvcom_config *)ifdc;

                    if (ifdc->ifdc_len != sizeof(nic_config_t)) {
                        error = EINVAL;
                        break;
                    }
                    memcpy(&dcfgp->dcom_config, &cpsw->cfg, sizeof(cpsw->cfg));
                    break;

                case    DRVCOM_STATS:
                    dstp = (struct drvcom_stats *)ifdc;

                    if (ifdc->ifdc_len != sizeof(nic_stats_t)) {
                        error = EINVAL;
                        break;
                    }
#if 1
                    CpswApp_GetStats(&cpsw->stats);
#endif
                    memcpy(&dstp->dcom_stats, &cpsw->stats, sizeof(cpsw->stats));
                    break;

                default:
                    error = ENOTTY;
                }
            break;


        case    SIOCSIFMEDIA:
        case    SIOCGIFMEDIA: {
            struct ifreq *ifr = (struct ifreq *)data;

            error = ifmedia_ioctl(ifp, ifr, &cpsw->bsd_mii.mii_media, cmd);
            break;
            }

        case SIOCSDRVSPEC:
        case SIOCGDRVSPEC:
            ifd = (struct ifdrv *)data;
            switch (ifd->ifd_cmd) {
                case PTP_GET_TX_TIMESTAMP:
                case PTP_GET_RX_TIMESTAMP:
                case PTP_GET_TIME:
                case PTP_SET_TIME:
                case PTP_SET_COMPENSATION:
                    error = cpsw_ptp_ioctl(cpsw, ifd);
                    break;
            }
            break;

        default:
            error = ether_ioctl(ifp, cmd, data);
            if (error == ENETRESET) {
                /*
                 * Multicast list has changed; set the
                 * hardware filter accordingly.
                 */
               //TODO
                error = 0;
            }
            break;
        }

    if (cpsw->cfg.verbose & DEBUG_TRACE) {
        slogf(_SLOGC_NETWORK, _SLOG_ERROR, "%s:%d <--", __FUNCTION__, __LINE__);
    }

    return error;
}

