/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw.h
 *
 * \brief    This file contains the main structure definitions and function
 *           prototypes of the CPSW driver.
 */

/**
 * \defgroup DRV_CPSW_MODULE CPSW Driver
 *
 * The Common Port Switch (CPSW) is present in the Jacinto 7 device
 * family.  This CPSW driver implementation supports the CPSW 2G and 9G
 * instances.
 *
 *  @{
 */

/*!
 * \defgroup CPSW_MAIN_API CPSW Main API
 */

/*!
 * \defgroup CPSW_DMA_API CPSW Data Path Header
 */

/*!
 * \defgroup CPSW_HOSTPORT_API CPSW Host Port API
 */

/*!
 * \defgroup CPSW_MACPORT_API CPSW MAC Port API
 */

/*!
 * \defgroup CPSW_MDIO_API CPSW Management Data I/O (MDIO) API
 */

/*!
 * \defgroup CPSW_STATS_API CPSW Statistics API
 */

/*!
 * \defgroup CPSW_CPTS_API CPSW CPTS API
 */

/*!
 * \defgroup CPSW_QUEUE_API CPSW Queue API
 */

/*!
 * \defgroup CPSW_RM_API CPSW RM API
 */

/*!
 * \defgroup CPSW_TRACE_API CPSW TRACE API
 */

/*!
 * \defgroup CPSW_PHY_API CPSW_PHY_API
 */

/* @} */

/*!
 * \addtogroup CPSW_MAIN_API
 * @{
 */

#ifndef CPSW_H_
#define CPSW_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/drv/cpsw/include/cpsw_types.h>
#include <ti/drv/cpsw/soc/cpsw_soc.h>

#include <ti/drv/cpsw/include/cpsw_ale.h>
#include <ti/drv/cpsw/include/cpsw_cpts.h>
#include <ti/drv/cpsw/include/cpsw_dma.h>
#include <ti/drv/cpsw/include/cpsw_hostport.h>
#include <ti/drv/cpsw/include/cpsw_macport.h>
#include <ti/drv/cpsw/include/cpsw_mdio.h>
#include <ti/drv/cpsw/include/cpsw_stats.h>
#include <ti/drv/cpsw/include/cpsw_phy.h>
#include <ti/drv/cpsw/include/cpsw_trace.h>
#include <ti/drv/cpsw/include/cpsw_utils.h>
#include <ti/drv/cpsw/include/cpsw_rm.h>

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

/*!
 * \brief Set null args for an IOCTL command that takes no arguments
 */
#define CPSW_IOCTL_SET_NO_ARGS(prms) \
    {                                \
        (prms)->inArgs      = NULL;  \
        (prms)->inArgsSize  = 0U;    \
        (prms)->outArgs     = NULL;  \
        (prms)->outArgsSize = 0U;    \
    }

/*!
 * \brief Set the input args for an IOCTL command
 */
#define CPSW_IOCTL_SET_IN_ARGS(prms, in)     \
    {                                        \
        (prms)->inArgs      = (in);          \
        (prms)->inArgsSize  = sizeof(*(in)); \
        (prms)->outArgs     = NULL;          \
        (prms)->outArgsSize = 0U;            \
    }

/*!
 * \brief Set the output args for an IOCTL command
 */
#define CPSW_IOCTL_SET_OUT_ARGS(prms, out)    \
    {                                         \
        (prms)->inArgs      = NULL;           \
        (prms)->inArgsSize  = 0U;             \
        (prms)->outArgs     = (out);          \
        (prms)->outArgsSize = sizeof(*(out)); \
    }

/*!
 * \brief Set the input and output args for an IOCTL command
 */
#define CPSW_IOCTL_SET_INOUT_ARGS(prms, in, out) \
    {                                            \
        (prms)->inArgs      = (in);              \
        (prms)->inArgsSize  = sizeof(*(in));     \
        (prms)->outArgs     = (out);             \
        (prms)->outArgsSize = sizeof(*(out));    \
    }

/*!
 * Helper macro for CPSW public IOCTL command ids
 */
#define CPSW_PUBLIC_IOCTL(x)                  (CPSW_IOCTL_CPSW_BASE | CPSW_IOCTL_MIN(x))

/*!
 * \brief Set the input and output args for an IOCTL command
 */
#define CPSW_DMA_TX_CH_NUM(chNum)             (chNum)

/**
 *  \name CPSW Intervlan ingress packet match types
 *
 *  Ingress packet match criteria for interVLAN routing.
 *  Each packet match criteria  is represented by a bit
 *  To enable multiple match criteria create bitmask
 *  ORing required ingress packet match criteria defines
 *  @{
 */
/*! Enable classifier match with PORT number */
#define CPSW_INTERVLAN_INGRESSPKT_MATCH_PORT (CPSW_ALE_POLICER_MATCH_PORT)
/*! Enable classifier match with MAC source address entry */
#define CPSW_INTERVLAN_INGRESSPKT_MATCH_MACSRC  (CPSW_ALE_POLICER_MATCH_MACSRC)
/*! Enable classifier match with MAC destination address entry */
#define CPSW_INTERVLAN_INGRESSPKT_MATCH_MACDST  (CPSW_ALE_POLICER_MATCH_MACDST)
/*! Enable classifier match with MAC destination address entry */
#define CPSW_INTERVLAN_INGRESSPKT_MATCH_ETHERTYPE (CPSW_ALE_POLICER_MATCH_ETHERTYPE)
/*! Enable classifier match with IPv4/IPv6 source address */
#define CPSW_INTERVLAN_INGRESSPKT_MATCH_IPSRC (CPSW_ALE_POLICER_MATCH_IPSRC)
/*! Enable classifier match with IPv4/IPv6 destination address */
#define CPSW_INTERVLAN_INGRESSPKT_MATCH_IPDST (CPSW_ALE_POLICER_MATCH_IPDST)
/*  @} */

/**
 *  \name InterVLAN ingress packet match aleEntry delete bitmask
 *
 *  Associated with some classifier types are ALE table entries.
 *  When deleting the classifier entry application needs to specify by means
 *  of below bitmask if the ALE entry associated with the classifier should also
 *  be deleted.If not set the ALE entry will remain.
 *  @{
 */
/*! Delete  MAC source address ALE table entry */
#define CPSW_INTERVLAN_PKTCLASSIFIER_DELETE_MACSRC (CPSW_ALE_POLICER_TABLEENTRY_DELETE_MACSRC)
/*! Delete  MAC destination address ALE table entry */
#define CPSW_INTERVLAN_PKTCLASSIFIER_DELETE_MACDST (CPSW_ALE_POLICER_TABLEENTRY_DELETE_MACDST)
/*! Delete  EtherType classifier associated ALE table entry */
#define CPSW_INTERVLAN_PKTCLASSIFIER_DELETE_ETHERTYPE (CPSW_ALE_POLICER_TABLEENTRY_DELETE_ETHERTYPE)
/*! Delete  IPv4/IPv6 source address classifier associated ALE table entry */
#define CPSW_INTERVLAN_PKTCLASSIFIER_DELETE_IPSRC (CPSW_ALE_POLICER_TABLEENTRY_DELETE_IPSRC)
/*! Delete  IPv4/IPv6 destination address classifier associated ALE table entry */
#define CPSW_INTERVLAN_PKTCLASSIFIER_DELETE_IPDST (CPSW_ALE_POLICER_TABLEENTRY_DELETE_IPDST)
/*! Delete  InnerVLAN classifier associated ALE table entry */
#define CPSW_INTERVLAN_PKTCLASSIFIER_DELETE_VLAN (CPSW_ALE_POLICER_TABLEENTRY_DELETE_IVLAN)

/*  @} */

/*! \brief Frequency in HZ for RGMII 250MHz clock needed to operate in 1Gbps mode */
#define CPSW_SOC_RGMII_MHZ_250_CLK_VAL        (250000000U)

/*! \brief Frequency in HZ for RGMII 50MHz clock needed to operate in 100Mbps mode */
#define CPSW_SOC_RGMII_MHZ_50_CLK_VAL         (50000000U)

/*! \brief Frequency in HZ for RGMII 5MHz clock needed to operate in 10Mbps mode */
#define CPSW_SOC_RGMII_MHZ_5_CLK_VAL          (5000000U)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 * \brief CPSW module IOCTL commands
 */
enum Cpsw_Ioctl_e
{
    /*!
     * \brief Get the version of CPSW
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: #Cpsw_GetVersionOutArgs
     */
    CPSW_IOCTL_GET_VERSION                        = CPSW_PUBLIC_IOCTL(0U),

    /*!
     * \brief Open port link (MAC port and PHY)
     *
     * IOCTL parameters:
     * -  inArgs: #Cpsw_OpenPortLinkInArgs
     * - outArgs: None
     */
    CPSW_IOCTL_OPEN_PORT_LINK                     = CPSW_PUBLIC_IOCTL(1U),

    /*!
     * \brief Close port link (MAC port and PHY)
     *
     * IOCTL parameters:
     * -  inArgs: #Cpsw_GenericPortLinkInArgs
     * - outArgs: None
     */
    CPSW_IOCTL_CLOSE_PORT_LINK                    = CPSW_PUBLIC_IOCTL(2U),

    /*!
     * \brief Check if port link is up
     *
     * IOCTL parameters:
     * -  inArgs: #Cpsw_GenericPortLinkInArgs
     * - outArgs: bool
     */
    CPSW_IOCTL_IS_PORT_LINK_UP                    = CPSW_PUBLIC_IOCTL(3U),

    /*!
     * \brief Get port link configuration (speed, duplexity)
     *
     * IOCTL parameters:
     * -  inArgs: #Cpsw_GenericPortLinkInArgs
     * - outArgs: #CpswMacPort_LinkConfig
     */
    CPSW_IOCTL_GET_PORT_LINK_CONFIG               = CPSW_PUBLIC_IOCTL(4U),

    /*!
     * \brief Set host port traffic shaping configuration
     *
     * Sets the traffic shaping configuration which not only depends on the host
     * port parameters but also on the number of free TX FIFO blocks of the MAC
     * ports.
     *
     * IOCTL parameters:
     * -  inArgs: #Cpsw_SetHostPortTrafficShapingInArgs
     * - outArgs: None
     */
    CPSW_IOCTL_SET_HOSTPORT_TRAFFIC_SHAPING       = CPSW_PUBLIC_IOCTL(5U),

    /*!
     * \brief Print CPSW and all submodule registers
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: None
     */
    CPSW_IOCTL_PRINT_REGISTERS                    = CPSW_PUBLIC_IOCTL(6U),

    /*!
     * \brief Setup interVLAN route for a single egress port
     *
     * IOCTL parameters:
     * -  inArgs: #Cpsw_SetInterVlanRouteUniEgressInArgs
     * - outArgs: #Cpsw_SetInterVlanRouteUniEgressOutArgs
     */
    CPSW_IOCTL_SET_INTERVLAN_ROUTE_UNI_EGRESS     = CPSW_PUBLIC_IOCTL(7U),

    /*!
     * \brief Setup interVLAN route for a single egress port
     *
     * IOCTL parameters:
     * -  inArgs: #Cpsw_ClearInterVlanRouteUniEgressInArgs
     * - outArgs: #Cpsw_ClearInterVlanRouteUniEgressOutArgs
     */
    CPSW_IOCTL_CLEAR_INTERVLAN_ROUTE_UNI_EGRESS   = CPSW_PUBLIC_IOCTL(8U),

    /*!
     * \brief Setup interVLAN route for a multiple egress port
     *
     * IOCTL parameters:
     * -  inArgs: #Cpsw_SetInterVlanRouteMultiEgressInArgs
     * - outArgs: #Cpsw_SetInterVlanRouteMultiEgressOutArgs
     */
    CPSW_IOCTL_SET_INTERVLAN_ROUTE_MULTI_EGRESS   = CPSW_PUBLIC_IOCTL(9U),

    /*!
     * \brief Clear interVLAN route for a multiple egress port route
     *
     * IOCTL parameters:
     * -  inArgs: #Cpsw_ClearInterVlanRouteMultiEgressInArgs
     * - outArgs: #Cpsw_ClearInterVlanRouteMultiEgressOutArgs
     */
    CPSW_IOCTL_CLEAR_INTERVLAN_ROUTE_MULTI_EGRESS = CPSW_PUBLIC_IOCTL(10U),

    /*!
     * \brief Configure the short interpacket gap feature
     *
     * IOCTL parameters:
     * -  inArgs: #Cpsw_SetShortTxIPGConfigInArgs
     * - outArgs: None
     */
    CPSW_IOCTL_SET_SHORT_IPG_CONFIG               = CPSW_PUBLIC_IOCTL(11U),

    /*!
     * \brief Get the current short interpacket gap configuration
     *
     * IOCTL parameters:
     * -  inArgs: None
     * - outArgs: #Cpsw_GetShortTxIPGConfigOutArgs
     */
    CPSW_IOCTL_GET_SHORT_IPG_CONFIG               = CPSW_PUBLIC_IOCTL(12U),

    /*!
     * \brief Attach Core with CPSW
     *
     * IOCTL parameters:
     * -  inArgs: uint32_t coreId
     * - outArgs: #Cpsw_AttachCoreOutArgs
     */
    CPSW_IOCTL_ATTACH_CORE                        = CPSW_PUBLIC_IOCTL(13U),

    /*!
     * \brief Detach core from CPSW
     *
     * IOCTL parameters:
     * -  inArgs: uint32_t coreKey
     * - outArgs: None
     */
    CPSW_IOCTL_DETACH_CORE                        = CPSW_PUBLIC_IOCTL(14U),

    /*!
     * \brief Register default flow to a specific rx flow
     *
     * IOCTL parameters:
     * -  inArgs: #Cpsw_DefaultFlowInfo
     * - outArgs: None
     */
    CPSW_IOCTL_REGISTER_RX_DEFAULT_FLOW           = CPSW_PUBLIC_IOCTL(15U),

    /*!
     * \brief Unregister default flow
     *
     * After unregistering default flow, default flow traffic will be directed
     * to CPSW internal reserved flow where they will be dropped.
     *
     * IOCTL parameters:
     * -  inArgs: #Cpsw_DefaultFlowInfo
     * - outArgs: None
     */
    CPSW_IOCTL_UNREGISTER_RX_DEFAULT_FLOW         = CPSW_PUBLIC_IOCTL(16U),

    /*!
     * \brief Register destination MAC address to a specific rx flow
     *
     * Note that CPSW supports associating multiple L2/L3 header fields to a
     * specific rx flow. Refer ALE API #CPSW_ALE_IOCTL_SET_POLICER for details
     * on how to associate fields other than DST MAC to a specific flow.
     *
     * IOCTL parameters:
     * -  inArgs: #Cpsw_MacDstFlowInfo
     * - outArgs: None
     */
    CPSW_IOCTL_REGISTER_DSTMAC_RX_FLOW            = CPSW_PUBLIC_IOCTL(17U),

    /*!
     * \brief Unregister destination MAC address to a specific flow
     *
     * After unregistering dstmac, traffic with associated MAC address will be
     * directed to default flow.
     *
     * IOCTL parameters:
     * -  inArgs: #Cpsw_MacDstFlowInfo
     * - outArgs: None
     */
    CPSW_IOCTL_UNREGISTER_DSTMAC_RX_FLOW          = CPSW_PUBLIC_IOCTL(18U),
};

/*!
 * \brief Output args for #CPSW_IOCTL_GET_VERSION command
 */
typedef struct Cpsw_GetVersionsOutArgs_s
{
    /*! CPSW subsystem version */
    Cpsw_Version ss;

    /*! CPSW module version */
    Cpsw_Version cpsw;

    /*! ALE version */
    Cpsw_Version ale;

    /*! CPTS version */
    Cpsw_Version cpts;

    /*! MDIO version */
    CpswMdio_Version mdio;
} Cpsw_GetVersionsOutArgs;

/*!
 * \brief Input args for #CPSW_IOCTL_OPEN_PORT_LINK command
 */
typedef struct Cpsw_OpenPortLinkInArgs_s
{
    /*! Port number */
    Cpsw_MacPort portNum;

    /*! MAC interface type */
    CpswMacPort_Interface interface;

    /*! Link configuration (speed, duplexity) common to MAC and PHY */
    CpswMacPort_LinkConfig linkConfig;

    /*! MAC port configuration */
    CpswMacPort_Config macConfig;

    /*! PHY configuration */
    CpswPhy_Config phyConfig;
} Cpsw_OpenPortLinkInArgs;

/*!
 * \brief Input args for #CPSW_IOCTL_CLOSE_PORT_LINK, #CPSW_IOCTL_IS_PORT_LINK_UP
 *        and #CPSW_IOCTL_GET_PORT_LINK_CONFIG command
 */
typedef struct Cpsw_GenericPortLinkInArgs_s
{
    /*! Port number */
    Cpsw_MacPort portNum;
} Cpsw_GenericPortLinkInArgs;

/*!
 * \brief Input args for #CPSW_IOCTL_SET_HOSTPORT_TRAFFIC_SHAPING command
 */
typedef struct Cpsw_SetHostPortTrafficShapingInArgs_s
{
    /*! Generic traffic shaping configuration */
    CpswPort_TrafficShapingConfig trafficShapingConfig;

    /*! Transmit FIFO blocks that must be free before a non rate-limited
     *  host port 0 receive thread can begin sending a packet */
    uint32_t txBlocksRem;
} Cpsw_SetHostPortTrafficShapingInArgs;

/*!
 * \brief InterVLAN route ingress packet match criteria configuration
 */
typedef struct Cpsw_InterVlanRouteIngressPktMatchCfg_s
{
    /*! Bitmask selecting classifier types to enable
     * @sa CPSW Intervlan ingress packet match types */
    uint32_t packetMatchEnableMask;

    /*! Flag to enable TTL check.
     *  If enabled a packet whose TTL is zero will be routed to host port
     *  instead of being routed to egress port */
    bool enableTTLCheck;

    /*! Port number to match.
     *  Must be valid if bit #CPSW_INTERVLAN_INGRESSPKT_MATCH_PORT in
     *  packetMatchEnableMask is set */
    Cpsw_MacPort ingressPort;

    /*! Source MAC  address to match.
     *  Must be valid if bit #CPSW_INTERVLAN_INGRESSPKT_MATCH_MACSRC in
     *  packetMatchEnableMask is set */
    CpswAle_MacSrcAddrClassifierInfo srcMacAddr;

    /*! Destination MAC address to match.
     *  Must valid if bit #CPSW_INTERVLAN_INGRESSPKT_MATCH_MACDST in
     *  packetMatchEnableMask is set */
    CpswAle_MacDstAddrClassifierInfo dstMacAddr;

    /*! Ingress packet VLAN ID to match */
    uint32_t vlanId;

    /*! Source IP address to match.
     *  Must be valid if bit #CPSW_INTERVLAN_INGRESSPKT_MATCH_IPSRC in
     *  policerMatchEnableMask is set */
    CpswAle_IpAddrClassifierInfo srcIp;

    /*! Destination IP address to match.
     *  Must be valid if bit #CPSW_INTERVLAN_INGRESSPKT_MATCH_IPDST in
     *  policerMatchEnableMask is set. */
    CpswAle_IpAddrClassifierInfo dstIp;

    /*! Frame Ethertype to match.
     *  Must be valid if bit #CPSW_INTERVLAN_INGRESSPKT_MATCH_ETHERTYPE in
     *  policerMatchEnableMask is set. */
    CpswAle_EtherTypeEntryInfo etherType;
} Cpsw_InterVlanRouteIngressPktMatchCfg;

/*!
 * \brief InterVLAN egress port configuration
 */
typedef struct Cpsw_InterVlanEgressPortCfg_s
{
    /*! Egress port for the interVLAN route */
    Cpsw_MacPort egressPort;

    /*! Egress packet modification for the interVLAN route.
     *  The packet that is routed to the egress port will be modified based on
     *  configuration specified in this parameter */
    CpswMacPort_InterVlanRoutingParams outPktModCfg;
} Cpsw_InterVlanEgressPortCfg;

/*!
 * \brief Input args for #CPSW_IOCTL_SET_INTERVLAN_ROUTE_UNI_EGRESS command
 */
typedef struct Cpsw_SetInterVlanRouteUniEgressInArgs_s
{
    /*! Ingress packet match criteria for the interVLAN route.
     *  If the packet matches the given criteria and the destination port
     *  of the packet is host port then the interVLAN route will take effect */
    Cpsw_InterVlanRouteIngressPktMatchCfg inPktMatchCfg;

    /*! Egress packet modification configuration for the interVLAN route */
    Cpsw_InterVlanEgressPortCfg egressCfg;
} Cpsw_SetInterVlanRouteUniEgressInArgs;

/*!
 * \brief Output args for #CPSW_IOCTL_SET_INTERVLAN_ROUTE_UNI_EGRESS command
 */
typedef struct Cpsw_SetInterVlanRouteUniEgressOutArgs_s
{
    /*! Egress port route index used for the interVLAN route */
    CpswMacPort_InterVLANRouteId egressPortRouteId;

    /*! Ingress packet match classifier info */
    CpswAle_SetInterVLANConfigOutArgs ingressPacketClassifierInfo;
} Cpsw_SetInterVlanRouteUniEgressOutArgs;

/*!
 * \brief Input args for #CPSW_IOCTL_CLEAR_INTERVLAN_ROUTE_UNI_EGRESS command
 */
typedef struct Cpsw_ClearInterVlanRouteUniEgressInArgs_s
{
    /*! Ingress packet match criteria for the interVLAN route.
     *  If the packet matches the given criteria and the destination port
     *  of the packet is host port then the interVLAN route will take effect */
    Cpsw_InterVlanRouteIngressPktMatchCfg inPktMatchCfg;

    /*! Egress packet modification configuration for the interVLAN route */
    Cpsw_InterVlanEgressPortCfg egressCfg;

    /*! Bitmask to indicate which ALE entries associated with the ingress
     *  packet match classifier needs to be deleted @sa InterVLAN ingress
     * packet match aleEntry delete bitmask */
    uint32_t delAleEntryMask;
} Cpsw_ClearInterVlanRouteUniEgressInArgs;

/*!
 * \brief Output args for #CPSW_IOCTL_CLEAR_INTERVLAN_ROUTE_UNI_EGRESS command
 */
typedef struct Cpsw_ClearInterVlanRouteUniEgressOutArgs_s
{
    /*! Egress port route index used for the interVLAN route */
    CpswMacPort_InterVLANRouteId egressPortRouteId;
} Cpsw_ClearInterVlanRouteUniEgressOutArgs;

/*!
 * \brief Input args for #CPSW_IOCTL_SET_INTERVLAN_ROUTE_MULTI_EGRESS command
 */
typedef struct Cpsw_SetInterVlanRouteMultiEgressInArgs_s
{
    /*! Number of egress ports the packet needs to be routed to */
    uint32_t numEgressPorts;

    /*! Egress packet modification configuration for the interVLAN route for
     *  each port the packet will be routed to.
     *  #numEgressPorts determine number of valid entries in this table */
    Cpsw_InterVlanEgressPortCfg egressCfg[CPSW_ALE_NUM_MAC_PORTS];

    /*! Ingress packet match criteria for the interVLAN route.
     *  If the packet matches the given criteria and the destination port
     *  of the packet is host port then the interVLAN route will take effect */
    Cpsw_InterVlanRouteIngressPktMatchCfg inPktMatchCfg;
} Cpsw_SetInterVlanRouteMultiEgressInArgs;

/*!
 * \brief Output args for #CPSW_IOCTL_SET_INTERVLAN_ROUTE_MULTI_EGRESS command
 */
typedef struct Cpsw_SetInterVlanRouteMultiEgressOutArgs_s
{
    /*! Egress port route index used for the interVLAN route */
    CpswMacPort_InterVLANRouteId egressPortRouteId;

    /*! Ingress packet match classifier info */
    CpswAle_SetInterVLANConfigOutArgs ingressPacketClassifierInfo;
} Cpsw_SetInterVlanRouteMultiEgressOutArgs;

/*!
 * \brief Input args for #CPSW_IOCTL_CLEAR_INTERVLAN_ROUTE_MULTI_EGRESS command
 */
typedef struct Cpsw_ClearInterVlanRouteMultiEgressInArgs_s
{
    /*! Number of egress ports the packet needs to be routed to */
    uint32_t numEgressPorts;

    /*! Egress packet modification configuration for the interVLAN route for
     *  each port the packet will be routed to.
     *  #numEgressPorts determine number of valid entries in this table */
    Cpsw_InterVlanEgressPortCfg egressCfg[CPSW_ALE_NUM_MAC_PORTS];

    /*! Ingress packet match criteria for the interVLAN route.
     *  If the packet matches the given criteria and the destination port
     *  of the packet is host port then the interVLAN route will take effect */
    Cpsw_InterVlanRouteIngressPktMatchCfg inPktMatchCfg;

    /*! Bitmask to indicate which ALE entries associated with the
     *  ingress packet match classifier needs to be deleted
     *  @sa InterVLAN ingress packet match aleEntry delete bitmask */
    uint32_t delAleEntryMask;
} Cpsw_ClearInterVlanRouteMultiEgressInArgs;

/*!
 * \brief Output args for #CPSW_IOCTL_CLEAR_INTERVLAN_ROUTE_MULTI_EGRESS command
 */
typedef struct Cpsw_ClearInterVlanRouteMultiEgressOutArgs_s
{
    /*! Egress port route index used for the interVLAN route */
    CpswMacPort_InterVLANRouteId egressPortRouteId;
} Cpsw_ClearInterVlanRouteMultiEgressOutArgs;

/*!
 * \brief Output args for #CPSW_IOCTL_ATTACH_CORE command
 */
typedef struct Cpsw_AttachCoreOutArgs_s
{
    /*! Core Key */
    uint32_t coreKey;

    /*! Host Port Rx MTU */
    uint32_t rxMtu;

    /*! Tx MTU per priority*/
    uint32_t txMtu[CPSW_PRIORITY_NUM];
}Cpsw_AttachCoreOutArgs;

/*!
 * \brief PHY link status change event information
 */
typedef struct Cpsw_MdioLinkStateChangeInfo_s
{
    /*! PHY address */
    uint32_t phyAddr;

    /*! Whether PHY alive status changed.
     *  Applicable only for Status Change Mode */
    bool aliveChanged;

    /*! Whether PHY is alive or not.
     *  Applicable only for Status Change Mode */
    bool isAlive;

    /*! Whether PHY link status changed */
    bool linkChanged;

    /*! Whether PHY is linked or not */
    bool isLinked;
} Cpsw_MdioLinkStateChangeInfo;

/*!
 * \brief Rx Default Flow Info used for default flow registration/unregistration
 */
typedef struct Cpsw_DefaultFlowInfo_s
{
    /*! Core Key */
    uint32_t coreKey;

    /*! Rx Flow Base or Start index */
    uint32_t startIdx;

    /*! Allocated flow's index (offset from #startIdx) */
    uint32_t flowIdx;
} Cpsw_DefaultFlowInfo;

/*!
 * \brief Output args for #CPSW_IOCTL_ATTACH_CORE command
 */
typedef struct Cpsw_MacDstFlowInfo_s
{
    /*! Core Key */
    uint32_t coreKey;

    /*! Rx Flow Base or Start index */
    uint32_t startIdx;

    /*! Allocated flow's index (offset from #startIdx) */
    uint32_t flowIdx;

    /*! Destination mac address associated with the flow */
    uint8_t macAddress[CPSW_MAC_ADDR_LEN];
} Cpsw_MacDstFlowInfo;

/*!
 * \brief CPSW PHY link state change callback function
 *
 * Callback for PHY link state change interrupt (MDIO_LINKINT).
 * This callback is invoked from interrupt context.
 */
typedef void (*Cpsw_MdioLinkStateChangeCb)(Cpsw_MdioLinkStateChangeInfo *info,
                                           void *appArg);

/*!
 * \brief CPSW port link status change callback function
 *
 * Callback for port link state change event.  This callback is invoked when
 * port link is fully functional (PHY linked, ALE ports enabled, etc) or
 * when port link is fully shutdown (ALE ports disabled).
 */
typedef void (*Cpsw_PortLinkStatusChangeCb)(Cpsw_MacPort portNum,
                                            bool isLinkUp,
                                            void *appArg);

/*!
 * \brief Short IPG configuration for CPSW IP
 */
typedef struct Cpsw_SetShortTxIPGConfigInArgs_s
{
    /*! Flag to indicate if the IPG trigger threshold has to be configured
     *  or not. If set to FALSE, the parameter #ipgTriggerThresholdBlkCount
     *  will be ignored */
    bool configureGapThreshold;

    /*! Ethernet port associated FIFO transmit block usage value for triggering
     *  transmit short gap when short gap is enabled */
    uint32_t ipgTriggerThresholdBlkCount;

    /*! Number of MAC ports to configure TX IPG */
    uint32_t numMacPorts;

    /*! Per Port short TX IPG configuration.
     *  Up to #numMacPorts entries should be populated */
    CpswMacPort_PortTxShortGapConfig portIPGCfg[CPSW_MAC_PORT_NUM];
} Cpsw_SetShortTxIPGConfigInArgs;

/*!
 * \brief Short IPG configuration for CPSW IP
 */
typedef struct Cpsw_GetShortTxIPGConfigOutArgs_s
{
    /*! Ethernet port associated FIFO transmit block usage value for triggering
     *  transmit short gap when short gap is enabled */
    uint32_t ipgTriggerThresholdBlkCount;

    /*! Number of MAC ports to configure TX IPG */
    uint32_t numMacPorts;

    /*! Per Port short TX IPG configuration.
     *  Up to numMacPorts entries should be populated */
    CpswMacPort_PortTxShortGapConfig portIPGCfg[CPSW_MAC_PORT_NUM];
} Cpsw_GetShortTxIPGConfigOutArgs;

/*!
 * \brief CPSW-level VLAN configuration
 */
typedef struct Cpsw_VlanConfig_s
{
    /*! Whether in VLAN aware or unaware mode */
    bool vlanAware;

    /*! VLAN customer or service switch */
    Cpsw_VlanSwitch vlanSwitch;

    /*! Outer VLAN ltype */
    uint16_t outerVlan;

    /*! Inner VLAN ltype */
    uint16_t innerVlan;
} Cpsw_VlanConfig;

/*!
 * \brief CPSW driver OSAL function pointers.
 */
typedef struct
{
    /*! OSAL all interrupt disable function pointer */
    CpswOsal_DisableAllIntrFxn disableAllIntr;

    /*! OSAL all interrupt restore function pointer */
    CpswOsal_RestoreAllIntrFxn restoreAllIntr;

    /*! OSAL interrupt disable function pointer */
    CpswOsal_DisableIntrFxn disableIntr;

    /*! OSAL interrupt disable function pointer */
    CpswOsal_RestoreIntrFxn restoreIntr;

    /*! Create mutex function pointer */
    CpswOsal_MutexCreateFxn createMutex;

    /*! Delete mutex function pointer */
    CpswOsal_MutexDeleteFxn deleteMutex;

    /*! Lock mutex function pointer */
    CpswOsal_MutexLockFxn lockMutex;

    /*! Unlock mutex function pointer */
    CpswOsal_MutexUnlockFxn unlockMutex;

    /*! Register interrupt function pointer */
    CpswOsal_RegisterIntrFxn registerIntr;

    /*! Unregister interrupt function pointer */
    CpswOsal_UnRegisterIntrFxn unRegisterIntr;

    /*! Timer read function pointer */
    CpswOsal_TimerReadFxn timerRead;
} CpswOsal_Prms;

/*!
 * \brief CPSW driver OSAL cache function pointers.
 *
 * Note: This structure is separated from the #CpswOsal_Prms so that
 * the cache API are set independent of driver handle. The cache API
 * function pointer is common to all handles.
 */
typedef struct
{
    /*! Cache Invalidate function pointer */
    CpswOsal_CacheInv cacheInv;

    /*! Cache writeback function pointer */
    CpswOsal_CacheWb cacheWb;

    /*! Cache writeback function pointer */
    CpswOsal_CacheWbInv cacheWbInv;
} CpswOsal_CachePrms;

/*!
 * \brief CPSW configuration
 *
 * Configuration information for the CPSW driver.
 */
typedef struct Cpsw_Config_s
{
    /*! Escalate priority load value */
    uint32_t escalatePriorityLoadVal;

    /*! Configuration of the CPSW DMA */
    CpswDma_Config dmaConfig;

    /*! VLAN configuration (inner/outer VLAN ltype, customer/service switch) */
    Cpsw_VlanConfig vlanConfig;

    /*! Max packet length transmitted on egress. Packets that are larger than
     *  this length will be dropped. This length excludes VLAN addition or
     *  removal */
    uint32_t txMtu[CPSW_PRIORITY_NUM];

    /*! Configuration of the host (CPPI) port */
    CpswHostPort_Config hostPortConfig;

    /*! Configure of the ALE submodule */
    CpswAle_Config aleConfig;

    /*! Configure of the CPTS submodule */
    CpswCpts_Config cptsConfig;

    /*! Configuration of the MDIO submodule */
    CpswMdio_Config mdioConfig;

    /*! Configuration of CPSW Resource Partition */
    CpswRm_ResourceConfig resourceConfig;

    /*! Interrupt priority */
    uint32_t intrPriority;

    /*! MDIO Link state change callback function pointer */
    Cpsw_MdioLinkStateChangeCb mdioLinkStateChangeCb;

    /*! Application data to be passed to the MDIO link state change callback */
    void *mdioLinkStateChangeCbArg;

    /*! Port link status change callback function pointer.  This callback is
     *  called when the port link is either fully functional or fully shutdown */
    Cpsw_PortLinkStatusChangeCb portLinkStatusChangeCb;

    /*! Application data to be passed to the port link status change callback */
    void *portLinkStatusChangeCbArg;

    /*! QSGMII0 Running Disparity Check (RDCD) enable - disables the rx running
     *  disparity so that errors do not propagate across lanes.
     *  It can be used when debugging a multi-port link to disable the disparity
     *  to isolate the receive errors */
    bool enableQsgmii0RDC;

    /*! QSGMII1 Running Disparity Check (RDCD) enable */
    bool enableQsgmii1RDC;
} Cpsw_Config;

/*!
 * \brief CPSW handle
 *
 * CPSW opaque handle.
 */
typedef struct Cpsw_Obj_s *Cpsw_Handle;

/* ========================================================================== */
/*                         Global Variables Declarations                      */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/*!
 * \brief CpswOsal_Prms structure init function.
 *
 * \param osalPrms     [IN] Pointer to #CpswOsal_Prms structure.
 *
 */
void CpswOsal_initPrms(CpswOsal_Prms *osalPrms);

/*!
 * \brief Set the OSAL cache paramaters in case user needs to override the
 *        default cache API.
 *
 *  Caution: This is common across all handles and should be set perferably
 *  once for a given core before calling #Cpsw_open.
 *
 *  Note: All the function pointer needs to be set.
 *
 *  Requirement: DOX_REQ_TAG(PDK-2984)
 *
 *  \param cachePrms    [IN] Pointer to #CpswOsal_CachePrms structure.
 *                           This parameter can't be NULL.
 *
 *  \return \ref Cpsw_ErrorCodes
 */
int32_t CpswOsal_setCachePrms(const CpswOsal_CachePrms *cachePrms);

/*!
 * \brief CpswOsal_CachePrms structure init function.
 *
 * \param cachePrms    [IN] Pointer to #CpswOsal_CachePrms structure
 */
void CpswOsal_initCachePrms(CpswOsal_CachePrms *cachePrms);

/*!
 * \brief Initialize CPSW configuration parameters with default values
 *
 * \param cfg    [IN] CPSW configuration to be initialized
 */
void Cpsw_initParams(Cpsw_Config *cfg);

/*!
 * \brief Initialize MAC port configuration parameters with default values
 *
 * \param cfg    [IN] MAC port configuration to be initialized
 */
void Cpsw_initMacPortParams(CpswMacPort_Config *cfg);

/*!
 * \brief Initialize PHY configuration parameters
 *
 * Initializes PHY configuration to driver defaults.
 *
 * \param cfg    [IN] Pointer to the PHY config structure
 */
void Cpsw_initPhyParams(CpswPhy_Config *cfg);

/*!
 * \brief Set PHY extended parameters
 *
 * Sets the PHY-specific extended parameters to the PHY config structure.
 *
 * \param cfg               [IN] Pointer to the PHY config
 * \param extendedCfg       [IN] Pointer to the PHY extended config
 * \param extendedCfgSize   [IN] Size of the PHY extended config
 */
void Cpsw_setPhyExtendedParams(CpswPhy_Config *cfg,
                               const void *extendedCfg,
                               uint32_t extendedCfgSize);

/*!
 * \brief Initialize OSAL parameters with default values
 *
 * Initializes CPSW OSAL parameters structure with default values
 *
 * \param pOsalPrms    [IN] OSAL parameters to be initialized
 */
void Cpsw_initOsalPrms(CpswOsal_Prms *pOsalPrms);

/*!
 * \brief Initialize CPSW driver
 *
 * Initializes the CPSW driver with OSAL and utils provided by the caller.
 *
 * \param cpswType     [IN] CPSW instance type
 * \param pOsalPrms    [IN] OSAL parameters
 * \param pUtilsPrms   [IN] Utils parameters
 */
void Cpsw_init(Cpsw_Type cpswType,
               const CpswOsal_Prms *pOsalPrms,
               CpswUtils_Prms *pUtilsPrms);

/*!
 * \brief De-initialize CPSW driver
 *
 * \param cpswType     [IN] CPSW instance type
 */
void Cpsw_deinit(Cpsw_Type cpswType);

/*!
 * \brief Set runtime trace level
 *
 * Sets the driver's runtime trace level.  The previous trace level is returned
 * by this function should the application want to restore previous level at a
 * later point in time.
 *
 * \param level    [IN] Trace level
 *
 * \return Previous trace level
 */
CpswTrace_TraceLevel Cpsw_setTraceLevel(CpswTrace_TraceLevel level);

/*!
 * \brief Get runtime trace level
 *
 * Gets the driver's runtime trace level.
 *
 * \return Current trace level
 */
CpswTrace_TraceLevel Cpsw_getTraceLevel(void);

/*!
 * \brief Get the CPSW opaque handle
 *
 * Gets the CPSW opaque handle corresponding to the given CPSW instance type
 *
 * \param cpswType     [IN] CPSW instance type
 *
 * \return CPSW opaque handle
 */
Cpsw_Handle Cpsw_getHandle(Cpsw_Type cpswType);

/*!
 * \brief Open and initialize CPSW peripheral
 *
 * Opens and initializes the CPSW peripheral with the configuration parameters
 * provided by the caller.
 *
 * It must be called only after CPSW driver has been initialized via Cpsw_init().
 *
 * \param cpswType     [IN] CPSW instance type
 * \param cfg          [IN] CPSW configuration
 *
 * \return CPSW opaque handle
 */
Cpsw_Handle Cpsw_open(Cpsw_Type cpswType,
                      const Cpsw_Config *cfg);

/*!
 * \brief Re-open CPSW peripheral
 *
 * Re-opens the CPSW peripheral, but doesn't perform any hardware initialization.
 * This function is expected to be called when application desires to attach to
 * a running CPSW switch.
 *
 * It must be called only after CPSW driver has been initialized via Cpsw_init().
 *
 * \param cpswType     [IN] CPSW instance type
 *
 * \return CPSW opaque handle
 */
Cpsw_Handle Cpsw_reopen(Cpsw_Type cpswType);

/*!
 * \brief Close CPSW peripheral
 *
 * Closes the CPSW peripheral.
 *
 * \param hCpsw        [IN] CPSW opaque handle
 */
void Cpsw_close(Cpsw_Handle hCpsw);

/*!
 * \brief Issue an operation on the CPSW peripheral/driver
 *
 * Issues a control operation on the CPSW peripheral and/or driver.  The IOCTL
 * commands fall in the following categories: CPSW, ALE, CPTS, HOSTPORT, MACPORT,
 * MDIO, SGMII, STATS, SS, PHY and RM.
 *
 * The IOCTL parameters should be built using the helper macros provided by the
 * driver:
 * - #CPSW_IOCTL_SET_NO_ARGS - If the IOCTL command doesn't take any arguments
 * - #CPSW_IOCTL_SET_IN_ARGS - If the IOCTL command takes only input arguments
 * - #CPSW_IOCTL_SET_OUT_ARGS - If the IOCTL command takes only output arguments
 * - #CPSW_IOCTL_SET_INOUT_ARGS - If the IOCTL command takes input and output arguments
 *
 * \param hCpsw        [IN] CPSW opaque handle
 * \param coreId       [IN] Core Id
 * \param cmd          [IN] IOCTL command Id
 * \param prms         [IN] IOCTL parameters
 *
 * \return \ref Cpsw_ErrorCodes
 */
int32_t Cpsw_ioctl(Cpsw_Handle hCpsw,
                   uint32_t coreId,
                   uint32_t cmd,
                   Cpsw_IoctlPrms *prms);

/*!
 * \brief Run PHY-related periodic tick
 *
 * Runs the PHY-related periodic tick operation which executes one step in the
 * PHY state machine and also handles link up or down events.
 *
 * The registered Cpsw_Config::portLinkStatusChangeCb callback is executed in
 * the caller's context when link up or down events occur.
 *
 * Note: This must be called from only task context. Calling in interrupt
 *       context can cause deadlock with other CPSW driver APIs as both share
 *       the same mutex.
 *
 * \param hCpsw        [IN] CPSW opaque handle
 */
void Cpsw_periodicTick(Cpsw_Handle hCpsw);

/*!
 * \brief Get the number of MAC ports in the CPSW instance
 *
 * \param cpswType     [IN] CPSW instance type
 *
 * \return Number of supported MAC ports
 */
uint32_t Cpsw_getMacPortMax(Cpsw_Type cpswType);

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_H_ */

/* @} */
