/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_appboardutils_am65xx_evm.c
 *
 * \brief    This file contains the board specific utilities for AM65xx EVM.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <ti/osal/osal.h>
#include <ti/drv/cpsw/include/dp83867.h>
#include <ti/board/board.h>
#include <ti/drv/cpsw/cpsw.h>

#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appboardutils.h>
#include <ti/board/src/am65xx_evm/include/board_ethernet_config.h>

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

static void CpswAppUtils_setConnectionType(void)
{
    uint32_t val;
    CSL_mcu_ctrl_mmr_cfg0Regs *mcuRegs;

    mcuRegs = (CSL_mcu_ctrl_mmr_cfg0Regs *)(uintptr_t)CSL_MCU_CTRL_MMR0_CFG0_BASE;

    val  = ENET_CTRL_RGMII_ID_NODELAY << ENET_CTRL_RGMII_ID_SHIFT;
    val |= CPSW_ENET_CTRL_MODE_RGMII;

    CpswAppUtils_mcuMmrCtrl(CPSWAPPUTILS_MMR_LOCK1, CPSWAPPUTILS_UNLOCK_MMR);

    /* Set MAC port interface in MMR */
    CSL_REG32_WR(&mcuRegs->MCU_ENET_CTRL, val);

    CpswAppUtils_mcuMmrCtrl(CPSWAPPUTILS_MMR_LOCK1, CPSWAPPUTILS_LOCK_MMR);
}

void CpswAppBoardUtils_init(void)
{
    Board_STATUS boardInitStatus;
    Board_initCfg boardCfg;

    boardCfg = BOARD_INIT_PINMUX_CONFIG |
               BOARD_INIT_MODULE_CLOCK |
               BOARD_INIT_UART_STDIO;

    boardInitStatus = Board_init(boardCfg);
    CpswAppUtils_assert(boardInitStatus == BOARD_SOK);

    CpswAppUtils_setConnectionType();
}

void CpswAppBoardUtils_deInit(void)
{
    /* Board DeInit does not exist for AM65xx */
}

uint32_t CpswAppBoardUtils_getPhyAddr(Cpsw_Type cpswType,
                                      Cpsw_MacPort portNum)
{
    uint32_t phyAddr = CPSW_PHY_INVALID_PHYADDR;

    switch (cpswType)
    {
        case CPSW_2G:
            phyAddr = 0x00U;
            break;

        default:
            CpswAppUtils_assert(false);
            break;
    }

    return phyAddr;
}

void CpswAppBoardUtils_setPhyConfigSgmii(Cpsw_MacPort portNum,
                                         CpswMacPort_Config *macConfig,
                                         CpswMacPort_Interface *interface,
                                         CpswPhy_Config *phyConfig)
{
    CpswAppUtils_assert(false);
}

void CpswAppBoardUtils_setPhyConfigQsgmii(Cpsw_Type cpswType,
                                          Cpsw_MacPort portNum,
                                          CpswMacPort_Config *macConfig,
                                          CpswMacPort_Interface *interface,
                                          CpswPhy_Config *phyConfig)
{
    CpswAppUtils_assert(false);
}

void CpswAppBoardUtils_setPhyConfigRmii(Cpsw_MacPort portNum,
                                        CpswMacPort_Config *macConfig,
                                        CpswMacPort_Interface *interface,
                                        CpswPhy_Config *phyConfig)
{
    CpswAppUtils_assert(false);
}

void CpswAppBoardUtils_setPhyConfigRgmii(Cpsw_Type cpswType,
                                         Cpsw_MacPort portNum,
                                         CpswMacPort_Config *macConfig,
                                         CpswMacPort_Interface *interface,
                                         CpswPhy_Config *phyConfig)
{
    Dp83867_Config extendedConfig;

    Dp83867_initParams(&extendedConfig);
    Cpsw_initPhyParams(phyConfig);
    phyConfig->phyAddr = CpswAppBoardUtils_getPhyAddr(cpswType, portNum);

    switch (cpswType)
    {
        case CPSW_2G:
            CpswAppUtils_assert(portNum == CPSW_MAC_PORT_0);

            /* DP83867 specific configuration */
            extendedConfig.txFifoDepth = 4U;

            extendedConfig.txClkShiftEnable     = true;
            extendedConfig.rxClkShiftEnable     = true;
            extendedConfig.impedanceInMilliOhms = 35000; /* 35 ohms */
            extendedConfig.gpio0Mode            = DP83867_GPIO0_LED3;
            extendedConfig.ledMode[1]           = DP83867_LED_LINKED_100BTX;
            extendedConfig.ledMode[2]           = DP83867_LED_RXTXACT;
            extendedConfig.ledMode[3]           = DP83867_LED_LINKED_1000BT;

            /* RGMII interface type */
            interface->layerType    = CPSW_MAC_LAYER_GMII;
            interface->sublayerType = CPSW_MAC_SUBLAYER_REDUCED;
            interface->variantType  = CPSW_MAC_VARIANT_NONE;
            break;

        case CPSW_9G:
        case CPSW_5G:
            CpswAppUtils_assert(false);
            break;
    }

    Cpsw_setPhyExtendedParams(phyConfig,
                              &extendedConfig,
                              sizeof(extendedConfig));
}

void CpswAppBoardUtils_setPhyConfig(Cpsw_Type cpswType,
                                    Cpsw_MacPort portNum,
                                    CpswMacPort_Config *macConfig,
                                    CpswMacPort_Interface *interface,
                                    CpswPhy_Config *phyConfig)
{
    return CpswAppBoardUtils_setPhyConfigRgmii(cpswType,
                                               portNum,
                                               macConfig,
                                               interface,
                                               phyConfig);
}

void CpswAppBoardUtils_setNoPhyConfig(CpswMacPort_Interface *interface,
                                      CpswPhy_Config *phyConfig)
{
    phyConfig->phyAddr      = CPSW_PHY_INVALID_PHYADDR;
    interface->layerType    = CPSW_MAC_LAYER_GMII;
    interface->sublayerType = CPSW_MAC_SUBLAYER_REDUCED;
    interface->variantType  = CPSW_MAC_VARIANT_FORCED;
}

void CpswAppBoardUtils_setLpbkCfg(bool phyLpbk,
                                  Cpsw_MacPort portNum,
                                  const CpswMacPort_Interface *interface,
                                  CpswMacPort_Config *macConfig,
                                  CpswPhy_Config *phyConfig,
                                  CpswMacPort_LinkConfig *linkConfig)
{
    /* Set speed according to the interface type */
    if (interface->layerType == CPSW_MAC_LAYER_MII)
    {
        linkConfig->speed = CPSW_SPEED_100MBIT;
    }
    else
    {
        linkConfig->speed = CPSW_SPEED_1GBIT;
    }

    linkConfig->duplexity = CPSW_DUPLEX_FULL;

    /* MAC and PHY loopbacks are mutually exclusive */
    phyConfig->enableLoopback = phyLpbk;
    macConfig->enableLoopback = !phyLpbk;
}

void CpswAppBoardUtils_getMacAddrList(uint8_t macAddr[][CPSW_MAC_ADDR_LEN],
                                      uint32_t maxMacEntries,
                                      uint32_t *pAvailMacEntries)
{
    /* No MAC address programmed in EEPROM of AM65xx EVM */
    *pAvailMacEntries = 0;
}

static void CpswAppBoardUtils_setMcuEnetControl(uint32_t modeSel)
{
    CSL_mcu_ctrl_mmr_cfg0Regs *mcuRegs;

    mcuRegs = (CSL_mcu_ctrl_mmr_cfg0Regs *)(uintptr_t)CSL_MCU_CTRL_MMR0_CFG0_BASE;
    CpswAppUtils_mcuMmrCtrl(CPSWAPPUTILS_MMR_LOCK1, CPSWAPPUTILS_UNLOCK_MMR);
    CSL_REG32_FINS(&mcuRegs->MCU_ENET_CTRL,
                   MCU_CTRL_MMR_CFG0_MCU_ENET_CTRL_MODE_SEL,
                   modeSel);
    CpswAppUtils_mcuMmrCtrl(CPSWAPPUTILS_MMR_LOCK1, CPSWAPPUTILS_LOCK_MMR);
}

void CpswAppBoardUtils_setEnetControl(Cpsw_Type cpswType,
                                      Cpsw_MacPort portNum,
                                      uint32_t modeSel)
{
    CpswAppUtils_assert(portNum == CPSW_MAC_PORT_0);
    CpswAppUtils_assert(cpswType == CPSW_2G);
    CpswAppUtils_assert(((modeSel == RMII) || (modeSel == RGMII)));
    CpswAppBoardUtils_setMcuEnetControl(modeSel);
}
