/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file cpsw_mcm.c
 *
 * \brief This file contains Multi-client Manager related functionality.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */
/* This is needed for memset/memcpy */
#include <string.h>

#include <xdc/std.h>
#include <xdc/runtime/Error.h>

#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Clock.h>
#include <ti/sysbios/knl/Semaphore.h>
#include <ti/sysbios/knl/Mailbox.h>
#include <ti/sysbios/gates/GateMutexPri.h>
#include <ti/sysbios/hal/Hwi.h>

#include "inc/cpsw_apputils.h"
#include "inc/cpsw_appmemutils.h"
#include "inc/cpsw_appsoc.h"
#include "inc/cpsw_mcm.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define CPSWMCM_TSK_STACK_MAIN         (8U * 1024U)
#define CPSWMCM_MBOX_MSG_COUNT         (10U)
#define CPSWMCM_TSK_PRIORITY           (2U)
#define CPSWMCM_PERIODICTSK_PRIORITY   (7U)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/*!
 * \brief CPSW MCM handle
 *
 * CPSW MCM opaque handle.
 */
typedef struct CpswMcm_Obj_s *CpswMcm_Handle;

typedef struct CpswMcm_CoreAttachTableEntry_s
{
    uint32_t coreId;
    Cpsw_AttachCoreOutArgs coreInfo;
}  CpswMcm_CoreAttachTableEntry;

typedef struct CpswMcm_CoreAttachTable_s
{
    uint32_t numCoresAttached;
    CpswMcm_CoreAttachTableEntry entry[CPSW_REMOTE_CLIENT_CORES_MAX];
}  CpswMcm_CoreAttachTable;

typedef struct CpswMcm_Obj_s
{
    bool isInitDone;

    uint8_t refCnt;

    Cpsw_Type cpswType;

    Cpsw_Handle hCpsw;

    Udma_DrvHandle hUdmaDrv;

    Cpsw_Config cpswCfg;

    uint32_t selfCoreId;

    CpswMcm_setPortLinkCfg setPortLinkCfg;

    Cpsw_MacPort macPortList[CPSW_MAC_PORT_NUM];

    uint8_t numMacPorts;

    Task_Handle task;

    Mailbox_Handle hMboxCmd;

    Mailbox_Handle hMboxResponse;

    Clock_Handle hTimer;

    Task_Handle task_periodicTick;

    Semaphore_Handle timerSem;

    volatile bool timerTaskShutDownFlag;

    GateMutexPri_Struct GateMutexPriStruct;

    GateMutexPri_Handle hMutex;

    uint32_t periodicTaskPeriod;

    CpswMcm_CoreAttachTable coreAttachTable;

    Cpsw_PrintFxnCb print;

    Cpsw_TraceFxnCb trace;
}CpswMcm_Obj;

typedef enum CpswMCM_Command_e
{
    /*! GET CPSW & UDMA HANDLE */
    MCM_GET_HANDLE,

    /*! RELEASE HANDLE */
    MCM_RELEASE_HANDLE,

    /*! ATTACH core */
    MCM_CORE_ATTACH,

    /*! DETACH core */
    MCM_CORE_DETACH,

    /*! SHUTDOWN MCM */
    MCM_SHUTDOWN,

    /*! Response to MCM_GET_HANDLE command */
    MCM_RESPONSE_GET_HANDLE,

    /*! Response to MCM_RELEASE_HANDLE command */
    MCM_RESPONSE_RELEASE_HANDLE,

    /*! Response to MCM_CORE_ATTACH command */
    MCM_RESPONSE_CORE_ATTACH,

    /*! Response to MCM_CORE_DETACH command */
    MCM_RESPONSE_CORE_DETACH,

    /*! Response to SHUTDOWN MCM */
    MCM_RESPONSE_SHUTDOWN,
}
CpswMCM_Command;

typedef struct CpswMcm_mailboxObj_s
{
    CpswMCM_Command cmd;

    struct CpswMcm_mailboxMsgs_s
    {
        uint32_t coreId;
        uint32_t coreKey;
        CpswMcm_HandleInfo handleInfo;
        Cpsw_AttachCoreOutArgs attachInfo;
    } msgBody;
}CpswMcm_mailboxObj;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static int32_t  CpswMcm_open(CpswMcm_Handle hMcm);

static void     CpswMcm_close(CpswMcm_Handle hMcm);

static void     CpswMcm_serverTask(UArg hMcm,
                                   UArg a1);

static int32_t  CpswMcm_enablePorts(CpswMcm_Handle hMcm);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */
static CpswMcm_Obj gMcmObj[CPSW_COUNT] =
{
    [CPSW_2G] =
    {
        .isInitDone            = false,
        .timerTaskShutDownFlag = false,
        .hMutex                = NULL,
        .task                  = NULL,
        .hMboxCmd              = NULL,
        .hMboxResponse         = NULL,
        .hTimer                = NULL,
        .task_periodicTick     = NULL,
        .timerSem              = NULL,
    },

    [CPSW_9G] =
    {
        .isInitDone            = false,
        .timerTaskShutDownFlag = false,
        .hMutex                = NULL,
        .task                  = NULL,
        .hMboxCmd              = NULL,
        .hMboxResponse         = NULL,
        .hTimer                = NULL,
        .task_periodicTick     = NULL,
        .timerSem              = NULL,
    },
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */
static void CpswMcm_initAttachTable(CpswMcm_CoreAttachTable *attachTbl)
{
    uint32_t i;

    attachTbl->numCoresAttached = 0;
    for (i = 0; i < CPSW_UTILS_ARRAYSIZE(attachTbl->entry); i++)
    {
        attachTbl->entry[i].coreId = CPSW_RM_INVALIDCORE;
    }
}

static CpswMcm_CoreAttachTableEntry *CpswMcm_getCoreAttachEntry(CpswMcm_CoreAttachTable *attachTbl,
                                                                uint32_t coreId)
{
    uint32_t i;
    CpswMcm_CoreAttachTableEntry *entry;

    entry = NULL;
    for (i = 0; i < attachTbl->numCoresAttached; i++)
    {
        if (attachTbl->entry[i].coreId == coreId)
        {
            entry = &attachTbl->entry[i];
            break;
        }
    }

    return entry;
}

static void CpswMcm_addCoreAttachEntry(CpswMcm_CoreAttachTable *attachTbl,
                                       uint32_t coreId,
                                       Cpsw_AttachCoreOutArgs *coreInfo)
{
    CpswMcm_CoreAttachTableEntry *entry;

    entry = CpswMcm_getCoreAttachEntry(attachTbl, coreId);
    if (entry == NULL)
    {
        CpswAppUtils_assert(attachTbl->numCoresAttached < CPSW_UTILS_ARRAYSIZE(attachTbl->entry));
        attachTbl->entry[attachTbl->numCoresAttached].coreId   = coreId;
        attachTbl->entry[attachTbl->numCoresAttached].coreInfo = *coreInfo;
        attachTbl->numCoresAttached++;
    }
    else
    {
        uint32_t i;

        CpswAppUtils_assert(entry->coreId == coreId);
        CpswAppUtils_assert(entry->coreInfo.coreKey == coreInfo->coreKey);
        CpswAppUtils_assert(entry->coreInfo.rxMtu == coreInfo->rxMtu);
        for (i = 0; i < CPSW_UTILS_ARRAYSIZE(entry->coreInfo.txMtu); i++)
        {
            CpswAppUtils_assert(entry->coreInfo.txMtu[i] == coreInfo->txMtu[i]);
        }
    }
}

static void CpswMcm_delCoreAttachEntry(CpswMcm_CoreAttachTable *attachTbl,
                                       uint32_t coreId,
                                       uint32_t coreKey)
{
    uint32_t i;
    CpswMcm_CoreAttachTableEntry *entry;

    entry = CpswMcm_getCoreAttachEntry(attachTbl, coreId);
    if (entry != NULL)
    {
        CpswAppUtils_assert(attachTbl->numCoresAttached > 0);
        for (i = 0; i < attachTbl->numCoresAttached; i++)
        {
            if (attachTbl->entry[i].coreId == coreId)
            {
                break;
            }
        }

        CpswAppUtils_assert(i < attachTbl->numCoresAttached);
        CpswAppUtils_assert(attachTbl->entry[i].coreId == coreId);
        CpswAppUtils_assert(attachTbl->entry[i].coreInfo.coreKey == coreKey);
        /* Move the last entry to the freed entry */
        attachTbl->numCoresAttached--;
        attachTbl->entry[i] = attachTbl->entry[attachTbl->numCoresAttached];
    }
}

static void CpswMcm_initMbox(CpswMcm_Handle hMcm)
{
    Task_Params params;

    Mailbox_Params mboxParams;
    Error_Block eb;

    Error_init(&eb);
    Mailbox_Params_init(&mboxParams);
    mboxParams.instance->name = "MCM Request MBX";

    hMcm->hMboxCmd = Mailbox_create(sizeof(CpswMcm_mailboxObj),
                                    CPSWMCM_MBOX_MSG_COUNT,
                                    &mboxParams,
                                    &eb);
    CpswAppUtils_assert(hMcm->hMboxCmd != NULL);

    Error_init(&eb);
    Mailbox_Params_init(&mboxParams);
    mboxParams.instance->name = "MCM Response MBX";

    hMcm->hMboxResponse = Mailbox_create(sizeof(CpswMcm_mailboxObj),
                                         CPSWMCM_MBOX_MSG_COUNT,
                                         &mboxParams,
                                         &eb);
    CpswAppUtils_assert(hMcm->hMboxResponse != NULL);

    Task_Params_init(&params);
    Error_init(&eb);

    params.priority  = CPSWMCM_TSK_PRIORITY;
    params.stack     = NULL;
    params.stackSize = CPSWMCM_TSK_STACK_MAIN;
    params.arg0      = (UArg)hMcm;

    switch (hMcm->cpswType)
    {
        case CPSW_2G:
            params.instance->name = "MCM2G_Task";
            break;

#if defined(SOC_J7200)
    case CPSW_5G:
            params.instance->name = "MCM5G_Task";
            break;
#endif
#if defined(SOC_J721E)
    case CPSW_9G:
            params.instance->name = "MCM9G_Task";
            break;
#endif

        default:
            CpswAppUtils_assert(false);
    }

    hMcm->task = Task_create(CpswMcm_serverTask, &params, &eb);

    CpswAppUtils_assert(hMcm->task != NULL);
}

static int32_t CpswMcm_validateInitCfg(const CpswMcm_InitConfig *pCpswMcmInitCfg)
{
    int32_t status = CPSW_SOK;
    Cpsw_Type cpswType;

    if (pCpswMcmInitCfg != NULL &&
        pCpswMcmInitCfg->pCpswCfg != NULL &&
        pCpswMcmInitCfg->pCpswCfg->dmaConfig.hUdmaDrv != NULL &&
        pCpswMcmInitCfg->setPortLinkCfg != NULL)
    {
        cpswType = pCpswMcmInitCfg->cpswType;

        if ((cpswType == CPSW_2G) &&
            ((pCpswMcmInitCfg->numMacPorts != Cpsw_getMacPortMax(cpswType)) ||
             (pCpswMcmInitCfg->macPortList[0] != CPSW_MAC_PORT_0)))
        {
            status = CPSW_EINVALID_PARAMS;
        }

        if ((cpswType == CPSW_9G) &&
            ((pCpswMcmInitCfg->numMacPorts > Cpsw_getMacPortMax(cpswType))))
        {
            status = CPSW_EINVALID_PARAMS;
        }
    }
    else
    {
        status = CPSW_EINVALID_PARAMS;
    }

    return status;
}

int32_t  CpswMcm_init(const CpswMcm_InitConfig *pCpswMcmInitCfg)
{
    int32_t status = CPSW_SOK;
    GateMutexPri_Params params;
    IArg key;
    Cpsw_Type cpswType  = pCpswMcmInitCfg->cpswType;
    CpswMcm_Handle hMcm = &gMcmObj[cpswType];

    key = Hwi_disable();

    if (hMcm->hMutex == NULL)
    {
        GateMutexPri_Params_init(&params);

        GateMutexPri_construct(&hMcm->GateMutexPriStruct, &params);
        hMcm->hMutex = GateMutexPri_handle(&hMcm->GateMutexPriStruct);
    }

    Hwi_restore(key);
    CpswAppUtils_assert(hMcm->hMutex != NULL);
    key = GateMutexPri_enter(hMcm->hMutex);

    if (hMcm->isInitDone == false)
    {
        status = CpswMcm_validateInitCfg(pCpswMcmInitCfg);

        if (status == CPSW_SOK)
        {
            hMcm->cpswCfg            = *pCpswMcmInitCfg->pCpswCfg;
            hMcm->refCnt             = 0U;
            hMcm->cpswType           = cpswType;
            hMcm->hCpsw              = NULL;
            hMcm->hUdmaDrv           = NULL;
            hMcm->selfCoreId         = pCpswMcmInitCfg->pCpswCfg->resourceConfig.selfCoreId;
            hMcm->setPortLinkCfg     = pCpswMcmInitCfg->setPortLinkCfg;
            hMcm->numMacPorts        = pCpswMcmInitCfg->numMacPorts;
            hMcm->periodicTaskPeriod = pCpswMcmInitCfg->periodicTaskPeriod;

            hMcm->print = pCpswMcmInitCfg->printFxn;
            if (hMcm->print == NULL)
            {
                hMcm->print = CpswAppUtils_print;
            }

            hMcm->trace = pCpswMcmInitCfg->traceFxn;
            if (hMcm->trace == NULL)
            {
                hMcm->trace = CpswAppUtils_print;
            }

            CpswMcm_initAttachTable(&hMcm->coreAttachTable);

            memcpy(&hMcm->macPortList[0U], &pCpswMcmInitCfg->macPortList[0U], sizeof(pCpswMcmInitCfg->macPortList));
            CpswMcm_initMbox(hMcm);

            hMcm->isInitDone = true;
        }
    }

    GateMutexPri_leave(hMcm->hMutex, key);

    return status;
}

void  CpswMcm_getCmdIf(Cpsw_Type cpswType,
                       CpswMcm_CmdIf *hMcmCmdIf)
{
    CpswMcm_Handle hMcm = &gMcmObj[cpswType];
    IArg key;

    CpswAppUtils_assert(hMcmCmdIf != NULL);
    CpswAppUtils_assert(hMcm->hMutex != NULL);

    key = GateMutexPri_enter(hMcm->hMutex);
    CpswAppUtils_assert(hMcm->isInitDone == true);

    hMcmCmdIf->hMboxCmd      = hMcm->hMboxCmd;
    hMcmCmdIf->hMboxResponse = hMcm->hMboxResponse;

    GateMutexPri_leave(hMcm->hMutex, key);

    CpswAppUtils_assert(hMcmCmdIf->hMboxCmd != NULL);
    CpswAppUtils_assert(hMcmCmdIf->hMboxResponse != NULL);
}

void CpswMcm_deInit(Cpsw_Type cpswType)
{
    CpswMcm_Handle hMcm = &gMcmObj[cpswType];
    IArg key;
    uint32_t numPendingmsgs;

    CpswAppUtils_assert(hMcm->hMutex != NULL);

    key = GateMutexPri_enter(hMcm->hMutex);

    if (hMcm->isInitDone == true)
    {
        CpswMcm_mailboxObj msg;

        msg.cmd = MCM_SHUTDOWN;

        Mailbox_post(hMcm->hMboxCmd, &msg, BIOS_WAIT_FOREVER);
        Mailbox_pend(hMcm->hMboxResponse, &msg, BIOS_WAIT_FOREVER);
        CpswAppUtils_assert(msg.cmd == MCM_RESPONSE_SHUTDOWN);

        while (Task_getMode(hMcm->task) != Task_Mode_TERMINATED)
        {
            Task_sleep(10);
        }

        if (Task_deleteTerminatedTasks == FALSE)
        {
            Task_delete(&hMcm->task);
        }

        numPendingmsgs = Mailbox_getNumPendingMsgs(hMcm->hMboxCmd);
        while (numPendingmsgs != 0U)
        {
            Mailbox_pend(hMcm->hMboxCmd, &msg, BIOS_WAIT_FOREVER);
            numPendingmsgs = Mailbox_getNumPendingMsgs(hMcm->hMboxCmd);
        }

        Mailbox_delete(&hMcm->hMboxCmd);

        numPendingmsgs = Mailbox_getNumPendingMsgs(hMcm->hMboxResponse);
        while (numPendingmsgs != 0U)
        {
            Mailbox_pend(hMcm->hMboxResponse, &msg, BIOS_WAIT_FOREVER);
            numPendingmsgs = Mailbox_getNumPendingMsgs(hMcm->hMboxResponse);
        }

        Mailbox_delete(&hMcm->hMboxResponse);

        CpswMcm_initAttachTable(&hMcm->coreAttachTable);
        hMcm->isInitDone = false;
    }

    GateMutexPri_leave(hMcm->hMutex, key);
}

static int32_t CpswMcm_enablePorts(CpswMcm_Handle hMcm)
{
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    CpswAle_SetPortStateInArgs setPortStateInArgs;
    Cpsw_Handle hCpsw = hMcm->hCpsw;
    uint32_t coreId   = hMcm->selfCoreId;
    uint8_t i;

    for (i = 0U; i < hMcm->numMacPorts; i++)
    {
        Cpsw_OpenPortLinkInArgs linkArgs;

        hMcm->setPortLinkCfg(&linkArgs, hMcm->macPortList[i]);

        CPSW_IOCTL_SET_IN_ARGS(&prms, &linkArgs);
        status = Cpsw_ioctl(hCpsw,
                            coreId,
                            CPSW_IOCTL_OPEN_PORT_LINK,
                            &prms);
        if (status != CPSW_SOK)
        {
            hMcm->print("CpswMcm_enablePorts() failed to open MAC port: %d\n", status);
        }
    }

    setPortStateInArgs.portNum   = CPSW_ALE_HOST_PORT_NUM;
    setPortStateInArgs.portState = CPSW_ALE_PORTSTATE_FORWARD;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &setPortStateInArgs);
    prms.outArgs = NULL;
    status       = Cpsw_ioctl(hCpsw,
                              coreId,
                              CPSW_ALE_IOCTL_SET_PORT_STATE,
                              &prms);
    if (status != CPSW_SOK)
    {
        hMcm->print("CpswMcm_enablePorts() failed CPSW_ALE_IOCTL_SET_PORT_STATE: %d\n", status);
    }

    if (status == CPSW_SOK)
    {
        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(hCpsw,
                            coreId,
                            CPSW_HOSTPORT_IOCTL_ENABLE,
                            &prms);
        if (status != CPSW_SOK)
        {
            hMcm->print("CpswMcm_enablePorts() Failed to enable host port: %d\n", status);
        }
    }

    /* Show alive PHYs */
    if (status == CPSW_SOK)
    {
        Cpsw_IoctlPrms prms;
        bool alive;
        int8_t i;

        for (i = 0U; i < CPSW_MDIO_MAX_PHY_CNT; i++)
        {
            CPSW_IOCTL_SET_INOUT_ARGS(&prms, &i, &alive);
            status = Cpsw_ioctl(hCpsw,
                                coreId,
                                CPSW_MDIO_IOCTL_GET_ALIVE_STATUS,
                                &prms);
            if (status == CPSW_SOK)
            {
                if (alive == true)
                {
                    hMcm->print("PHY %d is alive\n", i);
                }
            }
            else
            {
                hMcm->print("Failed to get PHY %d alive status: %d\n", i, status);
            }
        }
    }

    return status;
}

static void CpswMcm_timerCb(UArg arg)
{
    Semaphore_Handle timerSem = (Semaphore_Handle)arg;

    /* Tick! */
    Semaphore_post(timerSem);
}

static void CpswMcm_periodicTick(UArg hTimerSem,
                                 UArg mcmHandle)
{
    Semaphore_Handle timerSem = (Semaphore_Handle)hTimerSem;
    CpswMcm_Handle hMcm       = (CpswMcm_Handle)mcmHandle;

    while (hMcm->timerTaskShutDownFlag != true)
    {
        Semaphore_pend(timerSem, BIOS_WAIT_FOREVER);
        /* Cpsw_periodicTick should be called from only task context */
        Cpsw_periodicTick(hMcm->hCpsw);
    }
}

static void CpswMcm_createClock(CpswMcm_Handle hMcm)
{
    Task_Params params;
    Semaphore_Params semParams;
    Clock_Params clkParams;
    Error_Block eb;

    Semaphore_Params_init(&semParams);
    semParams.mode = Semaphore_Mode_COUNTING;
    hMcm->timerSem = Semaphore_create(0, &semParams, NULL);

    Clock_Params_init(&clkParams);
    clkParams.startFlag = FALSE;
    clkParams.period    = hMcm->periodicTaskPeriod;
    clkParams.arg       = (UArg)hMcm->timerSem;

    /* Creating timer and setting timer callback function*/
    hMcm->hTimer = Clock_create((Clock_FuncPtr) & CpswMcm_timerCb,
                                hMcm->periodicTaskPeriod,
                                &clkParams,
                                NULL);
    if (hMcm->hTimer != NULL)
    {
        /* Set timer expiry time in OS ticks */
        Clock_setTimeout(hMcm->hTimer, hMcm->periodicTaskPeriod);
        Clock_setPeriod(hMcm->hTimer, hMcm->periodicTaskPeriod);
        Clock_start(hMcm->hTimer);
        hMcm->timerTaskShutDownFlag = false;
    }
    else
    {
        hMcm->print("CpswMcm_createClock() failed to create clock\n");
    }

    /* Initialize the taskperiodicTick params. Set the task priority higher than the
     * default priority (1) */
    Task_Params_init(&params);
    Error_init(&eb);
    params.priority       = CPSWMCM_PERIODICTSK_PRIORITY;
    params.stack          = NULL;
    params.stackSize      = CPSWMCM_TSK_STACK_MAIN;
    params.arg0           = (UArg)hMcm->timerSem;
    params.arg1           = (UArg)hMcm;
    params.instance->name = "CPSW_PeriodicTickTask";

    hMcm->task_periodicTick = Task_create(CpswMcm_periodicTick, &params, &eb);
    CpswAppUtils_assert(hMcm->task_periodicTick != NULL);
}

static int32_t CpswMcm_open(CpswMcm_Handle hMcm)
{
    int32_t status = CPSW_SOK;
    CpswOsal_Prms osalPrms;
    CpswUtils_Prms utilsPrms;

    if (CPSW_9G == hMcm->cpswType)
    {
        hMcm->print("CPSW_9G Test on MAIN NAVSS\n");
    }
    else if (CPSW_5G == hMcm->cpswType)
    {
        hMcm->print("CPSW_5G Test on MAIN NAVSS\n");
    }
    else
    {
        hMcm->print("CPSW_2G Test on MCU NAVSS\n");
    }

    /* Initialize CPSW driver with default OSAL, utils */
    utilsPrms.printFxn     = hMcm->print;
    utilsPrms.traceFxn     = hMcm->print;
    utilsPrms.phyToVirtFxn = &CpswAppUtils_phyToVirtFxn;
    utilsPrms.virtToPhyFxn = &CpswAppUtils_virtToPhyFxn;

    Cpsw_initOsalPrms(&osalPrms);

    Cpsw_init(hMcm->cpswType, &osalPrms, &utilsPrms);

    status = CpswAppMemUtils_init();
    CpswAppUtils_assert(CPSW_SOK == status);

    hMcm->hUdmaDrv = hMcm->cpswCfg.dmaConfig.hUdmaDrv;

    hMcm->hCpsw = Cpsw_open(hMcm->cpswType, &hMcm->cpswCfg);
    CpswAppUtils_assert(hMcm->hCpsw != NULL);

    /* Enable Host and MAC Ports */
    status = CpswMcm_enablePorts(hMcm);
    CpswAppUtils_assert(CPSW_SOK == status);

    CpswMcm_createClock(hMcm);

    return status;
}

static void CpswMcm_coreAttachHandler(CpswMcm_Handle hMcm,
                                      uint32_t coreId,
                                      Cpsw_AttachCoreOutArgs *pAttachCoreOutArgs)
{
    Cpsw_IoctlPrms prms;
    CpswMcm_CoreAttachTableEntry *entry;
    int32_t status;

    if (NULL != pAttachCoreOutArgs)
    {
        entry = CpswMcm_getCoreAttachEntry(&hMcm->coreAttachTable, coreId);
        if (entry == NULL)
        {
            CPSW_IOCTL_SET_INOUT_ARGS(&prms, &coreId, pAttachCoreOutArgs);
            status = Cpsw_ioctl(hMcm->hCpsw,
                                coreId,
                                CPSW_IOCTL_ATTACH_CORE,
                                &prms);
            if (status != CPSW_SOK)
            {
                hMcm->print("CpswMcm_open failed CPSW_IOCTL_ATTACH_CORE: %d\n", status);
                CpswAppUtils_assert(false);
            }

            CpswMcm_addCoreAttachEntry(&hMcm->coreAttachTable, coreId, pAttachCoreOutArgs);
            entry = CpswMcm_getCoreAttachEntry(&hMcm->coreAttachTable, coreId);
        }

        CpswAppUtils_assert((entry != NULL));
        if (entry != NULL)
        {
            *pAttachCoreOutArgs = entry->coreInfo;
        }
    }
    else
    {
        CpswAppUtils_assert(false);
    }
}

static void CpswMcm_coreDetachHandler(CpswMcm_Handle hMcm,
                                      uint32_t coreId,
                                      uint32_t coreKey)
{
    Cpsw_IoctlPrms prms;
    CpswMcm_CoreAttachTableEntry *entry;

    entry = CpswMcm_getCoreAttachEntry(&hMcm->coreAttachTable, coreId);

    if (entry != NULL)
    {
        int32_t status;

        CPSW_IOCTL_SET_IN_ARGS(&prms, &coreKey);
        status = Cpsw_ioctl(hMcm->hCpsw,
                            coreId,
                            CPSW_IOCTL_DETACH_CORE,
                            &prms);
        if (status != CPSW_SOK)
        {
            hMcm->print("close() failed CPSW_IOCTL_DETACH_CORE: %d\n", status);
            CpswAppUtils_assert(false);
        }

        CpswMcm_delCoreAttachEntry(&hMcm->coreAttachTable, coreId, coreKey);
    }
}

static void CpswMcm_serverTask(UArg McmHandle,
                               UArg a1)
{
    volatile bool isShutdownMcm = false;
    CpswMcm_Handle hMcm         = (CpswMcm_Handle)McmHandle;

    while (!isShutdownMcm)
    {
        int32_t status = CPSW_SOK;
        CpswMcm_mailboxObj msg;

        Mailbox_pend(hMcm->hMboxCmd, &msg, BIOS_WAIT_FOREVER);

        switch (msg.cmd)
        {
            case MCM_GET_HANDLE:
                if (hMcm->refCnt == 0)
                {
                    status = CpswMcm_open(hMcm);
                    CpswAppUtils_assert(CPSW_SOK == status);
                }

                if (status == CPSW_SOK)
                {
                    hMcm->refCnt++;
                }

                msg.cmd                         = MCM_RESPONSE_GET_HANDLE;
                msg.msgBody.handleInfo.hCpsw    = hMcm->hCpsw;
                msg.msgBody.handleInfo.hUdmaDrv = hMcm->hUdmaDrv;
                break;

            case MCM_RELEASE_HANDLE:
                CpswAppUtils_assert(hMcm->refCnt > 0);
                hMcm->refCnt--;

                if (hMcm->refCnt == 0)
                {
                    CpswMcm_close(hMcm);
                }

                msg.cmd = MCM_RESPONSE_RELEASE_HANDLE;
                break;

            case MCM_CORE_ATTACH:
                CpswAppUtils_assert(hMcm->hCpsw != NULL);
                CpswMcm_coreAttachHandler(hMcm, msg.msgBody.coreId, &msg.msgBody.attachInfo);
                msg.cmd = MCM_RESPONSE_CORE_ATTACH;
                break;

            case MCM_CORE_DETACH:
                CpswAppUtils_assert(hMcm->hCpsw != NULL);
                CpswMcm_coreDetachHandler(hMcm, msg.msgBody.coreId, msg.msgBody.coreKey);
                msg.cmd = MCM_RESPONSE_CORE_DETACH;
                break;

            case MCM_SHUTDOWN:
                isShutdownMcm = true;
                msg.cmd       = MCM_RESPONSE_SHUTDOWN;
                break;

            default:
                /* Unahandle MCM command */
                CpswAppUtils_assert(false);
        }

        Mailbox_post(hMcm->hMboxResponse, &msg, BIOS_WAIT_FOREVER);
    }
}

static void CpswMcm_close(CpswMcm_Handle hMcm)
{
    Cpsw_IoctlPrms prms;
    uint8_t i;
    int32_t status;

    /* Disable host port */
    CPSW_IOCTL_SET_NO_ARGS(&prms);
    status = Cpsw_ioctl(hMcm->hCpsw,
                        hMcm->selfCoreId,
                        CPSW_HOSTPORT_IOCTL_DISABLE,
                        &prms);
    if (status != CPSW_SOK)
    {
        hMcm->print("Failed to disable host port: %d\n", status);
    }

    for (i = 0U; i < hMcm->numMacPorts; i++)
    {
        Cpsw_GenericPortLinkInArgs args;
        args.portNum = hMcm->macPortList[i];

        CPSW_IOCTL_SET_IN_ARGS(&prms, &args);
        status = Cpsw_ioctl(hMcm->hCpsw,
                            hMcm->selfCoreId,
                            CPSW_IOCTL_CLOSE_PORT_LINK,
                            &prms);
        if (status != CPSW_SOK)
        {
            hMcm->print("close() failed to close MAC port: %d\n", status);
        }
    }

    hMcm->timerTaskShutDownFlag = true;

    if (hMcm->hTimer != NULL)
    {
        Clock_stop(hMcm->hTimer);
        Clock_delete(&hMcm->hTimer);
    }

    /* Post Timer Sem once to get the Periodic Tick task terminated */
    Semaphore_post(hMcm->timerSem);

    while (Task_getMode(hMcm->task_periodicTick) != Task_Mode_TERMINATED)
    {
        Task_sleep(10);
    }

    if (Task_deleteTerminatedTasks == FALSE)
    {
        Task_delete(&hMcm->task_periodicTick);
    }

    if (hMcm->timerSem != NULL)
    {
        Semaphore_delete(&hMcm->timerSem);
    }

    Cpsw_close(hMcm->hCpsw);

    CpswAppMemUtils_deInit();

    Cpsw_deinit(hMcm->cpswType);

    hMcm->hCpsw = NULL;
}

void CpswMcm_acquireHandleInfo(const CpswMcm_CmdIf *hMcmCmdIf,
                               CpswMcm_HandleInfo *handleInfo)
{
    CpswMcm_mailboxObj msg;

    memset(&msg, 0, sizeof(msg));
    msg.cmd = MCM_GET_HANDLE;
    if ((hMcmCmdIf != NULL) && (handleInfo != NULL))
    {
        Mailbox_post(hMcmCmdIf->hMboxCmd, &msg, BIOS_WAIT_FOREVER);
        Mailbox_pend(hMcmCmdIf->hMboxResponse, &msg, BIOS_WAIT_FOREVER);
        CpswAppUtils_assert(msg.cmd == MCM_RESPONSE_GET_HANDLE);

        *handleInfo = msg.msgBody.handleInfo;
    }
    else
    {
        CpswAppUtils_assert(false);
    }
}

void CpswMcm_releaseHandleInfo(const CpswMcm_CmdIf *hMcmCmdIf)
{
    CpswMcm_mailboxObj msg;

    memset(&msg, 0, sizeof(msg));
    msg.cmd = MCM_RELEASE_HANDLE;
    CpswAppUtils_assert(hMcmCmdIf != NULL);

    Mailbox_post(hMcmCmdIf->hMboxCmd, &msg, BIOS_WAIT_FOREVER);
    Mailbox_pend(hMcmCmdIf->hMboxResponse, &msg, BIOS_WAIT_FOREVER);
    CpswAppUtils_assert(msg.cmd == MCM_RESPONSE_RELEASE_HANDLE);
}

void CpswMcm_coreAttach(const CpswMcm_CmdIf *hMcmCmdIf,
                        uint32_t coreId,
                        Cpsw_AttachCoreOutArgs *attachInfo)
{
    CpswMcm_mailboxObj msg;

    memset(&msg, 0, sizeof(msg));
    if ((hMcmCmdIf != NULL) && (attachInfo != NULL))
    {
        msg.cmd            = MCM_CORE_ATTACH;
        msg.msgBody.coreId = coreId;

        Mailbox_post(hMcmCmdIf->hMboxCmd, &msg, BIOS_WAIT_FOREVER);
        Mailbox_pend(hMcmCmdIf->hMboxResponse, &msg, BIOS_WAIT_FOREVER);
        CpswAppUtils_assert(msg.cmd == MCM_RESPONSE_CORE_ATTACH);
        *attachInfo = msg.msgBody.attachInfo;
    }
    else
    {
        CpswAppUtils_assert(false);
    }
}

void     CpswMcm_coreDetach(const CpswMcm_CmdIf *hMcmCmdIf,
                            uint32_t coreId,
                            uint32_t coreKey)
{
    CpswMcm_mailboxObj msg;

    memset(&msg, 0, sizeof(msg));
    CpswAppUtils_assert(hMcmCmdIf != NULL);
    msg.cmd             = MCM_CORE_DETACH;
    msg.msgBody.coreId  = coreId;
    msg.msgBody.coreKey = coreKey;

    Mailbox_post(hMcmCmdIf->hMboxCmd, &msg, BIOS_WAIT_FOREVER);
    Mailbox_pend(hMcmCmdIf->hMboxResponse, &msg, BIOS_WAIT_FOREVER);
    CpswAppUtils_assert(msg.cmd == MCM_RESPONSE_CORE_DETACH);
}
