/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_loopback.c
 *
 * \brief    This file contains the loopback example implementation.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include <ti/osal/osal.h>
#include <ti/drv/uart/UART_stdio.h>
#include <ti/drv/udma/udma.h>

#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils_cfg.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appboardutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpswapp_ethutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appsoc.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apprm.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils_baremetal.h>

#include <ti/board/board.h>
#if defined(SOC_J721E)
#include <ti/board/src/j721e_evm/include/board_ethernet_config.h>
#elif defined(SOC_J7200)
#include <ti/board/src/j7200_evm/include/board_ethernet_config.h>
#elif defined(SOC_AM65XX)
#include <ti/board/src/am65xx_evm/include/board_ethernet_config.h>
#endif
#include "cpsw_loopback_baremetal.h"
/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define TEST_LEN                    (500U)

#define TEST_NUM_LOOP               (10U)
#define TEST_PTK_NUM                (1000U)
#define APP_ENABLE_STATIC_CFG       (0U)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

typedef struct
{
    /* CPSW instance type */
    Cpsw_Type cpswType;

    /* Core Id */
    uint32_t coreId;
    /*Core Key */
    uint32_t coreKey;

    /* MAC port number */
    Cpsw_MacPort portNum;

    /* MAC mode (defined in board library) */
    emac_mode macMode;

    /* CPSW driver handle */
    Cpsw_Handle hCpsw;

    Udma_DrvHandle hUdmaDrv;

    CpswDma_RxFlowHandle hRxFlow;
    CpswDma_PktInfoQ rxFreeQ;
    CpswDma_PktInfoQ rxReadyQ;

    CpswDma_TxChHandle hTxCh;
    CpswDma_PktInfoQ txFreePktInfoQ;

    uint8_t hostMacAddr[CPSW_MAC_ADDR_LEN];

    /* Whether to test PHY loopback or MAC loopback */
    bool testPhyLoopback;

    /* Whether to print received Ethernet frames */
    bool printFrame;

    uint32_t rxStartFlowIdx;

    uint32_t rxFlowIdx;

    uint32_t txChNum;
} CpswApp_Obj;

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */

void CpswApp_getMacLpbkTestConfig(void);

void CpswApp_getPhyLpbkTestConfig(void);

int32_t  CpswApp_openCpsw(void);

void     CpswApp_closeCpsw(void);

void     CpswApp_showStats(void);

int32_t  CpswApp_openDma(void);

void     CpswApp_closeDma(void);

int32_t  CpswApp_pktRxTx(void);

uint32_t CpswApp_retrieveFreeTxPkts(void);

uint32_t CpswApp_receivePkts(void);

void     CpswApp_changeHostAleEntry(uint8_t macAddr[]);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */
/* Broadcast address */
static uint8_t bcastAddr[CPSW_MAC_ADDR_LEN] =
{0xffU, 0xffU, 0xffU, 0xffU, 0xffU, 0xffU};

/* CPSW configuration */
static CpswApp_Obj gCpswLpbkAppObj =
{
    .testPhyLoopback = false,
    .printFrame      = false,
};

static volatile bool txSem = false, rxSem = false;

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */


void CpswApp_getMacLpbkTestConfig(void)
{
    uint32_t choice;
    bool retry;

    do
    {
        CpswAppUtils_print("Select MAC interface type:\n");
        CpswAppUtils_print("0: RGMII\n");
        CpswAppUtils_print("1: RMII\n");
#if (defined(SOC_J721E) || defined(SOC_J7200)) && defined(BUILD_MCU2_0)
    /* SGMII is supported only on CPSW_9G & CPSW5G */
        CpswAppUtils_print("2: SGMII\n");
#endif

        retry = false;
        choice = CpswAppUtils_getNum();
        switch (choice)
        {
            case 0:
                gCpswLpbkAppObj.macMode = RGMII;
                break;
            case 1:
                gCpswLpbkAppObj.macMode = RMII;
                break;
#if (defined(SOC_J721E) || defined(SOC_J7200)) && defined(BUILD_MCU2_0)
            /* SGMII is supported only on CPSW_9G & CPSW5G */
            case 2:
                gCpswLpbkAppObj.macMode = SGMII;
                break;
#endif
            default:
                CpswAppUtils_print("Wrong interface type, try again...\n\n");
                retry = true;
                break;
        }
    }
    while (retry);

    /* Always use MAC port 0 for MAC loopback */
    gCpswLpbkAppObj.portNum = CPSW_MAC_PORT_0;

    /* For MAC loopback use case, update the mode_sel field of ENET_CTRL register
     * with user's selection */
    CpswAppBoardUtils_setEnetControl(gCpswLpbkAppObj.cpswType,
                                     gCpswLpbkAppObj.portNum,
                                     gCpswLpbkAppObj.macMode);
}

#if defined(BUILD_MCU1_0) || defined(BUILD_MCU2_1) || defined(BUILD_MPU1_0)
void CpswApp_getPhyLpbkTestConfig(void)
{
    uint32_t choice;
    bool retry;

    do
    {
        CpswAppUtils_print("Select CPSW MAC port:\n");
        CpswAppUtils_print("0: CPSW_MAC_PORT_0 - RGMII\n");

        retry = false;
        choice = CpswAppUtils_getNum();
        switch (choice)
        {
            case 0:
                gCpswLpbkAppObj.portNum = CPSW_MAC_PORT_0;
                gCpswLpbkAppObj.macMode = RGMII;
                break;

            default:
                CpswAppUtils_print("Wrong CPSW MAC port, try again...\n");
                retry = true;
                break;
        }
    }
    while (retry);
}
#endif

#if defined(SOC_J7200) && defined(BUILD_MCU2_0)
void CpswApp_getPhyLpbkTestConfig(void)
{
    uint32_t choice;
    bool retry;

    do
    {
        CpswAppUtils_print("Select CPSW MAC port:\n");
        CpswAppUtils_print("1: CPSW_MAC_PORT_1 - RGMII\n");

        retry = false;
        choice = CpswAppUtils_getNum();
        switch (choice)
        {
            case 1:
                gCpswLpbkAppObj.portNum = CPSW_MAC_PORT_1;
                gCpswLpbkAppObj.macMode = RGMII;
                break;

            default:
                CpswAppUtils_print("Wrong CPSW MAC port, try again...\n");
                retry = true;
                break;
        }
    }
    while (retry);
}
#endif

#if defined(SOC_J721E) && defined(BUILD_MCU2_0)
void CpswApp_getPhyLpbkTestConfig(void)
{
    uint32_t choice;
    bool retry;

    do
    {
        CpswAppUtils_print("Select CPSW MAC port:\n");
        CpswAppUtils_print("0: CPSW_MAC_PORT_0 - RGMII\n");
        CpswAppUtils_print("2: CPSW_MAC_PORT_2 - RGMII\n");
        CpswAppUtils_print("3: CPSW_MAC_PORT_3 - RGMII\n");
        CpswAppUtils_print("7: CPSW_MAC_PORT_7 - RMII\n");

        retry = false;
        choice = CpswAppUtils_getNum();
        switch (choice)
        {
            case 0:
                gCpswLpbkAppObj.portNum = CPSW_MAC_PORT_0;
                gCpswLpbkAppObj.macMode = RGMII;
                break;

            case 2:
                gCpswLpbkAppObj.portNum = CPSW_MAC_PORT_2;
                gCpswLpbkAppObj.macMode = RGMII;
                break;

            case 3:
                gCpswLpbkAppObj.portNum = CPSW_MAC_PORT_3;
                gCpswLpbkAppObj.macMode = RGMII;
                break;

            case 7:
                gCpswLpbkAppObj.portNum = CPSW_MAC_PORT_7;
                gCpswLpbkAppObj.macMode = RMII;
                break;

            default:
                CpswAppUtils_print("Wrong CPSW MAC port, try again...\n");
                retry = true;
                break;
        }
    }
    while (retry);
}
#endif

void CpswApp_loopbackTestInitCfg(void)
{
    uint32_t choice;
    bool retry;

#if (1U == APP_ENABLE_STATIC_CFG)
#if !defined(BUILD_MCU2_0)
    gCpswLpbkAppObj.cpswType        = CPSW_2G;
#else
    gCpswLpbkAppObj.cpswType        = CPSW_9G;
#endif
    gCpswLpbkAppObj.portNum         = CPSW_MAC_PORT_0;
    gCpswLpbkAppObj.testPhyLoopback = false;
    gCpswLpbkAppObj.macMode         = RGMII;
#else

    /* Get CPSW type */
    do
    {
        CpswAppUtils_print("Select CPSW type:\n");
#if defined(BUILD_MCU1_0) || defined(BUILD_MCU2_1) || defined(BUILD_MPU1_0)
        CpswAppUtils_print("0: CPSW2G\n");
#endif
#if defined(SOC_J7200) && defined(BUILD_MCU2_0)
        CpswAppUtils_print("1: CPSW5G\n");
#endif
#if defined(SOC_J721E) && defined(BUILD_MCU2_0)
        CpswAppUtils_print("2: CPSW9G\n");
#endif

        retry = false;
        choice = CpswAppUtils_getNum();
        switch (choice)
        {
#if defined(BUILD_MCU1_0) || defined(BUILD_MCU2_1) || defined(BUILD_MPU1_0)
            case 0:
                gCpswLpbkAppObj.cpswType = CPSW_2G;
                break;
#endif
#if defined(SOC_J7200) && defined(BUILD_MCU2_0)
            case 1:
                gCpswLpbkAppObj.cpswType = CPSW_5G;
                break;
#endif
#if defined(SOC_J721E) && defined(BUILD_MCU2_0)
            case 2:
                gCpswLpbkAppObj.cpswType = CPSW_9G;
                break;
#endif
            default:
                CpswAppUtils_print("Wrong CPSW type, try again...\n\n");
                retry = true;
                break;
        }
    }
    while (retry);

    /* Get loopback type */
    do
    {
        CpswAppUtils_print("Select loopback type:\n");
        CpswAppUtils_print("0: Internal (CPSW MAC port)\n");
        CpswAppUtils_print("1: External (PHY)\n");

        retry = false;
        choice = CpswAppUtils_getNum();
        switch (choice)
        {
            case 0:
                gCpswLpbkAppObj.testPhyLoopback = false;
                break;
            case 1:
                gCpswLpbkAppObj.testPhyLoopback = true;
                break;
            default:
                CpswAppUtils_print("Wrong loopback type, try again...\n\n");
                retry = true;
                break;
        }
    }
    while (retry);

    if (gCpswLpbkAppObj.testPhyLoopback)
    {
        CpswApp_getPhyLpbkTestConfig();
    }
    else
    {
        CpswApp_getMacLpbkTestConfig();
    }
#endif
}

int32_t CpswApp_loopbackTest(uint32_t iteration)
{
    Cpsw_IoctlPrms prms;
    CpswAle_SetPortStateInArgs setPortStateInArgs;
    bool alive;
    int8_t i;
    int32_t status;

    CpswAppUtils_enableClocks(gCpswLpbkAppObj.cpswType);

    CpswAppUtils_print("=================================\n");
    CpswAppUtils_print("Cpsw Loopback app: Iteration %d\n", iteration);
    CpswAppUtils_print("=================================\n");

    gCpswLpbkAppObj.coreId = CpswAppSoc_getCoreId();

    /* Open the CPSW */
    status = CpswApp_openCpsw();
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("Failed to open CPSW: %d\n", status);
    }

    /* Open CPSW DMA driver */
    if (status == CPSW_SOK)
    {
        status = CpswApp_openDma();
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("Failed to open CPSW DMA: %d\n", status);
        }
    }

    /* Enable host port */
    if (status == CPSW_SOK)
    {
        setPortStateInArgs.portNum   = CPSW_ALE_HOST_PORT_NUM;
        setPortStateInArgs.portState = CPSW_ALE_PORTSTATE_FORWARD;
        CPSW_IOCTL_SET_IN_ARGS(&prms, &setPortStateInArgs);
        prms.outArgs = NULL;
        status       = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                                  gCpswLpbkAppObj.coreId,
                                  CPSW_ALE_IOCTL_SET_PORT_STATE,
                                  &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswApp_openCpsw() failed CPSW_ALE_IOCTL_SET_PORT_STATE: %d\n", status);
        }

        if (status == CPSW_SOK)
        {
            CPSW_IOCTL_SET_NO_ARGS(&prms);
            status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                                gCpswLpbkAppObj.coreId,
                                CPSW_HOSTPORT_IOCTL_ENABLE,
                                &prms);
            if (status != CPSW_SOK)
            {
                CpswAppUtils_print("Failed to enable host port: %d\n", status);
            }
        }
    }

    /* Show alive PHYs */
    if (status == CPSW_SOK)
    {
        for (i = 0U; i < CPSW_MDIO_MAX_PHY_CNT; i++)
        {
            CPSW_IOCTL_SET_INOUT_ARGS(&prms, &i, &alive);
            status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                                gCpswLpbkAppObj.coreId,
                                CPSW_MDIO_IOCTL_GET_ALIVE_STATUS,
                                &prms);
            if (status == CPSW_SOK)
            {
                if (alive == true)
                {
                    CpswAppUtils_print("PHY %d is alive\n", i);
                }
            }
            else
            {
                CpswAppUtils_print("Failed to get PHY %d alive status: %d\n", i,
                                   status);
            }
        }
    }

    if (status == CPSW_SOK)
    {
        if (gCpswLpbkAppObj.testPhyLoopback)
        {
            while (!CpswAppUtils_isPortLinkUp(gCpswLpbkAppObj.hCpsw,
                                              gCpswLpbkAppObj.coreId,
                                              gCpswLpbkAppObj.portNum))
            {
                CpswAppUtils_wait(100);
                /* Cpsw_periodicTick should be called from non-ISR context.
                 * Calling Cpsw_periodicTick at regular intervals for port link detect before
                 * starting packet RX/TX */
                Cpsw_periodicTick(gCpswLpbkAppObj.hCpsw);
            }
        }

        status = CpswApp_pktRxTx();
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("Failed to enable host port: %d\n", status);
        }
    }

    /* Print CPSW statistics of all ports */
    if (status == CPSW_SOK)
    {
        CpswApp_showStats();
    }

    /* Disable host port */
    if (status == CPSW_SOK)
    {
        CPSW_IOCTL_SET_NO_ARGS(&prms);
        status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                            gCpswLpbkAppObj.coreId,
                            CPSW_HOSTPORT_IOCTL_DISABLE,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("Failed to disable host port: %d\n", status);
        }
    }

    if (status == CPSW_SOK)
    {
        /* Print RX flow DMA statistics */
        CpswAppUtils_showRxFlowStats(gCpswLpbkAppObj.hRxFlow);
        /* Print TX channel DMA statistics */
        CpswAppUtils_showTxChStats(gCpswLpbkAppObj.hTxCh);
        /* Close CPSW DMA driver */
        CpswApp_closeDma();
        /* Close CPSW */
        CpswApp_closeCpsw();
        /* Close UDMA driver */
        CpswAppUtils_udmaclose(gCpswLpbkAppObj.hUdmaDrv);
    }

    CpswAppUtils_disableClocks(gCpswLpbkAppObj.cpswType);

    return status;
}

void CpswApp_initAleConfig(CpswAle_Config *aleConfig)
{
    aleConfig->modeFlags =
        (CPSW_ALE_CONFIG_MASK_ALE_MODULE_ENABLE);
    aleConfig->agingConfig.enableAutoAging           = TRUE;
    aleConfig->agingConfig.agingPeriodInMs           = 1000;
    aleConfig->nwSecCfg.enableVid0Mode               = TRUE;
    aleConfig->vlanConfig.aleVlanAwareMode           = FALSE;
    aleConfig->vlanConfig.cpswVlanAwareMode          = FALSE;
    aleConfig->vlanConfig.unknownUnregMcastFloodMask = CPSW_ALE_ALL_PORTS_MASK;
    aleConfig->vlanConfig.unknownRegMcastFloodMask   = CPSW_ALE_ALL_PORTS_MASK;
    aleConfig->vlanConfig.unknownVlanMemberListMask  = CPSW_ALE_ALL_PORTS_MASK;
}

#if defined(BUILD_MCU1_0) || defined(BUILD_MCU2_1) || defined(BUILD_MPU1_0)
static int32_t CpswApp_setPhyConfig(Cpsw_Type cpswType,
                                    Cpsw_MacPort portNum,
                                    CpswMacPort_Config *macConfig,
                                    CpswMacPort_Interface *interface,
                                    CpswPhy_Config *phyConfig)
{
    int32_t status = CPSW_SOK;

    switch (portNum)
    {
        case CPSW_MAC_PORT_0:
            CpswAppBoardUtils_setPhyConfigRgmii(gCpswLpbkAppObj.cpswType,
                                                gCpswLpbkAppObj.portNum,
                                                macConfig,
                                                interface,
                                                phyConfig);
            break;

        default:
            CpswAppUtils_print("PHY loopback not supported in selected interface\n");
            status = CPSW_EINVALID_PARAMS;
            break;
    }

    return status;
}
#endif

#if defined(SOC_J7200) && defined(BUILD_MCU2_0)
static int32_t CpswApp_setPhyConfig(Cpsw_Type cpswType,
                                    Cpsw_MacPort portNum,
                                    CpswMacPort_Config *macConfig,
                                    CpswMacPort_Interface *interface,
                                    CpswPhy_Config *phyConfig)
{
    int32_t status = CPSW_SOK;

    switch (portNum)
    {
        case CPSW_MAC_PORT_1:
            CpswAppBoardUtils_setPhyConfigRgmii(gCpswLpbkAppObj.cpswType,
                                                gCpswLpbkAppObj.portNum,
                                                macConfig,
                                                interface,
                                                phyConfig);

            Board_pinmuxUpdate(gJ7200_MainPinmuxDataCpsw,
                               BOARD_SOC_DOMAIN_MAIN);
            Board_pinmuxUpdate(gJ7200_WkupPinmuxDataCpsw,
                               BOARD_SOC_DOMAIN_MAIN);
            break;

        default:
            CpswAppUtils_print("PHY loopback not supported in selected interface\n");
            status = CPSW_EINVALID_PARAMS;
            break;
    }

    return status;
}
#endif


#if defined(SOC_J721E) && defined(BUILD_MCU2_0)
static int32_t CpswApp_setPhyConfig(Cpsw_Type cpswType,
                                    Cpsw_MacPort portNum,
                                    CpswMacPort_Config *macConfig,
                                    CpswMacPort_Interface *interface,
                                    CpswPhy_Config *phyConfig)
{
    int32_t status = CPSW_SOK;

    switch (portNum)
    {
        case CPSW_MAC_PORT_0:
        case CPSW_MAC_PORT_2:
        case CPSW_MAC_PORT_3:
            CpswAppBoardUtils_setPhyConfigRgmii(gCpswLpbkAppObj.cpswType,
                                                gCpswLpbkAppObj.portNum,
                                                macConfig,
                                                interface,
                                                phyConfig);

            Board_pinmuxUpdate(gJ721E_MainPinmuxDataGesiCpsw9gQsgmii,
                               BOARD_SOC_DOMAIN_MAIN);
            break;

        case CPSW_MAC_PORT_7:
            CpswAppBoardUtils_setPhyConfigRmii(gCpswLpbkAppObj.portNum,
                                               macConfig,
                                               interface,
                                               phyConfig);

            Board_pinmuxUpdate(gJ721E_MainPinmuxDataGesiCpsw9g,
                               BOARD_SOC_DOMAIN_MAIN);
            break;

        default:
            CpswAppUtils_print("PHY loopback not supported in selected interface\n");
            status = CPSW_EINVALID_PARAMS;
            break;
    }

    return status;
}
#endif

int32_t CpswApp_openCpsw(void)
{
    Cpsw_Config cpswCfg;
    CpswOsal_Prms osalPrms;
    CpswUtils_Prms utilsPrms;
    int32_t status = CPSW_SOK;
    Cpsw_IoctlPrms prms;
    uint32_t coreId;
    Cpsw_AttachCoreOutArgs attachCoreOutArgs;

    /* Set configuration parameters */
    Cpsw_initParams(&cpswCfg);
    cpswCfg.vlanConfig.vlanAware          = false;
    cpswCfg.hostPortConfig.removeCrc      = true;
    cpswCfg.hostPortConfig.padShortPacket = true;
    cpswCfg.hostPortConfig.passCrcErrors  = true;

    /* Time stampping and MAC loopback are not supported together because of
     * IP limitation, so disabling timestampping for this application
     */
    cpswCfg.cptsConfig.hostRxTsEnable     = false;
    CpswAppUtils_initResourceConfig(gCpswLpbkAppObj.cpswType,
                                    gCpswLpbkAppObj.coreId,
                                    &cpswCfg.resourceConfig);
    CpswApp_initAleConfig(&cpswCfg.aleConfig);

    if (gCpswLpbkAppObj.cpswType == CPSW_9G)
    {
        CpswAppUtils_print("CPSW_9G Test on MAIN NAVSS\n");
    }
    else if (gCpswLpbkAppObj.cpswType == CPSW_5G)
    {
        CpswAppUtils_print("CPSW_5G Test on MAIN NAVSS\n");
    }
    else if (gCpswLpbkAppObj.cpswType == CPSW_2G)
    {
        CpswAppUtils_print("CPSW_2G Test on MCU NAVSS\n");
    }

    cpswCfg.dmaConfig.rxChInitPrms.dmaPriority = UDMA_DEFAULT_RX_CH_DMA_PRIORITY;

    /* App should open UDMA first as UDMA handle is needed to initialize
     * CPSW RX channel */
    gCpswLpbkAppObj.hUdmaDrv = CpswAppUtils_udmaOpen(gCpswLpbkAppObj.cpswType, NULL);
    CpswAppUtils_assert(NULL != gCpswLpbkAppObj.hUdmaDrv);

    cpswCfg.dmaConfig.hUdmaDrv = gCpswLpbkAppObj.hUdmaDrv;

    /* Initialize CPSW driver with default OSAL, utils */
    utilsPrms.printFxn     = UART_printf;
    utilsPrms.traceFxn     = UART_printf;
    utilsPrms.phyToVirtFxn = &CpswAppUtils_phyToVirtFxn;
    utilsPrms.virtToPhyFxn = &CpswAppUtils_virtToPhyFxn;

    Cpsw_initOsalPrms(&osalPrms);

    Cpsw_init(gCpswLpbkAppObj.cpswType, &osalPrms, &utilsPrms);

    /* Open the CPSW driver */
    gCpswLpbkAppObj.hCpsw = Cpsw_open(gCpswLpbkAppObj.cpswType, &cpswCfg);
    if (gCpswLpbkAppObj.hCpsw == NULL)
    {
        CpswAppUtils_print("CpswApp_openCpsw() failed to open: %d\n", status);
        status = CPSW_EFAIL;
    }

    if (status == CPSW_SOK)
    {
        /* Attach the core with RM */
        coreId = gCpswLpbkAppObj.coreId;

        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &coreId, &attachCoreOutArgs);
        status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                            gCpswLpbkAppObj.coreId,
                            CPSW_IOCTL_ATTACH_CORE,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswApp_loopbackTest failed CPSW_IOCTL_ATTACH_CORE: %d\n", status);
        }
        else
        {
            gCpswLpbkAppObj.coreKey = attachCoreOutArgs.coreKey;
        }
    }

    if (status == CPSW_SOK)
    {
        /* memutils open should happen after Cpsw is opened as it uses CpswUtils_Q
         * functions */
        status = CpswAppMemUtils_init();
        CpswAppUtils_assert(CPSW_SOK == status);
    }

    if (status == CPSW_SOK)
    {
        Cpsw_OpenPortLinkInArgs linkArgs;
        CpswMacPort_Config *macConfig      = &linkArgs.macConfig;
        CpswMacPort_Interface *interface   = &linkArgs.interface;
        CpswMacPort_LinkConfig *linkConfig = &linkArgs.linkConfig;
        CpswPhy_Config *phyConfig          = &linkArgs.phyConfig;

        linkArgs.portNum = gCpswLpbkAppObj.portNum;

        Cpsw_initMacPortParams(macConfig);

        if (gCpswLpbkAppObj.testPhyLoopback)
        {
            status = CpswApp_setPhyConfig(gCpswLpbkAppObj.cpswType,
                                          gCpswLpbkAppObj.portNum,
                                          macConfig,
                                          interface,
                                          phyConfig);
        }
        else
        {
            switch (gCpswLpbkAppObj.macMode)
            {
                case RMII:
                    CpswAppUtils_setNoPhyCfgRmii(interface, phyConfig);
                    break;

                case RGMII:
                    CpswAppUtils_setNoPhyCfgRgmii(interface, phyConfig);
                    break;

#if !defined(SOC_AM65XX)
                case SGMII:
                    CpswAppUtils_setNoPhyCfgSgmii(interface, macConfig, phyConfig);
                    break;
#endif

                default:
                    CpswAppUtils_print("MAC loopback not supported in selected interface\n");
                    status = CPSW_EINVALID_PARAMS;
                    break;
            }
        }

        CpswAppBoardUtils_setLpbkCfg(gCpswLpbkAppObj.testPhyLoopback,
                                     gCpswLpbkAppObj.portNum,
                                     interface,
                                     macConfig,
                                     phyConfig,
                                     linkConfig);

#if (defined(SOC_J721E) || defined(SOC_J7200)) && defined(BUILD_MCU2_0)
        /* CPSW5G/CPSW9G supported only on MCU2_0 */
        /* Overwrite the ENET control set by board lib */
        if (status == CPSW_SOK)
        {
            CpswAppBoardUtils_setEnetControl(gCpswLpbkAppObj.cpswType,
                                             gCpswLpbkAppObj.portNum,
                                             gCpswLpbkAppObj.macMode);
        }
#endif

        CPSW_IOCTL_SET_IN_ARGS(&prms, &linkArgs);
        status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                            gCpswLpbkAppObj.coreId,
                            CPSW_IOCTL_OPEN_PORT_LINK,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("CpswApp_openCpsw() failed to open MAC port: %d\n",
                               status);
        }
    }

    return status;
}

void CpswApp_closeCpsw(void)
{
    Cpsw_IoctlPrms prms;
    uint32_t coreKey;
    Cpsw_GenericPortLinkInArgs args =
    {
        .portNum = gCpswLpbkAppObj.portNum,
    };
    int32_t status;

    coreKey = gCpswLpbkAppObj.coreKey;
    CPSW_IOCTL_SET_IN_ARGS(&prms, &coreKey);

    status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                        gCpswLpbkAppObj.coreId,
                        CPSW_IOCTL_DETACH_CORE,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_closeCpsw() failed CPSW_IOCTL_DETACH_CORE: %d\n", status);
    }

    CPSW_IOCTL_SET_IN_ARGS(&prms, &args);
    status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                        gCpswLpbkAppObj.coreId,
                        CPSW_IOCTL_CLOSE_PORT_LINK,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("close() failed to close MAC port: %d\n", status);
    }

    Cpsw_close(gCpswLpbkAppObj.hCpsw);

    Cpsw_deinit(gCpswLpbkAppObj.cpswType);
}

void CpswApp_showStats(void)
{
    Cpsw_IoctlPrms prms;
    CpswStats_GenericMacPortInArgs inArgs;
    CpswStats_PortStats portStats;
    int32_t status = CPSW_SOK;

    CPSW_IOCTL_SET_OUT_ARGS(&prms, &portStats);
    status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                        gCpswLpbkAppObj.coreId,
                        CPSW_STATS_IOCTL_GET_HOSTPORT_STATS,
                        &prms);
    if (status == CPSW_SOK)
    {
        CpswAppUtils_print("\n Port 0 Statistics\n");
        CpswAppUtils_print("-----------------------------------------\n");
        CpswAppUtils_printHostPortStats2G((CpswStats_HostPort_2g *)&portStats);
        CpswAppUtils_print("\n");
    }
    else
    {
        CpswAppUtils_print("CpswApp_showStats() failed to get host stats: %d\n",
                           status);
    }

    if (status == CPSW_SOK)
    {
        inArgs.portNum = gCpswLpbkAppObj.portNum;
        CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, &portStats);
        status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                            gCpswLpbkAppObj.coreId,
                            CPSW_STATS_IOCTL_GET_MACPORT_STATS,
                            &prms);
        if (status == CPSW_SOK)
        {
            CpswAppUtils_print("\n Port 1 Statistics\n");
            CpswAppUtils_print("-----------------------------------------\n");
            CpswAppUtils_printMacPortStats2G((CpswStats_MacPort_2g *)&portStats);
            CpswAppUtils_print("\n");
        }
        else
        {
            CpswAppUtils_print("CpswApp_showStats() failed to get MAC stats: %d\n",
                               status);
        }
    }
}

void CpswApp_rxIsrFxn(void *appData)
{
    rxSem = true;
}

void CpswApp_txIsrFxn(void *appData)
{
    txSem = true;
}

void CpswApp_initTxFreePktQ(void)
{
    CpswDma_PktInfo *pPktInfo;
    uint32_t i;

    /* Initialize all queues */
    CpswUtils_initQ(&gCpswLpbkAppObj.txFreePktInfoQ);

    /* Initialize TX EthPkts and queue them to txFreePktInfoQ */
    for (i = 0U; i < CPSW_APPMEMUTILS_NUM_TX_PKTS; i++)
    {
        pPktInfo = CpswAppMemUtils_allocEthPktFxn(&gCpswLpbkAppObj,
                                                  CPSW_APPMEMUTILS_LARGE_POOL_PKT_SIZE,
                                                  UDMA_CACHELINE_ALIGNMENT);
        CpswAppUtils_assert(pPktInfo != NULL);
        CPSW_UTILS_SET_PKT_APP_STATE(&pPktInfo->pktState, CPSW_PKTSTATE_APP_WITH_FREEQ);

        CpswUtils_enQ(&gCpswLpbkAppObj.txFreePktInfoQ, &pPktInfo->node);
    }

    CpswAppUtils_print("initQs() txFreePktInfoQ initialized with %d pkts\n",
                       CpswUtils_getQCount(&gCpswLpbkAppObj.txFreePktInfoQ));
}

void CpswApp_initRxReadyPktQ(void)
{
    CpswDma_PktInfoQ rxReadyQ;
    CpswDma_PktInfo *pPktInfo;
    int32_t status;
    uint32_t i;

    CpswUtils_initQ(&gCpswLpbkAppObj.rxFreeQ);
    CpswUtils_initQ(&gCpswLpbkAppObj.rxReadyQ);
    CpswUtils_initQ(&rxReadyQ);

    for (i = 0U; i < CPSW_APPMEMUTILS_NUM_RX_PKTS; i++)
    {
        pPktInfo = CpswAppMemUtils_allocEthPktFxn(&gCpswLpbkAppObj,
                                                  CPSW_APPMEMUTILS_LARGE_POOL_PKT_SIZE,
                                                  UDMA_CACHELINE_ALIGNMENT);
        CpswAppUtils_assert(pPktInfo != NULL);
        CPSW_UTILS_SET_PKT_APP_STATE(&pPktInfo->pktState, CPSW_PKTSTATE_APP_WITH_FREEQ);
        CpswUtils_enQ(&gCpswLpbkAppObj.rxFreeQ, &pPktInfo->node);
    }

    /* Retrieve any CPSW packets which are ready */
    status = CpswDma_retrieveRxPackets(gCpswLpbkAppObj.hRxFlow, &rxReadyQ);
    CpswAppUtils_assert(status == CPSW_SOK);
    /* There should not be any packet with DMA during init */
    CpswAppUtils_assert(CpswUtils_getQCount(&rxReadyQ) == 0U);

    CpswAppUtils_validatePacketState(&gCpswLpbkAppObj.rxFreeQ,
                                     CPSW_PKTSTATE_APP_WITH_FREEQ,
                                     CPSW_PKTSTATE_APP_WITH_DRIVER);

    CpswAppUtils_submitRxPackets(gCpswLpbkAppObj.hRxFlow,
                                 &gCpswLpbkAppObj.rxFreeQ);

    /* Assert here as during init no. of DMA descriptors should be equal to
     * no. of free Ethernet buffers available with app */

    CpswAppUtils_assert(0U == CpswUtils_getQCount(&gCpswLpbkAppObj.rxFreeQ));
}

int32_t CpswApp_openDma(void)
{
    int32_t status = CPSW_SOK;
    CpswDma_OpenRxFlowPrms cpswRxFlowCfg;
    CpswDma_OpenTxChPrms cpswTxChCfg;

    /* Open the CPSW TX channel  */
    if (status == CPSW_SOK)
    {
        CpswDma_initTxChParams(&cpswTxChCfg);

        cpswTxChCfg.hUdmaDrv = gCpswLpbkAppObj.hUdmaDrv;
        cpswTxChCfg.hCbArg   = &gCpswLpbkAppObj;
        cpswTxChCfg.notifyCb = CpswApp_txIsrFxn;

        CpswAppUtils_setCommonTxChPrms(&cpswTxChCfg);

        CpswAppUtils_openTxCh(gCpswLpbkAppObj.hCpsw,
                              gCpswLpbkAppObj.coreKey,
                              gCpswLpbkAppObj.coreId,
                              &gCpswLpbkAppObj.txChNum,
                              &gCpswLpbkAppObj.hTxCh,
                              &cpswTxChCfg);

        CpswApp_initTxFreePktQ();

        if (NULL != gCpswLpbkAppObj.hTxCh)
        {
            status = CpswDma_enableTxEvent(gCpswLpbkAppObj.hTxCh);
            if (CPSW_SOK != status)
            {
                /* Free the Ch Num if enable event failed */
                CpswAppUtils_freeTxCh(gCpswLpbkAppObj.hCpsw,
                                      gCpswLpbkAppObj.coreKey,
                                      gCpswLpbkAppObj.coreId,
                                      gCpswLpbkAppObj.txChNum);
                CpswAppUtils_print("CpswDma_startTxCh() failed: %d\n", status);
                status = CPSW_EFAIL;
            }
        }
        else
        {
            /* Free the Ch Num if open Tx Ch failed */
            CpswAppUtils_freeTxCh(gCpswLpbkAppObj.hCpsw,
                                  gCpswLpbkAppObj.coreKey,
                                  gCpswLpbkAppObj.coreId,
                                  gCpswLpbkAppObj.txChNum);
            CpswAppUtils_print("CpswDma_openTxCh() failed to open: %d\n",
                               status);
            status = CPSW_EFAIL;
        }
    }

    /* Open the CPSW RX flow  */
    if (status == CPSW_SOK)
    {
        CpswDma_initRxFlowParams(&cpswRxFlowCfg);
        cpswRxFlowCfg.notifyCb = CpswApp_rxIsrFxn;
        cpswRxFlowCfg.hUdmaDrv = gCpswLpbkAppObj.hUdmaDrv;
        cpswRxFlowCfg.hCbArg   = &gCpswLpbkAppObj;

        CpswAppUtils_setCommonRxFlowPrms(&cpswRxFlowCfg);

        CpswAppUtils_openRxFlow(gCpswLpbkAppObj.hCpsw,
                                gCpswLpbkAppObj.coreKey,
                                gCpswLpbkAppObj.coreId,
                                true,
                                &gCpswLpbkAppObj.rxStartFlowIdx,
                                &gCpswLpbkAppObj.rxFlowIdx,
                                &gCpswLpbkAppObj.hostMacAddr[0U],
                                &gCpswLpbkAppObj.hRxFlow,
                                &cpswRxFlowCfg);

        if (NULL == gCpswLpbkAppObj.hRxFlow)
        {
            CpswAppUtils_freeRxFlow(gCpswLpbkAppObj.hCpsw,
                                    gCpswLpbkAppObj.coreKey,
                                    gCpswLpbkAppObj.coreId,
                                    gCpswLpbkAppObj.rxFlowIdx);
            CpswAppUtils_print("CpswDma_openRxFlow() failed to open: %d\n",
                               status);
            CpswAppUtils_assert(NULL != gCpswLpbkAppObj.hRxFlow);
        }
        else
        {
            CpswAppUtils_print("Host MAC address: ");
            CpswAppUtils_printMacAddr(gCpswLpbkAppObj.hostMacAddr);
            /* For MAC loopback disable secure flag for the host port entry */
            CpswApp_changeHostAleEntry(&gCpswLpbkAppObj.hostMacAddr[0U]);

            /* Submit all ready RX buffers to DMA.*/
            CpswApp_initRxReadyPktQ();
        }
    }

    return status;
}

void CpswApp_closeDma(void)
{
    CpswDma_PktInfoQ fqPktInfoQ;
    CpswDma_PktInfoQ cqPktInfoQ;

    CpswUtils_initQ(&fqPktInfoQ);
    CpswUtils_initQ(&cqPktInfoQ);

    /* There should not be any ready packet */
    CpswAppUtils_assert(0U == CpswUtils_getQCount(&gCpswLpbkAppObj.rxReadyQ));

    /* Close RX Flow */
    CpswAppUtils_closeRxFlow(gCpswLpbkAppObj.hCpsw,
                             gCpswLpbkAppObj.coreKey,
                             gCpswLpbkAppObj.coreId,
                             true,
                             &fqPktInfoQ,
                             &cqPktInfoQ,
                             gCpswLpbkAppObj.rxStartFlowIdx,
                             gCpswLpbkAppObj.rxFlowIdx,
                             gCpswLpbkAppObj.hostMacAddr,
                             gCpswLpbkAppObj.hRxFlow);
    CpswAppUtils_freePktInfoQ(&fqPktInfoQ);
    CpswAppUtils_freePktInfoQ(&cqPktInfoQ);

    /* Close TX channel */
    CpswUtils_initQ(&fqPktInfoQ);
    CpswUtils_initQ(&cqPktInfoQ);

    CpswAppUtils_closeTxCh(gCpswLpbkAppObj.hCpsw,
                           gCpswLpbkAppObj.coreKey,
                           gCpswLpbkAppObj.coreId,
                           &fqPktInfoQ,
                           &cqPktInfoQ,
                           gCpswLpbkAppObj.hTxCh,
                           gCpswLpbkAppObj.txChNum);
    CpswAppUtils_freePktInfoQ(&fqPktInfoQ);
    CpswAppUtils_freePktInfoQ(&cqPktInfoQ);

    CpswAppUtils_freePktInfoQ(&gCpswLpbkAppObj.rxFreeQ);
    CpswAppUtils_freePktInfoQ(&gCpswLpbkAppObj.txFreePktInfoQ);

    CpswAppMemUtils_deInit();
}

void CpswApp_changeHostAleEntry(uint8_t macAddr[])
{
    int32_t status;
    Cpsw_IoctlPrms prms;
    CpswAle_AddEntryOutArgs setUcastOutArgs;
    CpswAle_SetUcastEntryInArgs setUcastInArgs;

    memcpy(&setUcastInArgs.addr.addr[0U], macAddr, sizeof(setUcastInArgs.addr.addr));
    setUcastInArgs.addr.vlanId  = 0U;
    setUcastInArgs.info.portNum = 0U;
    setUcastInArgs.info.trunk   = false;
    setUcastInArgs.info.blocked = false;
    setUcastInArgs.info.secure  = false;
    setUcastInArgs.info.super   = 0U;
    setUcastInArgs.info.ageable = false;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &setUcastInArgs, &setUcastOutArgs);

    status = Cpsw_ioctl(gCpswLpbkAppObj.hCpsw,
                        gCpswLpbkAppObj.coreId,
                        CPSW_ALE_IOCTL_ADD_UNICAST,
                        &prms);
    if (status != CPSW_SOK)
    {
        CpswAppUtils_print("CpswApp_openCpsw() failed CPSW_ALE_IOCTL_SET_PORT_STATE: %d\n", status);
    }
}

int32_t CpswApp_pktRxTx(void)
{
    CpswDma_PktInfoQ txSubmitQ;
    CpswDma_PktInfo *pktInfo;
    EthFrame *frame;
    uint32_t txRetrievePktCnt, loopTxPktCnt, totalTxCnt;
    uint32_t rxReadyCnt;
    uint32_t loopCnt, pktCnt, loopRxPktCnt;
    int32_t status = CPSW_SOK;

    totalTxCnt = 0U;
    for (loopCnt = 0U; loopCnt < TEST_NUM_LOOP; loopCnt++)
    {
        pktCnt = 0U;
        while (pktCnt < TEST_PTK_NUM)
        {
            loopRxPktCnt = loopTxPktCnt = 0U;
            /* Transmit a single packet */
            CpswUtils_initQ(&txSubmitQ);

            /* Dequeue one free TX Eth packet */
            pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&gCpswLpbkAppObj.txFreePktInfoQ);

            while (NULL != pktInfo)
            {
                pktCnt++;
                /* Fill the TX Eth frame with test content */
                frame = (EthFrame *)pktInfo->bufPtr;
                memcpy(frame->hdr.dstMac, bcastAddr, CPSW_MAC_ADDR_LEN);
                memcpy(frame->hdr.srcMac, &gCpswLpbkAppObj.hostMacAddr[0U], CPSW_MAC_ADDR_LEN);
                frame->hdr.etherType = htons(ETHERTYPE_EXPERIMENTAL1);
                memset(&frame->payload[0U], (uint8_t)(0xA5 + pktCnt), TEST_LEN);
                pktInfo->userBufLen = TEST_LEN + sizeof(EthFrameHeader);
                pktInfo->appPriv    = &gCpswLpbkAppObj;
                CpswUtils_checkPktState(&pktInfo->pktState,
                                        CPSW_PKTSTATE_MODULE_APP,
                                        CPSW_PKTSTATE_APP_WITH_FREEQ,
                                        CPSW_PKTSTATE_APP_WITH_DRIVER);

                /* Enqueue the packet for later transmission */
                CpswUtils_enQ(&txSubmitQ, &pktInfo->node);

                if (pktCnt >= TEST_PTK_NUM)
                {
                    break;
                }

                /* Dequeue one free TX Eth packet */
                pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&gCpswLpbkAppObj.txFreePktInfoQ);
            }

            loopTxPktCnt = CpswUtils_getQCount(&txSubmitQ);
            while (0U != CpswUtils_getQCount(&txSubmitQ))
            {
                uint32_t txCnt = CpswUtils_getQCount(&txSubmitQ);
                status = CpswAppUtils_submitTxPackets(gCpswLpbkAppObj.hTxCh,
                                                      &txSubmitQ);
                while (txSem != true)
                {
                    ;
                }

                txSem = false;

                /* Retrieve TX free packets */
                if (status == CPSW_SOK)
                {
                    txCnt            = txCnt - CpswUtils_getQCount(&txSubmitQ);
                    txRetrievePktCnt = 0U;
                    while (txRetrievePktCnt != txCnt)
                    {
                        /* TODO this is not failure as HW is busy sending packets, we
                         * need to wait and again call retrieve packets */
                        CpswAppUtils_wait(1);
                        txRetrievePktCnt += CpswApp_retrieveFreeTxPkts();
                    }
                }
                else
                {
                    break;
                }
            }

            /* Wait for packet reception */
            do
            {
                while (rxSem != true)
                {
                    ;
                }

                rxSem = false;

                /* Get the packets received so far */
                rxReadyCnt = CpswApp_receivePkts();
                if (rxReadyCnt > 0U)
                {
                    /* Consume the received packets and release them */
                    pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&gCpswLpbkAppObj.rxReadyQ);
                    while (NULL != pktInfo)
                    {
                        CpswUtils_checkPktState(&pktInfo->pktState,
                                                CPSW_PKTSTATE_MODULE_APP,
                                                CPSW_PKTSTATE_APP_WITH_READYQ,
                                                CPSW_PKTSTATE_APP_WITH_FREEQ);

                        frame = (EthFrame *)pktInfo->bufPtr;

                        /* Consume the packet by just printing its content */
                        if (gCpswLpbkAppObj.printFrame)
                        {
                            CpswAppUtils_printFrame(frame,
                                                    pktInfo->userBufLen - sizeof(EthFrameHeader));
                        }

                        /* Release the received packet */
                        CpswUtils_enQ(&gCpswLpbkAppObj.rxFreeQ, &pktInfo->node);
                        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&gCpswLpbkAppObj.rxReadyQ);
                    }

                    /*Submit now processed buffers */
                    if (status == CPSW_SOK)
                    {
                        CpswAppUtils_validatePacketState(&gCpswLpbkAppObj.rxFreeQ,
                                                         CPSW_PKTSTATE_APP_WITH_FREEQ,
                                                         CPSW_PKTSTATE_APP_WITH_DRIVER);

                        CpswAppUtils_submitRxPackets(gCpswLpbkAppObj.hRxFlow,
                                                     &gCpswLpbkAppObj.rxFreeQ);
                    }
                }

                loopRxPktCnt += rxReadyCnt;
            }
            while (loopRxPktCnt < loopTxPktCnt);

            totalTxCnt += loopTxPktCnt;

            CpswAppUtils_print("Transmitted & Received %d packets\n", totalTxCnt);
        }
    }

    if (status == CPSW_SOK)
    {
        CpswAppUtils_print("Transmitted & Received %d packets\n", totalTxCnt);
    }
    else
    {
        CpswAppUtils_print("Failed to transmit/receive packets: %d, transmitted: %d \n", TEST_PTK_NUM, totalTxCnt);
    }

    return status;
}

uint32_t CpswApp_retrieveFreeTxPkts(void)
{
    CpswDma_PktInfoQ txFreeQ;
    CpswDma_PktInfo *pktInfo;
    int32_t status;
    uint32_t txFreeQCnt = 0U;

    CpswUtils_initQ(&txFreeQ);

    /* Retrieve any CPSW packets that may be free now */
    status = CpswDma_retrieveTxDonePackets(gCpswLpbkAppObj.hTxCh, &txFreeQ);
    if (status == CPSW_SOK)
    {
        txFreeQCnt = CpswUtils_getQCount(&txFreeQ);

        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&txFreeQ);
        while (NULL != pktInfo)
        {
            CpswUtils_checkPktState(&pktInfo->pktState,
                                    CPSW_PKTSTATE_MODULE_APP,
                                    CPSW_PKTSTATE_APP_WITH_DRIVER,
                                    CPSW_PKTSTATE_APP_WITH_FREEQ);

            CpswUtils_enQ(&gCpswLpbkAppObj.txFreePktInfoQ, &pktInfo->node);
            pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&txFreeQ);
        }
    }
    else
    {
        CpswAppUtils_print("retrieveFreeTxPkts() failed to retrieve pkts: %d\n",
                           status);
    }

    return txFreeQCnt;
}

uint32_t CpswApp_receivePkts(void)
{
    CpswDma_PktInfoQ rxReadyQ;
    CpswDma_PktInfo *pktInfo;
    int32_t status;
    uint32_t rxReadyCnt = 0U;

    CpswUtils_initQ(&rxReadyQ);

    /* Retrieve any CPSW packets which are ready */
    status = CpswDma_retrieveRxPackets(gCpswLpbkAppObj.hRxFlow, &rxReadyQ);
    if (status == CPSW_SOK)
    {
        rxReadyCnt = CpswUtils_getQCount(&rxReadyQ);

        /* Queue the received packet to rxReadyQ and pass new ones from rxFreeQ
        **/
        pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&rxReadyQ);
        while (pktInfo != NULL)
        {
            CpswUtils_checkPktState(&pktInfo->pktState,
                                    CPSW_PKTSTATE_MODULE_APP,
                                    CPSW_PKTSTATE_APP_WITH_DRIVER,
                                    CPSW_PKTSTATE_APP_WITH_READYQ);

            CpswUtils_enQ(&gCpswLpbkAppObj.rxReadyQ, &pktInfo->node);
            pktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(&rxReadyQ);
        }
    }
    else
    {
        CpswAppUtils_print("receivePkts() failed to retrieve pkts: %d\n",
                           status);
    }

    return rxReadyCnt;
}

Cpsw_Handle CpswApp_getCpswHandle(void)
{
    return gCpswLpbkAppObj.hCpsw;
}

/* end of file */
