/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     main_tirtos.c
 *
 * \brief    Main file for TI-RTOS build.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

/* XDCtools Header files */
#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/System.h>
#include <xdc/runtime/Memory.h>

/* BIOS Header files */
#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Clock.h>
#include <ti/sysbios/knl/Semaphore.h>
#include <ti/sysbios/heaps/HeapMem.h>

#include <ti/drv/sciclient/sciclient.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appboardutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils_rtos.h>
#include "cpsw_loopback_tirtos.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* Test application stack size */
#define APP_TSK_STACK_MAIN                    (10U * 1024U)
#define NUM_LOOPBACK_ITERATION                (5)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* Task function */
static Void taskFxn(UArg a0,
                    UArg a1);

/* CPSW loopback test */
extern Cpsw_Handle CpswApp_getCpswHandle(void);

static CpswApp_ClkHandle CpswApp_createClock(void);

static void CpswApp_deleteClock(void);

static void CpswApp_createTxTask(void);

static void CpswApp_createRxTask(void);

uint32_t Utils_memGetSystemHeapFreeSpace(void);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* Test application stack */
static uint8_t gAppTskStackMain[APP_TSK_STACK_MAIN]
__attribute__ ((aligned(32)));
static uint8_t gAppTskStackTick[APP_TSK_STACK_MAIN]
__attribute__ ((aligned(32)));
static uint8_t gAppTskStackTx[APP_TSK_STACK_MAIN]
__attribute__ ((aligned(32)));
static uint8_t gAppTskStackRx[APP_TSK_STACK_MAIN]
__attribute__ ((aligned(32)));
volatile uint32_t exitFlag = 0U;

static Clock_Handle hTimer = NULL;

static Task_Handle task_periodicTick = NULL;
static Task_Handle task_tx           = NULL;
static Task_Handle task_rx           = NULL;

static Semaphore_Handle timerSem = NULL;

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

int main(void)
{
    Task_Handle task;
    Task_Params params;
    Error_Block eb;

    Error_init(&eb);

    /* Initialize the task params. Set the task priority higher than the
     * default priority (1) */
    Task_Params_init(&params);
    params.priority       = 2U;
    params.stack          = gAppTskStackMain;
    params.stackSize      = sizeof(gAppTskStackMain);
    params.instance->name = "Loopback Main Task";

    task = Task_create(taskFxn, &params, &eb);
    if (task == NULL)
    {
        BIOS_exit(0);
    }

    /* Does not return */
    BIOS_start();

    return 0;
}

static Void taskFxn(UArg a0,
                    UArg a1)
{
    int32_t status;
    CpswApp_ClkHandle clkhandle = NULL;
    uint8_t i;
    uint32_t freeSizeBeforeLpbk, freeSizeAfterLpbk;

    CpswAppBoardUtils_init();

    CpswApp_loopbackTestInitCfg();

    for (i = 0; i < NUM_LOOPBACK_ITERATION; i++)
    {
        freeSizeBeforeLpbk = Utils_memGetSystemHeapFreeSpace();

        /* Note: Clock create/delete must be done per iteration to account for
         * memory allocation(from heap) done in snprintf for code reentrancy.
         * Moving this out will result in heap memory leak error in external
         * loopback mode on A72. */
        clkhandle = CpswApp_createClock();
        CpswAppUtils_assert(clkhandle != NULL);

        /* Run the loopback test */
        status = CpswApp_loopbackTest(clkhandle, i);

        CpswApp_deleteClock();

        freeSizeAfterLpbk = Utils_memGetSystemHeapFreeSpace();

        CpswAppUtils_assert(freeSizeBeforeLpbk == freeSizeAfterLpbk);
    }

    if (status == CPSW_SOK)
    {
        CpswAppUtils_print("Loopback application completed\n");
    }
    else
    {
        CpswAppUtils_print("Loopback application failed to complete\n");
    }

    CpswAppBoardUtils_deInit();
}

Void CpswApp_timerCallback(UArg arg)
{
    Semaphore_Handle timerSem = (Semaphore_Handle)arg;

    /* Tick! */
    Semaphore_post(timerSem);
}

void CpswApp_periodicTick(UArg a0,
                          UArg a1)
{
    Semaphore_Handle timerSem = (Semaphore_Handle)a0;

    while (!exitFlag)
    {
        Semaphore_pend(timerSem, BIOS_WAIT_FOREVER);
        /* Cpsw_periodicTick should be called from non-ISR context */
        Cpsw_periodicTick(CpswApp_getCpswHandle());
    }
}

static CpswApp_ClkHandle CpswApp_createClock(void)
{
    Task_Params params;
    Semaphore_Params semParams;
    Clock_Params clkParams;
    UInt period = 100U; /* msecs */
    Error_Block eb;

    Error_init(&eb);

    Semaphore_Params_init(&semParams);
    semParams.mode = Semaphore_Mode_COUNTING;
    timerSem       = Semaphore_create(0, &semParams, NULL);

    /* Initialize the taskperiodicTick params. Set the task priority higher than the
     * default priority (1) */
    Task_Params_init(&params);
    params.priority       = 7U;
    params.stack          = gAppTskStackTick;
    params.stackSize      = sizeof(gAppTskStackTick);
    params.arg0           = (UArg)timerSem;
    params.instance->name = "Periodic-tick Task";

    task_periodicTick = Task_create(CpswApp_periodicTick, &params, &eb);
    if (task_periodicTick == NULL)
    {
        BIOS_exit(0);
    }

    Clock_Params_init(&clkParams);
    clkParams.startFlag = FALSE;
    clkParams.period    = period;
    clkParams.arg       = (UArg)timerSem;

    /* Creating timer and setting timer callback function*/
    hTimer = Clock_create((Clock_FuncPtr) & CpswApp_timerCallback,
                          period,
                          &clkParams,
                          NULL);
    if (hTimer != NULL)
    {
        /* Set timer expiry time in OS ticks */
        Clock_setTimeout(hTimer, period);
        Clock_setPeriod(hTimer, period);
    }
    else
    {
        CpswAppUtils_print("CpswApp_createClock() failed to create clock: %d\n");
    }

    return (CpswApp_ClkHandle)hTimer;
}

void CpswApp_startClock(CpswApp_ClkHandle handle)
{
    Clock_Handle clkHandle = (Clock_Handle)handle;

    /* Start timer */
    Clock_start(clkHandle);
}

void CpswApp_stopClock(CpswApp_ClkHandle handle)
{
    Clock_Handle clkHandle = (Clock_Handle)handle;

    /* Stop and delete the tick timer */
    Clock_stop(clkHandle);
}

static void CpswApp_deleteClock(void)
{
    if (hTimer != NULL)
    {
        Clock_delete(&hTimer);
    }

    if (timerSem != NULL)
    {
        Semaphore_delete(&timerSem);
    }

    if (task_periodicTick != NULL)
    {
        Task_delete(&task_periodicTick);
    }
}

uint32_t Utils_memGetSystemHeapFreeSpace(void)
{
    Memory_Stats stats;
    uint32_t i;
    HeapMem_Object *obj    = NULL, *nextObj = NULL, *currObj = NULL;
    uint32_t totalFreeSize = 0U, totalFreeSizeStatic = 0U, totalFreeSizeDynamic = 0U;

    for (i = 0U; i < HeapMem_Object_count(); i++)
    {
        obj = HeapMem_Object_get(NULL, i);
        if (NULL != obj)
        {
            HeapMem_getStats(obj, &stats);
            totalFreeSizeStatic += stats.totalFreeSize;
        }
    }

    nextObj = HeapMem_Object_first();
    do
    {
        currObj = nextObj;
        if (NULL != currObj)
        {
            HeapMem_getStats(currObj, &stats);
            totalFreeSizeDynamic += stats.totalFreeSize;

            nextObj = HeapMem_Object_next(currObj);
        }
    }
    while (nextObj != NULL);

    totalFreeSize = totalFreeSizeStatic + totalFreeSizeDynamic;

    return totalFreeSize;
}

static void CpswApp_createTxTask(void)
{
    Task_Params params;
    Error_Block eb;

    Error_init(&eb);

    Task_Params_init(&params);
    params.priority       = 2U;
    params.stack          = gAppTskStackTx;
    params.stackSize      = sizeof(gAppTskStackTx);
    params.instance->name = "Tx Task";

    task_tx = Task_create((ti_sysbios_knl_Task_FuncPtr)CpswApp_txTask, &params, &eb);
    if (task_tx == NULL)
    {
        BIOS_exit(0);
    }
}

static void CpswApp_createRxTask(void)
{
    Task_Params params;
    Error_Block eb;

    Error_init(&eb);

    Task_Params_init(&params);
    params.priority       = 2U;
    params.stack          = gAppTskStackRx;
    params.stackSize      = sizeof(gAppTskStackRx);
    params.instance->name = "Rx Task";

    task_rx = Task_create((ti_sysbios_knl_Task_FuncPtr)CpswApp_rxTask, &params, &eb);
    if (task_rx == NULL)
    {
        BIOS_exit(0);
    }
}

void CpswApp_createRxTxTask(void)
{
    CpswApp_createTxTask();
    CpswApp_createRxTask();
}

void CpswApp_deleteRxTxTask(void)
{
    if (task_tx != NULL)
    {
        CpswAppUtils_assert(Task_getMode(task_tx) == Task_Mode_TERMINATED);
        Task_delete(&task_tx);
    }

    if (task_rx != NULL)
    {
        CpswAppUtils_assert(Task_getMode(task_rx) == Task_Mode_TERMINATED);
        Task_delete(&task_rx);
    }
}
