/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 *  \file main_tirtos.c
 *
 *  \brief Main file for TI-RTOS build
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <stdio.h>

/* XDCtools Header files */
#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/System.h>
#include <ti/drv/sciclient/sciclient.h>

/* BIOS Header files */
#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Clock.h>
#include <ti/sysbios/utils/Load.h>
#include <ti/sysbios/knl/Semaphore.h>

#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/udma/udma.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apputils_rtos.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appmemutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appboardutils.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_mcm.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_appsoc.h>
#include <ti/drv/cpsw/examples/cpsw_apputils/inc/cpsw_apprm.h>

#include <ti/drv/cpsw/nimucpsw/nimu_ndk.h>
#include <ti/drv/cpsw/nimucpsw/ndk2cpsw_appif.h>

/* NDK headers */
#include <ti/ndk/inc/netmain.h>
#include <ti/ndk/inc/stkmain.h>
#include <ti/ndk/inc/socket.h>
#include <ti/ndk/inc/_stack.h>
#include <ti/ndk/inc/tools/servers.h>
#include <ti/ndk/inc/tools/console.h>

/* OSAL Header files */
#include <ti/osal/osal.h>

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* NDK restart test count */
#define CPSWNIMUAPP_NUM_RESTART         (10U)
#define CPSWNIMUAPP_RESTART_PERIOD      (1000U * 10U)

/* Test application stack size */
#define CPSWNIMUAPP_TSK_STACK_MAIN      (2U * 1024U)

/* CPU load task stack size */
#define CPSWNIMUAPP_TSK_STACK_CPU_LOAD  (2U * 1024U)

#define CPSWNIMUAPP_PACKET_POLL_PERIOD_US (500U)

#define APP_ENABLE_STATIC_CFG             (0U)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

typedef struct
{
    /* CPSW instance type */
    Cpsw_Type cpswType;

    /* MAC port number */
    Cpsw_MacPort portNum;

    bool useDefaultRxFlow;
} CpswNimu_AppCfg;

typedef struct
{
    CpswMcm_CmdIf hMcmCmdIf[CPSW_COUNT];
    Udma_DrvHandle hUdmaDrv;

} CpswNimu_AppObj;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

#if (0U == APP_ENABLE_STATIC_CFG)
static void CpswApp_getTestConfig(Cpsw_Type *cpswType,
                                  Cpsw_MacPort *macPort);
#endif

static void CpswApp_setPhyConfig(Cpsw_Type cpswType,
                                 Cpsw_MacPort portNum,
                                 CpswMacPort_Config *macConfig,
                                 CpswMacPort_Interface *interface,
                                 CpswPhy_Config *phyConfig);

/* Task functions */
static void    CpswNimuApp_ndkRestartTask(UArg arg0,
                                          UArg arg1);

static void    CpswNimuApp_timerCallback(UArg arg);

static int32_t CpswNimuApp_init(void);

void           CpswNimuApp_deInit(void);

static void CpswNimuApp_cpuLoadTask(UArg a0,
                                    UArg a1);

/* ========================================================================== */
/*                          Extern variables                                  */
/* ========================================================================== */

extern xdc_Void ti_ndk_config_Global_stackThread(xdc_UArg,
                                                 xdc_UArg);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/*!
 * \brief
 *   NIMUDeviceTable
 *
 * \details
 *  This is the NIMU Device Table for the Platform.
 *  This should be defined for each platform. Since the current platform
 *  has a single network Interface; this has been defined here. If the
 *  platform supports more than one network interface this should be
 *  defined to have a list of "initialization" functions for each of the
 *  interfaces.
 */
NIMU_DEVICE_TABLE_ENTRY NIMUDeviceTable[2U] =
{
    /*!
     * \brief  NIMU_NDK_Init for this network device
     */
    {&NIMU_NDK_init},
    {NULL}
};

static void *hEcho    = 0;
static void *hEchoUdp = 0;
static void *hData    = 0;
static void *hNull    = 0;
static void *hOob     = 0;

/* App string used by Telnet server */
char *VerStr = "NIMU CPSW Example";

static uint8_t gCpswNimuAppTskStackMain[CPSWNIMUAPP_TSK_STACK_MAIN]
__attribute__ ((aligned(32)));
static uint8_t gCpswLoadTskStack[CPSWNIMUAPP_TSK_STACK_CPU_LOAD]
__attribute__ ((aligned(32)));

/* Semaphore to sync the NSP/NDK restart activity */
static Semaphore_Handle gNdkRestartSem = NULL;

/* gNdkRestartCnt needs to be initialized to 1 to avoid compilation error
 * "error: pointless comparison of unsigned integer with zero" when CPSWNIMUAPP_NUM_RESTART is zero */
volatile uint32_t gNdkRestartCnt = 1U;

static CpswNimu_AppCfg gCpswNimuAppCfg =
{
    .useDefaultRxFlow = true,
};

static CpswNimu_AppObj gCpswNimuAppObj =
{
    .hMcmCmdIf =
    {
        [CPSW_2G] = {.hMboxCmd = NULL, .hMboxResponse = NULL},
        [CPSW_9G] = {.hMboxCmd = NULL, .hMboxResponse = NULL},
    },
};

static Task_Handle hNdkRestartTask;
static Task_Handle hCpuLoadTask;

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

int main(void)
{
    Task_Params params;
    Error_Block eb;
    Semaphore_Params semParams;
    Clock_Params clkParams;

    CpswAppBoardUtils_init();

    CpswAppUtils_print("==========================\n");
    CpswAppUtils_print("      CPSW NIMU App       \n");
    CpswAppUtils_print("==========================\n");

#if (1U == APP_ENABLE_STATIC_CFG)
#if !defined(BUILD_MCU2_0)
    gCpswNimuAppCfg.cpswType = CPSW_2G;
    gCpswNimuAppCfg.portNum  = CPSW_MAC_PORT_0; /* RGMII port */
#else
#if defined(SOC_J721E)
    gCpswNimuAppCfg.cpswType = CPSW_9G;
    gCpswNimuAppCfg.portNum  = CPSW_MAC_PORT_2; /* RGMII port */
#elif defined(SOC_J7200)
    gCpswNimuAppCfg.cpswType = CPSW_5G;
    gCpswNimuAppCfg.portNum  = CPSW_MAC_PORT_0; /* Q/SGMII port */
#endif
#endif
#else
    CpswApp_getTestConfig(&gCpswNimuAppCfg.cpswType, &gCpswNimuAppCfg.portNum);
#endif

    CpswAppUtils_enableClocks(gCpswNimuAppCfg.cpswType);

    Error_init(&eb);
    Task_Params_init(&params);
    params.instance->name = "CPU_LOAD";
    params.priority       = 15U;
    params.stack          = gCpswLoadTskStack;
    params.stackSize      = sizeof(gCpswLoadTskStack);
    hCpuLoadTask          = Task_create(CpswNimuApp_cpuLoadTask, &params, &eb);
    if (hCpuLoadTask == NULL)
    {
        BIOS_exit(0);
    }

    if (CPSWNIMUAPP_NUM_RESTART > 0U)
    {
        Error_init(&eb);
        Semaphore_Params_init(&semParams);
        semParams.mode = Semaphore_Mode_BINARY;
        gNdkRestartSem = Semaphore_create(0, &semParams, NULL);
        CpswAppUtils_assert(NULL != gNdkRestartSem);

        /* Create NDK restart task.  */
        Task_Params_init(&params);

        /* Priority of restart NDK task should be lower that NDK stack thread,
         * otherwise NDK shutdown might not complete correctly as restart would kick
         * start before that */
        params.priority       = 2U;
        params.stack          = gCpswNimuAppTskStackMain;
        params.stackSize      = sizeof(gCpswNimuAppTskStackMain);
        params.instance->name = "NDK_RestartTask";
        hNdkRestartTask       = Task_create(CpswNimuApp_ndkRestartTask, &params, &eb);
        if (hNdkRestartTask == NULL)
        {
            BIOS_exit(0);
        }

        Clock_Handle hTimer;
        uint32_t period = CPSWNIMUAPP_RESTART_PERIOD;
        Clock_Params_init(&clkParams);
        clkParams.startFlag = TRUE;
        clkParams.period    = period;
        clkParams.arg       = (UArg)NULL;

        /* Creating timer and setting timer callback function*/
        hTimer = Clock_create((Clock_FuncPtr) & CpswNimuApp_timerCallback,
                              period,
                              &clkParams,
                              NULL);
        if (hTimer != NULL)
        {
            /* Set timer expiry time in OS ticks */
            Clock_setTimeout(hTimer, period);
            Clock_setPeriod(hTimer, period);
        }
        else
        {
            BIOS_exit(0);
        }
    }

    BIOS_start();    /* does not return */

    return(0);
}

#if (0U == APP_ENABLE_STATIC_CFG)
#if defined(BUILD_MCU1_0) || defined(BUILD_MCU2_1) || defined(BUILD_MPU1_0)
static Cpsw_MacPort CpswApp_getTestMacPort(void)
{
    Cpsw_MacPort macPort = CPSW_MAC_PORT_0;
    uint32_t choice = 0U;
    bool retry;
    static const char cpswMacPortSetting[] =
    {
        "0: CPSW_MAC_PORT_0 - RGMII\n"
    };

    do
    {
        CpswAppUtils_print("Select CPSW MAC Port\n");
        CpswAppUtils_print("%s\n", cpswMacPortSetting);

        retry = false;
        choice = CpswAppUtils_getNum();
        switch (choice)
        {
            case 0:
                macPort = CPSW_MAC_PORT_0;
                break;

            default:
                CpswAppUtils_print("Wrong CPSW MAC Port, enter again\n");
                retry = true;
                break;
        }
    }
    while (retry);

    return macPort;
}
#endif

#if defined(SOC_J7200) && defined(BUILD_MCU2_0)
static Cpsw_MacPort CpswApp_getTestMacPort(void)
{
    Cpsw_MacPort macPort = CPSW_MAC_PORT_0;
    uint32_t choice = 0U;
    bool retry;
    static const char cpswMacPortSetting[] =
    {
        "0: CPSW_MAC_PORT_0 - SGMII/QSGMII (selected based on board detected)\n"
        "1: CPSW_MAC_PORT_1 - RGMII\n"
    };

    do
    {
        CpswAppUtils_print("Select CPSW MAC Port\n");
        CpswAppUtils_print("%s\n", cpswMacPortSetting);

        retry = false;
        choice = CpswAppUtils_getNum();
        switch (choice)
        {
            case 0:
            case 1:
                macPort = (Cpsw_MacPort)(CPSW_MAC_PORT_FIRST + choice);
                break;
            default:
                CpswAppUtils_print("Wrong CPSW MAC Port, enter again\n");
                retry = true;
                break;
        }
    }
    while (retry);

    return macPort;
}
#endif

#if defined(SOC_J721E) && defined(BUILD_MCU2_0)
static Cpsw_MacPort CpswApp_getTestMacPort(void)
{
    Cpsw_MacPort macPort = CPSW_MAC_PORT_0;
    uint32_t choice = 0U;
    bool retry;
    static const char cpswMacPortSetting[] =
    {
        "0: CPSW_MAC_PORT_0 - RGMII\n"
        "1: CPSW_MAC_PORT_1 - SGMII/QSGMII\n"
        "2: CPSW_MAC_PORT_2 - RGMII\n"
        "3: CPSW_MAC_PORT_3 - RGMII\n"
        "7: CPSW_MAC_PORT_7 - RMII\n"
    };

    do
    {
        CpswAppUtils_print("Select CPSW MAC Port\n");
        CpswAppUtils_print("%s \n", cpswMacPortSetting);

        retry = false;
        choice = CpswAppUtils_getNum();
        switch (choice)
        {
            case 0:
            case 1:
            case 2:
            case 3:
            case 7:
                macPort = (Cpsw_MacPort)(CPSW_MAC_PORT_FIRST + choice);
                break;

            default:
                CpswAppUtils_print("Wrong CPSW MAC Port, enter again\n");
                retry = true;
                break;
        }
    }
    while (retry);

    return macPort;
}
#endif

static void CpswApp_getTestConfig(Cpsw_Type *cpswType,
                                  Cpsw_MacPort *macPort)
{
    uint32_t choice = 0U;
    bool retry;
    static const char cpswTypeSetting[] =
    {
#if defined(BUILD_MCU1_0) || defined(BUILD_MCU2_1) || defined(BUILD_MPU1_0)
        "0: CPSW2G\n"
#endif
#if defined(SOC_J7200) && defined(BUILD_MCU2_0)
        "1: CPSW5G\n"
#endif
#if defined(SOC_J721E) && defined(BUILD_MCU2_0)
        "2: CPSW9G\n"
#endif

    };

    do
    {
        CpswAppUtils_print("Select CPSW Type\n");
        CpswAppUtils_print("%s\n", cpswTypeSetting);

        retry = false;
        choice = CpswAppUtils_getNum();
        switch (choice)
        {
#if defined(BUILD_MCU1_0) || defined(BUILD_MCU2_1) || defined(BUILD_MPU1_0)
            case 0:
                *cpswType = CPSW_2G;
                break;
#endif
#if defined(SOC_J7200) && defined(BUILD_MCU2_0)
            case 1:
                *cpswType = CPSW_5G;
                break;
#endif
#if defined(SOC_J721E) && defined(BUILD_MCU2_0)
            case 2:
                *cpswType = CPSW_9G;
                break;
#endif
            default:
                CpswAppUtils_print("Wrong CPSW Type, enter again\n");
                retry = true;
                break;
        }
    }
    while (retry);

    *macPort = CpswApp_getTestMacPort();
}
#endif /* (0U == APP_ENABLE_STATIC_CFG) */

#if defined(BUILD_MCU1_0) || defined(BUILD_MCU2_1) || defined(BUILD_MPU1_0)
static void CpswApp_setPhyConfig(Cpsw_Type cpswType,
                                 Cpsw_MacPort portNum,
                                 CpswMacPort_Config *macConfig,
                                 CpswMacPort_Interface *interface,
                                 CpswPhy_Config *phyConfig)
{
    if (portNum == CPSW_MAC_PORT_0)
    {
        CpswAppBoardUtils_setPhyConfig(cpswType,
                                       portNum,
                                       macConfig,
                                       interface,
                                       phyConfig);
    }
    else
    {
        CpswAppUtils_print("Invalid port number - can't run test\n");
    }
}
#endif

#if defined(SOC_J7200) && defined(BUILD_MCU2_0)
static void CpswApp_setPhyConfig(Cpsw_Type cpswType,
                                 Cpsw_MacPort portNum,
                                 CpswMacPort_Config *macConfig,
                                 CpswMacPort_Interface *interface,
                                 CpswPhy_Config *phyConfig)
{
    if (portNum == CPSW_MAC_PORT_0)
    {
        CpswAppBoardUtils_setPhyConfig(cpswType,
                                       portNum,
                                       macConfig,
                                       interface,
                                       phyConfig);
    }
    else if (portNum == CPSW_MAC_PORT_1)
    {
        if (Board_detectBoard(BOARD_ID_GESI))
        {
            CpswAppBoardUtils_setPhyConfigRgmii(cpswType,
                                                portNum,
                                                macConfig,
                                                interface,
                                                phyConfig);
        }
        else
        {
            CpswAppUtils_print("GESI DB card not detected - can't run test\n");
        }
    }
    else
    {
        CpswAppUtils_print("Port number not supported by this test\n");
    }
}
#endif

#if defined(SOC_J721E) && defined(BUILD_MCU2_0)
static void CpswApp_setPhyConfig(Cpsw_Type cpswType,
                                 Cpsw_MacPort portNum,
                                 CpswMacPort_Config *macConfig,
                                 CpswMacPort_Interface *interface,
                                 CpswPhy_Config *phyConfig)
{
    int32_t enetCard;

    switch (portNum)
    {
        case CPSW_MAC_PORT_0:
        case CPSW_MAC_PORT_2:
        case CPSW_MAC_PORT_3:
            CpswAppBoardUtils_setPhyConfig(cpswType,
                                           portNum,
                                           macConfig,
                                           interface,
                                           phyConfig);
            break;

        case CPSW_MAC_PORT_1:
            enetCard = Board_detectEnetCard();
            if (BOARD_ENET_QSGMII == enetCard)
            {
                CpswAppBoardUtils_setPhyConfig(cpswType,
                                               portNum,
                                               macConfig,
                                               interface,
                                               phyConfig);
            }
            else if (BOARD_ENET_SGMII == enetCard)
            {
                CpswAppBoardUtils_setPhyConfigSgmii(portNum,
                                                    macConfig,
                                                    interface,
                                                    phyConfig);
            }
            else
            {
                CpswAppUtils_print("ENET DB card not detected - can't run test\n");
            }
            break;

        case CPSW_MAC_PORT_7:
            CpswAppBoardUtils_setPhyConfigRmii(portNum,
                                               macConfig,
                                               interface,
                                               phyConfig);
            break;

        default:
            CpswAppUtils_print("Port number not supported by this test\n");
            break;
    }
}
#endif

void CpswApp_initAleConfig(CpswAle_Config *aleConfig)
{
    int32_t status = CPSW_SOK;

    aleConfig->modeFlags =
        (CPSW_ALE_CONFIG_MASK_ALE_MODULE_ENABLE);
    aleConfig->agingConfig.enableAutoAging = TRUE;
    aleConfig->agingConfig.agingPeriodInMs = 1000;

    CpswAppUtils_assert(status == CPSW_SOK);

    aleConfig->nwSecCfg.enableVid0Mode               = TRUE;
    aleConfig->vlanConfig.aleVlanAwareMode           = FALSE;
    aleConfig->vlanConfig.cpswVlanAwareMode          = FALSE;
    aleConfig->vlanConfig.unknownUnregMcastFloodMask = CPSW_ALE_ALL_PORTS_MASK;
    aleConfig->vlanConfig.unknownRegMcastFloodMask   = CPSW_ALE_ALL_PORTS_MASK;
    aleConfig->vlanConfig.unknownVlanMemberListMask  = CPSW_ALE_ALL_PORTS_MASK;
    aleConfig->policerGlobalConfig.policingEnable    = true;
    aleConfig->policerGlobalConfig.yellowDropEnable  = false;
    /*! Enables the ALE to drop the red colored packets. */
    aleConfig->policerGlobalConfig.redDropEnable = false;
    /*! Policing match mode */
    aleConfig->policerGlobalConfig.policerNoMatchMode = CPSW_ALE_POLICER_NOMATCH_MODE_GREEN;
}

void CpswApp_initLinkArgs(Cpsw_OpenPortLinkInArgs *linkArgs,
                          Cpsw_MacPort macPort)
{
    CpswMacPort_Config *macConfig      = &linkArgs->macConfig;
    CpswMacPort_LinkConfig *linkConfig = &linkArgs->linkConfig;
    CpswMacPort_Interface *interface   = &linkArgs->interface;
    CpswPhy_Config *phyConfig          = &linkArgs->phyConfig;

    linkArgs->portNum = macPort;
    Cpsw_initMacPortParams(macConfig);

    CpswApp_setPhyConfig(gCpswNimuAppCfg.cpswType,
                         macPort,
                         macConfig,
                         interface,
                         phyConfig);

    linkConfig->speed     = CPSW_SPEED_AUTO;
    linkConfig->duplexity = CPSW_DUPLEX_AUTO;
}

static void CpswNimuApp_portLinkStatusChangeCb(Cpsw_MacPort portNum,
                                                  bool isLinkUp,
                                                  void *appArg)
{
    CpswAppUtils_print("MAC Port %u: link %s\n", portNum, isLinkUp ? "up" : "down");
}

static void CpswNimuApp_mdioLinkStatusChange(Cpsw_MdioLinkStateChangeInfo *info,
                                             void *appArg)
{
    static uint32_t linkUpCount = 0;
    if ((info->linkChanged) && (info->isLinked))
    {
        linkUpCount++;
    }
}

static void CpswNimuApp_initCpswLinkCbPrms(Cpsw_Config *cpswCfg)
{

    cpswCfg->mdioLinkStateChangeCb     = CpswNimuApp_mdioLinkStatusChange;
    cpswCfg->mdioLinkStateChangeCbArg  = &gCpswNimuAppObj;

    cpswCfg->portLinkStatusChangeCb    = &CpswNimuApp_portLinkStatusChangeCb;
    cpswCfg->portLinkStatusChangeCbArg = &gCpswNimuAppObj;
}

static int32_t CpswNimuApp_init(void)
{
    int32_t status = CPSW_SOK;
    CpswMcm_InitConfig cpswMcmCfg;
    Cpsw_Config cpswCfg;
    Cpsw_MacPort macPortList[] = {gCpswNimuAppCfg.portNum};
    uint8_t numMacPorts        = (sizeof(macPortList) / sizeof(macPortList[0U]));

    CpswAppUtils_assert(numMacPorts <= Cpsw_getMacPortMax(gCpswNimuAppCfg.cpswType));

    /* Set configuration parameters */
    Cpsw_initParams(&cpswCfg);
    cpswCfg.vlanConfig.vlanAware          = false;
    cpswCfg.hostPortConfig.removeCrc      = true;
    cpswCfg.hostPortConfig.padShortPacket = true;
    cpswCfg.hostPortConfig.passCrcErrors  = true;
    CpswAppUtils_initResourceConfig(gCpswNimuAppCfg.cpswType, CpswAppSoc_getCoreId(), &cpswCfg.resourceConfig);

    CpswNimuApp_initCpswLinkCbPrms(&cpswCfg);

    CpswApp_initAleConfig(&cpswCfg.aleConfig);

    cpswCfg.dmaConfig.rxChInitPrms.dmaPriority = UDMA_DEFAULT_RX_CH_DMA_PRIORITY;

    /* Open UDMA */
    gCpswNimuAppObj.hUdmaDrv = CpswAppUtils_udmaOpen(gCpswNimuAppCfg.cpswType, NULL);
    CpswAppUtils_assert(NULL != gCpswNimuAppObj.hUdmaDrv);

    cpswCfg.dmaConfig.hUdmaDrv = gCpswNimuAppObj.hUdmaDrv;

    cpswMcmCfg.cpswType           = gCpswNimuAppCfg.cpswType;
    cpswMcmCfg.pCpswCfg           = &cpswCfg;
    cpswMcmCfg.setPortLinkCfg     = CpswApp_initLinkArgs;
    cpswMcmCfg.numMacPorts        = numMacPorts;
    cpswMcmCfg.periodicTaskPeriod = CPSW_PHY_FSM_TICK_PERIOD_MS; /* msecs */
    cpswMcmCfg.printFxn           = CpswAppUtils_print;
    cpswMcmCfg.traceFxn           = CpswAppUtils_print;

    memcpy(&cpswMcmCfg.macPortList[0U], &macPortList[0U], sizeof(macPortList));

    status = CpswMcm_init(&cpswMcmCfg);

    return status;
}

void CpswNimuApp_stackInitHook(void *hCfg)
{
    uint32_t rc;

    /* increase stack size */
    rc = 16384;
    CfgAddEntry(hCfg, CFGTAG_OS, CFGITEM_OS_TASKSTKBOOT,
                CFG_ADDMODE_UNIQUE, sizeof(uint32_t), (uint8_t *)&rc, 0);

#if (CPSW_TRACE_CFG_TRACE_LEVEL >= CPSW_TRACE_CFG_LEVEL_DEBUG)
    /* We do not want to see debug messages less than WARNINGS */
    rc = DBG_INFO;
    CfgAddEntry(hCfg, CFGTAG_OS, CFGITEM_OS_DBGPRINTLEVEL,
                CFG_ADDMODE_UNIQUE, sizeof(uint32_t), (uint8_t *)&rc, 0);
#endif
}

void CpswNimuApp_stackDeleteHook(void *hCfg)
{
    // TODO enable restart via PC command

    /* As NDK shutdown is happening, we post the NDK restart task so it can
     * start NDK again. As CpswNimuApp_ndkRestartTask is at lower priority than
     * the NDK stack thread, we don't interrupt the NDK shutdown */
    Semaphore_post(gNdkRestartSem);
}

void CpswNimuApp_ipAddrHookFxn(uint32_t IPAddr,
                               uint32_t IfIdx,
                               uint32_t fAdd)
{
    char ipAddrStr[20];

    NtIPN2Str(IPAddr, ipAddrStr);

    if (1U == fAdd)
    {
        CpswAppUtils_print("CPSW NIMU App: Added Network IP address I/F %d: %s\n",
                           IfIdx, ipAddrStr);
    }
    else
    {
        CpswAppUtils_print("CPSW NIMU App: Removed Network IP address I/F %d: %s\n",
                           IfIdx, ipAddrStr);
    }
}

void CpswNimuApp_netOpenHook(void)
{
    /* Create our local servers */
    hEcho = DaemonNew(SOCK_STREAMNC, 0, 7, dtask_tcp_echo,
                      OS_TASKPRINORM, OS_TASKSTKNORM, 0, 3);
    hEchoUdp = DaemonNew(SOCK_DGRAM, 0, 7, dtask_udp_echo,
                         OS_TASKPRINORM, OS_TASKSTKNORM, 0, 1);
    hData = DaemonNew(SOCK_STREAM, 0, 1000, dtask_tcp_datasrv,
                      OS_TASKPRINORM, OS_TASKSTKNORM, 0, 3);
    hNull = DaemonNew(SOCK_STREAMNC, 0, 1001, dtask_tcp_nullsrv,
                      OS_TASKPRINORM, OS_TASKSTKNORM, 0, 3);
    hOob = DaemonNew(SOCK_STREAMNC, 0, 999, dtask_tcp_oobsrv,
                     OS_TASKPRINORM, OS_TASKSTKNORM, 0, 3);
}

void CpswNimuApp_netCloseHook(void)
{
    DaemonFree(hOob);
    DaemonFree(hNull);
    DaemonFree(hData);
    DaemonFree(hEchoUdp);
    DaemonFree(hEcho);
}

Void CpswNimuApp_timerCallback(UArg arg)
{
    if (gNdkRestartCnt <= CPSWNIMUAPP_NUM_RESTART)
    {
        /* Stop the network stack. */
        NC_NetStop(0);
    }
}

static void CpswNimuApp_ndkRestartTask(UArg arg0,
                                       UArg arg1)
{
    /* We complete restart test during init first CPSWNIMUAPP_NUM_RESTART runs
     * and then keep NDK enabled for NDK apps (send/recv) test */
    while (true)
    {
        Semaphore_pend(gNdkRestartSem, BIOS_WAIT_FOREVER);

        CpswAppUtils_printTaskStackUsage();
        CpswAppUtils_print("\nRestarting the NDK/CPSW, %d!\n", gNdkRestartCnt);
        gNdkRestartCnt++;

        /* Deinit the app as it would be again initialized in get handle call from next ndk2cpsw_open */
        CpswNimuApp_deInit();

        /* NDK Stack will be started. */
        {
            Task_Handle hNetStackTask = NULL;
            Task_Params taskParams;

            Task_Params_init(&taskParams);
            taskParams.instance->name = "ti_ndk_config_Global_stackThread";
            taskParams.priority       = 0x5;
            taskParams.arg0           = 0u;
            taskParams.arg1           = 0u;
            taskParams.stackSize      = (SizeT)0x2000;

            hNetStackTask = Task_create(&ti_ndk_config_Global_stackThread, &taskParams, NULL);
            if (hNetStackTask == NULL)
            {
                CpswAppUtils_print("Restarting ti_ndk_config_Global_stackThread failed\n");
                break;
            }
            else
            {
                /* Register with NDK. Only then socket open will succeed.*/
                fdOpenSession(hNetStackTask);
            }
        }
    }
}

static bool CpswApp_isPortLinked(Cpsw_Handle hCpsw)
{
    uint32_t coreId = CpswAppSoc_getCoreId();

    return CpswAppUtils_isPortLinkUp(hCpsw, coreId, gCpswNimuAppCfg.portNum);
}

void NimuCpswAppCb_getHandle(NimuCpswAppIf_GetHandleInArgs *inArgs,
                             NimuCpswAppIf_GetHandleOutArgs *outArgs)
{
    int32_t status;
    CpswMcm_HandleInfo handleInfo;
    Cpsw_AttachCoreOutArgs attachInfo;
    CpswDma_OpenRxFlowPrms cpswRxFlowCfg;
    CpswDma_OpenTxChPrms cpswTxChCfg;
    uint32_t coreId          = CpswAppSoc_getCoreId();
    bool useDefaultFlow      = gCpswNimuAppCfg.useDefaultRxFlow;
    CpswMcm_CmdIf *pMcmCmdIf = &gCpswNimuAppObj.hMcmCmdIf[gCpswNimuAppCfg.cpswType];
    bool useRingMon          = true;

    if (pMcmCmdIf->hMboxCmd == NULL)
    {
        status = CpswNimuApp_init();

        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("Failed to open CPSW: %d\n", status);
        }

        CpswAppUtils_assert(status == CPSW_SOK);
        CpswMcm_getCmdIf(gCpswNimuAppCfg.cpswType, pMcmCmdIf);
    }

    CpswAppUtils_assert(pMcmCmdIf->hMboxCmd != NULL);
    CpswAppUtils_assert(pMcmCmdIf->hMboxResponse != NULL);

    CpswMcm_acquireHandleInfo(pMcmCmdIf, &handleInfo);
    CpswMcm_coreAttach(pMcmCmdIf, coreId, &attachInfo);

    /* Confirm HW checksum offload is enabled as NIMU enables it by default */
    {
        Cpsw_IoctlPrms prms;
        bool csumOffloadFlg;
        CPSW_IOCTL_SET_OUT_ARGS(&prms, &csumOffloadFlg);
        status = Cpsw_ioctl(handleInfo.hCpsw,
                            coreId,
                            CPSW_HOSTPORT_IS_CSUM_OFFLOAD_ENABLE,
                            &prms);
        if (status != CPSW_SOK)
        {
            CpswAppUtils_print("() Failed to get checksum offload info: %d\n", status);
        }

        CpswAppUtils_assert(true == csumOffloadFlg);
    }

    /* Open TX channel */
    CpswDma_initTxChParams(&cpswTxChCfg);

    cpswTxChCfg.hUdmaDrv  = handleInfo.hUdmaDrv;
    cpswTxChCfg.numTxPkts = inArgs->txCfg.numPackets;
    cpswTxChCfg.hCbArg    = inArgs->txCfg.cbArg;
    cpswTxChCfg.notifyCb  = inArgs->txCfg.notifyCb;
    cpswTxChCfg.useProxy  = true;

    CpswAppUtils_setCommonTxChPrms(&cpswTxChCfg);

    CpswAppUtils_openTxCh(handleInfo.hCpsw,
                          attachInfo.coreKey,
                          coreId,
                          &outArgs->txInfo.txChNum,
                          &outArgs->txInfo.hTxChannel,
                          &cpswTxChCfg);

    /* Open RX Flow */
    CpswDma_initRxFlowParams(&cpswRxFlowCfg);
    cpswRxFlowCfg.notifyCb  = inArgs->rxCfg.notifyCb;
    cpswRxFlowCfg.numRxPkts = inArgs->rxCfg.numPackets;
    cpswRxFlowCfg.hUdmaDrv  = handleInfo.hUdmaDrv;
    cpswRxFlowCfg.hCbArg    = inArgs->rxCfg.cbArg;
    cpswRxFlowCfg.useProxy  = true;

    /* Use ring monitor for the CQ ring of RX flow */
    CpswDma_UdmaRingPrms *pFqRingPrms = &cpswRxFlowCfg.udmaChPrms.fqRingPrms;
    pFqRingPrms->useRingMon = useRingMon;
    pFqRingPrms->ringMonCfg.mode = TISCI_MSG_VALUE_RM_MON_MODE_THRESHOLD;
    /* Ring mon low threshold */
#if defined _DEBUG_
    /* In debug mode as CPU is processing lesser packets per event, keep threshold more */
    pFqRingPrms->ringMonCfg.data0 = (inArgs->rxCfg.numPackets - 10U);
#else
    pFqRingPrms->ringMonCfg.data0 = (inArgs->rxCfg.numPackets - 20U);
#endif
    /* Ring mon high threshold - to get only low  threshold event, setting high threshold as more than ring depth*/
    pFqRingPrms->ringMonCfg.data1 = inArgs->rxCfg.numPackets;

    CpswAppUtils_setCommonRxFlowPrms(&cpswRxFlowCfg);

    CpswAppUtils_openRxFlow(handleInfo.hCpsw,
                            attachInfo.coreKey,
                            coreId,
                            useDefaultFlow,
                            &outArgs->rxInfo.rxFlowStartIdx,
                            &outArgs->rxInfo.rxFlowIdx,
                            &outArgs->rxInfo.macAddr[0U],
                            &outArgs->rxInfo.hRxFlow,
                            &cpswRxFlowCfg);

    outArgs->coreId        = coreId;
    outArgs->coreKey       = attachInfo.coreKey;
    outArgs->hCpsw         = handleInfo.hCpsw;
    outArgs->hostPortRxMtu = attachInfo.rxMtu;
    CPSW_UTILS_ARRAY_COPY(outArgs->txMtu, attachInfo.txMtu);
    outArgs->hUdmaDrv       = handleInfo.hUdmaDrv;
    outArgs->printFxnCb     = &CpswAppUtils_print;
    outArgs->isPortLinkedFxn = &CpswApp_isPortLinked;

    outArgs->isRingMonUsed = useRingMon;
    outArgs->timerPeriodUs   = CPSWNIMUAPP_PACKET_POLL_PERIOD_US;

    /* Let NIMU use optimized processing where TX packets are relinquished in next
     * TX submit call */
    outArgs->disableTxEvent = true;
    CpswAppUtils_print("Host MAC address: ");
    CpswAppUtils_printMacAddr(&outArgs->rxInfo.macAddr[0U]);
}

void NimuCpswAppCb_releaseHandle(NimuCpswAppIf_ReleaseHandleInfo *releaseInfo)
{
    CpswDma_PktInfoQ fqPktInfoQ;
    CpswDma_PktInfoQ cqPktInfoQ;

    bool useDefaultFlow      = gCpswNimuAppCfg.useDefaultRxFlow;
    CpswMcm_CmdIf *pMcmCmdIf = &gCpswNimuAppObj.hMcmCmdIf[gCpswNimuAppCfg.cpswType];

    CpswAppUtils_assert(pMcmCmdIf->hMboxCmd != NULL);
    CpswAppUtils_assert(pMcmCmdIf->hMboxResponse != NULL);

    /* Close TX channel */
    {
        CpswUtils_initQ(&fqPktInfoQ);
        CpswUtils_initQ(&cqPktInfoQ);
        CpswAppUtils_closeTxCh(releaseInfo->hCpsw,
                               releaseInfo->coreKey,
                               releaseInfo->coreId,
                               &fqPktInfoQ,
                               &cqPktInfoQ,
                               releaseInfo->txInfo.hTxChannel,
                               releaseInfo->txInfo.txChNum);
        releaseInfo->txFreePktCb(releaseInfo->freePktCbArg, &fqPktInfoQ, &cqPktInfoQ);
    }

    {
        /* Close RX Flow */
        CpswUtils_initQ(&fqPktInfoQ);
        CpswUtils_initQ(&cqPktInfoQ);
        CpswAppUtils_closeRxFlow(releaseInfo->hCpsw,
                                 releaseInfo->coreKey,
                                 releaseInfo->coreId,
                                 useDefaultFlow,
                                 &fqPktInfoQ,
                                 &cqPktInfoQ,
                                 releaseInfo->rxInfo.rxFlowStartIdx,
                                 releaseInfo->rxInfo.rxFlowIdx,
                                 releaseInfo->rxInfo.macAddr,
                                 releaseInfo->rxInfo.hRxFlow);
        releaseInfo->rxFreePktCb(releaseInfo->freePktCbArg, &fqPktInfoQ, &cqPktInfoQ);
    }

    CpswMcm_coreDetach(pMcmCmdIf, releaseInfo->coreId, releaseInfo->coreKey);
    CpswMcm_releaseHandleInfo(pMcmCmdIf);
}

void CpswNimuApp_deInit(void)
{
    CpswAppUtils_udmaclose(gCpswNimuAppObj.hUdmaDrv);

    CpswMcm_deInit(gCpswNimuAppCfg.cpswType);

    memset(&gCpswNimuAppObj, 0U, sizeof(CpswNimu_AppObj));
}

static void CpswNimuApp_cpuLoadTask(UArg a0,
                                    UArg a1)
{
    volatile uint32_t enableLoad = 1U;

    while (enableLoad)
    {
        // TODO Add task load and more granular load support
        CpswAppUtils_print("CPU Load: %u%%\n", Load_getCPULoad());

        Task_sleep(30000U);
    }
}

void CpswNimuApp_putch(char c)
{
    CpswAppUtils_print("%c", c);
}
