/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_cfg.h
 *
 * \brief    CPSW configuration parameters.
 */

#ifndef CPSW_CFG_H_
#define CPSW_CFG_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/*
 * Default UDMA RM TX channels and RX flows corewise allocation for J721E.
 * Keep it updated from udma\soc\V1\udma_rmcfg.c
 * Use below when deciding the num Tx and flows configuration.
 *
 *  MAIN NAVSS
 *  |--------|-----------|-------------|
 *  | CORE   |   RX FLOW |  TX CHANNELS|
 *  |--------|-----------|-------------|
 *  | MPU1_0 |     16    |     28      |
 *  |--------|-----------|-------------|
 *  | MCU2_0 |     64    |     4       |
 *  |--------|-----------|-------------|
 *  | MCU2_1 |     8     |     4       |
 *  |--------|-----------|-------------|
 *  | MCU3_0 |     0     |     4       |
 *  |--------|-----------|-------------|
 *  | MCU3_1 |     0     |     4       |
 *  |--------|-----------|-------------|
 *  | MCU1_0 |     0     |     0       |
 *  |--------|-----------|-------------|
 *  | MCU1_1 |     0     |     0       |
 *  |--------|-----------|-------------|
 *
 *  MCU NAVSS
 *  |--------|-----------|-------------|
 *  | CORE   |   RX FLOW |  TX CHANNELS|
 *  |--------|-----------|-------------|
 *  | MPU1_0 |     8     |     9       |
 *  |--------|-----------|-------------|
 *  | MCU2_0 |     8     |     2       |
 *  |--------|-----------|-------------|
 *  | MCU2_1 |     4     |     1       |
 *  |--------|-----------|-------------|
 *  | MCU3_0 |     4     |     1       |
 *  |--------|-----------|-------------|
 *  | MCU3_1 |     4     |     1       |
 *  |--------|-----------|-------------|
 *  | MCU1_0 |     8     |     3       |
 *  |--------|-----------|-------------|
 *  | MCU1_1 |     4     |     2       |
 *  |--------|-----------|-------------|
 */

/*!
 * \brief Total number of TX channels used by application
 *
 * The total number of TX channels used by the application. This is used in the
 * driver for memory allocation of the TX channel objects.
 *
 * Note: Maximum supported by CPSW IP - 8.
 *
 * Important Note: To use required number of TX channels in CPSW app, you must
 * make sure those are allocated in the UDMA RM.
 * Basically #CPSW_CFG_TX_CHANNELS_NUM has to be less than or equal to UDMA RM
 * allocated Tx channels.
 *
 * - To find out total TX channels allocated to core when using CPSW2G refer to
 *   gUdmaRmDefCfg_McuNavss[YOUR_CORE].numTxCh
 * - To find out total TX channels allocated to core when using CPSW9G refer to
 *   gUdmaRmDefCfg_MainNavss[YOUR_CORE].numTxCh
 *
 * Important Note: Below config can change based on UDMA RM change or based on
 * CPSW instance used.
 */
#define CPSW_CFG_TX_CHANNELS_NUM                    (8U)

/*!
 * \brief Total number of RX flows used by application
 *
 * The total number of RX flows used by the application. This is used in the
 * driver for memory allocation of the RX Flow objects.
 *
 * Note: Maximum supported - 8 for CPSW2G and 64 for CPSW9G.
 *
 * Important Note: To use required number of RX flows in CPSW app, you must make
 * sure those are allocated in the UDMA RM.
 * Basically #CPSW_CFG_RX_FLOWS_NUM has to be less than or equal to UDMA RM
 * allocated RX flows.
 *
 * - To find out total RX flows allocated to core when using CPSW2G refer to
 *   gUdmaRmDefCfg_McuNavss[YOUR_CORE].numFreeFlow
 * - To find out total RX flows allocated to core when using CPSW9G refer to
 *   gUdmaRmDefCfg_MainNavss[YOUR_CORE].numFreeFlow
 *
 * Important Note: Below config can change based on UDMA RM change or based on
 * CPSW instance used.
 */
#define CPSW_CFG_RX_FLOWS_NUM                       (8U)

/*!
 * \brief Total number of Ring monitor objects used by application
 *
 * The total number of ring monitor objects used by the application.  This is
 * used in the driver for memory allocation of the RX Flow objects.
 *
 * Note: Make sure required ring monitor objects are allocated in UDMA RM
 */
#define CPSW_CFG_RING_MON_NUM                       (4U)

/*! \brief CPSW print buffer length */
#define CPSW_CFG_PRINT_BUF_LEN                      (200U)

/*! \brief Maximum length of a single Trace_print string */
#define CPSW_CFG_TRACE_BUF_LEN                      (200U)

/*! \brief Maximum Size of CPTS Events pool */
#define CPSW_CFG_CPTS_EVENTS_POOL_SIZE              (128U)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                       Static Function Definitions                          */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_CFG_H_ */
