/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_phy.h
 *
 * \brief    This file contains the structure definitions and function
 *           prototypes of the CPSW PHY driver.
 */

/*!
 * \addtogroup CPSW_PHY_API
 * @{
 */

#ifndef CPSW_PHY_H_
#define CPSW_PHY_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/csl/csl_types.h>

#include "cpsw_mdio.h"
#include "cpsw_types.h"
#include "cpsw_macport.h"

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

/*! \brief Max extended configuration size, arbitrarily chosen */
#define CPSW_PHY_EXTENDED_CFG_SIZE_MAX        (128U)

/*! \brief CPSW Phy State Machine tick period */
#define CPSW_PHY_FSM_TICK_PERIOD_MS           (100U)

/*!
 * \brief Invalid PHY address indicator
 *
 * For ports without PHY application but set PHYADDR as CPSW_PHY_INVALID_PHYADDR
 * to indicate NO PHY mode
 */
#define CPSW_PHY_INVALID_PHYADDR              (~0U)

/* TODO: Currently, always assume auto-detect */
/*! \brief Device ID is detected at runtime from PHYIDR1 and PHYIDR2 */
#define CPSW_PHY_DEVID_AUTODETECT             (~0U)

/*! \brief Indicates the driver that default value is to be used */
#define CPSW_PHY_USE_DEFAULT                  (~0U)

/*! \brief Auto MDIX enable */
#define CPSW_PHY_NWAY_AUTOMDIX_ENABLE         (1U << 15)

/*! \brief Auto MDIX enable */
#define CPSW_PHY_MIIMDIO_MDIXFLIP             (1U << 28)

/* These events are returned as result param by CpswPhy_tick() to allow the upper
 * layer to track PHY status */

/*! \brief No change from previous status */
#define CPSW_PHY_EVENT_NOCHANGE               (0U)

/*! \brief Link down event */
#define CPSW_PHY_EVENT_LINKDOWN               (1U)

/*! \brief NLink (or re-link) event */
#define CPSW_PHY_EVENT_LINKUP                 (2U)

/*! \brief No PHY connected */
#define CPSW_PHY_EVENT_PHYERROR               (3U)

/*! \brief PHY LPBK */
#define CPSW_PHY_EVENT_LPBK                   (4U)

/*! \brief Check if PHY address is valid (0 - 31) */
#define CPSW_PHY_IS_ADDR_VALID(addr)          ((addr) <= 31U)

/*! \brief Helper macro to create IOCTL commands for host port submodule */
#define CPSW_PHY_PUBLIC_IOCTL(x)              (CPSW_IOCTL_PHY_BASE | CPSW_IOCTL_MIN(x))

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 * \brief PHY IOCTL commands
 */
enum CpswPhy_Ioctl_e
{
    /*!
     * \brief Get PHY identification
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_GenericInArgs
     * - outArgs: uint32_t
     */
    CPSW_PHY_IOCTL_PHY_TICK                = CPSW_PHY_PUBLIC_IOCTL(0U),

    /*!
     * \brief Get PHY identification
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_GenericInArgs
     * - outArgs: #CpswPhy_Version
     */
    CPSW_PHY_IOCTL_PHY_GET_ID              = CPSW_PHY_PUBLIC_IOCTL(1U),

    /*!
     * \brief Get PHY supported modes by local PHY device
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_GenericInArgs
     * - outArgs: uint32_t
     */
    CPSW_PHY_IOCTL_PHY_GET_SUPPORTED_MODES = CPSW_PHY_PUBLIC_IOCTL(2U),

    /*!
     * \brief Check state-machine link state, that is, whether the state machine
     *        has reached the LINKED state
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_GenericInArgs
     * - outArgs: bool
     */
    CPSW_PHY_IOCTL_PHY_GET_LOOPBACK_STATE  = CPSW_PHY_PUBLIC_IOCTL(3U),

    /*!
     * \brief Check state-machine link state, that is, whether the state machine
     *        has reached the LINKED state
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_GenericInArgs
     * - outArgs: bool
     */
    CPSW_PHY_IOCTL_PHY_GET_LINK_STATE      = CPSW_PHY_PUBLIC_IOCTL(4U),

    /*!
     * \brief Get the link speed and duplexity state after the state machine
     *        has reached the LINKED state
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_GenericInArgs
     * - outArgs: #CpswMacPort_LinkConfig
     */
    CPSW_PHY_IOCTL_PHY_GET_LINK_MODE       = CPSW_PHY_PUBLIC_IOCTL(5U),

    /*!
     * \brief Set PHY mode
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_SetModeInArgs
     * - outArgs: None
     */
    CPSW_PHY_IOCTL_PHY_SET_MODE            = CPSW_PHY_PUBLIC_IOCTL(6U),

    /*!
     * \brief Reset PHY
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_GenericInArgs
     * - outArgs: None
     */
    CPSW_PHY_IOCTL_PHY_RESET               = CPSW_PHY_PUBLIC_IOCTL(7U),

    /*!
     * \brief Run compliance test
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_ComplianceTestInArgs
     * - outArgs: None
     */
    CPSW_PHY_IOCTL_RUN_COMPLIANCE_TEST     = CPSW_PHY_PUBLIC_IOCTL(8U),

    /*!
     * \brief Check PHY alive status
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_GenericInArgs
     * - outArgs: bool
     */
    CPSW_PHY_IOCTL_PHY_IS_ALIVE            = CPSW_PHY_PUBLIC_IOCTL(9U),

    /*!
     * \brief Check PHY link status from PHY register
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_GenericInArgs
     * - outArgs: bool
     */
    CPSW_PHY_IOCTL_PHY_IS_LINKED           = CPSW_PHY_PUBLIC_IOCTL(10U),

    /*!
     * \brief Read PHY register
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_ReadRegInArgs
     * - outArgs: uint16_t
     */
    CPSW_PHY_IOCTL_READ_REG                = CPSW_PHY_PUBLIC_IOCTL(11U),

    /*!
     * \brief Write PHY register
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_WriteRegInArgs
     * - outArgs: None
     */
    CPSW_PHY_IOCTL_WRITE_REG               = CPSW_PHY_PUBLIC_IOCTL(12U),

    /*!
     * \brief Read PHY extended register
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_ReadRegInArgs
     * - outArgs: uint16_t
     */
    CPSW_PHY_IOCTL_READ_REG_EXT            = CPSW_PHY_PUBLIC_IOCTL(13U),

    /*!
     * \brief Write PHY extended register
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_WriteRegInArgs
     * - outArgs: None
     */
    CPSW_PHY_IOCTL_WRITE_REG_EXT           = CPSW_PHY_PUBLIC_IOCTL(14U),

    /*!
     * \brief Read PHY register using Clause-45 frame
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_Clause45ReadRegInArgs
     * - outArgs: uint16_t
     */
    CPSW_PHY_IOCTL_READ_CLAUSE45_REG       = CPSW_PHY_PUBLIC_IOCTL(15U),

    /*!
     * \brief Write PHY register using Clause-45 frame
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_Clause45WriteRegInArgs
     * - outArgs: None
     */
    CPSW_PHY_IOCTL_WRITE_CLAUSE45_REG      = CPSW_PHY_PUBLIC_IOCTL(16U),

    /*!
     * \brief Print PHY registers
     *
     * IOCTL parameters:
     * -  inArgs: #CpswPhy_GenericInArgs
     * - outArgs: None
     */
    CPSW_PHY_IOCTL_PRINT_REGS              = CPSW_PHY_PUBLIC_IOCTL(17U),
};

/*!
 * \brief Generic input args
 */
typedef struct CpswPhy_GenericInArgs_s
{
    /*! Port number */
    Cpsw_MacPort portNum;
} CpswPhy_GenericInArgs;

/* TODO: Remove - it's unused */
/*!
 * \brief Output args for CPSW_PHY_IOCTL_PHY_GET_MODE command
 *
 * PHY current mode output argument.
 */
typedef struct CpswPhy_GetModeOutArgs_s
{
    /*! Speeed */
    Cpsw_Speed speed;

    /*! Duplexity */
    Cpsw_Duplexity duplexity;

    /*! Loopback enable */
    bool loopback;
} CpswPhy_GetModeOutArgs;

/*!
 * \brief Input args for #CPSW_PHY_IOCTL_PHY_SET_MODE command
 *
 * PHY set mode input argument.
 */
typedef struct CpswPhy_SetModeInArgs
{
    /*! Port number */
    Cpsw_MacPort portNum;

    /*! PHY mode */
    uint32_t mode;

    /*! Loopback enable */
    bool loopbackEn;
} CpswPhy_SetModeInArgs;

/*!
 * \brief Input args for #CPSW_PHY_IOCTL_RUN_COMPLIANCE_TEST command
 */
typedef struct CpswPhy_ComplianceTestInArgs_s
{
    /*! Port number */
    Cpsw_MacPort portNum;

    /*! Test mode */
    uint32_t testMode;
} CpswPhy_ComplianceTestInArgs;

/*!
 * \brief Input args for #CPSW_PHY_IOCTL_READ_REG_EXT command
 */
typedef struct CpswPhy_ReadRegInArgs_s
{
    /*! Port number */
    Cpsw_MacPort portNum;

    /*! Register address */
    uint16_t reg;
} CpswPhy_ReadRegInArgs;

/*!
 * \brief Input args for #CPSW_PHY_IOCTL_WRITE_REG_EXT command
 */
typedef struct CpswPhy_WriteRegInArgs_s
{
    /*! Port number */
    Cpsw_MacPort portNum;

    /*! Register address */
    uint16_t reg;

    /*! Value to be written */
    uint16_t val;
} CpswPhy_WriteRegInArgs;

/*!
 * \brief Input args for #CPSW_PHY_IOCTL_READ_CLAUSE45_REG command
 */
typedef struct CpswPhy_Clause45ReadRegInArgs_s
{
    /*! Port number */
    Cpsw_MacPort portNum;

    /*! MMD */
    CpswMdio_Clause45Mmd mmd;

    /*! Register address */
    uint16_t reg;
} CpswPhy_Clause45ReadRegInArgs;

/*!
 * \brief Input args for #CPSW_PHY_IOCTL_WRITE_CLAUSE45_REG command
 */
typedef struct CpswPhy_Clause45WriteRegInArgs_s
{
    /*! Port number */
    Cpsw_MacPort portNum;

    /*! MMD */
    CpswMdio_Clause45Mmd mmd;

    /*! Register address */
    uint16_t reg;

    /*! Value to be written */
    uint16_t val;
} CpswPhy_Clause45WriteRegInArgs;

/*!
 * \brief PHY version (ID)
 */
typedef struct CpswPhy_Version_s
{
    /*! Organizationally Unique Identifier (OUI) */
    uint32_t oui;

    /*! Manufacturer's model number */
    uint32_t model;

    /*! Revision number */
    uint32_t revision;
} CpswPhy_Version;

/*!
 * \brief PHY State-Machine time-out values
 */
typedef struct CpswPhy_FsmTimeoutConfig_s
{
    /*! \brief FINDING state timeout (in ticks) */
    uint32_t findingStateTicks;

    /*! \brief RESET_WAIT state timeout (in ticks) */
    uint32_t resetWaitStateTicks;

    /*! \brief RESET_WAIT state residence time (in ticks) */
    uint32_t resetWaitStateResidenceTicks;

    /*! \brief NWAY_START state timeout (in ticks) */
    uint32_t nwayStartStateTicks;

    /*! \brief NWAY_WAIT state timeout (in ticks) */
    uint32_t nwayWaitStateTicks;

    /*! \brief LINK_WAIT state timeout (in ticks) */
    uint32_t linkWaitStateTicks;

    /*! \brief Timeout if MDIX is enabled (in ticks) */
    uint32_t mdixTicks;
} CpswPhy_FsmTimeoutConfig;

/*!
 * \brief PHY configuration parameters
 */
typedef struct CpswPhy_Config_s
{
    /*! PHY device address */
    uint32_t phyAddr;

    /*! Auto-negotiation advertise capabilities */
    uint32_t nwayCaps;

    /*! MDIX enable */
    bool mdixEnable;

    /*! Whether PHY is strapped or not.  If strapping is present, no explicit
     *  PHY register configurations will take place and will only wait for
     *  link to be established */
    bool isStrapped;

    /*! Enable loopback once PHY is found */
    bool enableLoopback;

    /*! Enable master mode */
    bool masterMode;

    /*! Enable external clock source */
    bool extClkSource;

    /*! Skip PHY-specific extended configuration */
    bool skipExtendedConfig;

    /*! PHY state-machine timeout configuration */
    CpswPhy_FsmTimeoutConfig timeoutConfig;

    /*! Extended PHY-specific configuration */
    uint8_t extendedConfig[CPSW_PHY_EXTENDED_CFG_SIZE_MAX];

    /*! Size of the extended configuration */
    uint32_t extendedConfigSize;
} CpswPhy_Config;

/*!
 * \brief PHY driver object handle
 *
 * PHY driver opaque handle used to call any PHY related APIs.
 */
typedef struct CpswPhy_Obj_s *CpswPhy_Handle;

/* ========================================================================== */
/*                         Global Variables Declarations                      */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* #ifndef CPSW_PHY_H_ */

/*!
 *  @}
 */
