/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_queue.h
 *
 * \brief    This file contains the function prototypes of the Queue utility
 *           functions.
 */

/*!
 * \addtogroup CPSW_QUEUE_API
 * @{
 */

#ifndef CPSW_QUEUE_H_
#define CPSW_QUEUE_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include "cpsw_types.h"

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 * \brief A generic node structure for a single link list
 *
 * \ingroup group_avbtp
 */
typedef struct node_t_
{
    /*! Pointer to the next element in the list */
    struct node_t_ *pNext;
} node_t;

/*!
 * \brief Generic queue
 */
typedef struct queue_t_
{
    /*! The current number of entries in the queue */
    uint32_t count;

    /*! The top of the queue */
    node_t *pHead;

    /*! The end of the queue */
    node_t *pTail;

    /*! Magic number used to check Q initialization */
    uint32_t magic;
} queue_t;

/* ========================================================================== */
/*                         Global Variables Declarations                      */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/*!
 * \brief Init queue
 *
 * Initializes a queue.
 *
 * \param pQueue   Queue handle
 */
void CpswUtils_initQ(queue_t *pQueue);

/*!
 * \brief Copy queues
 *
 * Copies queues.  The copied queue will have the same head/tail pointers and
 * count as the source queue.
 *
 * \param pDstQueue      Destination queue handle
 * \param pSrcQueue      Source queue handle
 */
void CpswUtils_copyQ(queue_t *pDstQueue,
                     const queue_t *pSrcQueue);

/*!
 * \brief Enqueue a packet into the queue
 *
 * Enqueues a packet into the queue.
 *
 * \param pQueue    Queue handle
 * \param pNode     Queue node handle
 */
void CpswUtils_enQ(queue_t *pQueue,
                   node_t *pNode);

/*!
 * \brief Enqueue a packet at head into the queue
 *
 * Enqueues a packet into head of queue.
 *
 * \param pQueue    Queue handle
 * \param pNode     Queue node handle
 */
void CpswUtils_enQHead(queue_t *pQueue,
                       node_t *pNode);

/*!
 * \brief Dequeue a packet from the queue
 *
 * Dequeues a packet from the queue.  A null pointer is
 * returned if the queue was already empty.
 *
 * \param pQueue    Queue handle
 *
 * \return Queue node handle or NULL if queue was empty
 */
node_t *CpswUtils_deQ(queue_t *pQueue);

/*!
 * \brief Append queue
 *
 * Appends a packet queue into another queue.  The packets in the source
 * queue are queued to the tail of the destination queue.
 *
 * \param pDstQueue      Destination queue node handle
 * \param pSrcQueue      Source queue handle
 */
void CpswUtils_appendQ(queue_t *pDstQueue,
                       queue_t *pSrcQueue);

/*!
 * \brief Get queue count
 *
 * Gets the number of packets in the queue.
 *
 * \param pQueue    Queue handle
 *
 * \return Number of packets in the queue
 */
uint32_t CpswUtils_getQCount(queue_t *pQueue);

/*!
 * \brief Checks queue corruption
 *
 * Debug API which verifies the queue memory corruption by checking queue count
 * to actual elements in queue. Asserts if queue count doesn't match.
 *
 * \param pQueue    Queue handle
 */
void CpswUtils_verifyQCount(queue_t *pQueue);

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_QUEUE_H_ */

/* @} */
