/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_rm.h
 *
 * \brief    This file contains the structure definitions and function
 *           prototypes of the CPSW Resource manager module.
 */

/*!
 * \addtogroup CPSW_RM_API
 * @{
 */

#ifndef CPSW_RM_H_
#define CPSW_RM_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */
#include "cpsw_types.h"

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

/*!
 * \name Resource Manager Error Codes
 *
 * Error codes returned by RM functions
 * @{
 */

/*! \brief CPSW RM Invalid TX channel number */
#define CPSW_RM_TXCHNUM_INVALID                          (0xABCDABCDU)

/*! \brief CPSW RM Invalid RX flow id */
#define CPSW_RM_RXFLOWIDX_INVALID                        (0xABCDABCDU)

/*!\brief CPSW RM invalid core */
#define CPSW_RM_INVALIDCORE                              (0x5562AEFEU)

/*! \brief Maximum number of entries for IOCTL permission */
#define CPSW_RM_NUM_IOCTL_PERM_ENTRY_MAX                 (16)

/*! \brief Maximum number of macaddress to be managed by RM */
#define CPSW_RM_NUM_MACADDRESS_MAX                       (10)

/* @} */

/*!
 * \brief Helper macro to create IOCTL commands for CPSW Resource Manager
 */
#define CPSW_RM_PUBLIC_IOCTL(x)                          (CPSW_IOCTL_RM_BASE | CPSW_IOCTL_MIN(x))

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 * \name CPSW RM driver opaque handle
 *
 * Opaque handle typedefs for CPSW RM driver object.
 *  @{
 */

/*!
 * \brief Opaque structure to hold software state for CPSW RM.
 */
typedef struct CpswRm_Obj_s *CpswRm_Handle;
/* @} */

/*!
 * \brief CPSW RM IOCTL commands
 */
typedef enum CpswRm_Ioctls_e
{
    /*!
     * \brief Alloc MAC Address
     *
     * IOCTL parameters:
     * -  inArgs: uint32_t coreKey
     * - outArgs: #CpswRm_AllocMacAddrOutArgs
     */
    CPSW_RM_IOCTL_ALLOC_MAC_ADDR     = CPSW_RM_PUBLIC_IOCTL(0U),

    /*!
     * \brief Alloc RX Flow
     *
     * IOCTL parameters:
     * -  inArgs: uint32_t coreKey
     * - outArgs: #CpswRm_AllocRxFlowOutArgs
     */
    CPSW_RM_IOCTL_ALLOC_RX_FLOW      = CPSW_RM_PUBLIC_IOCTL(1U),

    /*!
     * \brief Alloc TX CH Peer Id
     *
     * IOCTL parameters:
     * -  inArgs: uint32_t coreKey
     * - outArgs: uint32_t txPSILThreadId
     */
    CPSW_RM_IOCTL_ALLOC_TX_CH_PEERID = CPSW_RM_PUBLIC_IOCTL(2U),

    /*!
     * \brief Free RX flow
     *
     * IOCTL parameters:
     * -  inArgs: #CpswRm_FreeRxFlowInArgs
     * - outArgs: None
     */
    CPSW_RM_IOCTL_FREE_RX_FLOW       = CPSW_RM_PUBLIC_IOCTL(3U),

    /*!
     * \brief Free TX channel peer Id
     *
     * IOCTL parameters:
     * -  inArgs: #CpswRm_FreeTxChInArgs
     * - outArgs: None
     */
    CPSW_RM_IOCTL_FREE_TX_CH_PEERID  = CPSW_RM_PUBLIC_IOCTL(4U),

    /*!
     * \brief Free MAC Address
     *
     * IOCTL parameters:
     * -  inArgs: #CpswRm_FreeMacAddrInArgs
     * - outArgs: None
     */
    CPSW_RM_IOCTL_FREE_MAC_ADDR      = CPSW_RM_PUBLIC_IOCTL(5U),
} CpswRm_Ioctls_e;

/*!
 * \brief CPSW RM Resource Information
 *
 * This structure holds the number of Tx channels, Rx flows, MAC addresses
 * partitioned to the core.
 */
typedef struct CpswRm_ResourceInfo_s
{
    /*! Core Id for which resource info is applicable */
    uint32_t coreId;

    /*! Number of Tx channels */
    uint32_t numTxCh;

    /*! Number of Rx Flows. The total number of Rx flows required for the Rx
     *  channel is the sum of all rx flows alloted to different cores.
     *
     *  The total number of Rx flows should be <= maximum number of Rx flows
     *  supported by CPSW. Refer CpswSoc_Config::rxFlowCount for maximum number
     *  of flows supported by SOC */
    uint32_t numRxFlows;

    /*! Number of MAC addresses */
    uint32_t numMacAddress;
} CpswRm_ResourceInfo;

/*!
 * \brief CPSW RM Resource parameters
 *
 * This structure has the information about reserved flow, default flow and
 * CPSW RM DMA Resource Information of all cores.
 */
typedef struct CpswRm_ResourcePrms_s
{
    /*! Number of cores */
    uint32_t numCores;

    /*! DMA Resource Information of all cores */
    CpswRm_ResourceInfo coreDmaResourceInfo[CPSW_REMOTE_CLIENT_CORES_MAX];
} CpswRm_ResourcePrms;

/*!
 * \brief CPSW RM IOCTL permission entry
 *
 * This structure has the information about IOCTL permissions for each IOCTL cmd.
 */
typedef struct CpswRm_IoctlPermissionEntry_s
{
    /*! IOCTL command */
    uint32_t cmd;

    /*! Cores from which this IOCTL call is allowed. This is a bitmask of allowed
     *  coreIds */
    uint32_t permittedCoreMask;
} CpswRm_IoctlPermissionEntry;

/*!
 * \brief CPSW RM IOCTL permission table
 *
 * This structure has the information about IOCTL permissions for all IOCTL cmd.
 */
typedef struct CpswRm_IoctlPermissionTable_s
{
    /*! Permitted core mask if IOCTL command not found in explicit permission
     *  entry table */
    uint32_t defaultPermittedCoreMask;

    /*! Number of explicit permission entry */
    uint32_t numEntries;

    /*! Permission table entries. Must have numEntries valid entries */
    CpswRm_IoctlPermissionEntry entry[CPSW_RM_NUM_IOCTL_PERM_ENTRY_MAX];
} CpswRm_IoctlPermissionTable;

/*!
 * \brief CPSW RM IOCTL permission table
 *
 * This structure has the information about IOCTL permissions for all IOCTL cmds
 */
typedef struct CpswRm_MacAddressPool_s
{
    /*! Number of MAC addresses */
    uint32_t numMacAddress;

    /*! Array of MAC addresses, not to exceed #CPSW_RM_NUM_MACADDRESS_MAX */ 
    uint8_t macAddress[CPSW_RM_NUM_MACADDRESS_MAX][CPSW_MAC_ADDR_LEN];
} CpswRm_MacAddressPool;

/*!
 * \brief Output args for #CPSW_RM_IOCTL_ALLOC_RX_FLOW IOCTL CMD
 */
typedef struct CpswRm_AllocRxFlowOutArgs_s
{
    /*! Rx flow base or start index */
    uint32_t startIdx;

    /*! Allocated flow's index (offset from #startIdx) */
    uint32_t flowIdx;
} CpswRm_AllocRxFlowOutArgs;

/*!
 * \brief Input args for #CPSW_RM_IOCTL_ALLOC_MAC_ADDR IOCTL command
 */
typedef struct CpswRm_AllocMacAddrOutArgs_s
{
    /*! MAC address */
    uint8_t macAddr[CPSW_MAC_ADDR_LEN];
} CpswRm_AllocMacAddrOutArgs;

/*!
 * \brief Input args for #CPSW_RM_IOCTL_FREE_TX_CH_PEERID IOCTL command
 */
typedef struct CpswRm_FreeTxChInArgs_s
{
    /*! Core key */
    uint32_t coreKey;

    /*! Tx channel PSIL Thread Id  */
    uint32_t txChNum;
} CpswRm_FreeTxChInArgs;

/*!
 * \brief Input args for #CPSW_RM_IOCTL_FREE_RX_FLOW IOCTL command
 */
typedef struct CpswRm_FreeRxFlowInArgs_s
{
    /*! Core key */
    uint32_t coreKey;

    /*! Rx flow index */
    uint32_t flowIdx;
} CpswRm_FreeRxFlowInArgs;

/*!
 * \brief Input args for #CPSW_RM_IOCTL_FREE_MAC_ADDR IOCTL command
 */
typedef struct CpswRm_FreeMacAddrInArgs_s
{
    /*! Core key */
    uint32_t coreKey;

    /*! MAC address */
    uint8_t macAddr[CPSW_MAC_ADDR_LEN];
} CpswRm_FreeMacAddrInArgs;

/*!
 * \brief CPSW driver resource configuration parameters
 *
 * Resources that will be managed by CPSW RM module for partitioning across cores
 */
typedef struct Cpsw_ResourceConfig_s
{
    /*! Self core Id */
    uint32_t selfCoreId;

    /*! Resource Partition Information */
    CpswRm_ResourcePrms resourcePartitionInfo;

    /*! IOCTL permission info */
    CpswRm_IoctlPermissionTable ioctlPermissionInfo;

    /*! MAC address list to be managed by RM */
    CpswRm_MacAddressPool macList;
} CpswRm_ResourceConfig;

/*---------------------------------------------------------------------------*\
 |                         Global Function Declarations                        |
 \*---------------------------------------------------------------------------*/

#ifdef __cplusplus
}
#endif

#endif /* CPSW_RM_H_ */

/*!
 *  @}
 */
