/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 *  \file cpsw_utils.h
 *
 *  \brief CPSW utility API to make TR, get TR descriptor memory requirement.
 */

#ifndef CPSW_UTILS_H_
#define CPSW_UTILS_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <stdarg.h>
#include <ti/drv/cpsw/include/cpsw_cfg.h>
#include <ti/drv/cpsw/include/cpsw_dma_pktutils.h>
#include <ti/drv/cpsw/soc/cpsw_soc.h>

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/*! \brief Macro to get container structure from one of its members' addresses */
#define container_of(ptr, type, member)                      \
    ({                                                       \
         const typeof(((type *)0)->member) * __mptr = (ptr); \
         (type *)((char *)__mptr - offsetof(type, member));  \
     })

/*! \brief Macro to get the size of an array */
#define CPSW_UTILS_ARRAYSIZE(x)  (sizeof(x) / sizeof(x[0]))

/*! \brief Get the index of the given element within an array */
#define CPSW_UTILS_GETARRAYINDEX(member, array)   (member - &array[0])

/*! \brief Macro to determine if a member is part of an array */
#define CPSW_UTILS_ARRAYISMEMBER(member, array)                         \
    (((((uint32)member - (uint32) & array[0]) % sizeof(array[0])) == 0) \
     && (member >= &array[0])                                           \
     && (CPSW_UTILS_GETARRAYINDEX(member, array) < CPSW_UTILS_ARRAYSIZE(array)))

/*! \brief Compile-time assertion */
#define CPSW_UTILS_COMPILETIME_ASSERT(condition)                 \
    do {                                                         \
        typedef char ErrorCheck[((condition) == TRUE) ? 1 : -1]; \
        ErrorCheck a = {0};                                      \
        a[0U] = a[0U];                                           \
    } while (0)

/*! \brief Macro to perform round-up division */
#define CPSW_UTILS_DIV_ROUNDUP(val, div)  (((val) + (div) - 1) / (div))

/*! \brief Macro to round up a value */
#define CPSW_UTILS_ROUNDUP(val, div)      (CPSW_UTILS_DIV_ROUNDUP((val), (div)) * (div))

/*! \brief Macro round down a value */
#define CPSW_UTILS_ROUNDDOWN(val, div)    (((val) / (div)) * (div))

/*!
 * \brief Generates a pattern of ones between end bit and start bit
 *        M_ONES(7,4) will give 0x000000F0
 */
#define CPSW_UTILS_MK_ONES(c_ebit, c_sbit) \
    (((1U << (((c_ebit) - (c_sbit)) + 1U)) - 1U) << (c_sbit))

/*!
 * \brief Generates a pattern of zeros between end bit and start bit
 *        M_MASK(7,4) will give 0xFFFFFF0F
 */
#define CPSW_UTILS_MK_MASK(c_ebit, c_sbit) (~(CPSW_UTILS_MK_ONES((c_ebit), (c_sbit))))


/* TODO: Generate a warning if extra bit found */
/*!
 * \brief Bounds the value before writing to register M_VAL_BOUND(0x1F, 7, 4)
 *        will remove the extra bit resulting in 0xF
 */
#define CPSW_UTILS_VAL_BOUND(val, c_ebit, c_sbit) \
    ((CPSW_UTILS_MK_ONES((c_ebit), (c_sbit)) >> (c_sbit)) & (val))

/*! \brief CPSW Ring Acc memory alignment */
#define CPSW_UTILS_CACHELINE_SIZE        (UDMA_CACHELINE_ALIGNMENT)

/*! \brief Align a value by performing a round-up operation */
#define CPSW_UTILS_ALIGN(x, y)            (((x + (y - 1)) / y) * y)

/*! \brief Macro to determine if an address is aligned to a given size */
#define CPSW_UTILS_IS_ALIGNED(addr, alignSize) \
    (((uintptr_t)addr & (alignSize - 1U)) == 0U)

/*! \brief Macro to copy arrays. They must be of the same size */
#define CPSW_UTILS_ARRAY_COPY(dst, src)                                  \
    do                                                                   \
    {                                                                    \
        /* dst argument of macro should be array and not pointer.*/      \
        CPSW_UTILS_COMPILETIME_ASSERT(sizeof(dst) != sizeof(uintptr_t)); \
        memcpy(dst, src, (CPSW_UTILS_ARRAYSIZE(dst) * sizeof(dst[0])));  \
    } while (0)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/*!
 * \brief CPSW utils parameters
 */
typedef struct
{
    /*! If not NULL, this function will be called to print debug/info message
     *  with appropriate string */
    Cpsw_PrintFxnCb printFxn;

    /*! If not NULL, this function will be called to log trace message
     *  with appropriate string */
    Cpsw_TraceFxnCb traceFxn;

    /*! If not NULL, this function will be called to convert virtual address
     *  to physical address to be provided to CPSW.
     *  If NULL, the driver will assume a one-one mapping.
     *
     *  Note: The init fxn will initialize this to the default one-one map
     *  function #CpswUtils_virtToPhyFxn */
    Cpsw_VirtToPhyFxn virtToPhyFxn;

    /*! If not NULL, this function will be called to convert physical address
     *  to virtual address to access the pointer returned by the CPSW.
     *  If NULL, the driver will assume a one-one mapping.
     *
     *  Note: The init fxn will initialize this to the default one-one map
     *  function #CpswUtils_phyToVirtFxn */
    Cpsw_PhyToVirtFxn phyToVirtFxn;
} CpswUtils_Prms;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/*!
 * \brief Prints to shared memory and CCS console
 *
 * This function prints the provided formatted string to shared memory and CCS
 * console.
 *
 * \param format       [IN] Formatted string followed by variable arguments
 */
void CpswUtils_printf(const char *format,
                      ...);

/*!
 * \brief Logs trace info
 *
 * This function logs the provided formatted string to the provided trace
 * callback function.
 *
 *  \param format       [IN] Formatted string
 *  \param vaArgPtr     [IN] Variadic argument list
 */
void CpswUtils_trace(const char *format,
                     va_list vaArgPtr);

/*!
 * \brief Initialize utils module
 *
 * Utils module initialization function. Should be only called from the
 * CPSW top level module.
 *
 * \param pUtilsPrms   [IN] Pointer to the initialization parameters
 */
void CpswUtils_init(CpswUtils_Prms *pUtilsPrms);

/*!
 * \brief De-initialize utils module
 */
void CpswUtils_deInit(void);

/*!
 * \brief Returns minimum of two numbers
 *
 * \param num1     [IN] First number
 * \param num2     [IN] Second number
 *
 * \return Minimum number
 */
uint32_t CpswUtils_min(uint32_t num1,
                       uint32_t num2);

/*!
 * \brief Returns minimum of two numbers
 *
 * \param num1     [IN] First number
 * \param num2     [IN] Second number
 *
 * \return Maximum number
 */
uint32_t CpswUtils_max(uint32_t num1,
                       uint32_t num2);

/*!
 * \brief Busy loop for a given amount of time
 *
 * \param delayVal  [IN] Delay time
 */
void CpswUtils_delay(uint32_t delayVal);

/*!
 * \brief Register an ISR for an interrupt
 *
 * \param isrFxn        [IN] Interrupt service routine
 * \param coreIntrNum   [IN] Interrupt number
 * \param intrPriority  [IN] Interrupt priority
 * \param arg           [IN] Argument to ISR function
 *
 * \return Interrupt handle pointer
 */
void *CpswUtils_registerIntr(CpswOsal_IsrFxn isrFxn,
                             uint32_t coreIntrNum,
                             uint32_t intrPriority,
                             void *arg);

/*!
 * \brief Unregister an interrupt
 *
 * \param hwiHandle    [IN] Interrupt handle pointer
 */
void CpswUtils_unRegisterIntr(void *hwiHandle);

/*!
 * \brief Enable interrupt
 *
 * \param coreIntrNum   [IN] Interrupt number
 */
void CpswUtils_enableIntr(uint32_t coreIntrNum);

/*!
 * \brief Disable interrupt
 *
 * \param coreIntrNum   [IN] Interrupt number
 */
void CpswUtils_disableIntr(uint32_t coreIntrNum);

/*!
 * \brief Create a mutex
 *
 * \return Mutex handle pointer
 */
void *CpswUtils_mutexCreate(void);

/*!
 * \brief Delete a mutex
 *
 * \param mutexHandle   [IN] Mutex handle pointer
 */
void CpswUtils_mutexDelete(void *mutexHandle);

/*!
 * \brief Lock a mutex
 *
 * \param mutexHandle   [IN] Mutex handle pointer
 */
void CpswUtils_mutexLock(void *mutexHandle);

/*!
 * \brief Unlock a mutex
 *
 * \param mutexHandle   [IN] Mutex handle pointer
 */
void CpswUtils_mutexUnlock(void *mutexHandle);

/*!
 * \brief Convert a virtual address to physical address
 *
 * \param virtAddr   [IN] Virtual address
 * \param appData    [IN] Auxiliary data
 *
 * \return Physical address
 */
uint64_t CpswUtils_virtToPhyFxn(const void *virtAddr,
                                void *appData);

/*!
 * \brief Convert a physical address to virtual address
 *
 * \param phyAddr    [IN] Physical address
 * \param appData    [IN] Auxiliary data
 *
 * \return Virtual address
 */
void *CpswUtils_phyToVirtFxn(uint64_t phyAddr,
                             void *appData);

/*!
 * \brief Initialize a timer
 *
 * \return \ref Cpsw_ErrorCodes
 */
int32_t CpswUtils_timerInit(void);

/*!
 * \brief De-initialize a timer
 *
 * \return \ref Cpsw_ErrorCodes
 */
int32_t CpswUtils_timerDeInit(void);

/*!
 * \brief Read timer value
 *
 * \return Timer's current value
 */
uint32_t CpswUtils_timerRead(void);

/*!
 * \brief Get time different between a starting and current time
 *
 * \return Time difference in timer's units
 */
uint32_t CpswOsal_timerGetDiff(uint32_t startTime);

/*!
 * \brief Returns CPSW IP featureBitmap
 *
 * Returns CPSW IP featureBitmap info from SOC for the given CPSW type
 *
 * \param  cpswType     [IN] CPSW instance type. Refer to #Cpsw_Type
 *
 * \return ipFeatureMap [OUT] IP feature bitmap for given CPSW type
 */
uint32_t CpswUtils_getIpFeatureBitMap(Cpsw_Type cpswType);

/*!
 * \brief Get CPSW IP information
 *
 * Returns CPSW IP information structure for the given CPSW type in the SoC.
 *
 * \param  cpswType  [IN] CPSW instance type. Refer to #Cpsw_Type
 *
 * \return ipInfoPtr [OUT] Pointer to #CpswSoc_CpswIpInfo structure
 */
const CpswSoc_CpswIpInfo *CpswUtils_getIpInfo(Cpsw_Type cpswType);

/*!
 * \brief Returns CPSW IP errataBitmap
 *
 * Returns CPSW IP errataBitmap info from SOC for the given CPSW type
 *
 * \param  cpswType     [IN] CPSW instance type. Refer to #Cpsw_Type
 *
 * \return errataBitmap [OUT] IP errata bitmap for given CPSW type
 */
uint32_t CpswUtils_getIpErrataBitMap(Cpsw_Type cpswType);

/* ========================================================================== */
/*                       Static Function Definitions                          */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* #ifndef CPSW_UTILS_H_ */
