/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_soc.c
 *
 * \brief    This file contains the CPSW driver SOC related configuration
 *           functions.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/drv/cpsw/include/cpsw_types.h>
#include <ti/drv/cpsw/include/cpsw_macport.h>
#include <ti/drv/cpsw/src/priv/cpsw_priv.h>
#include <ti/drv/cpsw/soc/cpsw_soc.h>

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/*! \brief CPSW ALE module RTL version expected reset value */
/* This should come from CSLR but RESETVALs have been deleted from CSLR */
#define CSL_ALE_MOD_VER_MODULE_ID_RESETVAL                       (0x00000029U)
#define CSL_ALE_MOD_VER_MAJOR_REVISION_RESETVAL                  (0x00000001U)
#define CSL_ALE_MOD_VER_MINOR_REVISION_RESETVAL                  (0x00000004U)
#define CSL_ALE_2G_MOD_VER_RTL_VERSION_RESETVAL                  (0x00000007U)

#define CSL_CPSW_SS_MOD_VER_IDENT_RESETVAL                       (0x00006BA0U)
#define CSL_CPSW_SS_MOD_VER_MAJOR_REVISION_RESETVAL              (0x00000001U)
#define CSL_CPSW_SS_MOD_VER_MINOR_REVISION_RESETVAL              (0x00000000U)
#define CSL_CPSW_SS_MOD_VER_MINOR_REVISION_RESETVAL_PG2          (0x00000002U)
#define CSL_CPSW_SS_2G_MOD_VER_RTL_VERSION_RESETVAL              (0x0000000EU)
#define CSL_CPSW_SS_2G_MOD_VER_RTL_VERSION_RESETVAL_PG2          (0x00000000U)

#define CSL_CPTS_MOD_VER_TX_IDENT_RESETVAL                       (0x00004E8AU)
#define CSL_CPTS_MOD_VER_MAJOR_REVISION_RESETVAL                 (0x00000001U)
#define CSL_CPTS_MOD_VER_MINOR_REVISION_RESETVAL                 (0x00000009U)
#define CSL_CPTS_MOD_VER_MINOR_REVISION_RESETVAL_PG2             (0x0000000BU)
#define CSL_CPTS_2G_MOD_VER_RTL_VERSION_RESETVAL                 (0x00000004U)
#define CSL_CPTS_2G_MOD_VER_RTL_VERSION_RESETVAL_PG2             (0x00000000U)

#define CSL_MDIO_MOD_VER_MODULE_ID_RESETVAL                      (0x00000007U)
#define CSL_MDIO_MOD_VER_MAJOR_REVISION_RESETVAL                 (0x00000001U)
#define CSL_MDIO_MOD_VER_MINOR_REVISION_RESETVAL                 (0x00000007U)
#define CSL_MDIO_2G_MOD_VER_RTL_VERSION_RESETVAL                 (0x00000001U)

#define CSL_XGE_CPSW_ID_VER_REG_IDENT_RESETVAL                   (0x00006BA8U)
#define CSL_XGE_CPSW_ID_VER_REG_MAJOR_VER_RESETVAL               (0x00000001U)
#define CSL_XGE_2G_CPSW_ID_VER_REG_MINOR_VER_RESETVAL            (0x00000000U)
#define CSL_XGE_CPSW_ID_VER_REG_RTL_VER_RESETVAL                 (0x0000000EU)


/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

CpswSoc_Config cpswSocConfig =
{
    /* As per the clocking specification(Maxwell_clocking_v0.46.pdf),
     * the mcu_sysclk0 is 400MHz and there is fixed /2 divider. */
    .cppiClkFreqHz = 200000000LLU,
    .dmscDevIdCpsw = TISCI_DEV_MCU_CPSW0,
    .intrs         =
    {
#if !defined(__aarch64__)
        .evntPend =
        {
            .coreIntNum = CSL_MCU0_INTR_CPSW0_EVNT_PEND,
            .srcIdx     = CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID,
        },
        .statPend0 =
        {
            .coreIntNum = CSL_MCU0_INTR_CPSW0_STAT_PEND,
            .srcIdx     = CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID,
        },
        .mdioIntr =
        {
            .coreIntNum = CSL_MCU0_INTR_CPSW0_MDIO_PEND,
            .srcIdx     = CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID,
        },
#else
        .evntPend =
        {
            .coreIntNum = CSL_GIC0_INTR_MCU_CPSW0_BUS_EVNT_PEND,
            .srcIdx     = CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID,
        },
        .statPend0 =
        {
            .coreIntNum = CSL_GIC0_INTR_MCU_CPSW0_BUS_STAT_PEND,
            .srcIdx     = CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID,
        },
        .mdioIntr =
        {
            .coreIntNum = CSL_GIC0_INTR_MCU_CPSW0_BUS_MDIO_PEND,
            .srcIdx     = CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID,
        },
#endif
    },
    .clocks =
    {
        .cppiClkId        = TISCI_DEV_MCU_CPSW0_BUS_CPPI_CLK_CLK,
        .rgmii250MHzClkId = TISCI_DEV_MCU_CPSW0_BUS_RGMII_MHZ_250_CLK,
        .rgmii50MHzClkId  = TISCI_DEV_MCU_CPSW0_BUS_RGMII_MHZ_50_CLK,
        .rgmii5MHzClkId   = TISCI_DEV_MCU_CPSW0_BUS_RGMII_MHZ_5_CLK,
    },
    .txChPeerThreadId = CSL_PSILCFG_NAVSS_MCU_CPSW0_PSILD_THREAD_OFFSET,
    .rxChPeerThreadId = CSL_PSILCFG_NAVSS_MCU_CPSW0_PSILS_THREAD_OFFSET,
    .txChCount        = CSL_PSILCFG_NAVSS_MCU_CPSW0_PSILD_THREAD_CNT,

    /* Note- Though CPSW supports 64 distinct flow Ids, there are only
     * 8 policer/classifier so can effectively assign only 8 flow ids in CPSW2G */
    .rxFlowCount = 8U,
    .cptsHwPushCount = 4U,
};

const CpswSoc_CpswIpInfoTbl cpswIpInfoTbl =
{
    .numCpswIPs = 1,
    .info       =
    {
        [0] =
        {
            .cpswType       = CPSW_2G,
            .aleInfo        =
                              {
                               .regBaseAddr = (CSL_MCU_CPSW0_NUSS_BASE + CPSW_ALE_OFFSET),
                               .versionInfo = {
                                                  .moduleId = CSL_ALE_MOD_VER_MODULE_ID_RESETVAL,
                                                  .majorVer = CSL_ALE_MOD_VER_MAJOR_REVISION_RESETVAL,
                                                  .minorVer = CSL_ALE_MOD_VER_MINOR_REVISION_RESETVAL,
                                                  .rtlVer   = CSL_ALE_2G_MOD_VER_RTL_VERSION_RESETVAL,
                                              },
                              },
            .mdioInfo        =
                              {
                               .regBaseAddr = (CSL_MCU_CPSW0_NUSS_BASE + CPSW_MDIO_OFFSET),
                               .versionInfo = {
                                                  .moduleId = CSL_MDIO_MOD_VER_MODULE_ID_RESETVAL,
                                                  .majorVer = CSL_MDIO_MOD_VER_MAJOR_REVISION_RESETVAL,
                                                  .minorVer = CSL_MDIO_MOD_VER_MINOR_REVISION_RESETVAL,
                                                  .rtlVer   = CSL_MDIO_2G_MOD_VER_RTL_VERSION_RESETVAL,
                                              },
                              },
            .cptsInfo        =
                              {
                               .regBaseAddr = (CSL_MCU_CPSW0_NUSS_BASE + CPSW_CPTS_OFFSET),
                               .versionInfo = {
                                                  .moduleId = CSL_CPTS_MOD_VER_TX_IDENT_RESETVAL,
                                                  .majorVer = CSL_CPTS_MOD_VER_MAJOR_REVISION_RESETVAL,
                                                  .minorVer = CSL_CPTS_MOD_VER_MINOR_REVISION_RESETVAL,
                                                  .rtlVer   = CSL_CPTS_2G_MOD_VER_RTL_VERSION_RESETVAL,
                                              },
                              },
            .sgmiiInfo      =
                              {
                               /* SGMII mode not supported for AM65xx CPSW2G */
                               .regBaseAddr = 0x00000000UL,
                              },
            .cpswSSInfo     =
                              {
                               .regBaseAddr = (CSL_MCU_CPSW0_NUSS_BASE + CPSW_NUSS_OFFSET),
                               .versionInfo = {
                                                  .moduleId = CSL_CPSW_SS_MOD_VER_IDENT_RESETVAL,
                                                  .majorVer = CSL_CPSW_SS_MOD_VER_MAJOR_REVISION_RESETVAL,
                                                  .minorVer = CSL_CPSW_SS_MOD_VER_MINOR_REVISION_RESETVAL,
                                                  .rtlVer   = CSL_CPSW_SS_2G_MOD_VER_RTL_VERSION_RESETVAL,
                                              },
                              },
            .cpswInfo     =
                              {
                               .regBaseAddr = (CSL_MCU_CPSW0_NUSS_BASE + CPSW_NU_OFFSET),
                               .versionInfo = {
                                                  .moduleId = CSL_XGE_CPSW_ID_VER_REG_IDENT_RESETVAL,
                                                  .majorVer = CSL_XGE_CPSW_ID_VER_REG_MAJOR_VER_RESETVAL,
                                                  .minorVer = CSL_XGE_2G_CPSW_ID_VER_REG_MINOR_VER_RESETVAL,
                                                  .rtlVer   = CSL_XGE_CPSW_ID_VER_REG_RTL_VER_RESETVAL,
                                              },
                              },
            .featureBitmap  = 0,
            .errataBitmap   = CPSW_ERRATA_CHKSUMOFFLOAD,
        }
    },
};

const CpswSoc_CpswIpInfoTbl cpswIpInfoTblPG2 =
{
    .numCpswIPs = 1,
    .info       =
    {
        [0] =
        {
            .cpswType       = CPSW_2G,
            .aleInfo        =
                              {
                               .regBaseAddr = (CSL_MCU_CPSW0_NUSS_BASE + CPSW_ALE_OFFSET),
                               .versionInfo = {
                                                  .moduleId = CSL_ALE_MOD_VER_MODULE_ID_RESETVAL,
                                                  .majorVer = CSL_ALE_MOD_VER_MAJOR_REVISION_RESETVAL,
                                                  .minorVer = CSL_ALE_MOD_VER_MINOR_REVISION_RESETVAL,
                                                  .rtlVer   = CSL_ALE_2G_MOD_VER_RTL_VERSION_RESETVAL,
                                              },
                              },
            .mdioInfo        =
                              {
                               .regBaseAddr = (CSL_MCU_CPSW0_NUSS_BASE + CPSW_MDIO_OFFSET),
                               .versionInfo = {
                                                  .moduleId = CSL_MDIO_MOD_VER_MODULE_ID_RESETVAL,
                                                  .majorVer = CSL_MDIO_MOD_VER_MAJOR_REVISION_RESETVAL,
                                                  .minorVer = CSL_MDIO_MOD_VER_MINOR_REVISION_RESETVAL,
                                                  .rtlVer   = CSL_MDIO_2G_MOD_VER_RTL_VERSION_RESETVAL,
                                              },
                              },
            .cptsInfo        =
                              {
                               .regBaseAddr = (CSL_MCU_CPSW0_NUSS_BASE + CPSW_CPTS_OFFSET),
                               .versionInfo = {
                                                  .moduleId = CSL_CPTS_MOD_VER_TX_IDENT_RESETVAL,
                                                  .majorVer = CSL_CPTS_MOD_VER_MAJOR_REVISION_RESETVAL,
                                                  .minorVer = CSL_CPTS_MOD_VER_MINOR_REVISION_RESETVAL_PG2,
                                                  .rtlVer   = CSL_CPTS_2G_MOD_VER_RTL_VERSION_RESETVAL_PG2,
                                              },
                              },
            .sgmiiInfo      =
                              {
                               /* SGMII mode not supported for AM65xx CPSW2G */
                               .regBaseAddr = 0x00000000UL,
                              },
            .cpswSSInfo     =
                              {
                               .regBaseAddr = (CSL_MCU_CPSW0_NUSS_BASE + CPSW_NUSS_OFFSET),
                               .versionInfo = {
                                                  .moduleId = CSL_CPSW_SS_MOD_VER_IDENT_RESETVAL,
                                                  .majorVer = CSL_CPSW_SS_MOD_VER_MAJOR_REVISION_RESETVAL,
                                                  .minorVer = CSL_CPSW_SS_MOD_VER_MINOR_REVISION_RESETVAL_PG2,
                                                  .rtlVer   = CSL_CPSW_SS_2G_MOD_VER_RTL_VERSION_RESETVAL_PG2,
                                              },
                              },
            .cpswInfo     =
                              {
                               .regBaseAddr = (CSL_MCU_CPSW0_NUSS_BASE + CPSW_NU_OFFSET),
                               .versionInfo = {
                                                  .moduleId = CSL_XGE_CPSW_ID_VER_REG_IDENT_RESETVAL,
                                                  .majorVer = CSL_XGE_CPSW_ID_VER_REG_MAJOR_VER_RESETVAL,
                                                  .minorVer = CSL_XGE_2G_CPSW_ID_VER_REG_MINOR_VER_RESETVAL,
                                                  .rtlVer   = CSL_XGE_CPSW_ID_VER_REG_RTL_VER_RESETVAL,
                                              },
                              },
            .featureBitmap  = 0,
            .errataBitmap   = CPSW_ERRATA_CHKSUMOFFLOAD,
        }
    },
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

const CpswSoc_Config *CpswSoc_getConfig(Cpsw_Type cpswType)
{
    return &cpswSocConfig;
}

void CpswSoc_setConfig(Cpsw_Type cpswType,
                       const CpswSoc_Config *cfg)
{
    cpswSocConfig = *cfg;
}

uint32_t CpswSoc_getCoreDevId(void)
{
    uint32_t coreDevId;

#if defined(BUILD_MPU1_0)
    coreDevId = TISCI_DEV_COMPUTE_CLUSTER_A53_0;
#elif defined(BUILD_MCU1_0)
    coreDevId = TISCI_DEV_MCU_ARMSS0_CPU0;
#elif defined(BUILD_MCU1_1)
    coreDevId = TISCI_DEV_MCU_ARMSS0_CPU1;
#else
#error "CPSW SOC V0: Core not supported!!"
#endif

    return(coreDevId);
}

uint32_t CpswSoc_getEnetControl(Cpsw_Type cpswType,
                                Cpsw_MacPort portNum,
                                CpswMac_LayerType *enetLayer,
                                CpswMac_SublayerType *enetSublayer)
{
    CSL_mcu_ctrl_mmr_cfg0Regs *regs =
        (CSL_mcu_ctrl_mmr_cfg0Regs *)(uintptr_t)CSL_MCU_CTRL_MMR0_CFG0_BASE;
    uint32_t modeSel;
    int32_t status = CPSW_SOK;

    if (portNum == CPSW_MAC_PORT_0)
    {
        modeSel = CSL_FEXT(regs->MCU_ENET_CTRL,
                           MCU_CTRL_MMR_CFG0_MCU_ENET_CTRL_MODE_SEL);
        switch (modeSel)
        {
            /* RMII */
            case CPSW_ENET_CTRL_MODE_RMII:
                *enetLayer    = CPSW_MAC_LAYER_MII;
                *enetSublayer = CPSW_MAC_SUBLAYER_REDUCED;
                break;

            /* RGMII */
            case CPSW_ENET_CTRL_MODE_RGMII:
                *enetLayer    = CPSW_MAC_LAYER_GMII;
                *enetSublayer = CPSW_MAC_SUBLAYER_REDUCED;
                break;

            default:
                status = CPSW_EINVALID_PARAMS;
                break;
        }
    }
    else
    {
        status = CPSW_EINVALID_PARAMS;
    }

    return status;
}

void CpswSoc_getEFusedMacAddr(uint8_t *macAddr)
{
    /* Update default config with the correct MAC addresses */
    CSL_mcu_ctrl_mmr_cfg0Regs *mmrRegs =
        (CSL_mcu_ctrl_mmr_cfg0Regs *)(uintptr_t)CSL_MCU_CTRL_MMR0_CFG0_BASE;
    uint32_t val;

    val        = CSL_REG32_RD(&mmrRegs->MAC_ID0);
    macAddr[5] = (uint8_t)((val & 0x000000FFU) >> 0U);
    macAddr[4] = (uint8_t)((val & 0x0000FF00U) >> 8U);
    macAddr[3] = (uint8_t)((val & 0x00FF0000U) >> 16U);
    macAddr[2] = (uint8_t)((val & 0xFF000000U) >> 24U);

    val        = CSL_REG32_RD(&mmrRegs->MAC_ID1);
    macAddr[1] = (uint8_t)((val & 0x000000FFU) >> 0U);
    macAddr[0] = (uint8_t)((val & 0x0000FF00U) >> 8U);

    return;
}

uint32_t CpswSoc_getMacCaps(Cpsw_Type cpswType,
                            Cpsw_MacPort portNum)
{
    uint32_t linkCaps = (CPSW_LINK_CAP_HD10 | CPSW_LINK_CAP_FD10 |
                         CPSW_LINK_CAP_HD100 | CPSW_LINK_CAP_FD100 |
                         CPSW_LINK_CAP_FD1000);

    if (cpswType == CPSW_2G)
    {
        if (portNum != CPSW_MAC_PORT_0)
        {
            linkCaps = 0U;
        }
    }

    return linkCaps;
}

const CpswSoc_CpswIpInfoTbl *CpswSoc_getIPInfoTbl(void)
{
    const CpswSoc_CpswIpInfoTbl *pInfoTbl;
    /* Need to check if device variant indicates Maxwell PG1.0 or PG2.0, IP table will depend on that */
    uint32_t deviceVariant = CSL_REG32_FEXT(CSL_WKUP_CTRL_MMR0_CFG0_BASE +
                                            CSL_WKUP_CTRL_MMR_CFG0_JTAGID,
                                            WKUP_CTRL_MMR_CFG0_JTAGID_VARIATN);

    if (deviceVariant == 0U)
    {
        /* PG1.0 case */
        pInfoTbl = &cpswIpInfoTbl;
    }
    else
    {
        /* PG2.0 case */
        pInfoTbl = &cpswIpInfoTblPG2;
    }
    return pInfoTbl;
}

int32_t CpswSoc_validateQsgmiiCfg(Cpsw_Type cpswType)
{
    return CPSW_ENOTSUPPORTED;
}

int32_t CpswSoc_mapPort2QsgmiiId(Cpsw_Type cpswType,
                                 Cpsw_MacPort portNum,
                                 uint32_t *qsgmiiId)
{
    return CPSW_ENOTSUPPORTED;
}

