/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_soc.h
 *
 * \brief    This file contains the CPSW low-level driver SOC specific
 *           function implementations.
 */

#ifndef CPSW_SOC_TOP_H_
#define CPSW_SOC_TOP_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/drv/cpsw/include/cpsw_types.h>
#include <ti/drv/cpsw/include/cpsw_macport.h>
#include <ti/drv/cpsw/include/cpsw_ale.h>

#if defined(SOC_AM65XX)
#include <ti/drv/cpsw/soc/V0/cpsw_soc.h>
#endif

#if defined(SOC_J721E) || defined(SOC_AM77X)
#include <ti/drv/cpsw/soc/V1/cpsw_soc.h>
#endif

#if defined (SOC_J7200)
#include <ti/drv/cpsw/soc/V2/cpsw_soc.h>
#endif
#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/*!
 * \brief SRC index invalid macro - use this for direct interrupts
 *
 * SRC index invalid macro as src_index is only needed for interrupts coming
 * through routers, we use below in a check whether to call sci-client or not.
 */
#define CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID                     (0xABCDU)

/**
 *  \name CPSW feature bitmap
 *
 *  Certain CPSW features are not present on all CPSW variants and all SoCs
 *  The feature bitmap lists there features
 *  The feature bitmap is populated in the CpswSoc_Config and used internally
 *  in the cpsw driver
 *  @{
 */
/*! InterVLAN routing feature in CPSW */
#define CPSW_FEATURE_INTERVLAN         (CPSW_SET_BIT(0U))
/*! IP header whitelist feature in CPSW */
#define CPSW_FEATURE_IPHDRWHITELIST    (CPSW_SET_BIT(1U))
/*! Support for ORing the packet priority with the flow id and disabling
 *  default flow id assignment if a port is configured in MAC only mode
 */
#define CPSW_FEATURE_FLOWPRIORITY      (CPSW_SET_BIT(2U))
/*! Support for SGMII interface in CPSW instance */
#define CPSW_FEATURE_SGMII             (CPSW_SET_BIT(3U))
/*! Support for QSGMII interface in CPSW instance */
#define CPSW_FEATURE_QSGMII            (CPSW_SET_BIT(4U))

/*  @} */

/**
 *  \name CPSW errata bitmap
 *
 *  The feature bitmap lists the silicon errata issues.
 *  The feature bitmap is populated in the CpswSoc_Config and used internally
 *  in the cpsw driver to decide on functionality.
 *  @{
 */
/*! Support for checksum offload in CPSW instance */
#define CPSW_ERRATA_CHKSUMOFFLOAD      (CPSW_SET_BIT(0U))

/*! Support for re-configuring CPTS GENFn/ESTFn */
#define CPSW_ERRATA_GENFN_RECONFIG     (CPSW_SET_BIT(1U))


/*  @} */

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/*!
 * \brief CPSW interrupt "SOC connection" config
 *
 *  In SOC, for any given core two types of interrupts exists, direct interrupts
 *   and interrupts routed through the IR (interrupt router).
 *  For direct interrupts as name suggests, there is direct connection between
 *   peripheral (ex-CPSW) and the Core. As connection is direct, no SYSFW configuration is
 *   required.
 *  The indirect interrupts are coming from IR. So peripheral interrupt output connects
 *   to IR on one end & Core interrupt input is connected to other end. Here the "any" of
 *   core interrupt connected to IR can be selected to get peripheral interrupt. So IR output
 *   number needs to be "resource managed" by the application to avoid conflicts. For indirect
 *   interrupts we need to ask SYSFW to make the connection as IR is shared resource at SOC
 *   level.
 *
 */
typedef struct CpswSoc_IntrConnCfg_s
{
    /*! Core Interrupt number in case of direct connection or Interrupt router index for
     *  indirect interrupt */
    uint16_t coreIntNum;

    /*! SysFw source index for the peripheral interrupt. Only used if interrupt is indirect
     *  via router. If interrupt is direct set this to CPSW_SOC_DIRECT_INTR_SRCIDX_INVALID,
     */
    uint16_t srcIdx;
} CpswSoc_IntrConnCfg;

/*!
 * \brief CPSW Interrupt config structure
 *
 * SOC CSL Interrupt Id information for CPSW interrupts.
 */
typedef struct CpswSoc_Interrupt_s
{
    /*! CPTS event interrupt */
    CpswSoc_IntrConnCfg evntPend;
    /*! Statistics interrupt */
    CpswSoc_IntrConnCfg statPend0;
    /*! MDIO interrupt (combined MDIO_LINKINT and MDIO_USERINT events) */
    CpswSoc_IntrConnCfg mdioIntr;
} CpswSoc_Interrupt;

/*!
 * \brief CPSW clocks config structure
 *
 * TI sci-client Clock ID information for CPSW clocks. This would be used to
 * configure interrupts using System firmware via sci-client.
 */
typedef struct CpswSoc_Clock_s
{
    /*! CPSW CPPI clock id */
    uint32_t cppiClkId;
    /*! CPSW RGMII 250MHz clock id */
    uint32_t rgmii250MHzClkId;
    /*! CPSW RGMII 50MHz clock id */
    uint32_t rgmii50MHzClkId;
    /*! CPSW RGMII 5MHz clock id */
    uint32_t rgmii5MHzClkId;
} CpswSoc_Clock;

/*!
 * \brief CPSW SoC configuration
 *
 * SoC-level configuration information for the CPSW driver.
 */
typedef struct CpswSoc_Config_s
{
    /*! CPSW main clock (CPPI_ICLK) frequency in Hz. CPPI packet streaming
     * interface clock
     * Note: Clock frequency variable needs to be uint64_t as PMLIB API takes
     * clkRate input as uint64_t */
    uint64_t cppiClkFreqHz;

    /*! SCI-Client module ID for CPSW */
    uint16_t dmscDevIdCpsw;

    /*! CPSW interrupts */
    CpswSoc_Interrupt intrs;

    /*! CPSW clocks */
    CpswSoc_Clock clocks;

    /* Tx Ch Peer Thread Id */
    uint32_t txChPeerThreadId;

    /* Rx Ch Peer Thread Id */
    uint32_t rxChPeerThreadId;

    /* Tx Ch Count*/
    uint32_t txChCount;

    /* Rx Flow Count */
    uint32_t rxFlowCount;

    /* CPTS Hardware Push Event Count */
    uint32_t cptsHwPushCount;

} CpswSoc_Config;

/*!
 * \brief CPSW IP submodule version info structure
 *
 * CPSW IP submodule version specific info
 */
typedef struct CpswSoc_CpswIpSubModuleVersionInfo_s
{
    /*! CPSW IP submodule module id */
    uint32_t moduleId;
    /*! CPSW IP submodule major version */
    uint32_t majorVer;
    /*! CPSW IP submodule minor version */
    uint32_t minorVer;
    /*! CPSW IP submodule RTL version */
    uint32_t rtlVer;
} CpswSoc_CpswIpSubModuleVersionInfo;

/*!
 * \brief CPSW IP submodule info structure
 *
 * CPSW IP submodule info
 */
typedef struct CpswSoc_CpswIpSubModuleInfo_s
{
    /*! CPSW IP submodule base address */
    uintptr_t regBaseAddr;
    /*! CPSW IP submodule version info */
    CpswSoc_CpswIpSubModuleVersionInfo versionInfo;
} CpswSoc_CpswIpSubModuleInfo;


/*!
 * \brief CPSW IP information structure
 *
 * CPSW IP variant specific info
 */
typedef struct CpswSoc_CpswIpInfo_s
{
    /*! CPSW IP type */
    Cpsw_Type cpswType;

    /*! CPSW IP submodule : ALE Info */
    CpswSoc_CpswIpSubModuleInfo aleInfo;

    /*! CPSW IP submodule : CPTS Info */
    CpswSoc_CpswIpSubModuleInfo cptsInfo;

    /*! CPSW IP submodule : MDIO Info */
    CpswSoc_CpswIpSubModuleInfo mdioInfo;

    /*! CPSW IP submodule : SGMII Info */
    CpswSoc_CpswIpSubModuleInfo sgmiiInfo;

    /*! CPSW IP submodule : CPSW SS Info */
    CpswSoc_CpswIpSubModuleInfo cpswSSInfo;

    /*! CPSW IP submodule : CPSW SS Info */
    CpswSoc_CpswIpSubModuleInfo cpswInfo;

    /*! Errata bitmap for silicon issues on all CPSW variants / SoCs.
     *  @sa CPSW errata bitmap
     */
    uint32_t errataBitmap;

    /*! Feature bitmap for certain features which do not exist on all
     *  CPSW variants / SoCs.
     *  @sa CPSW feature bitmap
     */
    uint32_t featureBitmap;

} CpswSoc_CpswIpInfo;

typedef struct CpswSoc_CpswIpInfoTbl_s
{
    /*! Number of CPSW IP variants supported on the SoC */
    uint32_t numCpswIPs;

    /*! Information related to each CPSW IP variant.
     *  numCpswIPs entries will be populated for the SoC
     */
    CpswSoc_CpswIpInfo info[CPSW_COUNT];
} CpswSoc_CpswIpInfoTbl;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/*!
 * \brief Get SoC specific configuration
 */
const CpswSoc_Config *CpswSoc_getConfig(Cpsw_Type cpswType);

/*!
 * \brief Set SoC specific configuration
 */
void CpswSoc_setConfig(Cpsw_Type cpswType,
                       const CpswSoc_Config *cfg);

/*!
 * \brief Get CPSW IP info table that is specific to SOC
 */
const CpswSoc_CpswIpInfoTbl *CpswSoc_getIPInfoTbl(void);

/*!
 * \brief Get MAC interface from ENET_CTRL register settings
 */
uint32_t CpswSoc_getEnetControl(Cpsw_Type cpswType,
                                Cpsw_MacPort portNum,
                                CpswMac_LayerType *enetLayer,
                                CpswMac_SublayerType *enetSublayer);

/*!
 * \brief Get Efused MAC address from SoC CTRL register.
 *
 * The efused MAC addr is typically used as CPSW2G host port mac addr
 */
void CpswSoc_getEFusedMacAddr(uint8_t *macAddr);

/*!
 * \brief Get SoC specific MAC port capabilities
 *
 */
uint32_t CpswSoc_getMacCaps(Cpsw_Type cpswType,
                            Cpsw_MacPort portNum);

/**
 *  \brief Returns the TI SCI-Client device ID for the core
 *
 *  \return Core device ID \ref Sciclient_PmDeviceIds
 */
uint32_t CpswSoc_getCoreDevId(void);

/**
 *  \brief Validates QSGMII configuration for CPSW instance.
 *
 *  \return \ref Cpsw_ErrorCodes
 */
int32_t CpswSoc_validateQsgmiiCfg(Cpsw_Type cpswType);

/**
 *  \brief Find QSGMII main port for given port
 *
 *  \return QSGMII id for given port
 */
int32_t CpswSoc_mapPort2QsgmiiId(Cpsw_Type cpswType,
                                 Cpsw_MacPort portNum,
                                 uint32_t *qsgmiiId);

/* ========================================================================== */
/*                       Static Function Definitions                          */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_SOC_TOP_H_ */
