/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 *  \file cpsw_dma.c
 *
 *  \brief This file contains CPPI-related functionality.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <ti/csl/arch/csl_arch.h>
#include <ti/drv/cpsw/include/cpsw_types.h>
#include <ti/drv/cpsw/include/cpsw_queue.h>
#include <ti/drv/cpsw/include/cpsw_utils.h>
#include <ti/drv/cpsw/include/cpsw_cfg.h>
#include <ti/drv/udma/udma.h>

#include "priv/cpsw_priv.h"
#include "priv/cpsw_dma_priv.h"
#include "priv/cpsw_osal.h"
#include "priv/cpsw_trace_priv.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define CPSW_DMA_RSVDFLOW_RX_PKTS_NUM    (2U)

/*! \brief Total number of Rx channels for CPSW (always 1 for Jacinto 7 family of devices) */
#define CPSW_DMA_RX_CHANNELS_NUM         (1U)

/*! \brief Source tag low mask of descriptor (used to get packet's received port number) */
#define CPSW_DMA_HPD_SRC_TAG_LOW_MASK    (0xFF)

typedef queue_t CpswDma_RxFlowObjMemQ;

typedef queue_t CpswDma_TxChObjMemQ;

typedef queue_t CpswDma_RingMonObjMemQ;

typedef queue_t CpswDma_tdCqRingObjMemQ;

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/**
 *  \brief
 */
typedef struct CpswDma_TxChObjMem_s
{
    /*! The node element so this packet can be added to a queue
     * Note- Keep node_t as first member always as driver uses generic Q functions
     *       and deferences to this member */
    node_t node;

    /*! Tx channel object mem element */
    CpswDma_TxChObj txChObj;
} CpswDma_TxChObjMem;

/**
 *  \brief
 */
typedef struct CpswDma_RxFlowObjMem_s
{
    /*! The node element so this packet can be added to a queue
     * Note- Keep node_t as first member always as driver uses generic Q functions
     *       and deferences to this member */
    node_t node;

    /*! Rx flow object mem element */
    CpswDma_RxFlowObj rxFlowObj;
} CpswDma_RxFlowObjMem;

/**
 *  \brief
 */
typedef struct CpswDma_RingMonObjMem_s
{
    /*! The node element so this packet can be added to a queue
     * Note- Keep node_t as first member always as driver uses generic Q functions
     *       and deferences to this member */
    node_t node;

    /*! Ring monitor object mem element */
    struct Udma_RingMonObj ringMonObj;
} CpswDma_RingMonObjMem;

/**
 *  \brief
 */
typedef struct CpswDma_TdCqRingObjMem_s
{
    /*! The node element so this packet can be added to a queue
     * Note- Keep node_t as first member always as driver uses generic Q functions
     *       and deferences to this member */
    node_t node;

    /*! Teardown Cq Ring object mem element. Allocating cahce aligned start address and size,
     *  to avoid cache operations on other memory regions while cache invalidating after
     *  popping from the ring. Only allocating single element i.e. CPSW_DMA_TDCQ_RING_ELE_CNT of 1,
     *  as it stores the return status */
    uint64_t tdCqRingMemObj [CPSW_UTILS_ALIGN(CPSW_DMA_TDCQ_RING_ELE_CNT, UDMA_CACHELINE_ALIGNMENT)]
    __attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT)));
} CpswDma_TdCqRingObjMem;

/**
 *  \brief
 */
typedef struct CpswDma_MemMgrObj_s
{
    /*! RX flow driver object Q */
    CpswDma_RxFlowObjMemQ rxFlowObjMemQ;

    /*! TX channel driver object Q */
    CpswDma_TxChObjMemQ txChObjMemQ;

    /*! TX channel driver object Q */
    CpswDma_RingMonObjMemQ ringMonObjMemQ;

    /*! Teardown CQ memory Q */
    CpswDma_tdCqRingObjMemQ tdCqRingObjMemQ;
} CpswDma_MemMgrObj;

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

#ifdef QNX_OS

typedef struct CpswDma_Mem_s
{
    /* RX flow object memories */
    CpswDma_RxFlowObjMem gCpswDmaRxFlowObjMem[CPSW_CFG_RX_FLOWS_NUM]
        __attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT)));

    /* Tx channel object memories */
    CpswDma_TxChObjMem gCpswDmaTxChObjMem[CPSW_CFG_TX_CHANNELS_NUM]
        __attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT)));

    /* Ring monitor object memories */
    CpswDma_RingMonObjMem gCpswDmaRinMonObjMem[CPSW_CFG_RING_MON_NUM]
        __attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT)));

    uint8_t gRsvdFlowRingMemFQ[CPSW_DMA_RSVDFLOW_RX_PKTS_NUM][CPSW_DMA_RING_MEM_SIZE]
        __attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT)));

    uint8_t gRsvdFlowRingMemCQ[CPSW_DMA_RSVDFLOW_RX_PKTS_NUM][CPSW_DMA_RING_MEM_SIZE]
        __attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT)));

    /* Teardown Cq ring object memories */
    CpswDma_TdCqRingObjMem gCpswDmaTdCqRingObjMem[CPSW_CFG_TX_CHANNELS_NUM + CPSW_DMA_RX_CHANNELS_NUM]
        __attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT)));
} CpswDma_Mem;

static void *gCpswDmaMemBuf        = NULL;
static uint64_t gCpswDmaMemBufPhys  = 0;
static CpswDma_Mem *gCpswDmaMem = NULL;

#else

/* RX flow object memories */
static CpswDma_RxFlowObjMem gCpswDmaRxFlowObjMem[CPSW_CFG_RX_FLOWS_NUM]
__attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT), section(".bss:CPSW_DMA_OBJ_MEM")));

/* Tx channel object memories */
static CpswDma_TxChObjMem gCpswDmaTxChObjMem[CPSW_CFG_TX_CHANNELS_NUM]
__attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT), section(".bss:CPSW_DMA_OBJ_MEM")));

/* Ring monitor object memories */
static CpswDma_RingMonObjMem gCpswDmaRinMonObjMem[CPSW_CFG_RING_MON_NUM]
__attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT), section(".bss:CPSW_DMA_OBJ_MEM")));

static uint8_t gRsvdFlowRingMemFQ[CPSW_DMA_RSVDFLOW_RX_PKTS_NUM][CPSW_DMA_RING_MEM_SIZE]
__attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT), section(".bss:CPSW_DMA_OBJ_MEM")));

static uint8_t gRsvdFlowRingMemCQ[CPSW_DMA_RSVDFLOW_RX_PKTS_NUM][CPSW_DMA_RING_MEM_SIZE]
__attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT), section(".bss:CPSW_DMA_OBJ_MEM")));

/* Teardown Cq ring object memories */
static CpswDma_TdCqRingObjMem gCpswDmaTdCqRingObjMem[CPSW_CFG_TX_CHANNELS_NUM + CPSW_DMA_RX_CHANNELS_NUM]
__attribute__ ((aligned(UDMA_CACHELINE_ALIGNMENT), section(".bss:CPSW_DMA_OBJ_MEM")));

#endif

/* Cpsw mem utils driver object */
static CpswDma_MemMgrObj gCpswDmaMemMgrObj;

#ifdef QNX_OS
extern uint64_t gDmaDescMemArrayBase;
extern uint64_t  gDmaDescMemArrayBasePhys;
#endif

#ifdef QNX_OS
/* SMMU enable flag.. define by cpsw_lld_if.c */
extern uint32_t gSMMU;
extern uint32_t gSMMU_virtid;
#endif

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

static int32_t CpswDma_processRetrievedDesc(CpswDma_DmaDesc *pDmaDesc,
                                            CpswDma_PktInfoQ *pFromHwQueue,
                                            CpswDma_DmaDescQ *pDmaDescQ,
                                            CpswDma_Dir transferDir)
{
    int32_t retVal = UDMA_SOK;
    CpswDma_CpswHpdDesc *pHpdDesc;
    CpswDma_PktInfo *pAppPktInfo;
    CpswDma_CppiTxStatus *cppiTxStatus;
    uint32_t srcTag;

    if (NULL != pDmaDesc)
    {
        pHpdDesc = &pDmaDesc->hpdDesc;

        /* Check that the desc address from cqRing is valid */
        if (NULL != pHpdDesc)
        {
            pAppPktInfo = pDmaDesc->pAppPktInfo;
            pAppPktInfo->userBufLen = CSL_udmapCppi5GetBufferLen(&pHpdDesc->hostDesc);

            if (CPSW_DMA_DIR_RX == transferDir)
            {
#if (1U == CPSW_CFG_DEV_ERROR)
                /* Protocol-specific data is should be in the descriptor (psLocation set to
                 * TISCI_MSG_VALUE_RM_UDMAP_RX_FLOW_PS_END_PD */
                Cpsw_assert(0U == CSL_udmapCppi5GetPsDataLoc(&pHpdDesc->hostDesc));
                /* EPI(Extended packet info) block should be present in the descriptor (einfoPresent set to
                 * TISCI_MSG_VALUE_RM_UDMAP_RX_FLOW_EINFO_PRESENT */
                Cpsw_assert(true == CSL_udmapCppi5IsEpiDataPresent(&pHpdDesc->hostDesc));
#endif
                cppiTxStatus                = (CpswDma_CppiTxStatus *)pHpdDesc->psInfo;
                pAppPktInfo->chkSumInfo     = cppiTxStatus->chkSumInfo;

                pAppPktInfo->tsInfo.rxPktTs = ((((uint64_t)cppiTxStatus->tsHigh) << 32U) |
                                                ((uint64_t)cppiTxStatus->tsLow));

                srcTag = CSL_udmapCppi5GetSrcTag(&pHpdDesc->hostDesc) &
                                  CPSW_DMA_HPD_SRC_TAG_LOW_MASK;
                pAppPktInfo->rxPortNum = (Cpsw_MacPort) CPSW_ALE_ALEPORT_TO_MACPORT(srcTag);
            }

#if (1U == CPSW_CFG_DEV_ERROR)
            Cpsw_assert(pAppPktInfo->orgBufLen == pHpdDesc->hostDesc.orgBufLen);
#ifdef QNX_OS
            // QNX: We working directly with PHYS buffer
            Cpsw_assert((uint64_t)pAppPktInfo->bufPtrPhys == CSL_udmapCppi5GetBufferAddr(&pHpdDesc->hostDesc));
            Cpsw_assert((uint64_t)pAppPktInfo->bufPtrPhys == pHpdDesc->hostDesc.orgBufPtr);
#else
            Cpsw_assert((uint64_t)pAppPktInfo->bufPtr == CSL_udmapCppi5GetBufferAddr(&pHpdDesc->hostDesc));
            Cpsw_assert((uint64_t)pAppPktInfo->bufPtr == pHpdDesc->hostDesc.orgBufPtr);
#endif
#endif

            CpswUtils_checkPktState(&pAppPktInfo->pktState,
                                    CPSW_PKTSTATE_MODULE_DRIVER,
                                    (uint32_t)CPSW_PKTSTATE_DMA_WITH_HW,
                                    (uint32_t)CPSW_PKTSTATE_DMA_NOT_WITH_HW);

            CpswUtils_enQ(pFromHwQueue, &pAppPktInfo->node);

            /* Enq free Tx packet in txFreeDmaDescQ and full Rx into rxReadyDmaDescQ */
            CpswDma_dmaDescEnque(pDmaDescQ, pDmaDesc);
        }
        else
        {
            retVal = UDMA_EFAIL;
        }
    }
    else
    {
        retVal = UDMA_EFAIL;
    }

    return retVal;
}

int32_t CpswDma_retrievePkts(Udma_RingHandle hUdmaRing,
                             CpswDma_PktInfoQ *pFromHwQueue,
                             CpswDma_DmaDescQ *pDmaDescQ,
                             bool disableCacheOpsFlag,
                             CpswDma_Dir transferDir)
{
    int32_t retVal = UDMA_SOK;
    CpswDma_DmaDesc *pDmaDesc;
    bool isExposedRing;
    uint32_t ringOcc = 0U;
#ifndef QNX_OS
    uint64_t *ringMemPtr = NULL, *currRingMemPtr = NULL;
    uint32_t ringMemEleCnt = 0U, ringRdIdx = 0U;
#endif
    uint32_t i = 0U;
    int32_t retrieveCnt = 0;

    CpswUtils_initQ(pFromHwQueue);
    isExposedRing = (Udma_ringGetMode(hUdmaRing) == CSL_RINGACC_RING_MODE_RING);

    if (isExposedRing == true)
    {
        /* Disable interrupts before retrieving descriptors using Prime APIs to prevent race condition.
         * As ring occupancy is updated by the driver only at the end of this function, CQ interrupts
         * occurring in b/w might result in context switch and update the occupancy value more than once
         * for same set of descriptors. */
        uintptr_t key = CpswOsal_disableAllIntr();

        ringOcc = Udma_ringGetReverseRingOcc(hUdmaRing);
        while (ringOcc != 0U)
        {
#ifndef QNX_OS
            ringMemPtr = (uint64_t *)Udma_ringGetMemPtr(hUdmaRing);
            ringMemEleCnt = Udma_ringGetElementCnt(hUdmaRing);
            ringRdIdx = Udma_ringGetRdIdx(hUdmaRing);
            currRingMemPtr = ringMemPtr + ringRdIdx;

            /* Invalidate the Ring memory before reading from the ring */
            if ((ringRdIdx + ringOcc) > ringMemEleCnt)
            {
                CpswOsal_cacheInv(currRingMemPtr,
                                  ((ringMemEleCnt - ringRdIdx) * CPSW_DMA_RING_MEM_SIZE));
                CpswOsal_cacheInv(ringMemPtr,
                                  ((ringOcc - ringMemEleCnt + ringRdIdx) * CPSW_DMA_RING_MEM_SIZE));
            }
            else
            {
                CpswOsal_cacheInv(currRingMemPtr,
                                  (ringOcc * CPSW_DMA_RING_MEM_SIZE));
            }
#endif

            for (i = 0U; i < ringOcc; i++)
            {
                retVal = CpswDma_ringDequeue(hUdmaRing,
                                             &pDmaDesc,
                                             disableCacheOpsFlag,
                                             transferDir);

                if ((UDMA_SOK == retVal) && (NULL != pDmaDesc))
                {
                    retVal = CpswDma_processRetrievedDesc(pDmaDesc,
                                                          pFromHwQueue,
                                                          pDmaDescQ,
                                                          transferDir);
                    if (UDMA_SOK != retVal)
                    {
                        break;
                    }
                }
                else
                {
                    break;
                }
            }

            if ((retVal == UDMA_SOK) && (ringOcc > 0U))
            {
                /* Make the count negative as the occupancy needs to be reduced after popping from ring */
                retrieveCnt = -1 * (int32_t)ringOcc;
                /* Set Ring door bell register with count of number of descriptors de-queued */
                Udma_ringSetDoorBell(hUdmaRing, retrieveCnt);
            }

            ringOcc = Udma_ringGetReverseRingOcc(hUdmaRing);
        }

        /* Re-enable interrupts */
        CpswOsal_restoreAllIntr(key);
    }
    /* If ring is in message mode */
    else
    {
        /* Dequeue descs from cqRing */
        retVal = CpswDma_ringDequeue(hUdmaRing,
                                     &pDmaDesc,
                                     disableCacheOpsFlag,
                                     transferDir);

        while ((retVal == UDMA_SOK) && (NULL != pDmaDesc))
        {
            retVal = CpswDma_processRetrievedDesc(pDmaDesc,
                                                  pFromHwQueue,
                                                  pDmaDescQ,
                                                  transferDir);
            if (UDMA_SOK != retVal)
            {
                break;
            }

            retVal = CpswDma_ringDequeue(hUdmaRing,
                                         &pDmaDesc,
                                         disableCacheOpsFlag,
                                         transferDir);
        }
    }

    /* It's ok if there are no more descs in cqRing */
    if (UDMA_ETIMEOUT == retVal)
    {
        retVal = UDMA_SOK;
    }

    return retVal;
}

int32_t CpswDma_submitPkts(Udma_RingHandle hUdmaRing,
                           CpswDma_PktInfoQ *pToHwQueue,
                           CpswDma_DmaDescQ *pDmaDescQ,
                           bool disableCacheOpsFlag,
                           CpswDma_Dir transferDir,
                           Udma_ProxyHandle hUdmaProxy)
{
    int32_t retVal = UDMA_SOK;
    CpswDma_DmaDesc *pDmaDesc;
    CpswDma_PktInfo *pAppPktInfo;
    uint32_t bufLen;
    CpswDma_CppiRxControl *cppiRxCntr;
    int32_t submitCnt = 0;
    bool isExposedRing;
#ifndef QNX_OS
    uint64_t *ringMemPtr = NULL, *currRingMemPtr = NULL;
    uint32_t ringWrIdx = 0U, ringMemEleCnt = 0U;
#endif

    isExposedRing = (Udma_ringGetMode(hUdmaRing) == CSL_RINGACC_RING_MODE_RING);
#ifndef QNX_OS
    if (isExposedRing == true)
    {
        ringMemPtr = (uint64_t *)Udma_ringGetMemPtr(hUdmaRing);
        ringWrIdx = Udma_ringGetWrIdx(hUdmaRing);
        ringMemEleCnt = Udma_ringGetElementCnt(hUdmaRing);
    }
#endif

    /* Enqueue packets until fqRing is full */
    pAppPktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(pToHwQueue);
    while (NULL != pAppPktInfo)
    {
        /* Enqueue desc to fqRing */
        pDmaDesc = CpswDma_dmaDescDeque(pDmaDescQ);

        if (NULL != pDmaDesc)
        {
            CpswDma_CpswHpdDesc *pHpdDesc = (CpswDma_CpswHpdDesc *)pDmaDesc;
            CSL_UdmapCppi5HMPD *pHostDesc = &pHpdDesc->hostDesc;

            pDmaDesc->pAppPktInfo = pAppPktInfo;
            CpswUtils_checkPktState(&pDmaDesc->pAppPktInfo->pktState,
                                    CPSW_PKTSTATE_MODULE_DRIVER,
                                    (uint32_t)CPSW_PKTSTATE_DMA_NOT_WITH_HW,
                                    (uint32_t)CPSW_PKTSTATE_DMA_WITH_HW);
#ifdef QNX_OS
            // QNX: We working directly with PHYS buffer

            CSL_udmapCppi5SetBufferAddr(pHostDesc, (uint64_t)pAppPktInfo->bufPtrPhys);

            CSL_udmapCppi5SetOrgBufferAddr(pHostDesc, (uint64_t)pAppPktInfo->bufPtrPhys);
#else
            CSL_udmapCppi5SetBufferAddr(pHostDesc, (uint64_t)pAppPktInfo->bufPtr);

            CSL_udmapCppi5SetOrgBufferAddr(pHostDesc, (uint64_t)pAppPktInfo->bufPtr);
#endif
            CSL_udmapCppi5SetOrgBufferLen(pHostDesc, pAppPktInfo->orgBufLen);

            CSL_udmapCppi5SetBufferLen(pHostDesc, pAppPktInfo->userBufLen);

            CSL_udmapCppi5HostSetPktLen(pHostDesc, pAppPktInfo->userBufLen);

            if (CPSW_DMA_DIR_TX == transferDir)
            {
#if (1U == CPSW_CFG_DEV_ERROR)
                /* For TX all defaults are set in CpswDma_buffDescInit function. We just confirm those
                 * are not changed here */
                Cpsw_assert(CSL_UDMAP_CPPI5_PD_DESCINFO_DTYPE_VAL_HOST == CSL_udmapCppi5GetDescType(pHostDesc));
                Cpsw_assert(0U == CSL_udmapCppi5GetPsDataLoc(pHostDesc));
                Cpsw_assert(true == CSL_udmapCppi5IsEpiDataPresent(pHostDesc));
                Cpsw_assert(CPSW_DMA_PROTOCOL_SPECIFIC_INFO_BLOCK_SIZE ==
                                CSL_udmapCppi5GetPsDataLen(pHostDesc));
#endif
                cppiRxCntr = (CpswDma_CppiRxControl *)pHpdDesc->psInfo;
                cppiRxCntr->chkSumInfo = pAppPktInfo->chkSumInfo;

                if (pAppPktInfo->tsInfo.enableHostTxTs == true)
                {
                    CPSWDMA_CPPIPSI_SET_TSEN(cppiRxCntr->tsInfo, 1U);
                    CPSWDMA_CPPIPSI_SET_DOMAIN(cppiRxCntr->tsInfo,
                                               pAppPktInfo->tsInfo.txPktDomain);
                    CPSWDMA_CPPIPSI_SET_MSGTYPE(cppiRxCntr->tsInfo,
                                                pAppPktInfo->tsInfo.txPktMsgType);
                    CPSWDMA_CPPIPSI_SET_SEQID(cppiRxCntr->tsInfo,
                                              pAppPktInfo->tsInfo.txPktSeqId);

                    /* Set host Tx timestamp flag back to false. */
                    pAppPktInfo->tsInfo.enableHostTxTs = false;
                }
                else
                {
                    CPSWDMA_CPPIPSI_SET_TSEN(cppiRxCntr->tsInfo, 0U);
                }

                if (pAppPktInfo->txPortNum != CPSW_MAC_PORT_INV)
                {
                    CSL_udmapCppi5SetDstTag(pHostDesc,
                                            CPSW_ALE_MACPORT_TO_ALEPORT(pAppPktInfo->txPortNum));
                    /* Set txPortNum back to invalid to reset directed packet configuration */
                    pAppPktInfo->txPortNum = CPSW_MAC_PORT_INV;
                }
                else
                {
                    CSL_udmapCppi5SetDstTag(pHostDesc, 0U);
                }
            }

            /* Check that the buffer pointer and length are correct */
            retVal = CpswDma_dmaDescCheck(pDmaDesc, transferDir);

            if (UDMA_SOK == retVal)
            {
                if (CPSW_DMA_DIR_RX == transferDir)
                {
                    bufLen = pAppPktInfo->orgBufLen;
                }
                else
                {
                    bufLen = pAppPktInfo->userBufLen;
                }

                retVal = CpswDma_ringEnqueue(hUdmaRing,
                                             pDmaDesc,
                                             bufLen,
                                             disableCacheOpsFlag,
                                             transferDir,
                                             hUdmaProxy);
            }

            /* Dequeue from toHwQ only if the packet was actually queued to the fqRing */
            if (UDMA_SOK == retVal)
            {
                pAppPktInfo = (CpswDma_PktInfo *)CpswUtils_deQ(pToHwQueue);
                submitCnt++;
            }
            else
            {
                // TODO - based on error we should add back dmadesc and pktInfo to orig Queues
                CpswUtils_checkPktState(&pDmaDesc->pAppPktInfo->pktState,
                                        CPSW_PKTSTATE_MODULE_DRIVER,
                                        (uint32_t)CPSW_PKTSTATE_DMA_WITH_HW,
                                        (uint32_t)CPSW_PKTSTATE_DMA_NOT_WITH_HW);
                break;
            }
        }
        else
        {
            /* Return dequeued packet buffer as couldn't get free dma desc to attach
             *  packet to */
            CpswUtils_enQHead(pToHwQueue, &pAppPktInfo->node);
            break;
        }
    }

    /* Wb the Ring memory cache before commiting to the ring in case of exposed ring mode */
    if ((UDMA_SOK == retVal) && (isExposedRing == true))
    {
#ifndef QNX_OS
        if ((ringWrIdx + submitCnt) > ringMemEleCnt)
        {
            currRingMemPtr = ringMemPtr + ringWrIdx;
            CpswOsal_cacheWb(currRingMemPtr,
                             (ringMemEleCnt - ringWrIdx) * CPSW_DMA_RING_MEM_SIZE);
            CpswOsal_cacheWb(ringMemPtr,
                             (submitCnt - ringMemEleCnt + ringWrIdx) * CPSW_DMA_RING_MEM_SIZE);
        }
        else
        {
            CpswOsal_cacheWb(currRingMemPtr,
                             submitCnt * CPSW_DMA_RING_MEM_SIZE);
        }
#endif
        /* Set Ring door bell register with count of number of descriptors queued */
        Udma_ringSetDoorBell(hUdmaRing, submitCnt);
    }

    /* If fqRing ran out of space is not an error, packets will be re-submitted from toHwQ */
    if (UDMA_ETIMEOUT == retVal)
    {
        retVal = UDMA_SOK;
    }

    return retVal;
}

int32_t CpswDma_flushRxFlowRing(CpswDma_RxFlowHandle hRxFlow,
                                Udma_RingHandle hUdmaRing,
                                CpswDma_PktInfoQ *pPktInfoQ)
{
    int32_t retVal = UDMA_SOK;
    uint64_t pDesc = 0;
    CpswDma_DmaDesc *pDmaDesc;

    CpswUtils_initQ(pPktInfoQ);

    /* Dequeue the packets still sitting in the cqRing */
    retVal = Udma_ringDequeueRaw(hUdmaRing, &pDesc);
    while ((UDMA_SOK == retVal) && (0ULL != pDesc))
    {
#ifdef QNX_OS
        //
        // QNX: Use the physical memory offset to calculate the virtual
        // memory address
        //
        pDmaDesc = (CpswDma_DmaDesc *)(gDmaDescMemArrayBase + (pDesc - gDmaDescMemArrayBasePhys));
#else
        pDmaDesc = (CpswDma_DmaDesc *)CpswUtils_phyToVirtFxn(pDesc, NULL);
#endif
        CpswUtils_checkPktState(&pDmaDesc->pAppPktInfo->pktState,
                                CPSW_PKTSTATE_MODULE_DRIVER,
                                (uint32_t)CPSW_PKTSTATE_DMA_WITH_HW,
                                (uint32_t)CPSW_PKTSTATE_DMA_NOT_WITH_HW);
        CpswUtils_enQ(pPktInfoQ, &pDmaDesc->pAppPktInfo->node);

        hRxFlow->rxFlowPrms.dmaDescFreeFxn(hRxFlow->rxFlowPrms.hCbArg,
                                           pDmaDesc);

        retVal = Udma_ringDequeueRaw(hUdmaRing, &pDesc);
    }

    /* It's ok if there are no more descs in cqRing */
    if (UDMA_ETIMEOUT == retVal)
    {
        retVal = UDMA_SOK;
    }

    return retVal;
}

int32_t CpswDma_flushTxChRing(CpswDma_TxChHandle hTxCh,
                              Udma_RingHandle hUdmaRing,
                              CpswDma_PktInfoQ *pFqPktInfoQ)
{
    int32_t retVal = UDMA_SOK;
    uint64_t pDesc = 0;
    CpswDma_DmaDesc *pDmaDesc;

    CpswUtils_initQ(pFqPktInfoQ);

    /* Dequeue the packets still sitting in the ring */
    retVal = Udma_ringDequeueRaw(hUdmaRing, &pDesc);
    while ((UDMA_SOK == retVal) && (0ULL != pDesc))
    {
#ifdef QNX_OS
        //
        // QNX: Use the physical memory offset to calculate the virtual
        // memory address
        //
        pDmaDesc = (CpswDma_DmaDesc *)(gDmaDescMemArrayBase + (pDesc - gDmaDescMemArrayBasePhys));
#else
        pDmaDesc = (CpswDma_DmaDesc *)CpswUtils_phyToVirtFxn(pDesc, NULL);
#endif
        CpswUtils_checkPktState(&pDmaDesc->pAppPktInfo->pktState,
                                CPSW_PKTSTATE_MODULE_DRIVER,
                                (uint32_t)CPSW_PKTSTATE_DMA_WITH_HW,
                                (uint32_t)CPSW_PKTSTATE_DMA_NOT_WITH_HW);
        CpswUtils_enQ(pFqPktInfoQ, &pDmaDesc->pAppPktInfo->node);

        CpswDma_dmaDescEnque(hTxCh->hDmaDescPool, pDmaDesc);
        hTxCh->txChPrms.dmaDescFreeFxn(hTxCh->txChPrms.hCbArg,
                                       pDmaDesc);

        retVal = Udma_ringDequeueRaw(hUdmaRing, &pDesc);
    }

    /* It's ok if there are no more descs in ring */
    if (UDMA_ETIMEOUT == retVal)
    {
        retVal = UDMA_SOK;
    }

    return retVal;
}

int32_t CpswDma_freeRing(Udma_RingHandle hUdmaRing,
                        uint32_t numPkts,
                        CpswDma_FreeRingMemFxn ringMemFreeFxn,
                        void *hCbArg)
{
    uint8_t *ringMemPtr;
    int32_t retVal = UDMA_SOK;

    ringMemPtr = (uint8_t *)Udma_ringGetMemPtr(hUdmaRing);
    retVal     = Udma_ringFree(hUdmaRing);

    if (NULL != ringMemFreeFxn)
    {
        ringMemFreeFxn(hCbArg, ringMemPtr, numPkts);
    }

    return retVal;
}

int32_t CpswDma_allocRing(Udma_DrvHandle hUdmaDrv,
                          Udma_RingHandle hUdmaRing,
                          Udma_RingPrms *pRingPrms,
                          CpswDma_ringAllocInfo *pRingAllocInfo)
{
    int32_t retVal = UDMA_SOK;
    bool freeMem   = false;

    if ((pRingAllocInfo != NULL ) && (pRingAllocInfo->allocRingMem))
    {
        Cpsw_assert(NULL != pRingAllocInfo);
        Cpsw_assert(NULL != pRingAllocInfo->ringMemAllocFxn);
        Cpsw_assert(NULL != pRingAllocInfo->ringMemFreeFxn);

        pRingPrms->ringMem = pRingAllocInfo->ringMemAllocFxn(pRingAllocInfo->hCbArg,
                                             pRingPrms->elemCnt,
                                             UDMA_CACHELINE_ALIGNMENT);
        if (pRingPrms->ringMem == NULL)
        {
            CPSWTRACE_ERROR("[CPSW DMA Error] RX ring memory allocation failed !!\n");
            retVal = UDMA_EALLOC;
        }
        else if (!CPSW_UTILS_IS_ALIGNED(pRingPrms->ringMem, UDMA_CACHELINE_ALIGNMENT))
        {
            CPSWTRACE_ERROR("[CPSW DMA] TX fq ring memory not aligned!!\n");
            freeMem = true;
            retVal  = UDMA_EALLOC;
        }
        else
        {
        }
    }
    else
    {
        /* ring memory should not be NULL if allocRingMem flag is false */
        Cpsw_assert (NULL != pRingPrms->ringMem);
    }


    if ((UDMA_SOK == retVal) && (pRingAllocInfo != NULL))
    {
        /* Clear and invalidate ring memory */
        memset(pRingPrms->ringMem, 0, (pRingPrms->elemCnt * CPSW_DMA_RING_MEM_SIZE));
        if (FALSE == Udma_isCacheCoherent())
        {
            CpswOsal_cacheWbInv(pRingPrms->ringMem, (pRingPrms->elemCnt * CPSW_DMA_RING_MEM_SIZE));
        }

        /* need to allocate rxfree/rxcompletion ring pair */
        retVal = Udma_ringAlloc(hUdmaDrv,
                                hUdmaRing,
                                pRingAllocInfo->ringNum,
                                pRingPrms);
        if (UDMA_SOK != retVal)
        {
            freeMem = true;
        }
    }

    if ((pRingAllocInfo != NULL) && (pRingAllocInfo->allocRingMem) && (freeMem))
    {
        pRingAllocInfo->ringMemFreeFxn(pRingAllocInfo->hCbArg, pRingPrms->ringMem, pRingPrms->elemCnt);
        pRingPrms->ringMem = NULL;
    }

    return retVal;
}

uint32_t CpswDma_getRxChPeerThreadId(Cpsw_Type cpswType)
{
    uint32_t peerChNum;
    const CpswSoc_Config *pSocCfg;

    pSocCfg = CpswSoc_getConfig(cpswType);

    peerChNum = pSocCfg->rxChPeerThreadId;

    return peerChNum;
}

uint32_t CpswDma_getTxChPeerThreadId(Cpsw_Type cpswType,
                                     uint32_t chNum)
{
    int32_t peerChNum = CPSW_RM_TXCHNUM_INVALID;
    const CpswSoc_Config *pSocCfg;

    pSocCfg = CpswSoc_getConfig(cpswType);
    /* Get PSI-L destination thread offset for Tx channel */
    peerChNum = (pSocCfg->txChPeerThreadId + chNum);

    return peerChNum;
}

void CpswDma_getTxChUdmaInfo(CpswDma_TxChObj *pTxCh,
                             CpswDma_udmaInfo *pUdmaInfo)
{
    pUdmaInfo->hUdmaDrv     = pTxCh->hUdmaDrv;
    pUdmaInfo->hUdmaCh      = pTxCh->hUdmaCh;
    pUdmaInfo->hUdmaEvt     = pTxCh->hUdmaEvt;
    pUdmaInfo->hUdmaRingMon = pTxCh->hUdmaRingMon;
}

void CpswDma_getRxFlowUdmaInfo(CpswDma_RxFlowObj *pRxFlow,
                               CpswDma_udmaInfo *pUdmaInfo)
{
    pUdmaInfo->hUdmaDrv     = pRxFlow->hUdmaDrv;
    pUdmaInfo->hUdmaEvt     = pRxFlow->hUdmaEvt;
    pUdmaInfo->hUdmaRingMon = pRxFlow->hUdmaRingMon;
}

static int32_t CpswDma_ringProxyEnqueue(Udma_RingHandle hUdmaRing,
                                        Udma_ProxyHandle hUdmaProxy,
                                        uint64_t phyDescMem)
{
    int32_t retVal = UDMA_SOK;
    uint32_t ringHwOcc;
    uintptr_t key;

#if (1U == CPSW_CFG_DEV_ERROR)
    if ((NULL == hUdmaProxy) ||
        (NULL == hUdmaRing))
    {
        CPSWTRACE_ERROR_IF((NULL == hUdmaProxy), "[CPSW DMA] hUdmaProxy is NULL!!\n");
        CPSWTRACE_ERROR_IF((NULL == hUdmaRing), "[CPSW DMA] hUdmaRing is NULL!!\n");
        retVal = UDMA_EBADARGS;
    }
    else
#endif
    {
        key = CpswOsal_disableAllIntr();

        /* Get ring occupancy. We should write to FIFO (through proxy) only
         * when there is room available. Otherwise ring will overflow */
        ringHwOcc = CSL_ringaccGetRingHwOcc(&hUdmaRing->drvHandle->raRegs,
                                            hUdmaRing->ringNum);
        if (ringHwOcc >= hUdmaRing->cfg.elCnt)
        {
            /* Ring full */
            retVal = UDMA_EFAIL;
        }
        else
        {
            Udma_proxyQueue(hUdmaProxy, phyDescMem);
            retVal = UDMA_SOK;
        }

        CpswOsal_restoreAllIntr(key);
    }

    return(retVal);
}

CpswDma_RxFlowHandle CpswDma_openRxRsvdFlow(CpswDma_RsvdRxFlowPrms *pRxFlowPrms)
{
    int32_t retVal = UDMA_SOK;
    Udma_FlowPrms flowPrms;
    Udma_RingPrms ringPrms;
    CpswDma_RxFlowObj *pRxFlow;
    CpswDma_ringAllocInfo ringAllocInfo;
    bool allocFlowObj = false, allocFqRing = false;
    bool allocCqRing  = false, flowAttachFlag = false;
    uintptr_t intrKey;
    uint8_t flowStart;

    intrKey = CpswOsal_disableAllIntr();

    /* Set to NULL so if error condition we return NULL */
    pRxFlow = NULL;

    /* Error check */
    retVal = UDMA_EBADARGS;
    if (NULL != pRxFlowPrms)
    {
        if (NULL != pRxFlowPrms->hUdmaDrv)
        {
            retVal = UDMA_SOK;
        }
        else
        {
            CPSWTRACE_ERROR("[CPSW DMA Error] pRxFlowPrms has null UDMA handle!!\n");
        }
    }
    else
    {
        CPSWTRACE_ERROR("[CPSW DMA Error] pRxFlowPrms NULL !!\n");
    }

    if (UDMA_SOK == retVal)
    {
        pRxFlow = CpswDma_memMgrAllocRxFlowObj();

        if (pRxFlow == NULL)
        {
            CPSWTRACE_ERROR("[CPSW DMA Error] Rx flow Memory alloc object failed !!\n");
            retVal = UDMA_EALLOC;
        }
        else
        {
            allocFlowObj = true;
            memset(pRxFlow, 0U, sizeof(*pRxFlow));
            /* Save Flow config */
            pRxFlow->rxFlowPrms.startIdx = pRxFlowPrms->startIdx;
            pRxFlow->rxFlowPrms.flowIdx  = pRxFlowPrms->flowIdx;
            pRxFlow->rxFlowPrms.hUdmaDrv = pRxFlowPrms->hUdmaDrv;

            /* Initialize local variables to be used in reminder of this function  */
            pRxFlow->hUdmaDrv  = pRxFlowPrms->hUdmaDrv;
            pRxFlow->hUdmaFlow = &pRxFlow->rxFlowMemObj.flowUdmaObj;
            pRxFlow->fqRing    = &pRxFlow->rxFlowMemObj.fqRingObj;
            pRxFlow->cqRing    = &pRxFlow->rxFlowMemObj.cqRingObj;

        }
    }

    if (UDMA_SOK == retVal)
    {
        UdmaRingPrms_init(&ringPrms);
#ifdef QNX_OS
        ringPrms.elemCnt = CPSW_UTILS_ARRAYSIZE(gCpswDmaMem->gRsvdFlowRingMemFQ);
        ringPrms.mode    = CSL_RINGACC_RING_MODE_MESSAGE;
        ringPrms.ringMem = &gCpswDmaMem->gRsvdFlowRingMemFQ[0U][0U];
#else
        ringPrms.elemCnt = CPSW_UTILS_ARRAYSIZE(gRsvdFlowRingMemFQ);
        ringPrms.mode    = CSL_RINGACC_RING_MODE_MESSAGE;
        ringPrms.ringMem = &gRsvdFlowRingMemFQ[0U][0U];
#endif

#ifdef QNX_OS
        if (gSMMU)
            ringPrms.virtId  = gSMMU_virtid;
#endif

        ringAllocInfo.allocRingMem = false;
        ringAllocInfo.ringNum      = UDMA_RING_ANY;

        retVal = CpswDma_allocRing(pRxFlow->hUdmaDrv,
                                   pRxFlow->fqRing,
                                   &ringPrms,
                                   &ringAllocInfo);
        CPSWTRACE_ERROR_IF((retVal != UDMA_SOK),
                           "[CPSW DMA Error] Fq ring allocation failed !!: %d\n", retVal);
        if (UDMA_SOK == retVal)
        {
            allocFqRing = true;
        }
    }

    if (UDMA_SOK == retVal)
    {
        UdmaRingPrms_init(&ringPrms);
#ifdef QNX_OS
        ringPrms.elemCnt = CPSW_UTILS_ARRAYSIZE(gCpswDmaMem->gRsvdFlowRingMemCQ);
        ringPrms.mode    = CSL_RINGACC_RING_MODE_MESSAGE;
        ringPrms.ringMem = &gCpswDmaMem->gRsvdFlowRingMemCQ[0U][0U];
#else
        ringPrms.elemCnt = CPSW_UTILS_ARRAYSIZE(gRsvdFlowRingMemCQ);
        ringPrms.mode    = CSL_RINGACC_RING_MODE_MESSAGE;
        ringPrms.ringMem = &gRsvdFlowRingMemCQ[0U][0U];
#endif

#ifdef QNX_OS
        if (gSMMU)
            ringPrms.virtId  = gSMMU_virtid;
#endif

        ringAllocInfo.allocRingMem = false;
        ringAllocInfo.ringNum      = UDMA_RING_ANY;

        retVal = CpswDma_allocRing(pRxFlow->hUdmaDrv,
                                   pRxFlow->cqRing,
                                   &ringPrms,
                                   &ringAllocInfo);
        CPSWTRACE_ERROR_IF((retVal != UDMA_SOK),
                           "[CPSW DMA Error] Cq ring allocation failed !!: %d\n", retVal);
        if (UDMA_SOK == retVal)
        {
            allocCqRing = true;
        }
    }

    if (UDMA_SOK == retVal)
    {
        memset(&flowPrms, 0, sizeof(flowPrms));
        UdmaFlowPrms_init(&flowPrms, UDMA_CH_TYPE_RX);

        /* UdmaFlowPrms_init sets errorHandling to retry(1) by default,
         * in case of descriptor/buffer starvation UDMA retries unless it
         * gets a descriptor. In case of multi flow, this results in bottom
         * of FIFO drop, to avoid this errorHandling must be set to drop(0).
         */
        flowPrms.errorHandling = TISCI_MSG_VALUE_RM_UDMAP_RX_FLOW_ERR_DROP;
        flowPrms.psInfoPresent = TISCI_MSG_VALUE_RM_UDMAP_RX_FLOW_PSINFO_PRESENT;

        flowPrms.defaultRxCQ = pRxFlow->cqRing->ringNum;

        flowPrms.fdq0Sz0Qnum = pRxFlow->fqRing->ringNum;
        flowPrms.fdq0Sz1Qnum = pRxFlow->fqRing->ringNum;
        flowPrms.fdq0Sz2Qnum = pRxFlow->fqRing->ringNum;
        flowPrms.fdq0Sz3Qnum = pRxFlow->fqRing->ringNum;
        flowPrms.fdq1Qnum    = pRxFlow->fqRing->ringNum;
        flowPrms.fdq2Qnum    = pRxFlow->fqRing->ringNum;
        flowPrms.fdq3Qnum    = pRxFlow->fqRing->ringNum;

        flowStart = pRxFlowPrms->startIdx + pRxFlowPrms->flowIdx;

        /* Attach and configure the flows */
        retVal = Udma_flowAttach(pRxFlow->hUdmaDrv,
                                 pRxFlow->hUdmaFlow,
                                 flowStart, /* flowStart */
                                 1U /* flowCnt */);

        CPSWTRACE_ERROR_IF((retVal != UDMA_SOK), "[CPSW DMA Error] Flow attach failed!!: %d\n", retVal);
        if (UDMA_SOK == retVal)
        {
            flowAttachFlag = true;
            retVal         = Udma_flowConfig(pRxFlow->hUdmaFlow,
                                             0U, /* Flow Index */
                                             &flowPrms);
        }
    }

    /* Save flow config */
    if (UDMA_SOK == retVal)
    {
        pRxFlow->initFlag    = true;
        pRxFlow->evtInitFlag = false;
    }
    else
    {
        /* Error. Free-up resources if allocated */

        if (allocFqRing)
        {
            retVal = CpswDma_freeRing(pRxFlow->fqRing, 0U, NULL, NULL);
            Cpsw_assert(UDMA_SOK == retVal);
        }

        if (allocCqRing)
        {
            retVal = CpswDma_freeRing(pRxFlow->cqRing, 0U, NULL, NULL);
            Cpsw_assert(UDMA_SOK == retVal);
        }

        if (flowAttachFlag)
        {
            retVal = Udma_flowDetach(pRxFlow->hUdmaFlow);
            Cpsw_assert(UDMA_SOK == retVal);
        }

        /* Free the rxFlowObj last */
        if (allocFlowObj)
        {
            CpswDma_memMgrFreeRxFlowObj(pRxFlow);
        }

        /* As flow open is failed return NULL */
        pRxFlow = NULL;
    }

    CpswOsal_restoreAllIntr(intrKey);

    return pRxFlow;
}

int32_t CpswDma_closeRxRsvdFlow(CpswDma_RxFlowHandle hRxFlow)
{
    int32_t retVal = UDMA_SOK;
    uintptr_t intrKey;

#if (1U == CPSW_CFG_DEV_ERROR)
    if (NULL == hRxFlow)
    {
        CPSWTRACE_ERROR("[CPSW DMA Error] Input params can't be NULL!!\n");
        Cpsw_assert(false);
        retVal = UDMA_EBADARGS;
    }
    else
#endif
    {
        intrKey = CpswOsal_disableAllIntr();

        /* Error check */
        if (hRxFlow == NULL)
        {
            retVal = UDMA_EBADARGS;
        }

        if (UDMA_SOK == retVal)
        {
            Cpsw_assert(hRxFlow->evtInitFlag == false);

            retVal = CpswDma_freeRing(hRxFlow->fqRing, 0U, NULL, NULL);
            Cpsw_assert(UDMA_SOK == retVal);

            retVal = CpswDma_freeRing(hRxFlow->cqRing, 0U, NULL, NULL);
            Cpsw_assert(UDMA_SOK == retVal);

            retVal = Udma_flowDetach(hRxFlow->hUdmaFlow);
            Cpsw_assert(UDMA_SOK == retVal);
            hRxFlow->initFlag = false;

            /* Free Rx Flow driver object memory */
            CpswDma_memMgrFreeRxFlowObj(hRxFlow);
        }

        CpswOsal_restoreAllIntr(intrKey);
    }

    return retVal;
}

int32_t CpswDma_ringEnqueue(Udma_RingHandle hUdmaRing,
                            CpswDma_DmaDesc *pDmaDesc,
                            uint32_t packetSize,
                            bool disableCacheOpsFlag,
                            CpswDma_Dir transferDir,
                            Udma_ProxyHandle hUdmaProxy)
{
    int32_t retVal = UDMA_SOK;
    bool isExposedRing;
#ifndef QNX_OS
    void *virtBufPtr;
#endif
    uint64_t physDescPtr;

    if ((pDmaDesc != NULL) && (hUdmaRing != NULL))
    {
        CpswDma_CpswHpdDesc *pHpdDesc = (CpswDma_CpswHpdDesc *)pDmaDesc;

        isExposedRing = (Udma_ringGetMode(hUdmaRing) == CSL_RINGACC_RING_MODE_RING);
#ifdef QNX_OS
        // QNX: Use the virtual memory offset to calculate the physical memory address
        physDescPtr = gDmaDescMemArrayBasePhys + ((uint64_t)&pHpdDesc->hostDesc - gDmaDescMemArrayBase);
#else
        virtBufPtr = (void *)(uintptr_t)pHpdDesc->hostDesc.bufPtr;

        pHpdDesc->hostDesc.bufPtr =
            CpswUtils_virtToPhyFxn((void *)(uintptr_t)pHpdDesc->hostDesc.bufPtr, NULL);

        pHpdDesc->hostDesc.orgBufPtr =
            CpswUtils_virtToPhyFxn((void *)(uintptr_t)pHpdDesc->hostDesc.orgBufPtr, NULL);

        physDescPtr = (uint64_t)CpswUtils_virtToPhyFxn((void *)&pHpdDesc->hostDesc, NULL);

        /* Wb Invalidate data cache */
        if ((false == disableCacheOpsFlag) &&
            (FALSE == Udma_isCacheCoherent()))
        {
            if (CPSW_DMA_DIR_RX == transferDir)
            {
                CpswOsal_cacheInv(virtBufPtr, packetSize);
            }
            else
            {
                CpswOsal_cacheWbInv(virtBufPtr, packetSize);
            }
        }

        if (FALSE == Udma_isCacheCoherent())
        {
            CpswOsal_cacheWbInv((void *)pHpdDesc, sizeof(*pHpdDesc));
        }
#endif

        if (isExposedRing == true)
        {
           /* Udma_ringPrime doesn't check for invalid parameters and also doesn't return
            * any errors */
           Udma_ringPrime(hUdmaRing, physDescPtr);
        }
        else
        {
            if (hUdmaProxy == NULL)
            {
                retVal = Udma_ringQueueRaw(hUdmaRing, physDescPtr);
            }
            else
            {
                retVal = CpswDma_ringProxyEnqueue(hUdmaRing, hUdmaProxy, physDescPtr);
            }
        }
    }
    else
    {
        retVal = UDMA_EFAIL;
    }

    return retVal;
}

int32_t CpswDma_ringDequeue(Udma_RingHandle hUdmaRing,
                            CpswDma_DmaDesc **pDmaDesc,
                            bool disableCacheOpsFlag,
                            CpswDma_Dir transferDir)
{
    int32_t retVal = UDMA_SOK;
    bool isExposedRing;
    uint64_t pDesc = 0;
#ifndef QNX_OS
    CpswDma_CpswHpdDesc *pHpdDesc;
#endif
    CpswDma_DmaDesc *pVirtDmaDesc;

    if ((pDmaDesc != NULL) && (hUdmaRing != NULL))
    {
        isExposedRing = (Udma_ringGetMode(hUdmaRing) == CSL_RINGACC_RING_MODE_RING);
        if (isExposedRing == true)
        {
            /* Udma_ringPrimeRead doesn't check for invalid parameters and also doesn't return
             * any errors */
            Udma_ringPrimeRead(hUdmaRing, &pDesc);
            if (pDesc == 0U)
            {
                *pDmaDesc = NULL;
                retVal = UDMA_ETIMEOUT;
            }
        }
        else
        {
            retVal = Udma_ringDequeueRaw(hUdmaRing, &pDesc);
            if (UDMA_SOK != retVal)
            {
                *pDmaDesc = NULL;
            }
        }

        if (UDMA_SOK == retVal)
        {
#ifdef QNX_OS
            //
            // QNX: Use the physical memory offset to calculate the virtual
            // memory address
            //
            pVirtDmaDesc = (CpswDma_DmaDesc *)(gDmaDescMemArrayBase + (pDesc - gDmaDescMemArrayBasePhys));
#else
            pVirtDmaDesc = (CpswDma_DmaDesc *)CpswUtils_phyToVirtFxn(pDesc, NULL);
            pHpdDesc     = &pVirtDmaDesc->hpdDesc;
#endif

            /* Note - Not doing cache invalidate here as it is not required as the descriptor
               * memory was writeback-invalidated during enqueue.*/
            *pDmaDesc    = pVirtDmaDesc;
#ifdef QNX_OS
            // QNX: Not required since we are working directly with PHYS buffer
#else
            pHpdDesc->hostDesc.bufPtr =
                (uint64_t)CpswUtils_phyToVirtFxn(pHpdDesc->hostDesc.bufPtr, NULL);
            pHpdDesc->hostDesc.orgBufPtr =
                (uint64_t)CpswUtils_phyToVirtFxn(pHpdDesc->hostDesc.orgBufPtr, NULL);
#endif
            /* Debug point -> The Cortex A cores can do speculative reads which
               * can mandate to do cache_invalidate before packets are received.
               * If Cortex-A doesn't support coherency, invalidate the data buffer
               * so the CPU is forced to read new data from HW.
               * We only need to so this on the Cortex v7-A & v8-A devices because of possible
               * speculative reads done by the CPU core while the buffers are owned by
               * the DMA hardware.*/
        }
    }
    else
    {
        retVal = UDMA_EFAIL;
    }

    return retVal;
}

void CpswDma_reconfigureDescQ(CpswDma_TxChObj *pTxCh)
{
    CpswDma_DmaDescQ tempQ;
    CpswDma_DmaDesc *dmaDesc;
    CpswDma_CpswHpdDesc *pHpdDesc;

    Cpsw_assert(pTxCh != NULL);
    Cpsw_assert(pTxCh->hDmaDescPool != NULL);

    CpswDma_DmaDescQ *descQPtr = pTxCh->hDmaDescPool;
    uint32_t retQIdx = pTxCh->cqRing->ringNum;

    CpswDma_dmaDescQInit(&tempQ);

    while (CpswDma_dmaDescQCount(descQPtr) > 0U)
    {
        dmaDesc = CpswDma_dmaDescDeque(descQPtr);
        Cpsw_assert(dmaDesc != NULL);

        pHpdDesc = (CpswDma_CpswHpdDesc *)dmaDesc;
        CSL_udmapCppi5SetReturnPolicy(&pHpdDesc->hostDesc,
                                      CPSW_DMA_UDMAP_DESC_TYPE,
                                      CSL_UDMAP_CPPI5_PD_PKTINFO2_RETPOLICY_VAL_ENTIRE_PKT,
                                      CSL_UDMAP_CPPI5_PD_PKTINFO2_EARLYRET_VAL_NO,
                                      CSL_UDMAP_CPPI5_PD_PKTINFO2_RETPUSHPOLICY_VAL_TO_TAIL,
                                      retQIdx);
        CpswDma_dmaDescEnque(&tempQ, dmaDesc);
    }

    CpswDma_dmaDescCopyQ(descQPtr, &tempQ);

    return;
}

void CpswDma_initTxFreeDescQ(CpswDma_TxChObj *pTxCh)
{
    uint32_t i;
    uint32_t descType;
    uint32_t alignSize = UDMA_CACHELINE_ALIGNMENT;

    Cpsw_assert(pTxCh != NULL);
    Cpsw_assert(pTxCh->hDmaDescPool != NULL);

    CpswDma_dmaDescQInit(pTxCh->hDmaDescPool);

    for (i = 0; i < pTxCh->txChPrms.numTxPkts; i++)
    {
        CpswDma_DmaDesc *dmaDesc =
            pTxCh->txChPrms.dmaDescAllocFxn(pTxCh->txChPrms.hCbArg, alignSize);
        if (dmaDesc == NULL)
        {
            CPSWTRACE_ERROR("[CPSW DMA Error] Tx DMA descriptor memory allocation failed !!\n");
            Cpsw_assert(false);
        }
        else if (!CPSW_UTILS_IS_ALIGNED(dmaDesc, alignSize))
        {
            CPSWTRACE_ERROR("[CPSW DMA Error] dmaDesc memory not aligned!!\n");
            Cpsw_assert(false);
        }
        else
        {
            CpswDma_CpswHpdDesc *pHpdDesc = (CpswDma_CpswHpdDesc *)dmaDesc;
            CSL_UdmapCppi5HMPD *hostDesc  = &pHpdDesc->hostDesc;

            descType = CPSW_DMA_UDMAP_DESC_TYPE;

            CpswDma_dmaDescInit(dmaDesc);

            CSL_udmapCppi5SetDescType(hostDesc, descType);

            CSL_udmapCppi5SetReturnPolicy(hostDesc,
                                          descType,
                                          CSL_UDMAP_CPPI5_PD_PKTINFO2_RETPOLICY_VAL_ENTIRE_PKT,
                                          CSL_UDMAP_CPPI5_PD_PKTINFO2_EARLYRET_VAL_NO,
                                          CSL_UDMAP_CPPI5_PD_PKTINFO2_RETPUSHPOLICY_VAL_TO_TAIL,
                                          pTxCh->cqRing->ringNum);
            CpswDma_dmaDescEnque(pTxCh->hDmaDescPool, dmaDesc);
        }
    }
}

void CpswDma_deInitTxFreeDescQ(CpswDma_TxChObj *pTxCh)
{
    CpswDma_DmaDesc *pDmaDesc;

    pDmaDesc = CpswDma_dmaDescDeque(pTxCh->hDmaDescPool);
    while (NULL != pDmaDesc)
    {
        pTxCh->txChPrms.dmaDescFreeFxn(pTxCh->txChPrms.hCbArg, pDmaDesc);
        pDmaDesc = CpswDma_dmaDescDeque(pTxCh->hDmaDescPool);
    }
}

void CpswDma_initRxFreeDescQ(CpswDma_RxFlowObj *pRxFlow)
{
    uint32_t i;
    uint32_t descType;
    uint32_t alignSize = UDMA_CACHELINE_ALIGNMENT;

    CpswDma_dmaDescQInit(pRxFlow->hDmaDescPool);

    for (i = 0; i < pRxFlow->rxFlowPrms.numRxPkts; i++)
    {
        CpswDma_DmaDesc *dmaDesc =
            pRxFlow->rxFlowPrms.dmaDescAllocFxn(pRxFlow->rxFlowPrms.hCbArg, alignSize);
        if (dmaDesc == NULL)
        {
            CPSWTRACE_ERROR("[CPSW DMA Error] Tx DMA descriptor memory allocation failed !!\n");
            Cpsw_assert(dmaDesc != NULL);
        }
        else if (!CPSW_UTILS_IS_ALIGNED(dmaDesc, alignSize))
        {
            CPSWTRACE_ERROR("[CPSW DMA Error] dmaDesc memory not aligned!!\n");
            Cpsw_assert(false);
        }
        else
        {
            CpswDma_CpswHpdDesc *pHpdDesc = (CpswDma_CpswHpdDesc *)dmaDesc;
            CSL_UdmapCppi5HMPD *hostDesc  = &pHpdDesc->hostDesc;

            CpswDma_dmaDescInit(dmaDesc);

            descType = CPSW_DMA_UDMAP_DESC_TYPE;

            CSL_udmapCppi5SetDescType(hostDesc, descType);

            CSL_udmapCppi5SetReturnPolicy(hostDesc,
                                          descType,
                                          CSL_UDMAP_CPPI5_PD_PKTINFO2_RETPOLICY_VAL_ENTIRE_PKT,
                                          CSL_UDMAP_CPPI5_PD_PKTINFO2_EARLYRET_VAL_NO,
                                          CSL_UDMAP_CPPI5_PD_PKTINFO2_RETPUSHPOLICY_VAL_TO_TAIL,
                                          pRxFlow->cqRing->ringNum);
            CpswDma_dmaDescEnque(pRxFlow->hDmaDescPool, dmaDesc);
        }
    }
}

void CpswDma_deInitRxFreeDescQ(CpswDma_RxFlowObj *pRxFlow)
{
    CpswDma_DmaDesc *pDmaDesc;

    pDmaDesc = CpswDma_dmaDescDeque(pRxFlow->hDmaDescPool);

    while (NULL != pDmaDesc)
    {
        pRxFlow->rxFlowPrms.dmaDescFreeFxn(pRxFlow->rxFlowPrms.hCbArg, pDmaDesc);
        pDmaDesc = CpswDma_dmaDescDeque(pRxFlow->hDmaDescPool);
    }
}

/*!
 * \defgroup CpswDma_HELPERS_BUFF_DESC_QUEUE CpswDma Buffer Descriptor Queue Helpers
 * @{
 */

/*! Zero init the buffer descriptor queue */
void CpswDma_buffDescInit(CpswDma_CpswHpdDesc *hpdDesc)
{
    CSL_UdmapCppi5HMPD *hostDesc = &hpdDesc->hostDesc;

    uint32_t descType = (uint32_t)CSL_UDMAP_CPPI5_PD_DESCINFO_DTYPE_VAL_HOST;

    CSL_udmapCppi5SetDescType(hostDesc, descType);
    CSL_udmapCppi5SetEpiDataPresent(hostDesc, true);
    CSL_udmapCppi5SetPsDataLoc(hostDesc, TISCI_MSG_VALUE_RM_UDMAP_RX_FLOW_PS_END_PD);
    CSL_udmapCppi5SetPsDataLen(hostDesc, CPSW_DMA_PROTOCOL_SPECIFIC_INFO_BLOCK_SIZE);
    CSL_udmapCppi5HostSetPktLen(hostDesc, 0U);
    CSL_udmapCppi5SetPsFlags(hostDesc, 0U);
    CSL_udmapCppi5SetIds(hostDesc, descType, 0U, 0U);
    CSL_udmapCppi5SetSrcTag(hostDesc, 0x0000U);
    CSL_udmapCppi5SetDstTag(hostDesc, 0x0000U);
    CSL_udmapCppi5LinkDesc(hostDesc, 0U);
    CSL_udmapCppi5SetBufferAddr(hostDesc, (uint64_t)NULL);
    CSL_udmapCppi5SetBufferLen(hostDesc, 0U);
    CSL_udmapCppi5SetOrgBufferAddr(hostDesc, (uint64_t)NULL);
    CSL_udmapCppi5SetOrgBufferLen(hostDesc, 0U);
}

/*  Packet operations  */

/*!
 * \defgroup CpswDma_HELPERS_PACKET CpswDma Packet Helpers
 * @{
 */
/*! DMA descriptor Q init */
void CpswDma_dmaDescQInit(CpswDma_DmaDescQ *pDmaDescQ)
{
    pDmaDescQ->pHead = NULL;
    pDmaDescQ->pTail = NULL;
    pDmaDescQ->count = 0;
}

uint32_t CpswDma_dmaDescQCount(CpswDma_DmaDescQ *pDmaDescQ)
{
    return pDmaDescQ->count;
}

/*! Zero init the packet */
void CpswDma_dmaDescInit(CpswDma_DmaDesc *pDmaDesc)
{
    memset(pDmaDesc, 0, sizeof(*pDmaDesc));
    pDmaDesc->pNextDesc   = NULL;
    pDmaDesc->pAppPktInfo = NULL;

    CpswDma_buffDescInit(&pDmaDesc->hpdDesc);
}

void CpswDma_dmaDescEnque(CpswDma_DmaDescQ *pDmaDescQ,
                          CpswDma_DmaDesc *pDmaDesc)
{
    if (pDmaDesc != NULL)
    {
        uintptr_t key = CpswOsal_disableAllIntr();

        pDmaDesc->pNextDesc = NULL;

        if (NULL == pDmaDescQ->pHead)
        {
            /* Queue is empty, make head point to new packet */
            pDmaDescQ->pHead = pDmaDesc;
        }
        else
        {
            /* Queue is not empty, make previous tail point to new packet */
            pDmaDescQ->pTail->pNextDesc = pDmaDesc;
        }

        /* Make tail of queue point to new */
        pDmaDescQ->count++;
        pDmaDescQ->pTail = pDmaDesc;

        CpswOsal_restoreAllIntr(key);
    }
    else
    {
        Cpsw_assert(false);
    }
}

CpswDma_DmaDesc *CpswDma_dmaDescDeque(CpswDma_DmaDescQ *pDmaDescQ)
{
    CpswDma_DmaDesc *pDmaDesc = NULL;

    if (NULL != pDmaDescQ)
    {
        uintptr_t key = CpswOsal_disableAllIntr();
        pDmaDesc = pDmaDescQ->pHead;

        if (NULL != pDmaDesc)
        {
            pDmaDescQ->pHead = pDmaDesc->pNextDesc;
            if (NULL == pDmaDescQ->pHead)
            {
                pDmaDescQ->pTail = NULL;
            }

            pDmaDescQ->count--;
            pDmaDesc->pNextDesc = NULL;
        }

        CpswOsal_restoreAllIntr(key);
    }

    return(pDmaDesc);
}

void CpswDma_dmaDescCopyQ(CpswDma_DmaDescQ *pDstQueue,
                          const CpswDma_DmaDescQ *pSrcQueue)
{
    uintptr_t key = CpswOsal_disableAllIntr();

    pDstQueue->pHead = pSrcQueue->pHead;
    pDstQueue->pTail = pSrcQueue->pTail;
    pDstQueue->count = pSrcQueue->count;

    CpswOsal_restoreAllIntr(key);
}

int32_t CpswDma_registerEvent(CpswDma_udmaInfo *pUdmaInfo,
                              Udma_RingHandle hUdmaRing,
                              Udma_EventCallback eventCb,
                              void *hCbArg,
                              uint32_t eventType)
{
    int32_t retVal = UDMA_SOK;

    Udma_EventPrms evtPrms;

    /* Register for the UDMA completion event. The UDMA driver sets up
     * the Int Aggr and Int Router according to our preferred interrupt
     * number.  The interrupt registration itself doesn't happen here */
    UdmaEventPrms_init(&evtPrms);

    switch (eventType)
    {
        case UDMA_EVENT_TYPE_DMA_COMPLETION:
            evtPrms.chHandle = pUdmaInfo->hUdmaCh;
            break;

        case UDMA_EVENT_TYPE_RING:
            Cpsw_assert(NULL != hUdmaRing);
            evtPrms.ringHandle = hUdmaRing;
            break;

        case UDMA_EVENT_TYPE_RING_MON:
            Cpsw_assert(NULL != hUdmaRing);
            Cpsw_assert(NULL != pUdmaInfo->hUdmaRingMon);
            evtPrms.ringHandle = hUdmaRing;
            evtPrms.monHandle  = pUdmaInfo->hUdmaRingMon;
            break;

        default:
            CPSWTRACE_ERROR("[CPSW DMA Error] Event registration failed allocation failed !!\n");
            Cpsw_assert(false);
    }

    evtPrms.eventType         = eventType;;
    evtPrms.eventMode         = UDMA_EVENT_MODE_SHARED;
    evtPrms.masterEventHandle = NULL;
    evtPrms.eventCb           = eventCb;
    evtPrms.appData           = hCbArg;

    retVal = Udma_eventRegister(pUdmaInfo->hUdmaDrv, pUdmaInfo->hUdmaEvt, &evtPrms);

    return retVal;
}

int32_t CpswDma_unregisterEvent(Udma_EventHandle hUdmaEvt)
{
    int32_t retVal = UDMA_SOK;

    retVal = Udma_eventUnRegister(hUdmaEvt);

    return retVal;
}

int32_t CpswDma_dmaDescCheck(CpswDma_DmaDesc *dmaDesc,
                             CpswDma_Dir transferDir)
{
    int32_t retVal = UDMA_SOK;

#if (1U == CPSW_CFG_DEV_ERROR)
    CpswDma_CpswHpdDesc *pHpdDesc = (CpswDma_CpswHpdDesc *)dmaDesc;
    CSL_UdmapCppi5HMPD *hostDesc  = &pHpdDesc->hostDesc;

    /* Is the desc memory aligned? */
    if (0ULL != ((uint64_t)hostDesc % CPSW_DMA_HPD_SIZE))
    {
        retVal = UDMA_EBADARGS;
    }

    /* Is the desc of Host Packet type? */
    if (retVal == UDMA_SOK)
    {
        if ((uint32_t)CSL_UDMAP_CPPI5_PD_DESCINFO_DTYPE_VAL_HOST !=
            CSL_udmapCppi5GetDescType(hostDesc))
        {
            retVal = UDMA_EBADARGS;
        }
    }

    /* Is the buffer address not NULL? */
    if (retVal == UDMA_SOK)
    {
        if (0ULL == CSL_udmapCppi5GetBufferAddr(hostDesc))
        {
            retVal = UDMA_EBADARGS;
        }
    }

    /* Is the buffer length > 0? */
    if (retVal == UDMA_SOK)
    {
        if (0U == CSL_udmapCppi5GetBufferLen(hostDesc))
        {
            if (CPSW_DMA_DIR_TX == transferDir)
            {
                retVal = UDMA_EBADARGS;
            }
        }
    }
#endif
    return retVal;
}

/*! Rx flow object allocation function  */
CpswDma_RxFlowObj *CpswDma_memMgrAllocRxFlowObj(void)
{
    CpswDma_RxFlowObj *pRxFlow = NULL;
    CpswDma_RxFlowObjMem *pRxFlowMemObj;

    pRxFlowMemObj = (CpswDma_RxFlowObjMem *)CpswUtils_deQ(&gCpswDmaMemMgrObj.rxFlowObjMemQ);
    pRxFlow    = &pRxFlowMemObj->rxFlowObj;

    return pRxFlow;
}

/*! Rx flow object free function  */
void CpswDma_memMgrFreeRxFlowObj(CpswDma_RxFlowObj *pRxFlow)
{
    CpswDma_RxFlowObjMem *pRxFlowMemObj;

#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(NULL != pRxFlow);
#endif

    pRxFlowMemObj = container_of(pRxFlow, CpswDma_RxFlowObjMem, rxFlowObj);
    CpswUtils_enQ(&gCpswDmaMemMgrObj.rxFlowObjMemQ, &pRxFlowMemObj->node);
    pRxFlowMemObj = NULL;

    return;
}

/*! TX channel object allocation function  */
CpswDma_TxChObj *CpswDma_memMgrAllocTxChObj(void)
{
    CpswDma_TxChObj *pTxChObj = NULL;
    CpswDma_TxChObjMem *pTxChMemObj;

    pTxChMemObj = (CpswDma_TxChObjMem *)CpswUtils_deQ(&gCpswDmaMemMgrObj.txChObjMemQ);
    pTxChObj    = &pTxChMemObj->txChObj;

    return pTxChObj;
}

/*! TX channel object free function  */
void CpswDma_memMgrFreeTxChObj(CpswDma_TxChObj *pTxChObj)
{
    CpswDma_TxChObjMem *pTxChMemObj;

#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(NULL != pTxChObj);
#endif

    pTxChMemObj = container_of(pTxChObj, CpswDma_TxChObjMem, txChObj);
    CpswUtils_enQ(&gCpswDmaMemMgrObj.txChObjMemQ, &pTxChMemObj->node);
    pTxChObj = NULL;

    return;
}

/*! Ring Monitor object allocation function  */
Udma_RingMonHandle CpswDma_memMgrAllocRingMonObj(void)
{
    struct Udma_RingMonObj *pRingMonObj = NULL;
    CpswDma_RingMonObjMem *pRingMemObj;

    pRingMemObj = (CpswDma_RingMonObjMem *)
                  CpswUtils_deQ(&gCpswDmaMemMgrObj.ringMonObjMemQ);
    pRingMonObj = &pRingMemObj->ringMonObj;

    return pRingMonObj;
}

/*! Ring Monitor object free function  */
void CpswDma_memMgrFreeRingMonObj(Udma_RingMonHandle hUdmaRingMon)
{
    CpswDma_RingMonObjMem *pRingMemObj;

#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(NULL != hUdmaRingMon);
#endif

    pRingMemObj = container_of(hUdmaRingMon, CpswDma_RingMonObjMem, ringMonObj);
    CpswUtils_enQ(&gCpswDmaMemMgrObj.ringMonObjMemQ, &pRingMemObj->node);
    hUdmaRingMon = NULL;

    return;
}

/*! TD Cq Ring memory object allocation function  */
void *CpswDma_memMgrAllocTdCqRingMemObj(void)
{
    void *pTdCqRingMem = NULL;
    CpswDma_TdCqRingObjMem *pTdCqRingMemObj;

    pTdCqRingMemObj = (CpswDma_TdCqRingObjMem *)
                      CpswUtils_deQ(&gCpswDmaMemMgrObj.tdCqRingObjMemQ);
#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(NULL != pTdCqRingMemObj);
#endif
    pTdCqRingMem = (void *)&pTdCqRingMemObj->tdCqRingMemObj[0U];
    Cpsw_assert(CPSW_UTILS_IS_ALIGNED(pTdCqRingMem, UDMA_CACHELINE_ALIGNMENT));
    memset(pTdCqRingMem, 0U, sizeof(uint64_t));

    return pTdCqRingMem;
}

/*! TD Cq Ring memory object free function  */
void CpswDma_memMgrFreeTdCqRingMemObj(void *pTdCqRingMem)
{
    CpswDma_TdCqRingObjMem *pTdCqRingMemObj;

#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(NULL != pTdCqRingMem);
#endif

    pTdCqRingMemObj = container_of((const uint64_t (*)[CPSW_UTILS_ALIGN(CPSW_DMA_TDCQ_RING_ELE_CNT,
                                                                        UDMA_CACHELINE_ALIGNMENT)]) pTdCqRingMem,
                                   CpswDma_TdCqRingObjMem,
                                   tdCqRingMemObj);
    CpswUtils_enQ(&gCpswDmaMemMgrObj.tdCqRingObjMemQ, &pTdCqRingMemObj->node);
    pTdCqRingMem = NULL;
}

#ifdef QNX_OS
extern void *cpsw_alloc(size_t size, paddr64_t *paddr);
extern void cpsw_free(void *addr, size_t size);
#endif

void CpswDma_memMgrInit(void)
{
#ifdef QNX_OS
    uint32_t i;
    int32_t retVal = CPSW_SOK;
    uint32_t alignment_offset = 0;

    gCpswDmaMemBuf = cpsw_alloc(sizeof(CpswDma_Mem) + UDMA_CACHELINE_ALIGNMENT, (paddr64_t *)&gCpswDmaMemBufPhys);
    if (gCpswDmaMemBuf == MAP_FAILED)
    {
        CPSWTRACE_WARN("Error: alloc fail");
        retVal  = CPSW_EFAIL;
        Cpsw_assert(retVal == CPSW_SOK);
    }

    if (!CPSW_UTILS_IS_ALIGNED(gCpswDmaMemBuf, UDMA_CACHELINE_ALIGNMENT))
    {
        CPSWTRACE_WARN("Error: addr=0x%lx is not aligned", (uintptr_t)gCpswDmaMemBuf);
        gCpswDmaMem = (CpswDma_Mem *)(((uintptr_t)gCpswDmaMemBuf + UDMA_CACHELINE_ALIGNMENT) & ~(UDMA_CACHELINE_ALIGNMENT-1));
        alignment_offset = (uintptr_t)gCpswDmaMem - (uintptr_t)gCpswDmaMemBuf;
        gCpswDmaMemBufPhys += alignment_offset;
    }
    else
    {
        CPSWTRACE_WARN("addr=0x%lx is  aligned", (uintptr_t)gCpswDmaMemBuf);
        gCpswDmaMem = gCpswDmaMemBuf;
    }


    memset(&gCpswDmaMemMgrObj, 0U, sizeof(gCpswDmaMemMgrObj));
    memset(&gCpswDmaMem->gCpswDmaRxFlowObjMem, 0U, sizeof(gCpswDmaMem->gCpswDmaRxFlowObjMem));
    memset(&gCpswDmaMem->gCpswDmaTxChObjMem, 0U, sizeof(gCpswDmaMem->gCpswDmaTxChObjMem));
    memset(&gCpswDmaMem->gCpswDmaRinMonObjMem, 0U, sizeof(gCpswDmaMem->gCpswDmaRinMonObjMem));
    memset(&gCpswDmaMem->gCpswDmaTdCqRingObjMem, 0U, sizeof(gCpswDmaMem->gCpswDmaTdCqRingObjMem));

    /*********************** Rx flow object memory  Q ************************/
    CpswUtils_initQ(&gCpswDmaMemMgrObj.rxFlowObjMemQ);
    for (i = 0U; i < CPSW_CFG_RX_FLOWS_NUM; i++)
    {
        CpswUtils_enQ(&gCpswDmaMemMgrObj.rxFlowObjMemQ, &gCpswDmaMem->gCpswDmaRxFlowObjMem[i].node);
    }

    /*********************** Tx channel object memory  Q ************************/
    CpswUtils_initQ(&gCpswDmaMemMgrObj.txChObjMemQ);
    for (i = 0U; i < CPSW_CFG_TX_CHANNELS_NUM; i++)
    {
        CpswUtils_enQ(&gCpswDmaMemMgrObj.txChObjMemQ, &gCpswDmaMem->gCpswDmaTxChObjMem[i].node);
    }

    /*********************** Ring monitor object memory  Q ************************/
    CpswUtils_initQ(&gCpswDmaMemMgrObj.ringMonObjMemQ);
    for (i = 0U; i < CPSW_CFG_RING_MON_NUM; i++)
    {
        CpswUtils_enQ(&gCpswDmaMemMgrObj.ringMonObjMemQ, &gCpswDmaMem->gCpswDmaRinMonObjMem[i].node);
    }

    /*********************** Teardown CQ object memory  Q ************************/
    CpswUtils_initQ(&gCpswDmaMemMgrObj.tdCqRingObjMemQ);
    for (i = 0U; i < (CPSW_CFG_TX_CHANNELS_NUM + CPSW_DMA_RX_CHANNELS_NUM); i++)
    {
        CpswUtils_enQ(&gCpswDmaMemMgrObj.tdCqRingObjMemQ, &gCpswDmaMem->gCpswDmaTdCqRingObjMem[i].node);
    }

#else
    uint32_t i;

    memset(&gCpswDmaMemMgrObj, 0U, sizeof(gCpswDmaMemMgrObj));
    memset(&gCpswDmaRxFlowObjMem, 0U, sizeof(gCpswDmaRxFlowObjMem));
    memset(&gCpswDmaTxChObjMem, 0U, sizeof(gCpswDmaTxChObjMem));
    memset(&gCpswDmaRinMonObjMem, 0U, sizeof(gCpswDmaRinMonObjMem));
    memset(&gCpswDmaTdCqRingObjMem, 0U, sizeof(gCpswDmaTdCqRingObjMem));

    /*********************** Rx flow object memory  Q ************************/
    CpswUtils_initQ(&gCpswDmaMemMgrObj.rxFlowObjMemQ);
    for (i = 0U; i < CPSW_CFG_RX_FLOWS_NUM; i++)
    {
        CpswUtils_enQ(&gCpswDmaMemMgrObj.rxFlowObjMemQ, &gCpswDmaRxFlowObjMem[i].node);
    }

    /*********************** Tx channel object memory  Q ************************/
    CpswUtils_initQ(&gCpswDmaMemMgrObj.txChObjMemQ);
    for (i = 0U; i < CPSW_CFG_TX_CHANNELS_NUM; i++)
    {
        CpswUtils_enQ(&gCpswDmaMemMgrObj.txChObjMemQ, &gCpswDmaTxChObjMem[i].node);
    }

    /*********************** Ring monitor object memory  Q ************************/
    CpswUtils_initQ(&gCpswDmaMemMgrObj.ringMonObjMemQ);
    for (i = 0U; i < CPSW_CFG_RING_MON_NUM; i++)
    {
        CpswUtils_enQ(&gCpswDmaMemMgrObj.ringMonObjMemQ, &gCpswDmaRinMonObjMem[i].node);
    }

    /*********************** Teardown CQ object memory  Q ************************/
    CpswUtils_initQ(&gCpswDmaMemMgrObj.tdCqRingObjMemQ);
    for (i = 0U; i < (CPSW_CFG_TX_CHANNELS_NUM + CPSW_DMA_RX_CHANNELS_NUM); i++)
    {
        CpswUtils_enQ(&gCpswDmaMemMgrObj.tdCqRingObjMemQ, &gCpswDmaTdCqRingObjMem[i].node);
    }
#endif
}

void CpswDma_memMgrDeInit(void)
{
    memset(&gCpswDmaMemMgrObj, 0U, sizeof(gCpswDmaMemMgrObj));
#ifdef QNX_OS
    cpsw_free(gCpswDmaMemBuf, (sizeof(CpswDma_Mem) + UDMA_CACHELINE_ALIGNMENT));
#endif
}

void CpswDma_txCqIsr(Udma_EventHandle hUdmaEvt,
                     uint32_t eventType,
                     void *appData)
{
    int32_t retVal;
    CpswDma_TxChObj *pTxCh = (CpswDma_TxChObj *)appData;
    queue_t curIsrCQ;

    CpswUtils_initQ(&curIsrCQ);

    retVal = CpswDma_retrievePkts(pTxCh->cqRing,
                                  &curIsrCQ,
                                  pTxCh->hDmaDescPool,
                                  pTxCh->txChPrms.disableCacheOpsFlag,
                                  CPSW_DMA_DIR_TX);
    Cpsw_assert(retVal == UDMA_SOK);

    CpswUtils_appendQ(&pTxCh->cqIsrQ, &curIsrCQ);

    if ((pTxCh->txChPrms.notifyCb != NULL)
        &&
        (CpswUtils_getQCount(&curIsrCQ) != 0))
    {
        pTxCh->txChPrms.notifyCb(pTxCh->txChPrms.hCbArg);
    }
}

void CpswDma_rxCqIsr(Udma_EventHandle hUdmaEvt,
                     uint32_t eventType,
                     void *appData)
{
    int32_t retVal;
    CpswDma_RxFlowObj *pRxFlow = (CpswDma_RxFlowObj *)appData;
    queue_t curIsrCQ;

    CpswUtils_initQ(&curIsrCQ);

    retVal = CpswDma_retrievePkts(pRxFlow->cqRing,
                                  &curIsrCQ,
                                  pRxFlow->hDmaDescPool,
                                  pRxFlow->rxFlowPrms.disableCacheOpsFlag,
                                  CPSW_DMA_DIR_RX);
    Cpsw_assert(UDMA_SOK == retVal);

    CpswUtils_appendQ(&pRxFlow->cqIsrQ, &curIsrCQ);

    if ((pRxFlow->rxFlowPrms.notifyCb != NULL) &&
        (CpswUtils_getQCount(&curIsrCQ) != 0))
    {
        pRxFlow->rxFlowPrms.notifyCb(pRxFlow->rxFlowPrms.hCbArg);
    }
}

int32_t CpswDma_allocProxy(Udma_DrvHandle hUdmaDrv,
                           Udma_ProxyHandle hUdmaProxy,
                           Udma_RingHandle hUdmaRing)
{
    int32_t retVal;
    Udma_ProxyCfg proxyCfg;

    /* Allocate a proxy for queue operation */
    retVal = Udma_proxyAlloc(hUdmaDrv, hUdmaProxy, UDMA_PROXY_ANY);
    CPSWTRACE_ERROR_IF((UDMA_SOK != retVal), "CPSWDMA:Proxy Alloc Failed Status=0x%x\n", retVal);

    if (UDMA_SOK == retVal)
    {
        /* Config proxy for queue operation */
        proxyCfg.proxyMode = CSL_PROXY_QUEUE_ACCESS_MODE_TAIL;
        proxyCfg.elemSize  = UDMA_RING_ES_8BYTES;
        proxyCfg.ringNum   = Udma_ringGetNum(hUdmaRing);

        retVal = Udma_proxyConfig(hUdmaProxy, &proxyCfg);

        if (UDMA_SOK != retVal)
        {
            CPSWTRACE_ERROR("CPSWDMA:Proxy Config Failed Status=0x%x\n", retVal);
            Cpsw_assert(UDMA_SOK == Udma_proxyFree(hUdmaProxy));
        }
    }

    return retVal;
}

int32_t CpswDma_freeProxy(Udma_ProxyHandle hUdmaProxy)
{
    int32_t retVal = UDMA_EBADARGS;

    if (NULL != hUdmaProxy)
    {
        retVal = Udma_proxyFree(hUdmaProxy);
    }

    return retVal;
}

Udma_RingMonHandle CpswDma_allocRingMon(Udma_DrvHandle hUdmaDrv,
                                        Udma_RingHandle hUdmaRing,
                                        uint32_t elemCnt,
                                        CpswDma_RingMonCfg *pRingMonCfg)
{
    int32_t retVal = UDMA_SOK;
    Udma_RingMonPrms monPrms;
    Udma_RingMonHandle hRingMon;

    /* Allocate ring monitor object memory*/
    hRingMon = CpswDma_memMgrAllocRingMonObj();
    if (NULL == hRingMon)
    {
        CPSWTRACE_ERROR("Ring monitor object memory allocation failed!!\n");
        retVal = UDMA_EALLOC;
    }

    if (UDMA_SOK == retVal)
    {
        /* Allocate and configure a ring monitor */
        retVal = Udma_ringMonAlloc(hUdmaDrv, hRingMon, UDMA_RING_MON_ANY);
        if (UDMA_SOK != retVal)
        {
            CPSWTRACE_ERROR("Ring monitor alloc failed Status=0x%x\n", retVal);
            CpswDma_memMgrFreeRingMonObj(hRingMon);
        }
    }

    if (UDMA_SOK == retVal)
    {
        UdmaRingMonPrms_init(&monPrms);
        monPrms.source  = TISCI_MSG_VALUE_RM_MON_SRC_ELEM_CNT;
        monPrms.mode    = pRingMonCfg->mode;
        monPrms.ringNum = Udma_ringGetNum(hUdmaRing);
        monPrms.data0   = pRingMonCfg->data0;
        monPrms.data1   = pRingMonCfg->data1;

        retVal = Udma_ringMonConfig(hRingMon, &monPrms);
        if (UDMA_SOK != retVal)
        {
            CPSWTRACE_ERROR(" Ring monitor config failed Status=0x%x\n", retVal);
            /*If free also fails, we can't do much but assert */
            Cpsw_assert(UDMA_SOK == Udma_ringMonFree(hRingMon));
            hRingMon = NULL;
        }
    }

    return hRingMon;
}

int32_t CpswDma_freeRingMon(Udma_RingMonHandle hRingMon)
{
    int32_t retVal = UDMA_EBADARGS;

    if (NULL != hRingMon)
    {
        retVal = Udma_ringMonFree(hRingMon);
        CpswDma_memMgrFreeRingMonObj(hRingMon);
    }

    return retVal;
}

void CpswDmaStats_updateNotifyStats(CpswDma_CbFxnStats *pktStats,
                                    uint32_t pktCnt,
                                    uint32_t timeDiff)
{
#if defined(CPSWDMA_INSTRUMENTATION_ENABLED)
    uint32_t notifyCount;
    uint32_t timePerPkt = timeDiff / pktCnt;

    if (pktCnt == 0)
    {
        pktStats->zeroNotifyCnt++;
    }

    notifyCount = pktStats->dataNotifyCnt & (CPSW_DMA_STATS_HISTORY_CNT - 1U);
    pktStats->dataNotifyCnt++;

    pktStats->totalPktCnt   += pktCnt;
    pktStats->totalCycleCnt += timeDiff;

    pktStats->cycleCntPerNotify[notifyCount] = timeDiff;
    pktStats->pktsPerNotify[notifyCount]     = pktCnt;
    pktStats->cycleCntPerPkt[notifyCount]    = timePerPkt;

    if (notifyCount != 0)
    {
        /* As app submits all packets for Rx during init, we ignore that to get run-time stats */
        if (timeDiff > pktStats->cycleCntPerNotifyMax)
        {
            pktStats->cycleCntPerNotifyMax = timeDiff;
        }

        if (pktCnt > pktStats->pktsPerNotifyMax)
        {
            pktStats->pktsPerNotifyMax = pktCnt;
        }

        if (timePerPkt > pktStats->cycleCntPerPktMax)
        {
            pktStats->cycleCntPerPktMax = timePerPkt;
        }
    }
#endif
}

/*!
 * @}
 */
