/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file cpsw_hostport.c
 *
 * \brief This file contains the Host Port submodule related functionality.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/csl/csl_cpswitch.h>
#include <ti/drv/cpsw/include/cpsw_types.h>
#include <ti/drv/cpsw/include/cpsw_hostport.h>
#include "priv/cpsw_hostport_priv.h"
#include "priv/cpsw_priv.h"
#include "priv/cpsw_trace_priv.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define CPSW_HOSTPORT_START_REG_OFFSET              (CSL_XGE_CPSW_P0_CONTROL_REG)
#define CPSW_HOSTPORT_END_REG_OFFSET                (CSL_XGE_CPSW_P0_HOST_BLKS_PRI_REG)

/*! \brief Rate limiting count to transfer rate divider */
#define CPSW_HOSTPORT_RATELIM_DIV             (32768U)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static CpswHostPort_Handle CpswHostPort_getObject(Cpsw_Type cpswType);

static void CpswHostPort_getFifoStats(CpswHostPort_Handle hPort,
                                      CpswHostPort_FifoStats *stats);

static int32_t CpswHostPort_setTxPriority(CpswHostPort_Handle hPort,
                                          CpswPort_TxPriority *txPriority);

static void CpswHostPort_getTxPriority(CpswHostPort_Handle hPort,
                                       CpswPort_TxPriority *txPriority);

static int32_t CpswHostPort_setRxPriority(CpswHostPort_Handle hPort,
                                          CpswPort_RxPriority *rxPriority);

static void CpswHostPort_getRxPriority(CpswHostPort_Handle hPort,
                                       CpswPort_RxPriority *rxPriority);

static int32_t CpswHostPort_setDscpPriority(CpswHostPort_Handle hPort,
                                            CpswPort_DscpPriority *dscpPriority);

static void CpswHostPort_getDscpPriority(CpswHostPort_Handle hPort,
                                         CpswPort_DscpPriority *dscpPriority);

static uint64_t CpswHostPort_mapBwToCnt(uint64_t rateInBps,
                                        uint32_t cppiClkFreqHz);

static uint64_t CpswHostPort_mapCntToBw(uint32_t cntVal,
                                        uint32_t cppiClkFreqHz);

static int32_t CpswHostPort_setTrafficShaping(CpswHostPort_Handle hPort,
                                              const CpswPort_TrafficShapingConfig *config);

static int32_t CpswHostPort_getTrafficShaping(CpswHostPort_Handle hPort,
                                              CpswPort_TrafficShapingConfig *config);

static void CpswHostPort_printRegs(CSL_Xge_cpswRegs *cpswRegs);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/*! \brief CPSW host port objects */
CpswHostPort_Obj hostPortObjs[CPSW_COUNT];

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

void CpswHostPort_initParams(CpswHostPort_Config *cfg)
{
    cfg->crcType                     = CPSW_CRC_ETHERNET;
    cfg->removeCrc                   = false;
    cfg->padShortPacket              = true; /* Short packets were dropped by default */
    cfg->passCrcErrors               = false;
    cfg->passPriorityTaggedUnchanged = false;
    /* Keep checksum offload enabled by default as no side-effect if app doesn't use it */
    cfg->enableCsumOffload           = true;
    cfg->enableRxVlanRemap           = false;
    cfg->enableRxDscpIPv4Remap       = false;
    cfg->enableRxDscpIPv6Remap       = false;
    cfg->rxMtu                       = CPSW_HOSTPORT_RX_MTU_DEFAULT;
    cfg->vlanCfg.portPri             = 0U;
    cfg->vlanCfg.portCfi             = 0U;
    cfg->vlanCfg.portVID             = 0U;
    cfg->rxPriorityType              = CPSW_RX_PRIORITY_TYPE_FIXED;
    cfg->txPriorityType              = CPSW_TX_PRIORITY_TYPE_FIXED;
}

CpswHostPort_Handle CpswHostPort_open(Cpsw_Type cpswType,
                                      const CpswHostPort_Config *cfg)
{
    CpswHostPort_Handle hPort = CpswHostPort_getObject(cpswType);
    CSL_CPSW_CONTROL control;
    CSL_CPSW_PTYPE pType;
    CSL_CPSW_CPPI_P0_CONTROL cppiP0ControlCfg;

    memset(hPort, 0U, sizeof(*hPort));

    hPort->cpswRegs = Cpsw_getCslRegs(cpswType);
    hPort->cpswType = cpswType;
    Cpsw_assert(hPort->cpswRegs != NULL);

    if (cfg->crcType == CPSW_CRC_ETHERNET)
    {
        CSL_CPSW_disableP0TxCastagnoliCRC(hPort->cpswRegs);
    }
    else
    {
        CSL_CPSW_enableP0TxCastagnoliCRC(hPort->cpswRegs);
    }

    CSL_CPSW_getCpswControlReg(hPort->cpswRegs, &control);
    control.p0Enable       = false;
    control.p0PassPriTag   = cfg->passPriorityTaggedUnchanged;
    control.p0TxCrcRemove  = cfg->removeCrc;
    control.p0RxPad        = cfg->padShortPacket;
    control.p0RxPassCrcErr = cfg->passCrcErrors;
    CSL_CPSW_setCpswControlReg(hPort->cpswRegs, &control);

    CSL_CPSW_setPort0VlanReg(hPort->cpswRegs,
                             cfg->vlanCfg.portVID,
                             cfg->vlanCfg.portCfi,
                             cfg->vlanCfg.portPri);

    CSL_CPSW_setPort0RxMaxLen(hPort->cpswRegs, cfg->rxMtu);

    CSL_CPSW_getPTypeReg(hPort->cpswRegs, &pType);
    if (cfg->txPriorityType == CPSW_TX_PRIORITY_TYPE_FIXED)
    {
        pType.port0PriorityTypeEscalateEnable = FALSE;
    }
    else
    {
        pType.port0PriorityTypeEscalateEnable = TRUE;
    }

    CSL_CPSW_setPTypeReg(hPort->cpswRegs, &pType);

    CSL_CPSW_getCppiP0Control(hPort->cpswRegs, &cppiP0ControlCfg);

    if (cfg->enableCsumOffload)
    {
        cppiP0ControlCfg.p0RxChksumEn = TRUE;
    }
    else
    {
        cppiP0ControlCfg.p0RxChksumEn = FALSE;
    }

    cppiP0ControlCfg.p0RxRemapVlan     = cfg->enableRxVlanRemap ? TRUE : FALSE;
    cppiP0ControlCfg.p0RxRemapDscpIpv4 = cfg->enableRxDscpIPv4Remap ? TRUE : FALSE;
    cppiP0ControlCfg.p0RxRemapDscpIpv6 = cfg->enableRxDscpIPv6Remap ? TRUE : FALSE;

    CSL_CPSW_setCppiP0Control(hPort->cpswRegs, &cppiP0ControlCfg);

    hPort->magic = CPSW_HOSTPORT_MAGIC;

    return hPort;
}

CpswHostPort_Handle CpswHostPort_reopen(Cpsw_Type cpswType)
{
    CpswHostPort_Handle hPort = CpswHostPort_getObject(cpswType);

    memset(hPort, 0U, sizeof(*hPort));

    hPort->cpswType = cpswType;
    hPort->cpswRegs = Cpsw_getCslRegs(cpswType);
    Cpsw_assert(hPort->cpswRegs != NULL);

    hPort->magic = CPSW_HOSTPORT_MAGIC;

    return hPort;
}

void CpswHostPort_close(CpswHostPort_Handle hPort)
{
    if (hPort->magic == CPSW_HOSTPORT_MAGIC)
    {
        CSL_CPSW_disablePort0(hPort->cpswRegs);

        hPort->magic    = CPSW_NO_MAGIC;
        hPort->cpswRegs = NULL;
    }
    else
    {
        CPSWTRACE_ERROR("Host port is not open (handle=%p)\n", hPort);
    }
}

int32_t CpswHostPort_ioctl(CpswHostPort_Handle hPort,
                           uint32_t cmd,
                           Cpsw_IoctlPrms *prms)
{
    int32_t status = CPSW_SOK;

    if (hPort->magic != CPSW_HOSTPORT_MAGIC)
    {
        CPSWTRACE_ERROR("Host port is not open (handle=%p)\n", hPort);
        status = CPSW_EINVALID_PARAMS;
    }

    if (status == CPSW_SOK)
    {
        switch (cmd)
        {
            case CPSW_HOSTPORT_IOCTL_ENABLE:
            {
                CSL_CPSW_enablePort0(hPort->cpswRegs);
            }
            break;

            case CPSW_HOSTPORT_IOCTL_DISABLE:
            {
                CSL_CPSW_disablePort0(hPort->cpswRegs);
            }
            break;

            case CPSW_HOSTPORT_IOCTL_GET_FIFO_STATS:
            {
                status = Cpsw_checkOutArgs(prms, sizeof(CpswHostPort_FifoStats));
                if (status == CPSW_SOK)
                {
                    CpswHostPort_FifoStats *outArgs =
                        (CpswHostPort_FifoStats *)prms->outArgs;

                    CpswHostPort_getFifoStats(hPort, outArgs);
                }
                else
                {
                    CPSWTRACE_ERROR("invalid GET_FIFO_STATS params\n");
                }
            }
            break;

            case CPSW_HOSTPORT_IOCTL_SET_TX_PRIORITY:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswPort_TxPriority));
                if (status == CPSW_SOK)
                {
                    CpswPort_TxPriority *txPriority =
                        (CpswPort_TxPriority *)prms->inArgs;

                    status = CpswHostPort_setTxPriority(hPort, txPriority);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "failed to set TX priority: %d\n", status);
                }
                else
                {
                    CPSWTRACE_ERROR("invalid SET_TX_PRIORITY params\n");
                }
            }
            break;

            case CPSW_HOSTPORT_IOCTL_GET_TX_PRIORITY:
            {
                status = Cpsw_checkOutArgs(prms, sizeof(CpswPort_TxPriority));
                if (status == CPSW_SOK)
                {
                    CpswPort_TxPriority *txPriority =
                        (CpswPort_TxPriority *)prms->outArgs;

                    CpswHostPort_getTxPriority(hPort, txPriority);
                }
                else
                {
                    CPSWTRACE_ERROR("invalid GET_TX_PRIORITY params\n");
                }
            }
            break;

            case CPSW_HOSTPORT_IOCTL_SET_RX_PRIORITY:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswPort_RxPriority));
                if (status == CPSW_SOK)
                {
                    CpswPort_RxPriority *rxPriority =
                        (CpswPort_RxPriority *)prms->inArgs;

                    status = CpswHostPort_setRxPriority(hPort, rxPriority);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "failed to set RX priority: %d\n", status);
                }
                else
                {
                    CPSWTRACE_ERROR("invalid SET_RX_PRIORITY params\n");
                }
            }
            break;

            case CPSW_HOSTPORT_IOCTL_GET_RX_PRIORITY:
            {
                status = Cpsw_checkOutArgs(prms, sizeof(CpswPort_RxPriority));
                if (status == CPSW_SOK)
                {
                    CpswPort_RxPriority *rxPriority =
                        (CpswPort_RxPriority *)prms->outArgs;

                    CpswHostPort_getRxPriority(hPort, rxPriority);
                }
                else
                {
                    CPSWTRACE_ERROR("invalid GET_RX_PRIORITY params\n");
                }
            }
            break;

            case CPSW_HOSTPORT_IOCTL_SET_RX_DSCP_PRIORITY:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswPort_DscpPriority));
                if (status == CPSW_SOK)
                {
                    CpswPort_DscpPriority *dscpPriority =
                        (CpswPort_DscpPriority *)prms->inArgs;

                    status = CpswHostPort_setDscpPriority(hPort, dscpPriority);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "failed to set DSCP priority: %d\n", status);
                }
                else
                {
                    CPSWTRACE_ERROR("invalid SET_RX_DSCP_PRIORITY params\n");
                }
            }
            break;

            case CPSW_HOSTPORT_IOCTL_GET_RX_DSCP_PRIORITY:
            {
                status = Cpsw_checkOutArgs(prms, sizeof(CpswPort_DscpPriority));
                if (status == CPSW_SOK)
                {
                    CpswPort_DscpPriority *dscpPriority =
                        (CpswPort_DscpPriority *)prms->outArgs;

                    CpswHostPort_getDscpPriority(hPort, dscpPriority);
                }
                else
                {
                    CPSWTRACE_ERROR("invalid GET_RX_DSCP_PRIORITY params\n");
                }
            }
            break;

            case CPSW_HOSTPORT_IOCTL_SET_TRAFFIC_SHAPING:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswPort_TrafficShapingConfig));
                if (status == CPSW_SOK)
                {
                    CpswPort_TrafficShapingConfig *trafficShapingConfig =
                        (CpswPort_TrafficShapingConfig *)prms->inArgs;

                    status = CpswHostPort_setTrafficShaping(hPort, trafficShapingConfig);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "failed to set traffic shaping: %d\n", status);
                }
                else
                {
                    CPSWTRACE_ERROR("invalid SET_TRAFFIC_SHAPING params\n");
                }
            }
            break;

            case CPSW_HOSTPORT_IOCTL_GET_TRAFFIC_SHAPING:
            {
                status = Cpsw_checkOutArgs(prms, sizeof(CpswPort_TrafficShapingConfig));
                if (status == CPSW_SOK)
                {
                    CpswPort_TrafficShapingConfig *trafficShapingConfig =
                        (CpswPort_TrafficShapingConfig *)prms->outArgs;

                    status = CpswHostPort_getTrafficShaping(hPort, trafficShapingConfig);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "failed to get traffic shaping: %d\n", status);
                }
                else
                {
                    CPSWTRACE_ERROR("invalid GET_TRAFFIC_SHAPING params\n");
                }
            }
            break;

            case CPSW_HOSTPORT_IOCTL_PRINT_REGS:
            {
                CpswHostPort_printRegs(hPort->cpswRegs);
            }
            break;

            case CPSW_HOSTPORT_SET_FLOW_ID_OFFSET:
            {
                status = Cpsw_checkInArgs(prms, sizeof(uint32_t));
                if (status == CPSW_SOK)
                {
                    uint32_t *flowIdOffset = (uint32_t *)prms->inArgs;

                    CSL_CPSW_setPort0FlowIdOffset(hPort->cpswRegs, *flowIdOffset);
                }
                else
                {
                    CPSWTRACE_ERROR("invalid SET_FLOW_ID_OFFSET params\n");
                }
            }
            break;

            case CPSW_HOSTPORT_IOCTL_GET_RX_MTU:
            {
                status = Cpsw_checkOutArgs(prms, sizeof(uint32_t));
                if (status == CPSW_SOK)
                {
                    uint32_t *rxMtu = (uint32_t *)prms->outArgs;

                    *rxMtu = CSL_CPSW_getPort0RxMaxLen(hPort->cpswRegs);
                }
                else
                {
                    CPSWTRACE_ERROR("invalid GET_RX_MTU params\n");
                }
            }
            break;

            case CPSW_HOSTPORT_GET_FLOW_ID_OFFSET:
            {
                status = Cpsw_checkOutArgs(prms, sizeof(uint32_t));
                if (status == CPSW_SOK)
                {
                    uint32_t *flowIdOffset = (uint32_t *)prms->outArgs;

                    *flowIdOffset = CSL_CPSW_getPort0FlowIdOffset(hPort->cpswRegs);
                }
                else
                {
                    CPSWTRACE_ERROR("invalid SET_FLOW_ID_OFFSET params\n");
                }
            }
            break;

            case CPSW_HOSTPORT_IS_CSUM_OFFLOAD_ENABLE:
            {
                status = Cpsw_checkOutArgs(prms, sizeof(bool));
                if (status == CPSW_SOK)
                {
                    bool *csumOffloadFlg = (bool *)prms->outArgs;
                    CSL_CPSW_CPPI_P0_CONTROL cppiP0ControlCfg;

                    CSL_CPSW_getCppiP0Control(hPort->cpswRegs, &cppiP0ControlCfg);
                    if (cppiP0ControlCfg.p0RxChksumEn)
                    {
                        *csumOffloadFlg = true;
                    }
                    else
                    {
                        *csumOffloadFlg = false;
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("invalid CPSW_HOSTPORT_IS_CSUM_OFFLOAD_ENABLE params\n");
                }
            }
            break;

            default:
            {
                CPSWTRACE_ERROR("invalid IOCTL cmd: %u\n", cmd);
                status = CPSW_EINVALID_PARAMS;
            }
            break;
        }
    }

    return status;
}

static CpswHostPort_Handle CpswHostPort_getObject(Cpsw_Type cpswType)
{
    CpswHostPort_Handle hPort = NULL;

    switch (cpswType)
    {
        case CPSW_2G:
        case CPSW_5G:
        case CPSW_9G:
            hPort = &hostPortObjs[cpswType];
            break;

        default:
            Cpsw_assert(false);
    }

    return hPort;
}

static void CpswHostPort_getFifoStats(CpswHostPort_Handle hPort,
                                      CpswHostPort_FifoStats *stats)
{
    CSL_CPSW_THRURATE thruRate;
    CSL_CPSW_CPPI_P0_FIFOSTATUS p0FifoStatus;

    CSL_CPSW_getPortBlockCountReg(hPort->cpswRegs, 0U,
                                  &stats->rxBlockCountExpress,
                                  &stats->rxBlockCountPreempt,
                                  &stats->txBlockCount);
    CSL_CPSW_getThruRateReg(hPort->cpswRegs, &thruRate);

    stats->rxThroughputRate = thruRate.cppiRxThruRate;

    CSL_CPSW_getP0FifoStatus(hPort->cpswRegs, &p0FifoStatus);

    stats->txActiveFifo[0] = p0FifoStatus.p0TxPriActivePri0;
    stats->txActiveFifo[1] = p0FifoStatus.p0TxPriActivePri1;
    stats->txActiveFifo[2] = p0FifoStatus.p0TxPriActivePri2;
    stats->txActiveFifo[3] = p0FifoStatus.p0TxPriActivePri3;
    stats->txActiveFifo[4] = p0FifoStatus.p0TxPriActivePri4;
    stats->txActiveFifo[5] = p0FifoStatus.p0TxPriActivePri5;
    stats->txActiveFifo[6] = p0FifoStatus.p0TxPriActivePri6;
    stats->txActiveFifo[7] = p0FifoStatus.p0TxPriActivePri7;
}

static int32_t CpswHostPort_setTxPriority(CpswHostPort_Handle hPort,
                                          CpswPort_TxPriority *txPriority)
{
    int32_t status = CPSW_SOK;
    uint32_t i;

    for (i = 0U; i < CPSW_HEADER_PACKET_PRIORITY_NUM; i++)
    {
        if (txPriority->priorityMap[i] >= CPSW_PRIORITY_NUM)
        {
            CPSWTRACE_ERROR("invalid priority map %u -> %u\n",
                            i, txPriority->priorityMap[i]);
            status = CPSW_EINVALID_PARAMS;
            break;
        }
    }

    if (status == CPSW_SOK)
    {
        CSL_CPSW_setPortTxPriMapReg(hPort->cpswRegs, 0U,
                                    txPriority->priorityMap);
    }

    return status;
}

static void CpswHostPort_getTxPriority(CpswHostPort_Handle hPort,
                                       CpswPort_TxPriority *txPriority)
{
    CSL_CPSW_getPortTxPriMapReg(hPort->cpswRegs, 0U,
                                txPriority->priorityMap);
}

static int32_t CpswHostPort_setRxPriority(CpswHostPort_Handle hPort,
                                          CpswPort_RxPriority *rxPriority)
{
    int32_t status = CPSW_SOK;
    uint32_t i;

    for (i = 0U; i < CPSW_PACKET_PRIORITY_NUM; i++)
    {
        if (rxPriority->priorityMap[i] >= CPSW_PRIORITY_NUM)
        {
            CPSWTRACE_ERROR("invalid priority map %u -> %u\n",
                            i, rxPriority->priorityMap[i]);
            status = CPSW_EINVALID_PARAMS;
            break;
        }
    }

    if (status == CPSW_SOK)
    {
        CSL_CPSW_setPortRxPriMapReg(hPort->cpswRegs, 0U,
                                    rxPriority->priorityMap);
    }

    return status;
}

static void CpswHostPort_getRxPriority(CpswHostPort_Handle hPort,
                                       CpswPort_RxPriority *rxPriority)
{
    CSL_CPSW_getPortRxPriMapReg(hPort->cpswRegs, 0U,
                                rxPriority->priorityMap);
}

static int32_t CpswHostPort_setDscpPriority(CpswHostPort_Handle hPort,
                                            CpswPort_DscpPriority *dscpPriority)
{
    CSL_CPSW_PORT_CONTROL control;
    int32_t status = CPSW_SOK;
    uint32_t i;

    for (i = 0U; i < CPSW_TOS_PRIORITY_NUM; i++)
    {
        if (dscpPriority->tosMap[i] >= CPSW_PRIORITY_NUM)
        {
            CPSWTRACE_ERROR("invalid priority map %u -> %u\n",
                            i, dscpPriority->tosMap[i]);
            status = CPSW_EINVALID_PARAMS;
            break;
        }
    }

    if (status == CPSW_SOK)
    {
        CSL_CPSW_setPortRxDscpMap(hPort->cpswRegs, 0U, dscpPriority->tosMap);

        CSL_CPSW_getPortControlReg(hPort->cpswRegs, 0U, &control);

        control.dscpIpv4Enable = dscpPriority->enableDscpIPv4;
        control.dscpIpv6Enable = dscpPriority->enableDscpIPv6;

        CSL_CPSW_setPortControlReg(hPort->cpswRegs, 0U, &control);
    }

    return status;
}

static void CpswHostPort_getDscpPriority(CpswHostPort_Handle hPort,
                                         CpswPort_DscpPriority *dscpPriority)
{
    CSL_CPSW_PORT_CONTROL control;

    CSL_CPSW_getPortRxDscpMap(hPort->cpswRegs, 0U,
                              dscpPriority->tosMap);

    CSL_CPSW_getPortControlReg(hPort->cpswRegs, 0U,
                               &control);

    dscpPriority->enableDscpIPv4 = control.dscpIpv4Enable;
    dscpPriority->enableDscpIPv6 = control.dscpIpv6Enable;
}

static uint64_t CpswHostPort_mapBwToCnt(uint64_t rateInBps,
                                        uint32_t cppiClkFreqHz)
{
    /*
     * Transfer rate to CIR or EIR count value equation:
     *
     *                Transfer rate (Mbps) * 32768
     * Count value = ------------------------------
     *                   CPPI_ICLK freq (MHz)
     *
     * Round division to prevent transfer rates below traffic shaping resolution
     * end up with no shaping at all.
     */
    return CPSW_UTILS_DIV_ROUNDUP(rateInBps * CPSW_HOSTPORT_RATELIM_DIV, cppiClkFreqHz);
}

static uint64_t CpswHostPort_mapCntToBw(uint32_t cntVal,
                                        uint32_t cppiClkFreqHz)
{
    /*
     * CIR or EIR count value to transfer rate equation:
     *
     *                         count value * CPPI_ICLK freq (MHz)
     * Transfer rate (Mbps) = ------------------------------------
     *                                      32768
     */
    return ((uint64_t)cntVal * cppiClkFreqHz) / CPSW_HOSTPORT_RATELIM_DIV;
}

static int32_t CpswHostPort_setTrafficShaping(CpswHostPort_Handle hPort,
                                              const CpswPort_TrafficShapingConfig *config)
{
    const CpswSoc_Config *pSocCfg = CpswSoc_getConfig(hPort->cpswType);
    uint32_t cppiClkFreqHz        = pSocCfg->cppiClkFreqHz;
    uint64_t cirBps;
    uint64_t eirBps;
    uint64_t cir;
    uint64_t eir;
    uint32_t i;
    int32_t status      = CPSW_SOK;
    bool rateLimEnabled = false;

    for (i = 0U; i < CPSW_PRIORITY_NUM; i++)
    {
        cirBps = config->rates[i].committedRateBitsPerSec;
        eirBps = config->rates[i].excessRateBitsPerSec;
        cir    = CpswHostPort_mapBwToCnt(cirBps, cppiClkFreqHz);
        eir    = CpswHostPort_mapBwToCnt(eirBps, cppiClkFreqHz);

        /* CIR must be non-zero if EIR is non-zero */
        if ((cir == 0ULL) && (eir != 0ULL))
        {
            CPSWTRACE_ERROR("EIR is enabled (%llubps = %llu) but CIR is not (%llubps = %llu) "
                            "for priority %u\n",
                            eirBps, eir, cirBps, cir, i);
            status = CPSW_EINVALID_PARAMS;
            break;
        }

        /* Rate limit must be the highest priority channels */
        if ((cir == 0ULL) && rateLimEnabled)
        {
            CPSWTRACE_ERROR("rate limiting disabled for priority %u\n", i);
            status = CPSW_EINVALID_PARAMS;
            break;
        }

        /* Out of range */
        if ((cir > 0x0FFFFFFFULL) || (eir > 0x0FFFFFFFULL))
        {
            CPSWTRACE_ERROR("invalid CIR=%llu (%llubps) EIR=%llu (%llubps) for priority %u\n",
                            cir, cirBps, eir, eirBps, i);
            status = CPSW_EINVALID_PARAMS;
            break;
        }

        CPSWTRACE_DEBUG("rate limiting %s for priority %u, CIR=%ubps (%llu) EIR=%ubps (%llu)\n",
                        (cir != 0ULL) ? "enabled" : "disabled",
                        i, cirBps, cir, eirBps, eir);

        CSL_CPSW_setCppiPriCirEir(hPort->cpswRegs,
                                  i, (uint32_t)cir, (uint32_t)eir);

        if (cir != 0ULL)
        {
            rateLimEnabled = true;
        }
    }

    return status;
}

static int32_t CpswHostPort_getTrafficShaping(CpswHostPort_Handle hPort,
                                              CpswPort_TrafficShapingConfig *config)
{
    const CpswSoc_Config *pSocCfg = CpswSoc_getConfig(hPort->cpswType);
    uint32_t cppiClkFreqHz        = pSocCfg->cppiClkFreqHz;
    uint32_t cir;
    uint32_t eir;
    uint64_t cirBps;
    uint64_t eirBps;
    uint32_t i;
    int32_t status = CPSW_SOK;

    for (i = 0U; i < CPSW_PRIORITY_NUM; i++)
    {
        CSL_CPSW_getCppiPriCirEir(hPort->cpswRegs, i, &cir, &eir);

        cirBps = CpswHostPort_mapCntToBw(cir, cppiClkFreqHz);
        eirBps = CpswHostPort_mapCntToBw(eir, cppiClkFreqHz);

        /* CIR must be non-zero if EIR is non-zero */
        if ((cirBps == 0ULL) && (eirBps != 0ULL))
        {
            CPSWTRACE_ERROR("EIR is enabled (%llubps = %u) but CIR is not (%llubps = %u) "
                            "for priority %u\n",
                            eirBps, eir, cirBps, cir, i);
            status = CPSW_EUNEXPECTED;
            break;
        }

        CPSWTRACE_DEBUG("rate limiting %s for priority %u, CIR=%llubps (%u) EIR=%llubps (%u)\n",
                        (cir != 0U) ? "enabled" : "disabled",
                        i, cirBps, cir, eirBps, eir);

        config->rates[i].committedRateBitsPerSec = cirBps;
        config->rates[i].excessRateBitsPerSec    = eirBps;
    }

    return status;
}

static void CpswHostPort_printRegs(CSL_Xge_cpswRegs *cpswRegs)
{
    uint32_t *regAddr = (uint32_t *)((uintptr_t)cpswRegs + CPSW_HOSTPORT_START_REG_OFFSET);
    uint32_t regIdx   = 0;

    while (((uintptr_t)regAddr) <= ((uintptr_t)cpswRegs + CPSW_HOSTPORT_END_REG_OFFSET))
    {
        if (*regAddr)
        {
            CpswUtils_printf("HOSTPORT: %u: 0x%08x\n", regIdx, *regAddr);
        }

        regAddr++;
        regIdx++;
    }
}
