/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file cpsw_macport.c
 *
 * \brief This file contains MAC Port submodule related functionality.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/csl/csl_cpswitch.h>
#include <ti/drv/cpsw/cpsw.h>
#include "priv/cpsw_macport_priv.h"
#include "priv/cpsw_priv.h"
#include "priv/cpsw_trace_priv.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define CPSW_MACPORT_RX_MTU_DEFAULT           (1518U)

#define CPSW_MACPORT_INTERVLAN_ROUTEID_BASE   (1U)

#define CPSW_MACPORT_START_REG_OFFSET         (offsetof(CSL_Xge_cpswEnetportRegs, PN_CONTROL_REG))

#define CPSW_MACPORT_END_REG_OFFSET           (offsetof(CSL_Xge_cpswEnetportRegs, PN_INTERVLAN_OPX_D_REG))

/*! \brief Rate limiting count to transfer rate divider */
#define CPSW_MACPORT_RATELIM_DIV              (32768U)

/*! \brief PTP sequence ID offset minimum value */
#define CPSW_MACPORT_PTP_SEQID_OFFSET_MIN_VAL (0x6U)


/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static void CpswMacPort_validateSgmiiIPVersion(CpswMacPort_Handle hPort);

static void CpswMacPort_resetSgmiiPort(CpswMacPort_Handle hPort);

static int32_t CpswMacPort_configSgmii(CpswMacPort_Handle hPort,
                                       const CpswMacPort_Config *macCfg,
                                       CpswMacPort_LinkConfig *linkCfg);

static int32_t CpswMacPort_checkSgmiiAutoNegStatus(CpswMacPort_Handle hPort);

static CpswMacPort_Handle CpswMacPort_getMacPortObject(Cpsw_Type cpswType,
                                                       Cpsw_MacPort portNum);

static void CpswMacPort_reset(CpswMacPort_Handle hPort);

static bool CpswMacPort_matchesEnetCtrl(CpswMacPort_Handle hPort,
                                        const CpswMacPort_Interface *interface);

static int32_t CpswMacPort_setInterface(CpswMacPort_Handle hPort,
                                        const CpswMacPort_Config *macCfg,
                                        const CpswMacPort_Interface *interface,
                                        CpswMacPort_LinkConfig *linkConfig);

static int32_t CpswMacPort_enablePort(CpswMacPort_Handle hPort,
                                      CpswMacPort_LinkConfig *cfg);

static void CpswMacPort_disablePort(CpswMacPort_Handle hPort);

static bool CpswMacPort_isPortEnabled(CpswMacPort_Handle hPort);

static void CpswMacPort_getFifoStats(CpswMacPort_Handle hPort,
                                     CpswMacPort_FifoStats *stats);

static void CpswMacPort_getLinkConfig(CpswMacPort_Handle hPort,
                                      CpswMacPort_LinkConfig *linkCfg);

static int32_t CpswMacPort_isInBand(CpswMacPort_Handle hPort,
                                    bool *inBand);

static int32_t CpswMacPort_setTxPriority(CpswMacPort_Handle hPort,
                                         CpswPort_TxPriority *txPriority);

static void CpswMacPort_getTxPriority(CpswMacPort_Handle hPort,
                                      CpswPort_TxPriority *txPriority);

static int32_t CpswMacPort_setRxPriority(CpswMacPort_Handle hPort,
                                         CpswPort_RxPriority *rxPriority);

static void CpswMacPort_getRxPriority(CpswMacPort_Handle hPort,
                                      CpswPort_RxPriority *rxPriority);

static int32_t CpswMacPort_setDscpPriority(CpswMacPort_Handle hPort,
                                           CpswPort_DscpPriority *dscpPriority);

static void CpswMacPort_getDscpPriority(CpswMacPort_Handle hPort,
                                        CpswPort_DscpPriority *dscpPriority);

static uint64_t CpswMacPort_mapBwToCnt(uint64_t rateInBps,
                                       uint32_t cppiClkFreqHz);

static uint64_t CpswMacPort_mapCntToBw(uint32_t cntVal,
                                       uint32_t cppiClkFreqHz);

static int32_t CpswMacPort_setTrafficShaping(CpswMacPort_Handle hPort,
                                             const CpswPort_TrafficShapingConfig *config);

static int32_t CpswMacPort_getTrafficShaping(CpswMacPort_Handle hPort,
                                             CpswPort_TrafficShapingConfig *config);

static void CpswMacPort_setSwitchTxScheduler(CpswMacPort_Handle hPort,
                                             CpswPort_TxPriorityType txPriorityType);

static int32_t CpswMacPort_setInterVLANParams(const CpswMacPort_Handle hPort,
                                              CpswMacPort_InterVLANRouteId routeId,
                                              const CpswMacPort_InterVlanRoutingParams *interVLANConfig);

static int32_t CpswMacPort_getInterVLANParams(const CpswMacPort_Handle hPort,
                                              CpswMacPort_InterVLANRouteId routeId,
                                              CpswMacPort_InterVlanRoutingParams *interVLANConfig);

static int32_t CpswMacPort_getInterVLANFreeRoutes(const CpswMacPort_Handle hPort,
                                                  CpswMacPort_InterVlanFreeRouteInfo *freeRoutes);

static bool CpswMacPort_isInterVLANRouteIdFree(const CpswMacPort_Handle hPort,
                                               CpswMacPort_InterVLANRouteId routeId);

static int32_t CpswMacPort_clearInterVLANRouteId(const CpswMacPort_Handle hPort,
                                                 CpswMacPort_InterVLANRouteId routeId);

static int32_t CpswMacPort_allocSetInterVLANParams(const CpswMacPort_Handle hPort,
                                                   const CpswMacPort_InterVlanRoutingParams *interVLANConfig,
                                                   CpswMacPort_InterVLANRouteId *pRouteId);

static int32_t CpswMacPort_findInterVLANRouteId(const CpswMacPort_Handle hPort,
                                                const CpswMacPort_InterVlanRoutingParams *routeMatchParams,
                                                CpswMacPort_InterVLANRouteId *pRouteId);

static void CpswMacPort_clearAllInterVLANRoutes(const CpswMacPort_Handle hPort);

static int32_t CpswMacPort_delInterVLANRoute(const CpswMacPort_Handle hPort,
                                             const CpswMacPort_InterVlanRoutingParams *interVLANConfig,
                                             CpswMacPort_InterVLANRouteId *pRouteId);

static int32_t CpswMacPort_checkInterVLANRouteIsFree(const CpswMacPort_Handle hPort,
                                                     CpswMacPort_InterVLANRouteId routeId,
                                                     bool *pIsFree);

static bool CpswMacPort_isInterVLANSupported(Cpsw_Type cpswType);

static void CpswMacPort_printRegs(Cpsw_MacPort portNum,
                                  CSL_Xge_cpswEnetportRegs *cpswMacPortRegs);

static int32_t CpswMacPort_setShortIPGCfg(const CpswMacPort_Handle hPort,
                                          const CpswMacPort_TxShortGapConfig *txShortIPGCfg);

static int32_t CpswMacPort_getShortIPGCfg(const CpswMacPort_Handle hPort,
                                          CpswMacPort_TxShortGapConfig *txShortIPGCfg);

static int32_t CpswMacPort_enableCptsPortEvents(const CpswMacPort_Handle hPort,
                                                CpswMacPort_EnableTsEventInArgs *portEventCfg);

static int32_t CpswMacPort_disableCptsPortEvents(const CpswMacPort_Handle hPort,
                                                 Cpsw_MacPort portNum);

static bool CpswMacPort_getSgmiiStatus(CpswMacPort_Handle hPort);
/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/*! \brief CPSW MAC port objects */
CpswMacPort_Obj gCpswMacPortObjs[CPSW_COUNT][CPSW_MAC_PORT_NUM];

#if (CPSW_TRACE_CFG_TRACE_LEVEL >= CPSW_TRACE_CFG_LEVEL_ERROR)
static const char *CpswMacPort_gSgmiiSpeedNames[] =
{
    [CSL_SGMII_10_MBPS]  = "10-Mbps",
    [CSL_SGMII_100_MBPS] = "100-Mbps",
    [CSL_SGMII_1000_MBPS]   = "1-Gbps",
};

static const char *CpswMacPort_gSgmiiDuplexNames[] =
{
    [CSL_SGMII_HALF_DUPLEX] = "Half-Duplex",
    [CSL_SGMII_FULL_DUPLEX] = "Full-Duplex",
};
#endif
/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

void CpswMacPort_initParams(CpswMacPort_Config *cfg)
{
    cfg->enableLoopback              = false;
    cfg->crcType                     = CPSW_CRC_ETHERNET;
    cfg->rxMtu                       = CPSW_MACPORT_RX_MTU_DEFAULT;
    cfg->passPriorityTaggedUnchanged = false;
    cfg->vlanConfig.portPri          = 0U;
    cfg->vlanConfig.portCfi          = 0U;
    cfg->vlanConfig.portVID          = 0U;
    cfg->txPriorityType              = CPSW_TX_PRIORITY_TYPE_FIXED;
    cfg->sgmiiConfig.sgmiiMode       = CPSW_MAC_SGMIIMODE_INVALID;
}

CpswMacPort_Handle CpswMacPort_open(Cpsw_Type cpswType,
                                    Cpsw_MacPort portNum,
                                    const CpswMacPort_Config *cfg,
                                    const CpswMacPort_Interface *interface,
                                    CpswMacPort_LinkConfig *linkConfig)
{
    CpswMacPort_Handle hPort = CpswMacPort_getMacPortObject(cpswType, portNum);
    int32_t status           = CPSW_SOK;
    bool sgmiiLpbk           = false;
    bool match;

    memset(hPort, 0U, sizeof(*hPort));

    hPort->cpswRegs = Cpsw_getCslRegs(cpswType);
    Cpsw_assert(hPort->cpswRegs != NULL);

    hPort->cpswType = cpswType;

    /* Soft-reset the Ethernet MAC logic */
    hPort->portNum = portNum;
    CpswMacPort_reset(hPort);

    /* Make sure that interface parameters matches ENET_CONTROL setting */
    match = CpswMacPort_matchesEnetCtrl(hPort, interface);
    CPSWTRACE_ERROR_IF(!match,
                       "MAC %u: Mac mode doesn't match with ENET_CONTROL config\n",
                       hPort->portNum);
    Cpsw_assert(match);

    if (interface->layerType == CPSW_MAC_LAYER_GMII)
    {
        if ((interface->sublayerType == CPSW_MAC_SUBLAYER_SERIAL) ||
            (interface->sublayerType == CPSW_MAC_SUBLAYER_QUAD_SERIAL_MAIN) ||
            (interface->sublayerType == CPSW_MAC_SUBLAYER_QUAD_SERIAL_SUB))
        {
            /* Assert if CPSW instance doesn't support SGMII */
            Cpsw_assert(0U != (CPSW_FEATURE_SGMII & CpswUtils_getIpFeatureBitMap(cpswType)));

            /* Get SGMII register base address */
            hPort->sgmiiRegs = Cpsw_getCslSgmiiRegs(cpswType);
            Cpsw_assert(hPort->sgmiiRegs != NULL);
            /* Validate SGMII IP version */
            CpswMacPort_validateSgmiiIPVersion(hPort);
            /* Soft-reset SGMII port */
            CpswMacPort_resetSgmiiPort(hPort);
        }
    }

    if (cfg->crcType == CPSW_CRC_ETHERNET)
    {
        CSL_CPGMAC_SL_disableCastagnoliCRC(hPort->cpswRegs, portNum);
    }
    else
    {
        CSL_CPGMAC_SL_enableCastagnoliCRC(hPort->cpswRegs, portNum);
    }

    CSL_CPGMAC_SL_setRxMaxLen(hPort->cpswRegs, portNum, cfg->rxMtu);

    if (cfg->passPriorityTaggedUnchanged == true)
    {
        CSL_CPSW_enablePortPassPriTag(hPort->cpswRegs, portNum + 1U);
    }
    else
    {
        CSL_CPSW_disablePortPassPriTag(hPort->cpswRegs, portNum + 1U);
    }

    CSL_CPSW_setPortVlanReg(hPort->cpswRegs,
                            portNum + 1U,
                            cfg->vlanConfig.portVID,
                            cfg->vlanConfig.portCfi,
                            cfg->vlanConfig.portPri);

    CpswMacPort_setSwitchTxScheduler(hPort, cfg->txPriorityType);

    if (cfg->enableLoopback == true)
    {
        switch (interface->layerType)
        {
            /* MII layer type */
            case CPSW_MAC_LAYER_MII:
            {
                /* RMII */
                if ((interface->sublayerType == CPSW_MAC_SUBLAYER_REDUCED) &&
                    (interface->variantType == CPSW_MAC_VARIANT_NONE))
                {
                    status = CPSW_SOK;
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: Loopback not supported in MII mode\n", hPort->portNum);
                    status = CPSW_ENOTSUPPORTED;
                }
            }
            break;

            /* GMII layer type */
            case CPSW_MAC_LAYER_GMII:
            {
                switch (interface->sublayerType)
                {
                    /* RGMII */
                    case CPSW_MAC_SUBLAYER_REDUCED:
                    {
                        if (interface->variantType != CPSW_MAC_VARIANT_FORCED)
                        {
                            CPSWTRACE_ERROR("MAC %u: Variant type should be forced for Loopback mode \n", hPort->portNum);
                            status = CPSW_EINVALID_PARAMS;
                        }
                    }
                    break;

                    /* SGMII */
                    case CPSW_MAC_SUBLAYER_SERIAL:
                    {
                        if (NULL == hPort->sgmiiRegs)
                        {
                            CPSWTRACE_ERROR("MAC %u: Serial Layer not supported \n", hPort->portNum);
                            status = CPSW_EINVALID_PARAMS;
                        }
                        else
                        {
                            CSL_SGMII_disableAutoNegotiation(hPort->sgmiiRegs, portNum);
                            CSL_SGMII_startRxTxSoftReset(hPort->sgmiiRegs, hPort->portNum);
                            CSL_SGMII_enableLoopback(hPort->sgmiiRegs, portNum);
                            CSL_SGMII_endRxTxSoftReset(hPort->sgmiiRegs, hPort->portNum);

                            if (0U != CSL_SGMII_getRxTxSoftResetStatus (hPort->sgmiiRegs, hPort->portNum))
                            {
                                /* Software reset is not completed. We dont wait in while loop here as SGMII
                                 * reset happens immediately */
                                status = CPSW_EFAIL;
                            }
                            else
                            {
                                /* For SGMII loopback we don't set MAC interface as it is digital loopback
                                 * (before the SERDES) from the CPSGMII transmit to the CPSGMII receive.
                                 * The SGMII modes are complementary to loopback so configuring MAC interface
                                 * in loopback mode will cause conflicting configuration */
                                sgmiiLpbk = true;
                            }
                        }
                    }
                    break;

                    default:
                        CPSWTRACE_ERROR("MAC %u: Loopback config for QSGMII/XGMII is not supported \n", hPort->portNum);
                        status = CPSW_ENOTSUPPORTED;
                        break;
                }
            }
            break;

            default:
                status = CPSW_EINVALID_PARAMS;
                break;
        }

        if (status == CPSW_SOK)
        {
            CSL_CPGMAC_SL_enableLoopback(hPort->cpswRegs, portNum);
        }
    }
    else
    {
        CSL_CPGMAC_SL_disableLoopback(hPort->cpswRegs, portNum);
    }

    if (status == CPSW_SOK)
    {
        if (!sgmiiLpbk)
        {
            status = CpswMacPort_setInterface(hPort, cfg, interface, linkConfig);
        }

        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "MAC %u: failed to set interface config: %d\n",
                           portNum, status);

        CpswMacPort_clearAllInterVLANRoutes(hPort);
    }

    if (status == CPSW_SOK)
    {
        hPort->magic = CPSW_MACPORT_MAGIC;
    }
    else
    {
        hPort = NULL;
    }

    return hPort;
}

CpswMacPort_Handle CpswMacPort_reopen(Cpsw_Type cpswType,
                                      Cpsw_MacPort portNum)
{
    CpswMacPort_Handle hPort = CpswMacPort_getMacPortObject(cpswType, portNum);

    memset(hPort, 0U, sizeof(*hPort));
    hPort->cpswType = cpswType;

    hPort->cpswRegs = Cpsw_getCslRegs(cpswType);
    Cpsw_assert(hPort->cpswRegs != NULL);

    hPort->magic   = CPSW_MACPORT_MAGIC;
    hPort->portNum = portNum;

    return hPort;
}

void CpswMacPort_close(CpswMacPort_Handle hPort)
{
    bool enabled;

    if (hPort->magic == CPSW_MACPORT_MAGIC)
    {
        enabled = CpswMacPort_isPortEnabled(hPort);
        if (enabled)
        {
            CpswMacPort_disablePort(hPort);
        }

        hPort->magic = CPSW_NO_MAGIC;
    }
    else
    {
        CPSWTRACE_ERROR("MAC %u: MAC port is not open (handle=%p)\n", hPort->portNum, hPort);
    }
}

int32_t CpswMacPort_ioctl(CpswMacPort_Handle hPort,
                          uint32_t cmd,
                          Cpsw_IoctlPrms *prms)
{
    int32_t status = CPSW_SOK;

    if (hPort->magic != CPSW_MACPORT_MAGIC)
    {
        CPSWTRACE_ERROR("MAC %u: MAC port is not open (handle=%p)\n", hPort->portNum, hPort);
        status = CPSW_EINVALID_PARAMS;
    }

    if (status == CPSW_SOK)
    {
        switch (cmd)
        {
            case CPSW_MACPORT_IOCTL_ENABLE:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswMacPort_LinkConfig));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_LinkConfig *linkCfg =
                        (CpswMacPort_LinkConfig *)prms->inArgs;

                    status = CpswMacPort_enablePort(hPort, linkCfg);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "MAC %u: failed to enable port: %d\n",
                                       hPort->portNum, status);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid ENABLE params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_DISABLE:
            {
                CpswMacPort_disablePort(hPort);
            }
            break;

            case CPSW_MACPORT_IOCTL_GET_STATUS:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswMacPort_GenericInArgs),
                                             sizeof(CpswMacPort_PortStatus));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_GenericInArgs *inArgs =
                        (CpswMacPort_GenericInArgs *)prms->inArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid GET_STATUS params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_GET_FIFO_STATS:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswMacPort_GenericInArgs),
                                             sizeof(CpswMacPort_FifoStats));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_GenericInArgs *inArgs =
                        (CpswMacPort_GenericInArgs *)prms->inArgs;
                    CpswMacPort_FifoStats *stats =
                        (CpswMacPort_FifoStats *)prms->outArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    CpswMacPort_getFifoStats(hPort, stats);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid GET_FIFO_STATS params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_GET_LINK_CONFIG:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswMacPort_GenericInArgs),
                                             sizeof(CpswMacPort_LinkConfig));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_GenericInArgs *inArgs =
                        (CpswMacPort_GenericInArgs *)prms->inArgs;
                    CpswMacPort_LinkConfig *linkCfg =
                        (CpswMacPort_LinkConfig *)prms->outArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    CpswMacPort_getLinkConfig(hPort, linkCfg);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid GET_LINK_CONFIG params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_IS_INBAND:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswMacPort_GenericInArgs),
                                             sizeof(bool));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_GenericInArgs *inArgs =
                        (CpswMacPort_GenericInArgs *)prms->inArgs;
                    bool *inBand = (bool *)prms->outArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    status = CpswMacPort_isInBand(hPort, inBand);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "MAC %u: failed to get in-band state: %d\n",
                                       hPort->portNum, status);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid IS_INBAND params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_SET_TX_PRIORITY:
            {
                status = Cpsw_checkInArgs(prms,
                                          sizeof(CpswMacPort_SetTxPriorityInArgs));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_SetTxPriorityInArgs *inArgs =
                        (CpswMacPort_SetTxPriorityInArgs *)prms->inArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    status = CpswMacPort_setTxPriority(hPort,
                                                       &inArgs->txPriority);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "MAC %u: failed to set TX priority: %d\n",
                                       hPort->portNum, status);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid SET_TX_PRIORITY params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_GET_TX_PRIORITY:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswMacPort_GenericInArgs),
                                             sizeof(CpswPort_TxPriority));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_GenericInArgs *inArgs =
                        (CpswMacPort_GenericInArgs *)prms->inArgs;
                    CpswPort_TxPriority *txPriority =
                        (CpswPort_TxPriority *)prms->outArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    CpswMacPort_getTxPriority(hPort, txPriority);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid GET_TX_PRIORITY params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_SET_RX_PRIORITY:
            {
                status = Cpsw_checkInArgs(prms,
                                          sizeof(CpswMacPort_SetRxPriorityInArgs));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_SetRxPriorityInArgs *inArgs =
                        (CpswMacPort_SetRxPriorityInArgs *)prms->inArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    status = CpswMacPort_setRxPriority(hPort,
                                                       &inArgs->rxPriority);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "MAC %u: failed to set RX priority: %d\n",
                                       hPort->portNum, status);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid SET_RX_PRIORITY params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_GET_RX_PRIORITY:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswMacPort_GenericInArgs),
                                             sizeof(CpswPort_RxPriority));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_GenericInArgs *inArgs =
                        (CpswMacPort_GenericInArgs *)prms->inArgs;
                    CpswPort_RxPriority *rxPriority =
                        (CpswPort_RxPriority *)prms->outArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    CpswMacPort_getRxPriority(hPort, rxPriority);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid GET_RX_PRIORITY params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_SET_RX_DSCP_PRIORITY:
            {
                status = Cpsw_checkInArgs(prms,
                                          sizeof(CpswMacPort_SetRxDscpPriorityInArgs));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_SetRxDscpPriorityInArgs *inArgs =
                        (CpswMacPort_SetRxDscpPriorityInArgs *)prms->inArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    status = CpswMacPort_setDscpPriority(hPort,
                                                         &inArgs->dscpPriority);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "MAC %u: failed to set DSCP priority: %d\n",
                                       hPort->portNum, status);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid SET_RX_DSCP_PRIORITY params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_GET_RX_DSCP_PRIORITY:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswMacPort_GenericInArgs),
                                             sizeof(CpswPort_DscpPriority));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_GenericInArgs *inArgs =
                        (CpswMacPort_GenericInArgs *)prms->inArgs;
                    CpswPort_DscpPriority *dscpPriority =
                        (CpswPort_DscpPriority *)prms->outArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    CpswMacPort_getDscpPriority(hPort, dscpPriority);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid GET_RX_DSCP_PRIORITY params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_SET_TRAFFIC_SHAPING:
            {
                status = Cpsw_checkInArgs(prms,
                                          sizeof(CpswMacPort_SetTrafficShapingInArgs));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_SetTrafficShapingInArgs *inArgs =
                        (CpswMacPort_SetTrafficShapingInArgs *)prms->inArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    status = CpswMacPort_setTrafficShaping(hPort,
                                                           &inArgs->trafficShapingConfig);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "MAC %u: failed to set traffic shaping: %d\n",
                                       hPort->portNum, status);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid SET_TRAFFIC_SHAPING params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_GET_TRAFFIC_SHAPING:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswMacPort_GenericInArgs),
                                             sizeof(CpswPort_TrafficShapingConfig));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_GenericInArgs *inArgs =
                        (CpswMacPort_GenericInArgs *)prms->inArgs;
                    CpswPort_TrafficShapingConfig *trafficShapingConfig =
                        (CpswPort_TrafficShapingConfig *)prms->outArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    status = CpswMacPort_getTrafficShaping(hPort, trafficShapingConfig);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "MAC %u: failed to get traffic shaping: %d\n",
                                       hPort->portNum, status);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid GET_TRAFFIC_SHAPING params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_SET_TX_BLOCKS_REM:
            {
                status = Cpsw_checkInArgs(prms,
                                          sizeof(CpswMacPort_SetTxBlocksRemInArgs));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_SetTxBlocksRemInArgs *inArgs =
                        (CpswMacPort_SetTxBlocksRemInArgs *)prms->inArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    CSL_CPSW_setTxHostBlksRem(hPort->cpswRegs,
                                              CPSW_NORMALIZE_MACPORT(hPort->portNum),
                                              inArgs->txBlocksRemaining);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid SET_TX_BLOCKS_REM params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_GET_TX_BLOCKS_REM:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswMacPort_GenericInArgs),
                                             sizeof(uint32_t));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_GenericInArgs *inArgs =
                        (CpswMacPort_GenericInArgs *)prms->inArgs;
                    uint32_t *txBlocksRemaining = (uint32_t *)prms->outArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    *txBlocksRemaining = CSL_CPSW_getTxHostBlksRem(hPort->cpswRegs,
                                                                   CPSW_NORMALIZE_MACPORT(hPort->portNum));
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid GET_TX_BLOCKS_REM params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_PRINT_REGS:
            {
                status = Cpsw_checkInArgs(prms,
                                          sizeof(CpswMacPort_GenericInArgs));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_GenericInArgs *inArgs =
                        (CpswMacPort_GenericInArgs *)prms->inArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);
                    CpswMacPort_printRegs(hPort->portNum,
                                          &hPort->cpswRegs->ENETPORT[CPSW_NORMALIZE_MACPORT(hPort->portNum)]);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid PRINT_REGS params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_SET_INTERVLAN_ROUTE:
            {
                if (CpswMacPort_isInterVLANSupported(hPort->cpswType))
                {
                    status = Cpsw_checkInOutArgs(prms,
                                                 sizeof(CpswMacPort_InterVlanRoutingParams),
                                                 sizeof(CpswMacPort_InterVlanRouteIdArgs));
                    if (status == CPSW_SOK)
                    {
                        const CpswMacPort_InterVlanRoutingParams *inArgs =
                            (CpswMacPort_InterVlanRoutingParams *)prms->inArgs;
                        CpswMacPort_InterVlanRouteIdArgs *outArgs =
                            (CpswMacPort_InterVlanRouteIdArgs *)prms->outArgs;

                        status = CpswMacPort_allocSetInterVLANParams(hPort,
                                                                     inArgs,
                                                                     &outArgs->routeId);
                        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                           "MAC %u: failed to set interVLAN params: %d\n",
                                           hPort->portNum, status);
                    }
                    else
                    {
                        CPSWTRACE_ERROR("MAC %u: invalid SET_INTERVLAN_ROUTE params\n", hPort->portNum);
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: interVLAN is not supported\n", hPort->portNum);
                    status = CPSW_ENOTSUPPORTED;
                }
            }
            break;

            /* Having separate IOCTL instead of using
             * CPSW_MACPORT_IOCTL_SET_INTERVLAN_ROUTE_PARAMS with
             * additional param to indicate specific route id needs to be used.
             * This is because the usecase for setting interVLAN and setting
             * interVLAN at specific route index differs
             * Set interVLAN config at specific route index is used only in
             * special case of Multicast interVLAN route.
             */
            case CPSW_MACPORT_IOCTL_SET_SPECIFIC_INTERVLAN_ROUTE:
            {
                if (CpswMacPort_isInterVLANSupported(hPort->cpswType))
                {
                    status = Cpsw_checkInArgs(prms,
                                              sizeof(CpswMacPort_SetSpecificInterVlanRouteInArgs));
                    if (status == CPSW_SOK)
                    {
                        const CpswMacPort_SetSpecificInterVlanRouteInArgs *inArgs =
                            (const CpswMacPort_SetSpecificInterVlanRouteInArgs *)prms->inArgs;

                        status = CpswMacPort_setInterVLANParams(hPort,
                                                                inArgs->routeId,
                                                                &inArgs->routeCfg);
                        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                           "MAC %u: failed to set specific interVLAN params: %d\n",
                                           hPort->portNum, status);
                    }
                    else
                    {
                        CPSWTRACE_ERROR("MAC %u: invalid SET_SPECIFIC_INTERVLAN_ROUTE params\n",
                                        hPort->portNum);
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: interVLAN is not supported\n", hPort->portNum);
                    status = CPSW_ENOTSUPPORTED;
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_GET_INTERVLAN_ROUTE:
            {
                if (CpswMacPort_isInterVLANSupported(hPort->cpswType))
                {
                    status = Cpsw_checkInOutArgs(prms,
                                                 sizeof(CpswMacPort_InterVlanRouteIdArgs),
                                                 sizeof(CpswMacPort_InterVlanRoutingParams));
                    if (status == CPSW_SOK)
                    {
                        const CpswMacPort_InterVlanRouteIdArgs *inArgs =
                            (const CpswMacPort_InterVlanRouteIdArgs *)prms->inArgs;
                        CpswMacPort_InterVlanRoutingParams *outArgs =
                            (CpswMacPort_InterVlanRoutingParams *)prms->outArgs;

                        status = CpswMacPort_getInterVLANParams(hPort,
                                                                inArgs->routeId,
                                                                outArgs);
                        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                           "MAC %u: failed to get interVLAN params: %d\n",
                                           hPort->portNum, status);
                    }
                    else
                    {
                        CPSWTRACE_ERROR("MAC %u: invalid GET_INTERVLAN_ROUTE params\n", hPort->portNum);
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: interVLAN is not supported\n", hPort->portNum);
                    status = CPSW_ENOTSUPPORTED;
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_GET_INTERVLAN_FREEROUTES:
            {
                if (CpswMacPort_isInterVLANSupported(hPort->cpswType))
                {
                    status = Cpsw_checkOutArgs(prms,
                                               sizeof(CpswMacPort_InterVlanFreeRouteInfo));
                    if (status == CPSW_SOK)
                    {
                        CpswMacPort_InterVlanFreeRouteInfo *outArgs =
                            (CpswMacPort_InterVlanFreeRouteInfo *)prms->outArgs;

                        status = CpswMacPort_getInterVLANFreeRoutes(hPort, outArgs);
                        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                           "MAC %u: failed to get free interVLAN routes: %d\n",
                                           hPort->portNum, status);
                    }
                    else
                    {
                        CPSWTRACE_ERROR("MAC %u: invalid GET_INTERVLAN_FREEROUTES params\n", hPort->portNum);
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: interVLAN is not supported\n", hPort->portNum);
                    status = CPSW_ENOTSUPPORTED;
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_GET_RX_MTU:
            {
                status = Cpsw_checkOutArgs(prms, sizeof(uint32_t));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_GenericInArgs *inArgs =
                        (CpswMacPort_GenericInArgs *)prms->inArgs;

                    uint32_t *rxMtu = (uint32_t *)prms->outArgs;
                    *rxMtu = CSL_CPGMAC_SL_getRxMaxLen(hPort->cpswRegs,
                                                       inArgs->portNum);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid GET_RX_MTU params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_DELETE_INTERVLAN_ROUTE:
            {
                if (CpswMacPort_isInterVLANSupported(hPort->cpswType))
                {
                    status = Cpsw_checkInOutArgs(prms,
                                                 sizeof(CpswMacPort_InterVlanRoutingParams),
                                                 sizeof(CpswMacPort_InterVlanRouteIdArgs));
                    if (status == CPSW_SOK)
                    {
                        const CpswMacPort_InterVlanRoutingParams *inArgs =
                            (const CpswMacPort_InterVlanRoutingParams *)prms->inArgs;
                        CpswMacPort_InterVlanRouteIdArgs *outArgs =
                            (CpswMacPort_InterVlanRouteIdArgs *)prms->outArgs;

                        status = CpswMacPort_delInterVLANRoute(hPort,
                                                               inArgs,
                                                               &outArgs->routeId);
                        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                           "MAC %u: failed to delete interVLAN route: %d\n",
                                           hPort->portNum, status);
                    }
                    else
                    {
                        CPSWTRACE_ERROR("MAC %u: invalid DELETE_INTERVLAN_ROUTE params\n", hPort->portNum);
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: interVLAN is not supported\n", hPort->portNum);
                    status = CPSW_ENOTSUPPORTED;
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_FIND_INTERVLAN_ROUTE:
            {
                if (CpswMacPort_isInterVLANSupported(hPort->cpswType))
                {
                    status = Cpsw_checkInOutArgs(prms,
                                                 sizeof(CpswMacPort_InterVlanRoutingParams),
                                                 sizeof(CpswMacPort_InterVlanRouteIdArgs));
                    if (status == CPSW_SOK)
                    {
                        const CpswMacPort_InterVlanRoutingParams *inArgs =
                            (CpswMacPort_InterVlanRoutingParams *)prms->inArgs;
                        CpswMacPort_InterVlanRouteIdArgs *outArgs =
                            (CpswMacPort_InterVlanRouteIdArgs *)prms->outArgs;

                        status = CpswMacPort_findInterVLANRouteId(hPort,
                                                                  inArgs,
                                                                  &outArgs->routeId);
                        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                           "MAC %u: failed to find interVLAN route: %d\n",
                                           hPort->portNum, status);
                    }
                    else
                    {
                        CPSWTRACE_ERROR("MAC %u: invalid FIND_INTERVLAN_ROUTE params\n", hPort->portNum);
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: interVLAN is not supported\n", hPort->portNum);
                    status = CPSW_ENOTSUPPORTED;
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_IS_INTERVLAN_ROUTE_FREE:
            {
                if (CpswMacPort_isInterVLANSupported(hPort->cpswType))
                {
                    status = Cpsw_checkInOutArgs(prms,
                                                 sizeof(CpswMacPort_InterVlanRouteIdArgs),
                                                 sizeof(bool));
                    if (status == CPSW_SOK)
                    {
                        const CpswMacPort_InterVlanRouteIdArgs *inArgs =
                            (CpswMacPort_InterVlanRouteIdArgs *)prms->inArgs;
                        bool *outArgs = (bool *)prms->outArgs;

                        status = CpswMacPort_checkInterVLANRouteIsFree(hPort,
                                                                       inArgs->routeId,
                                                                       outArgs);
                        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                           "MAC %u: failed to check interVLAN route state: %d\n",
                                           hPort->portNum, status);
                    }
                    else
                    {
                        CPSWTRACE_ERROR("MAC %u: invalid IS_INTERVLAN_ROUTE_FREE params\n", hPort->portNum);
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: interVLAN is not supported\n", hPort->portNum);
                    status = CPSW_ENOTSUPPORTED;
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_SET_SHORT_IPG:
            {
                status = Cpsw_checkInArgs(prms,
                                          sizeof(CpswMacPort_PortTxShortGapConfig));
                if (status == CPSW_SOK)
                {
                    const CpswMacPort_PortTxShortGapConfig *inArgs =
                        (CpswMacPort_PortTxShortGapConfig *)prms->inArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    status = CpswMacPort_setShortIPGCfg(hPort, &inArgs->shortGapCfg);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "MAC %u: failed to set short IPG config: %d\n",
                                       hPort->portNum, status);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid SET_SHORT_IPG params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_GET_SHORT_IPG:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswMacPort_GenericInArgs),
                                             sizeof(CpswMacPort_TxShortGapConfig));
                if (status == CPSW_SOK)
                {
                    const CpswMacPort_GenericInArgs *inArgs =
                        (CpswMacPort_GenericInArgs *)prms->inArgs;
                    CpswMacPort_TxShortGapConfig *outArgs =
                        (CpswMacPort_TxShortGapConfig *)prms->outArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    status = CpswMacPort_getShortIPGCfg(hPort, outArgs);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "MAC %u: failed to get short IPG config: %d\n",
                                       hPort->portNum, status);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid GET_SHORT_IPG params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_GET_SGMII_AN_LINK_STATUS:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswMacPort_GenericInArgs),
                                             sizeof(bool));
                if (status == CPSW_SOK)
                {
                    const CpswMacPort_GenericInArgs *inArgs = (CpswMacPort_GenericInArgs *)prms->inArgs;
                    bool *outArgs = (bool *)prms->outArgs;

                    Cpsw_assert(hPort->portNum == inArgs->portNum);

                    status = CpswMacPort_checkSgmiiAutoNegStatus(hPort);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "MAC %u: failed to get SGMII link status: %d\n",
                                       hPort->portNum, status);
                    if (CPSW_SOK == status)
                    {
                        *outArgs = true;
                    }
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid GET_SGMII_AN_LINK_STATUS params\n", hPort->portNum);
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_ENABLE_CPTS_EVENT:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswMacPort_EnableTsEventInArgs));
                if (status == CPSW_SOK)
                {
                    CpswMacPort_EnableTsEventInArgs *portEventCfg = (CpswMacPort_EnableTsEventInArgs *)prms->inArgs;
                    status = CpswMacPort_enableCptsPortEvents(hPort, portEventCfg);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "MAC %u: failed to enable Port event: %d\n", hPort->portNum, status);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid ENABLE_PORT_EVENT params\n", hPort->portNum);
                    status = CPSW_EINVALID_PARAMS;
                }

            }
            break;

            case CPSW_MACPORT_IOCTL_DISABLE_CPTS_EVENT:
            {
                status = Cpsw_checkInArgs(prms, sizeof(Cpsw_MacPort));
                if (status == CPSW_SOK)
                {
                    Cpsw_MacPort portNum = *((Cpsw_MacPort *)prms->inArgs);
                    status = CpswMacPort_disableCptsPortEvents(hPort, portNum);
                    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                       "MAC %u: failed to disable Port event: %d\n", hPort->portNum, status);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid DISABLE_PORT_EVENT params\n", hPort->portNum);
                    status = CPSW_EINVALID_PARAMS;
                }
            }
            break;

            case CPSW_MACPORT_IOCTL_GET_SGMII_LINK_STATUS:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswMacPort_GenericInArgs),
                                             sizeof(bool));
                if (status == CPSW_SOK)
                {
                    const CpswMacPort_GenericInArgs *inArgs = (CpswMacPort_GenericInArgs *)prms->inArgs;
                    bool *outArgs = (bool *)prms->outArgs;
                    Cpsw_assert(hPort->portNum == inArgs->portNum);
                    *outArgs = CpswMacPort_getSgmiiStatus(hPort);
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: invalid GET_SGMII_LINK_STATUS params\n", hPort->portNum);
                    status = CPSW_EINVALID_PARAMS;
                }
            }
            break;

            default:
            {
                CPSWTRACE_ERROR("MAC %u: invalid IOCTL cmd: %u\n", hPort->portNum, cmd);
                status = CPSW_EINVALID_PARAMS;
            }
            break;
        }
    }

    return status;
}

static void CpswMacPort_resetSgmiiPort(CpswMacPort_Handle hPort)
{

    CSL_SGMII_startRxTxSoftReset(hPort->sgmiiRegs, hPort->portNum);
    CSL_SGMII_endRxTxSoftReset(hPort->sgmiiRegs, hPort->portNum);

    while (0U != CSL_SGMII_getRxTxSoftResetStatus (hPort->sgmiiRegs, hPort->portNum))
    {
        /* Software reset is not completed. We dont wait in while loop here as SGMII
         * reset happens immediately */
         CpswUtils_delay(0U);
    }
}

static void CpswMacPort_validateSgmiiIPVersion(CpswMacPort_Handle hPort)
{
    CSL_SGMII_VERSION version;
    const CpswSoc_CpswIpInfo *ipInfo = CpswUtils_getIpInfo(hPort->cpswType);

    Cpsw_assert((ipInfo != NULL) && (ipInfo->cpswType == hPort->cpswType));

    CSL_SGMII_getVersionInfo(hPort->sgmiiRegs, CPSW_NORMALIZE_MACPORT(hPort->portNum), &version);

    if (0U == version.major_version)
    {
        /* The SGMII registers will be reset to zero if the SERDES clock is not initialized
         * i.e. the SERDES must be configured for the SGMII module to be configured */
        CPSWTRACE_ERROR("MAC %u: SGMII port not ready as SERDES PLL not locked %d\n", 
                           hPort->portNum, hPort->cpswType);
        Cpsw_assert(false);
    }
    else
    {
        Cpsw_assert(version.ident_val     == ipInfo->sgmiiInfo.versionInfo.moduleId);
        Cpsw_assert(version.major_version == ipInfo->sgmiiInfo.versionInfo.majorVer);
        Cpsw_assert(version.minor_version == ipInfo->sgmiiInfo.versionInfo.minorVer);
        Cpsw_assert(version.rtl_version   == ipInfo->sgmiiInfo.versionInfo.rtlVer);
    }
}

static CpswMacPort_Handle CpswMacPort_getMacPortObject(Cpsw_Type cpswType,
                                                       Cpsw_MacPort portNum)
{
    CpswMacPort_Handle hPort = NULL;

    switch (cpswType)
    {
        case CPSW_2G:
        case CPSW_5G:
        case CPSW_9G:
            hPort = &gCpswMacPortObjs[cpswType][portNum];
            break;

        default:
            Cpsw_assert(false);
    }

    return hPort;
}

static void CpswMacPort_reset(CpswMacPort_Handle hPort)
{
    CSL_CPGMAC_SL_MACSTATUS macStatus;
    uint32_t gmiiEn;
    uint32_t done;

    gmiiEn = CSL_CPGMAC_SL_isGMIIEnabled(hPort->cpswRegs, hPort->portNum);

    /* Idle MAC port */
    CSL_CPGMAC_SL_enableIdleMode(hPort->cpswRegs, hPort->portNum);
    do
    {
        CSL_CPGMAC_SL_getMacStatusReg(hPort->cpswRegs, hPort->portNum, &macStatus);
        done = (macStatus.idle == 1U);
        /* TX idle is set only if GMII clock is enabled */
        if (gmiiEn == 1U)
        {
            done = done && (macStatus.macTxIdle == 1U);
        }
    }
    while (done == FALSE);

    /* Soft-reset the Ethernet MAC logic */
    CSL_CPGMAC_SL_resetMac(hPort->cpswRegs, hPort->portNum);
    do
    {
        done = CSL_CPGMAC_SL_isMACResetDone(hPort->cpswRegs, hPort->portNum);
    }
    while (done == FALSE);
}

static bool CpswMacPort_matchesEnetCtrl(CpswMacPort_Handle hPort,
                                        const CpswMacPort_Interface *interface)
{
    CpswMacPort_Interface iface = *interface;
    CpswMac_LayerType enetLayer;
    CpswMac_SublayerType enetSublayer;
    uint32_t status;
    bool match = true;

    status = CpswSoc_getEnetControl(hPort->cpswType,
                                    hPort->portNum,
                                    &enetLayer,
                                    &enetSublayer);
    if (status != CPSW_SOK)
    {
        match = false;
    }

    if ((iface.layerType != enetLayer) ||
        (iface.sublayerType != enetSublayer))
    {
        match = false;
    }

    return match;
}

static int32_t CpswMacPort_checkSgmiiAutoNegStatus(CpswMacPort_Handle hPort)
{
    int32_t status = CPSW_SOK;
    CSL_SGMII_STATUS sgmiiStatus;

    /* Wait for SGMII Autonegotiation to complete without error */
    do
    {
        CSL_SGMII_getStatus(hPort->sgmiiRegs, hPort->portNum, &sgmiiStatus);
        if (sgmiiStatus.bIsAutoNegError)
        {
            /* Autonegotiation error */
            CPSWTRACE_ERROR("MAC %u: SGMII Auto-Negotiation failed: %d\n",
                            hPort->portNum,
                            status);
            status = CPSW_EFAIL;
            break;
        }
    }
    while (1U != sgmiiStatus.bIsAutoNegComplete);

    /* Check SGMII link status*/
    if (status == CPSW_SOK)
    {
        /* Link indicator is not valid until the lock status bit is asserted, so
         * check for lock first */
        if (1U != CSL_SGMII_getSerdesPLLLockStatus(hPort->sgmiiRegs, hPort->portNum))
        {
            CPSWTRACE_ERROR("MAC %u: SGMII SERDES PLL not locked: %d\n",
                            hPort->portNum,
                            status);
            status = CPSW_EUNEXPECTED;
        }
        else
        {
            /* Wait for SGMII link */
            do
            {
                CSL_SGMII_getStatus(hPort->sgmiiRegs, hPort->portNum, &sgmiiStatus);
            }
            while (1U != sgmiiStatus.bIsLinkUp);
        }
    }

    return status;
}

static void CpswMacPort_mapSgmiiLinkCfg(CSL_SGMII_ADVABILITY *sgmiiAdvAbility,
                                           CpswMacPort_LinkConfig *cfg)

{
    /* Set SGMII duplexity config */
    if (CPSW_DUPLEX_HALF == cfg->duplexity)
    {
        sgmiiAdvAbility->duplexMode = CSL_SGMII_HALF_DUPLEX;
    }
    else
    {
        sgmiiAdvAbility->duplexMode = CSL_SGMII_FULL_DUPLEX;
    }

    /* Set SGMII speed config */
    if (CPSW_SPEED_10MBIT == cfg->speed)
    {
        sgmiiAdvAbility->linkSpeed  = CSL_SGMII_10_MBPS;
    }
    else if (CPSW_SPEED_100MBIT == cfg->speed)
    {
        sgmiiAdvAbility->linkSpeed  = CSL_SGMII_100_MBPS;
    }
    else
    {
        sgmiiAdvAbility->linkSpeed  = CSL_SGMII_1000_MBPS;
    }

    return;
}

static int32_t CpswMacPort_checkSgmiiStatus(CpswMacPort_Handle hPort)
{
    int32_t status;
    CSL_SGMII_ADVABILITY sgmiiAdvAbility;

    if (1U == CSL_SGMII_isAutoNegotiationEnabled(hPort->sgmiiRegs, hPort->portNum))
    {
        status = CpswMacPort_checkSgmiiAutoNegStatus(hPort);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                          "MAC %u: SGMII AUTO NEG failed for MAC port: %d\n", hPort->portNum, status);
        if (status == CPSW_SOK)
        {
            if (1U == CSL_SGMII_isMasterModeEnabled(hPort->sgmiiRegs, hPort->portNum))
            {
                /* In SGMII AUTONEG with MASTER SLAVE config, the master tells the slave
                 * the rate so the slave would auto negotiate to master config. Hence we report
                 * master configured speed/duplexity */
                CSL_SGMII_getAdvAbility(hPort->sgmiiRegs, hPort->portNum, &sgmiiAdvAbility);
            }
            else
            {
                /* Get link partner speed/duplexity */
                CSL_SGMII_getLinkPartnerAdvAbility(hPort->sgmiiRegs, hPort->portNum, &sgmiiAdvAbility);
            }

            CPSWTRACE_INFO("MAC %u: SGMII Link Parter Config port: Link %s: %s %s\n",
                            hPort->portNum,
                            sgmiiAdvAbility.bLinkUp?"Up":"Down",
                            CpswMacPort_gSgmiiSpeedNames[sgmiiAdvAbility.linkSpeed],
                            CpswMacPort_gSgmiiDuplexNames[sgmiiAdvAbility.duplexMode]);
        }
    }
    else
    {
        /* Forced or loopback mode */
        status = CPSW_SOK;
    }

    return status;
}

static int32_t CpswMacPort_configSgmii(CpswMacPort_Handle hPort,
                                       const CpswMacPort_Config *macCfg,
                                       CpswMacPort_LinkConfig *linkCfg)
{
    int32_t status = CPSW_SOK;
    CSL_SGMII_ADVABILITY sgmiiAdvAbility;
    CSL_SGMII_STATUS sgmiiStatus;

    memset(&sgmiiAdvAbility, 0U, sizeof(sgmiiAdvAbility));

    if (macCfg->sgmiiConfig.sgmiiMode == CPSW_MAC_SGMIIMODE_FIBER_WITH_PHY)
    {
        CPSWTRACE_INFO("MAC %u: Configuring SGMII in FIBER_WITH_PHY mode \n",
                        hPort->portNum);
        /* In fiber mode Advertise full-duplex only */
        sgmiiAdvAbility.duplexMode = CSL_SGMII_FULL_DUPLEX;
        sgmiiAdvAbility.sgmiiMode  = CSL_SGMII_MODE_FIBER;

        CSL_SGMII_setAdvAbility(hPort->sgmiiRegs, hPort->portNum, &sgmiiAdvAbility);
        CSL_SGMII_enableAutoNegotiation(hPort->sgmiiRegs, hPort->portNum);
    }
    else if (macCfg->sgmiiConfig.sgmiiMode == CPSW_MAC_SGMIIMODE_SGMII_WITH_PHY)
    {
        CPSWTRACE_INFO("MAC %u: Configuring SGMII in SGMII_WITH_PHY mode \n",
                        hPort->portNum);
        /* Set highest speed when auto-negotiating with PHY. We don't need to configure
         * user provided speed/duplexity here as it is set in PHY and we auto-negotiate
         * with PHY  Ex - App wants to set 100Mbps full-duplex, it will be set in PHY
         * and PHY would auto-negotiate with with remote PHY with 100Mbps or lower which
         * will be fine with SGMII as we are at highest config
         */
        sgmiiAdvAbility.duplexMode = CSL_SGMII_FULL_DUPLEX;
        sgmiiAdvAbility.linkSpeed  = CSL_SGMII_1000_MBPS;
        sgmiiAdvAbility.bLinkUp    = 1U;
        sgmiiAdvAbility.sgmiiMode  = CSL_SGMII_MODE_SGMII;

        CSL_SGMII_setAdvAbility(hPort->sgmiiRegs, hPort->portNum, &sgmiiAdvAbility);
        /* Note - For SGMII connection with PHY, auto negotiate should always be a ‘1'
         * (master/slave should be selected) */
        CSL_SGMII_enableAutoNegotiation(hPort->sgmiiRegs, hPort->portNum);
    }
    else if (macCfg->sgmiiConfig.sgmiiMode == CPSW_MAC_SGMIIMODE_SGMII_AUTONEG_MASTER)
    {
        CPSWTRACE_INFO("MAC %u: Configuring SGMII in SGMII_AUTONEG_MASTER mode \n",
                        hPort->portNum);
        /* For SGMII master, advertise full-duplex gigabit*/
        sgmiiAdvAbility.linkSpeed  = CSL_SGMII_1000_MBPS;
        sgmiiAdvAbility.duplexMode = CSL_SGMII_FULL_DUPLEX;
        sgmiiAdvAbility.bLinkUp    = 1U;
        sgmiiAdvAbility.sgmiiMode  = CSL_SGMII_MODE_SGMII;

        CSL_SGMII_setAdvAbility(hPort->sgmiiRegs, hPort->portNum, &sgmiiAdvAbility);
        CSL_SGMII_enableMasterMode(hPort->sgmiiRegs, hPort->portNum);
        CSL_SGMII_enableAutoNegotiation(hPort->sgmiiRegs, hPort->portNum);
    }
    else if (macCfg->sgmiiConfig.sgmiiMode == CPSW_MAC_SGMIIMODE_SGMII_AUTONEG_SLAVE)
    {
        CPSWTRACE_INFO("MAC %u: Configuring SGMII in SGMII_AUTONEG_SLAVE mode \n",
                        hPort->portNum);
        /* To write 1 to tx_config_reg[0] bit, we pass empty ability structure */
        sgmiiAdvAbility.sgmiiMode  = CSL_SGMII_MODE_SGMII;

        CSL_SGMII_setAdvAbility(hPort->sgmiiRegs, hPort->portNum, &sgmiiAdvAbility);
        CSL_SGMII_disableMasterMode(hPort->sgmiiRegs, hPort->portNum);
        CSL_SGMII_enableAutoNegotiation(hPort->sgmiiRegs, hPort->portNum);
    }
    else if (macCfg->sgmiiConfig.sgmiiMode == CPSW_MAC_SGMIIMODE_SGMII_FORCEDLINK)
    {
        CPSWTRACE_INFO("MAC %u: Configuring SGMII in SGMII_FORCEDLINK mode \n",
                        hPort->portNum);
        CpswMacPort_mapSgmiiLinkCfg (&sgmiiAdvAbility, linkCfg);
        sgmiiAdvAbility.bLinkUp    = 1U;
        sgmiiAdvAbility.sgmiiMode  = CSL_SGMII_MODE_SGMII;

        CSL_SGMII_setAdvAbility(hPort->sgmiiRegs, hPort->portNum, &sgmiiAdvAbility);
        CSL_SGMII_enableMasterMode(hPort->sgmiiRegs, hPort->portNum);
        CSL_SGMII_disableAutoNegotiation(hPort->sgmiiRegs, hPort->portNum);

        /* Wait for SGMII link */
        do
        {
            CSL_SGMII_getStatus(hPort->sgmiiRegs, hPort->portNum, &sgmiiStatus);
        }
        while (1U != sgmiiStatus.bIsLinkUp);

        status = CPSW_SOK;
    }
    else
    {
        CPSWTRACE_ERROR("MAC %u: Invalid SGMII mode config\n", hPort->portNum);
        status = CPSW_EINVALID_PARAMS;
    }

    return status;
}

static bool CpswMacPort_getSgmiiStatus(CpswMacPort_Handle hPort)
{
    bool status;
    CSL_SGMII_STATUS sgmiiStatus;

    CSL_SGMII_getStatus(hPort->sgmiiRegs, hPort->portNum, &sgmiiStatus);
    status = (sgmiiStatus.bIsLinkUp == 0U) ? false : true;
    return status;
}

static int32_t CpswMacPort_setInterface(CpswMacPort_Handle hPort,
                                        const CpswMacPort_Config *macCfg,
                                        const CpswMacPort_Interface *interface,
                                        CpswMacPort_LinkConfig *linkConfig)
{
    uint32_t macControl;
    int32_t status = CPSW_EINVALID_PARAMS;

    macControl = CSL_CPGMAC_SL_getMacControlReg(hPort->cpswRegs,
                                                hPort->portNum);

    /* Not supported by hardware */
    CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_EXT_RX_FLOW_EN, 0U);
    CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_EXT_TX_FLOW_EN, 0U);

    switch (interface->layerType)
    {
        /* MII layer type */
        case CPSW_MAC_LAYER_MII:
        {
            /* RMII */
            if ((interface->sublayerType == CPSW_MAC_SUBLAYER_REDUCED) &&
                (interface->variantType == CPSW_MAC_VARIANT_NONE))
            {
                status = CPSW_SOK;
            }
        }
        break;

        /* GMII layer type */
        case CPSW_MAC_LAYER_GMII:
        {
            switch (interface->sublayerType)
            {
                /* RGMII */
                case CPSW_MAC_SUBLAYER_REDUCED:
                    if (interface->variantType == CPSW_MAC_VARIANT_FORCED)
                    {
                        /* Forced */
                        CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_EXT_EN, 0U);
                        CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_GIG_FORCE, 1U);
                        status = CPSW_SOK;
                    }
                    else if (interface->variantType == CPSW_MAC_VARIANT_NONE)
                    {
                        /* In-band */
                        CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_EXT_EN, 1U);
                        CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_GIG_FORCE, 0U);
                        status = CPSW_SOK;
                    }
                    else
                    {
                        status = CPSW_EINVALID_PARAMS;
                    }

                    break;

                /* SGMII */
                case CPSW_MAC_SUBLAYER_SERIAL:
                /* QSGMII main/sub*/
                case CPSW_MAC_SUBLAYER_QUAD_SERIAL_MAIN:
                case CPSW_MAC_SUBLAYER_QUAD_SERIAL_SUB:
                    Cpsw_assert(interface->variantType != CPSW_MAC_VARIANT_FORCED);

                    /* In all SGMII modes Ext_En bit in the Mac_Control register must be set to
                     * allow the speed and duplexity to be set by the signals from the CPSGMII */
                    CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_EXT_EN, 1U);

                    CSL_SGMII_DisableTestPattern(hPort->sgmiiRegs, hPort->portNum);
                    CSL_SGMII_disableMasterMode(hPort->sgmiiRegs, hPort->portNum);
                    CSL_SGMII_disableLoopback(hPort->sgmiiRegs, hPort->portNum);
                    CSL_SGMII_disableAutoNegotiation(hPort->sgmiiRegs, hPort->portNum);

                    /* In case of SGMII, the MDIO controls the auto-negotiation of the PHY to the
                     * other side of the fiber or wire (ie. to the remote system). The CPSGMII is
                     * the link to the PHY.  The PHY auto-negotiates with the other side and
                     * then the CPSGMII auto-negotiates with the PHY */


                    /* Confirm SERDES PLL is locked before configuring the port*/
                    Cpsw_assert(1U == CSL_SGMII_getSerdesPLLLockStatus(hPort->sgmiiRegs, hPort->portNum));

                    status = CpswMacPort_configSgmii(hPort, macCfg, linkConfig);

                    break;

                default:
                    break;
            }
        }
        break;

        /* XGMII layer type */
        case CPSW_MAC_LAYER_XGMII:
        {
            if (hPort->cpswType != CPSW_2G)
            {
                switch (interface->sublayerType)
                {
                    /* XFI */
                    case CPSW_MAC_SUBLAYER_STANDARD:
                        if (interface->variantType == CPSW_MAC_VARIANT_FORCED)
                        {
                            /* Forced */
                            CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_EXT_EN_XGIG, 0U);
                            status = CPSW_SOK;
                        }

                        break;

                    default:
                        break;
                }
            }
        }
        break;

        default:
            break;
    }

    /* Set interface layer, sublayer and variant */
    if (status == CPSW_SOK)
    {
        CSL_CPGMAC_SL_setMacControlReg(hPort->cpswRegs,
                                       hPort->portNum,
                                       macControl);
    }

    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "MAC %u: failed to set interface layer %u sublayer %u: %d\n",
                       hPort->portNum,
                       interface->layerType,
                       interface->sublayerType,
                       status);

    return status;
}

static int32_t CpswMacPort_enablePort(CpswMacPort_Handle hPort,
                                      CpswMacPort_LinkConfig *cfg)
{
    CpswMac_LayerType enetLayer;
    CpswMac_SublayerType enetSublayer;
    bool forced = true;
    uint32_t extCtrl;
    uint32_t macControl;
    int32_t status;

    status = CpswSoc_getEnetControl(hPort->cpswType,
                                    hPort->portNum,
                                    &enetLayer,
                                    &enetSublayer);

    if (status == CPSW_SOK)
    {
        macControl = CSL_CPGMAC_SL_getMacControlReg(hPort->cpswRegs,
                                                    hPort->portNum);

        /* Disable all fields, enable as needed based on link config */
        CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_FULLDUPLEX, 0U);
        CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_GMII_EN, 0U);
        CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_GIG, 0U);
        CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_XGMII_EN, 0U);
        CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_IFCTL_A, 0U);
        CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_IFCTL_B, 0U);

        status = CPSW_EINVALID_PARAMS;

        switch (enetLayer)
        {
            /* MII layer type */
            case CPSW_MAC_LAYER_MII:
            {
                switch (enetSublayer)
                {
                    /* RMII */
                    case CPSW_MAC_SUBLAYER_REDUCED:
                        /* In RMII mode, ifctl_a bit determines the RMII link speed (0=10mbps, 1=100mbps) */
                        if ((cfg->speed == CPSW_SPEED_100MBIT) ||
                            (cfg->speed == CPSW_SPEED_10MBIT))
                        {
                            if (cfg->speed == CPSW_SPEED_100MBIT)
                            {
                                CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_IFCTL_A, 1U);
                            }

                            CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_GMII_EN, 1U);
                            status = CPSW_SOK;
                        }
                        else
                        {
                            status = CPSW_EINVALID_PARAMS;
                        }

                        forced = true;
                        break;

                    default:
                        break;
                }
            }
            break;

            /* GMII layer type */
            case CPSW_MAC_LAYER_GMII:
            {
                switch (enetSublayer)
                {
                    /* RGMII */
                    case CPSW_MAC_SUBLAYER_REDUCED:
                        extCtrl = CSL_FEXT(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_EXT_EN);
                        forced = (extCtrl == TRUE) ? false : true;

                        if (forced)
                        {
                            if (cfg->speed == CPSW_SPEED_1GBIT)
                            {
                                CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_GIG, 1U);
                            }
                        }

                        CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_GMII_EN, 1U);
                        status = CPSW_SOK;
                        break;

                    /* SGMII, QSGMII */
                    case CPSW_MAC_SUBLAYER_SERIAL:
                    case CPSW_MAC_SUBLAYER_QUAD_SERIAL_MAIN:
                    case CPSW_MAC_SUBLAYER_QUAD_SERIAL_SUB:
                        extCtrl = CSL_FEXT(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_EXT_EN);
                        forced = (extCtrl == TRUE) ? false : true;

                        if (cfg->speed == CPSW_SPEED_1GBIT)
                        {
                            CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_GIG, 1U);
                        }

                        CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_GMII_EN, 1U);

                        status = CpswMacPort_checkSgmiiStatus(hPort);

                        break;

                    default:
                        break;
                }
            }
            break;

            /* XGMII layer type */
            case CPSW_MAC_LAYER_XGMII:
            {
                extCtrl = CSL_FEXT(macControl,XGE_CPSW_PN_MAC_CONTROL_REG_EXT_EN_XGIG);
                forced = (extCtrl == TRUE) ? false : true;

                /* XGMII iIn-band mode is not supported in hardware */
                Cpsw_assert(forced);

                if (forced)
                {
                    CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_XGMII_EN, 1U);
                    status = CPSW_SOK;
                }
            }
            break;

            default:
                break;
        }
    }

    /* Half-duplex mode is supported only at 10/100 Mbps */
    if (status == CPSW_SOK)
    {
        if ((cfg->duplexity == CPSW_DUPLEX_HALF) &&
            (cfg->speed != CPSW_SPEED_10MBIT) &&
            (cfg->speed != CPSW_SPEED_100MBIT))
        {
            CPSWTRACE_ERROR("MAC %u: 1Gbps half-duplex is not supported\n",
                            hPort->portNum);
            status = CPSW_EINVALID_PARAMS;
        }
    }

    /* Set duplexity, speed related fields in MAC control */
    if (status == CPSW_SOK)
    {
        if (forced && (cfg->duplexity == CPSW_DUPLEX_FULL))
        {
            CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_FULLDUPLEX, 1U);
        }

        /* Set interface layer, sublayer, speed */
        CSL_CPGMAC_SL_setMacControlReg(hPort->cpswRegs,
                                       hPort->portNum,
                                       macControl);
    }

    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "MAC %u: failed to enable MAC port: %d\n",
                       hPort->portNum, status);

    return status;
}

static void CpswMacPort_disablePort(CpswMacPort_Handle hPort)
{
    uint32_t macControl;

    /* Save MAC_CONTROL register context before soft-reset */
    macControl = CSL_CPGMAC_SL_getMacControlReg(hPort->cpswRegs, hPort->portNum);

    CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_FULLDUPLEX, 0U);
    CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_GMII_EN, 0U);
    CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_GIG, 0U);
    CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_XGMII_EN, 0U);
    CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_IFCTL_A, 0U);
    CSL_FINS(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_IFCTL_B, 0U);

    /* Soft-reset the Ethernet MAC logic */
    CpswMacPort_reset(hPort);

    /* Restore context of fields in MAC_CONTROL register that are configured
     * at open time via CpswMacPort_Config and that are not related to link
     * configuration */
    CSL_CPGMAC_SL_setMacControlReg(hPort->cpswRegs, hPort->portNum, macControl);
}

static bool CpswMacPort_isPortEnabled(CpswMacPort_Handle hPort)
{
    uint32_t macControl;
    uint32_t gmiiEn;
    uint32_t xgmiiEn;

    macControl = CSL_CPGMAC_SL_getMacControlReg(hPort->cpswRegs, hPort->portNum);
    gmiiEn = CSL_FEXT(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_GMII_EN);
    xgmiiEn = CSL_FEXT(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_XGMII_EN);

    return((gmiiEn == 1U) || (xgmiiEn == 1U));
}

static void CpswMacPort_getFifoStats(CpswMacPort_Handle hPort,
                                     CpswMacPort_FifoStats *stats)
{
    CSL_CPSW_THRURATE thruRate;
    CSL_CPGMAC_SL_FIFOSTATUS fifoStatus;
    uint32_t i;

    CSL_CPSW_getPortBlockCountReg(hPort->cpswRegs,
                                  hPort->portNum + 1U,
                                  &stats->rxExpressBlockCount,
                                  &stats->rxPreemptBlockCount,
                                  &stats->txBlockCount);

    CSL_CPSW_getPortMaxBlksReg(hPort->cpswRegs,
                               hPort->portNum + 1U,
                               &stats->rxMaxBlocks,
                               &stats->txMaxBlocks);

    CSL_CPSW_getThruRateReg(hPort->cpswRegs, &thruRate);

    stats->rxThroughputRate = thruRate.enetRxThruRate;

    stats->txStartWords = CSL_CPSW_getTxStartWords(hPort->cpswRegs);

    CSL_CPGMAC_SL_getFifoStatus(hPort->cpswRegs, hPort->portNum, &fifoStatus);
    for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(stats->txActiveFifo); i++)
    {
        Cpsw_assert(i < CSL_XGE_CPSW_PN_FIFO_STATUS_REG_TX_PRI_ACTIVE_MAX);

        stats->txActiveFifo[i] = (fifoStatus.txPriActive >> i) & 0x1U;
    }
}

static void CpswMacPort_getLinkConfig(CpswMacPort_Handle hPort,
                                      CpswMacPort_LinkConfig *linkCfg)
{
    CSL_CPGMAC_SL_MACSTATUS macStatus;

    CSL_CPGMAC_SL_getMacStatusReg(hPort->cpswRegs,
                                  hPort->portNum,
                                  &macStatus);

    /* TODO: How to identify 10Mbps */
    if (macStatus.extGigabitEnabled == TRUE)
    {
        linkCfg->speed = CPSW_SPEED_1GBIT;
    }
    else
    {
        linkCfg->speed = CPSW_SPEED_100MBIT;
    }

    if (macStatus.extFullDuplexEnabled == TRUE)
    {
        linkCfg->duplexity = CPSW_DUPLEX_FULL;
    }
    else
    {
        linkCfg->duplexity = CPSW_DUPLEX_HALF;
    }
}

static int32_t CpswMacPort_isInBand(CpswMacPort_Handle hPort,
                                    bool *inBand)
{
    CpswMac_LayerType enetLayer;
    CpswMac_SublayerType enetSublayer;
    uint32_t macControl;
    uint32_t extCtrl;
    int32_t status;

    status = CpswSoc_getEnetControl(hPort->cpswType,
                                    hPort->portNum,
                                    &enetLayer,
                                    &enetSublayer);
    if (status == CPSW_SOK)
    {
        switch (enetLayer)
        {
            case CPSW_MAC_LAYER_GMII:
                extCtrl = CSL_CPGMAC_SL_isExtControlEnabled(hPort->cpswRegs,
                                                            hPort->portNum);
                break;

            case CPSW_MAC_LAYER_XGMII:
                macControl = CSL_CPGMAC_SL_getMacControlReg(hPort->cpswRegs,
                                                            hPort->portNum);
                extCtrl = CSL_FEXT(macControl, XGE_CPSW_PN_MAC_CONTROL_REG_EXT_EN_XGIG);
                break;

            case CPSW_MAC_LAYER_MII:
            default:
                extCtrl = FALSE;
                break;
        }

        *inBand = (extCtrl == TRUE) ? true : false;
    }

    return status;
}

static int32_t CpswMacPort_setTxPriority(CpswMacPort_Handle hPort,
                                         CpswPort_TxPriority *txPriority)
{
    int32_t status = CPSW_SOK;
    uint32_t i;

    for (i = 0U; i < CPSW_HEADER_PACKET_PRIORITY_NUM; i++)
    {
        if (txPriority->priorityMap[i] >= CPSW_PRIORITY_NUM)
        {
            CPSWTRACE_ERROR("MAC %u: invalid priority map %u -> %u\n",
                            hPort->portNum, i, txPriority->priorityMap[i]);
            status = CPSW_EINVALID_PARAMS;
            break;
        }
    }

    if (status == CPSW_SOK)
    {
        CSL_CPSW_setPortTxPriMapReg(hPort->cpswRegs,
                                    hPort->portNum + 1U,
                                    txPriority->priorityMap);
    }

    return status;
}

static void CpswMacPort_getTxPriority(CpswMacPort_Handle hPort,
                                      CpswPort_TxPriority *txPriority)
{
    CSL_CPSW_getPortTxPriMapReg(hPort->cpswRegs,
                                hPort->portNum + 1U,
                                txPriority->priorityMap);
}

static int32_t CpswMacPort_setRxPriority(CpswMacPort_Handle hPort,
                                         CpswPort_RxPriority *rxPriority)
{
    int32_t status = CPSW_SOK;
    uint32_t i;

    for (i = 0U; i < CPSW_PACKET_PRIORITY_NUM; i++)
    {
        if (rxPriority->priorityMap[i] >= CPSW_PRIORITY_NUM)
        {
            CPSWTRACE_ERROR("MAC %u: invalid priority map %u -> %u\n",
                            hPort->portNum, i, rxPriority->priorityMap[i]);
            status = CPSW_EINVALID_PARAMS;
            break;
        }
    }

    if (status == CPSW_SOK)
    {
        CSL_CPSW_setPortRxPriMapReg(hPort->cpswRegs,
                                    hPort->portNum + 1U,
                                    rxPriority->priorityMap);
    }

    return status;
}

static void CpswMacPort_getRxPriority(CpswMacPort_Handle hPort,
                                      CpswPort_RxPriority *rxPriority)
{
    CSL_CPSW_getPortRxPriMapReg(hPort->cpswRegs,
                                hPort->portNum + 1U,
                                rxPriority->priorityMap);
}

static int32_t CpswMacPort_setDscpPriority(CpswMacPort_Handle hPort,
                                           CpswPort_DscpPriority *dscpPriority)
{
    CSL_CPSW_PORT_CONTROL control;
    int32_t status = CPSW_SOK;
    uint32_t i;

    for (i = 0U; i < CPSW_TOS_PRIORITY_NUM; i++)
    {
        if (dscpPriority->tosMap[i] >= CPSW_PRIORITY_NUM)
        {
            CPSWTRACE_ERROR("MAC %u: invalid priority map %u -> %u\n",
                            hPort->portNum, i, dscpPriority->tosMap[i]);
            status = CPSW_EINVALID_PARAMS;
            break;
        }
    }

    if (status == CPSW_SOK)
    {
        CSL_CPSW_setPortRxDscpMap(hPort->cpswRegs,
                                  hPort->portNum + 1U,
                                  dscpPriority->tosMap);

        CSL_CPSW_getPortControlReg(hPort->cpswRegs,
                                   hPort->portNum + 1U,
                                   &control);

        control.dscpIpv4Enable = dscpPriority->enableDscpIPv4;
        control.dscpIpv6Enable = dscpPriority->enableDscpIPv6;

        CSL_CPSW_setPortControlReg(hPort->cpswRegs,
                                   hPort->portNum + 1U,
                                   &control);
    }

    return status;
}

static void CpswMacPort_getDscpPriority(CpswMacPort_Handle hPort,
                                        CpswPort_DscpPriority *dscpPriority)
{
    CSL_CPSW_PORT_CONTROL control;

    CSL_CPSW_getPortRxDscpMap(hPort->cpswRegs,
                              hPort->portNum + 1U,
                              dscpPriority->tosMap);

    CSL_CPSW_getPortControlReg(hPort->cpswRegs,
                               hPort->portNum + 1U,
                               &control);

    dscpPriority->enableDscpIPv4 = control.dscpIpv4Enable;
    dscpPriority->enableDscpIPv6 = control.dscpIpv6Enable;
}

static uint64_t CpswMacPort_mapBwToCnt(uint64_t rateInBps,
                                       uint32_t cppiClkFreqHz)
{
    /*
     * Transfer rate to CIR or EIR count value equation:
     *
     *                Transfer rate (Mbps) * 32768
     * Count value = ------------------------------
     *                   CPPI_ICLK freq (MHz)
     *
     * Round division to prevent transfer rates below traffic shaping resolution
     * end up with no shaping at all.
     */
    return CPSW_UTILS_DIV_ROUNDUP(rateInBps * CPSW_MACPORT_RATELIM_DIV, cppiClkFreqHz);
}

static uint64_t CpswMacPort_mapCntToBw(uint32_t cntVal,
                                       uint32_t cppiClkFreqHz)
{
    /*
     * CIR or EIR count value to transfer rate equation:
     *
     *                         count value * CPPI_ICLK freq (MHz)
     * Transfer rate (Mbps) = ------------------------------------
     *                                      32768
     */
    return ((uint64_t)cntVal * cppiClkFreqHz) / CPSW_MACPORT_RATELIM_DIV;
}

static int32_t CpswMacPort_setTrafficShaping(CpswMacPort_Handle hPort,
                                             const CpswPort_TrafficShapingConfig *config)
{
    const CpswSoc_Config *pSocCfg = CpswSoc_getConfig(hPort->cpswType);
    uint32_t cppiClkFreqHz        = pSocCfg->cppiClkFreqHz;
    uint64_t cirBps;
    uint64_t eirBps;
    uint64_t cir;
    uint64_t eir;
    uint32_t i;
    int32_t status      = CPSW_SOK;
    bool rateLimEnabled = false;

    for (i = 0U; i < CPSW_PRIORITY_NUM; i++)
    {
        cirBps = config->rates[i].committedRateBitsPerSec;
        eirBps = config->rates[i].excessRateBitsPerSec;
        cir    = CpswMacPort_mapBwToCnt(cirBps, cppiClkFreqHz);
        eir    = CpswMacPort_mapBwToCnt(eirBps, cppiClkFreqHz);

        /* CIR must be non-zero if EIR is non-zero */
        if ((cir == 0ULL) && (eir != 0ULL))
        {
            CPSWTRACE_ERROR("MAC %u: EIR is enabled (%ubps = %llu) but CIR is not (%ubps = %llu) "
                            "for priority %u\n",
                            hPort->portNum, eirBps, eir, cirBps, cir, i);
            status = CPSW_EINVALID_PARAMS;
            break;
        }

        /* Rate limit must be the highest priority channels */
        if ((cir == 0ULL) && rateLimEnabled)
        {
            CPSWTRACE_ERROR("MAC %u: rate limiting disabled for priority %u\n",
                            hPort->portNum, i);
            status = CPSW_EINVALID_PARAMS;
            break;
        }

        /* Out of range */
        if ((cir > 0x0FFFFFFFULL) || (eir > 0x0FFFFFFFULL))
        {
            CPSWTRACE_ERROR("MAC %u: invalid CIR=%llu (%llubps) EIR=%llu (%llubps) for priority %u\n",
                            hPort->portNum, cir, cirBps, eir, eirBps, i);
            status = CPSW_EINVALID_PARAMS;
            break;
        }

        CPSWTRACE_DEBUG("MAC %u: rate limiting %s for priority %u, CIR=%llubps (%llu) EIR=%llubps (%llu)\n",
                        hPort->portNum,
                        (cir != 0ULL) ? "enabled" : "disabled",
                        i, cirBps, cir, eirBps, eir);

        CSL_CPSW_setPriCirEir(hPort->cpswRegs,
                              hPort->portNum,
                              i, (uint32_t)cir, (uint32_t)eir);

        if (cir != 0ULL)
        {
            rateLimEnabled = true;
        }
    }

    return status;
}

static int32_t CpswMacPort_getTrafficShaping(CpswMacPort_Handle hPort,
                                             CpswPort_TrafficShapingConfig *config)
{
    const CpswSoc_Config *pSocCfg = CpswSoc_getConfig(hPort->cpswType);
    uint32_t cppiClkFreqHz        = pSocCfg->cppiClkFreqHz;
    uint32_t cir;
    uint32_t eir;
    uint64_t cirBps;
    uint64_t eirBps;
    uint32_t i;
    int32_t status = CPSW_SOK;

    for (i = 0U; i < CPSW_PRIORITY_NUM; i++)
    {
        CSL_CPSW_getPriCirEir(hPort->cpswRegs,
                              hPort->portNum,
                              i,
                              &cir, &eir);

        cirBps = CpswMacPort_mapCntToBw(cir, cppiClkFreqHz);
        eirBps = CpswMacPort_mapCntToBw(eir, cppiClkFreqHz);

        /* CIR must be non-zero if EIR is non-zero */
        if ((cirBps == 0ULL) && (eirBps != 0ULL))
        {
            CPSWTRACE_ERROR("MAC %u: EIR is enabled (%llubps = %u) but CIR is not (%llubps = %u) "
                            "for priority %u\n",
                            hPort->portNum, eirBps, eir, cirBps, cir, i);
            status = CPSW_EUNEXPECTED;
            break;
        }

        CPSWTRACE_DEBUG("MAC %u: rate limiting %s for priority %u, CIR=%llubps (%u) EIR=%llubps (%u)\n",
                        hPort->portNum,
                        (cir != 0U) ? "enabled" : "disabled",
                        i, cirBps, cir, eirBps, eir);

        config->rates[i].committedRateBitsPerSec = cirBps;
        config->rates[i].excessRateBitsPerSec    = eirBps;
    }

    return status;
}

static void CpswMacPort_setSwitchTxScheduler(CpswMacPort_Handle hPort,
                                             CpswPort_TxPriorityType txPriorityType)
{
    CSL_CPSW_PTYPE pType;

    CSL_CPSW_getPTypeReg(hPort->cpswRegs, &pType);
    if (txPriorityType == CPSW_TX_PRIORITY_TYPE_FIXED)
    {
        switch (hPort->portNum)
        {
            case CPSW_MAC_PORT_0:
                pType.port1PriorityTypeEscalateEnable = FALSE;
                break;

            case CPSW_MAC_PORT_1:
                pType.port2PriorityTypeEscalateEnable = FALSE;
                break;

            case CPSW_MAC_PORT_2:
                pType.port3PriorityTypeEscalateEnable = FALSE;
                break;

            case CPSW_MAC_PORT_3:
                pType.port4PriorityTypeEscalateEnable = FALSE;
                break;

            case CPSW_MAC_PORT_4:
                pType.port5PriorityTypeEscalateEnable = FALSE;
                break;

            case CPSW_MAC_PORT_5:
                pType.port6PriorityTypeEscalateEnable = FALSE;
                break;

            case CPSW_MAC_PORT_6:
                pType.port7PriorityTypeEscalateEnable = FALSE;
                break;

            case CPSW_MAC_PORT_7:
                pType.port8PriorityTypeEscalateEnable = FALSE;
                break;

            default:
                CPSWTRACE_ERROR("MAC %u: invalid MAC port \n", hPort->portNum);
                Cpsw_assert(FALSE);
        }
    }
    else
    {
        switch (hPort->portNum)
        {
            case CPSW_MAC_PORT_0:
                pType.port1PriorityTypeEscalateEnable = TRUE;
                pType.escPriLoadVal                   = CPSW_CFG_ESC_PRI_LD_VAL;
                break;

            case CPSW_MAC_PORT_1:
                pType.port2PriorityTypeEscalateEnable = TRUE;
                pType.escPriLoadVal                   = CPSW_CFG_ESC_PRI_LD_VAL;
                break;

            case CPSW_MAC_PORT_2:
                pType.port3PriorityTypeEscalateEnable = TRUE;
                pType.escPriLoadVal                   = CPSW_CFG_ESC_PRI_LD_VAL;
                break;

            case CPSW_MAC_PORT_3:
                pType.port4PriorityTypeEscalateEnable = TRUE;
                pType.escPriLoadVal                   = CPSW_CFG_ESC_PRI_LD_VAL;
                break;

            case CPSW_MAC_PORT_4:
                pType.port5PriorityTypeEscalateEnable = TRUE;
                pType.escPriLoadVal                   = CPSW_CFG_ESC_PRI_LD_VAL;
                break;

            case CPSW_MAC_PORT_5:
                pType.port6PriorityTypeEscalateEnable = TRUE;
                pType.escPriLoadVal                   = CPSW_CFG_ESC_PRI_LD_VAL;
                break;

            case CPSW_MAC_PORT_6:
                pType.port7PriorityTypeEscalateEnable = TRUE;
                pType.escPriLoadVal                   = CPSW_CFG_ESC_PRI_LD_VAL;
                break;

            case CPSW_MAC_PORT_7:
                pType.port8PriorityTypeEscalateEnable = TRUE;
                pType.escPriLoadVal                   = CPSW_CFG_ESC_PRI_LD_VAL;
                break;

            default:
                CPSWTRACE_ERROR("MAC %u: invalid MAC port \n", hPort->portNum);
                Cpsw_assert(FALSE);
        }
    }

    CSL_CPSW_setPTypeReg(hPort->cpswRegs, &pType);
}

static bool  CpswMacPort_isMcastAddr(const uint8_t *macAddr)
{
    if ((macAddr[0] & 0x1) == 1)
    {
        return TRUE;
    }
    else
    {
        return FALSE;
    }
}

static int32_t CpswMacPort_validateInterVLANRouteId(const CpswMacPort_Handle hPort,
                                                    CpswMacPort_InterVLANRouteId routeId)
{
    int32_t status = CPSW_SOK;

    if (routeId > CPSW_MACPORT_INTERVLAN_ROUTEID_LAST)
    {
        CPSWTRACE_ERROR("MAC %u: invalid route id %u\n", hPort->portNum, routeId);
        status = CPSW_EINVALID_PARAMS;
    }

    return status;
}

static int32_t CpswMacPort_validateSetInterVLANParams(const CpswMacPort_Handle hPort,
                                                      CpswMacPort_InterVLANRouteId routeId,
                                                      const CpswMacPort_InterVlanRoutingParams *interVLANConfig)
{
    int32_t status;

    status = CpswMacPort_validateInterVLANRouteId(hPort, routeId);
    if (CpswMacPort_isMcastAddr(interVLANConfig->srcAddr))
    {
        CPSWTRACE_ERROR("MAC %u: source addr is multicast\n", hPort->portNum);
        status = CPSW_EINVALID_PARAMS;
    }

    if (interVLANConfig->vlanId > CPSW_VLAN_ID_MAX)
    {
        CPSWTRACE_ERROR("MAC %u: interVLAN VLAN ID %u exceeds max\n",
                        hPort->portNum, interVLANConfig->vlanId);
        status = CPSW_EINVALID_PARAMS;
    }

    return status;
}

static int32_t CpswMacPort_setInterVLANParams(const CpswMacPort_Handle hPort,
                                              CpswMacPort_InterVLANRouteId routeId,
                                              const CpswMacPort_InterVlanRoutingParams *interVLANConfig)
{
    int32_t status;

    status = CpswMacPort_validateSetInterVLANParams(hPort, routeId, interVLANConfig);
    if (CPSW_SOK == status)
    {
        CSL_CPSW_INTERVLANCFG cslInterVlanCfg;
        uint32_t interVLANPointer = CpswMacPort_mapInterVLANRouteId2Pointer(routeId);

        cslInterVlanCfg.decrementTtl = interVLANConfig->decrementTTL;
        cslInterVlanCfg.destForceUntaggedEgress = interVLANConfig->forceUntaggedEgress;
        CPSW_UTILS_ARRAY_COPY(cslInterVlanCfg.dstMacAddress, interVLANConfig->dstAddr);
        cslInterVlanCfg.replaceDaSa = interVLANConfig->replaceDASA;

        /* CPSW IP required replaceVid irrespective of whether VID needs to
         * be replaced or not. Else DA/SA/ VLAN strip is not functional.
         * Hence always enable replaceVid
         */
        cslInterVlanCfg.replaceVid = TRUE;
        CPSW_UTILS_ARRAY_COPY(cslInterVlanCfg.srcMacAddress, interVLANConfig->srcAddr);
        cslInterVlanCfg.vid = interVLANConfig->vlanId;
        CSL_CPGMAC_SL_setInterVLANCfg(hPort->cpswRegs,
                                      hPort->portNum,
                                      interVLANPointer,
                                      &cslInterVlanCfg);
    }
    else
    {
        CPSWTRACE_ERROR("MAC %u: failed to set interVLAN params: %d\n",
                        hPort->portNum, status);
    }

    return status;
}

static int32_t CpswMacPort_getInterVLANParams(const CpswMacPort_Handle hPort,
                                              CpswMacPort_InterVLANRouteId routeId,
                                              CpswMacPort_InterVlanRoutingParams *interVLANConfig)
{
    int32_t status;

    status = CpswMacPort_validateInterVLANRouteId(hPort, routeId);
    if (CPSW_SOK == status)
    {
        CSL_CPSW_INTERVLANCFG cslInterVlanCfg;
        uint32_t interVLANPointer = CpswMacPort_mapInterVLANRouteId2Pointer(routeId);

        CSL_CPGMAC_SL_getInterVLANCfg(hPort->cpswRegs, hPort->portNum, interVLANPointer, &cslInterVlanCfg);

        interVLANConfig->decrementTTL = cslInterVlanCfg.decrementTtl;
        interVLANConfig->forceUntaggedEgress = cslInterVlanCfg.destForceUntaggedEgress;
        CPSW_UTILS_COMPILETIME_ASSERT(sizeof(interVLANConfig->dstAddr) == sizeof(cslInterVlanCfg.dstMacAddress));
        CPSW_UTILS_ARRAY_COPY(interVLANConfig->dstAddr, cslInterVlanCfg.dstMacAddress);
        interVLANConfig->replaceDASA = cslInterVlanCfg.replaceDaSa;
        CPSW_UTILS_COMPILETIME_ASSERT(sizeof(interVLANConfig->srcAddr) == sizeof(cslInterVlanCfg.srcMacAddress));
        CPSW_UTILS_ARRAY_COPY(interVLANConfig->srcAddr, cslInterVlanCfg.srcMacAddress);
        interVLANConfig->vlanId = cslInterVlanCfg.vid;
    }
    else
    {
        CPSWTRACE_ERROR("MAC %u: failed to get interVLAN params: %d\n",
                        hPort->portNum, status);
    }

    return status;
}

static int32_t CpswMacPort_clearInterVLANRouteId(const CpswMacPort_Handle hPort,
                                                 CpswMacPort_InterVLANRouteId routeId)
{
    int32_t status;

    status = CpswMacPort_validateInterVLANRouteId(hPort, routeId);
    if (CPSW_SOK == status)
    {
        CSL_CPSW_INTERVLANCFG cslInterVlanCfg;
        uint32_t interVLANPointer = CpswMacPort_mapInterVLANRouteId2Pointer(routeId);

        cslInterVlanCfg.decrementTtl = FALSE;
        cslInterVlanCfg.destForceUntaggedEgress = FALSE;
        memset(&cslInterVlanCfg.dstMacAddress[0], 0, sizeof(cslInterVlanCfg.dstMacAddress));
        cslInterVlanCfg.replaceDaSa = FALSE;
        cslInterVlanCfg.replaceVid  = FALSE;
        memset(&cslInterVlanCfg.srcMacAddress[0], 0, sizeof(cslInterVlanCfg.srcMacAddress));
        cslInterVlanCfg.vid = FALSE;
        CSL_CPGMAC_SL_setInterVLANCfg(hPort->cpswRegs, hPort->portNum, interVLANPointer, &cslInterVlanCfg);
    }
    else
    {
        CPSWTRACE_ERROR("MAC %u: failed to clear interVLAN route %u: %d\n",
                        hPort->portNum, routeId, status);
    }

    return status;
}

static bool CpswMacPort_isInterVLANRouteIdFree(const CpswMacPort_Handle hPort,
                                               CpswMacPort_InterVLANRouteId routeId)
{
    CSL_CPSW_INTERVLANCFG cslInterVlanCfg;
    uint32_t interVLANPointer = CpswMacPort_mapInterVLANRouteId2Pointer(routeId);

    CSL_CPGMAC_SL_getInterVLANCfg(hPort->cpswRegs, hPort->portNum, interVLANPointer, &cslInterVlanCfg);

    if ((cslInterVlanCfg.replaceVid == 0) &&
        (cslInterVlanCfg.destForceUntaggedEgress == 0))
    {
        /* If a route does not have replace VId enabled it is not used.
         * Use that as an indicator to identify free routes.
         * We cannot use any software state to maintain interVLAN route allocation
         * as software state may get lost during warm boot */
        return TRUE;
    }
    else
    {
        return FALSE;
    }
}

static int32_t CpswMacPort_checkInterVLANRouteIsFree(const CpswMacPort_Handle hPort,
                                                     CpswMacPort_InterVLANRouteId routeId,
                                                     bool *pIsFree)
{
    int32_t status;

    status = CpswMacPort_validateInterVLANRouteId(hPort, routeId);
    if (CPSW_SOK == status)
    {
        *pIsFree = CpswMacPort_isInterVLANRouteIdFree(hPort, routeId);
    }
    else
    {
        CPSWTRACE_ERROR("MAC %u: failed to check interVLAN route %u state: %d\n",
                        hPort->portNum, routeId, status);
    }

    return status;
}

static int32_t CpswMacPort_getInterVLANFreeRoutes(const CpswMacPort_Handle hPort,
                                                  CpswMacPort_InterVlanFreeRouteInfo *freeRoutes)
{
    CpswMacPort_InterVLANRouteId i;

    freeRoutes->numFreeRoutes = 0;
    for (i = CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST;
         i <= CPSW_MACPORT_INTERVLAN_ROUTEID_LAST;
         i = (CpswMacPort_InterVLANRouteId)(i + 1))
    {
        if (CpswMacPort_isInterVLANRouteIdFree(hPort, i))
        {
            Cpsw_assert(freeRoutes->numFreeRoutes < CPSW_UTILS_ARRAYSIZE(freeRoutes->freeRouteId));
            freeRoutes->freeRouteId[freeRoutes->numFreeRoutes] = i;
            freeRoutes->numFreeRoutes++;
        }
    }

    return CPSW_SOK;
}

static bool CpswMacPort_compareRoutes(const CpswMacPort_InterVlanRoutingParams *params1,
                                      const CpswMacPort_InterVlanRoutingParams *params2)
{
    bool isEqual;

    isEqual = (params1->decrementTTL == params2->decrementTTL) &&
              (params1->forceUntaggedEgress == params2->forceUntaggedEgress) &&
              (params1->replaceDASA == params2->replaceDASA);
    if ((isEqual) && (params1->forceUntaggedEgress == false))
    {
        isEqual = isEqual && (params1->vlanId == params2->vlanId);
    }

    if ((isEqual) && (params1->replaceDASA))
    {
        isEqual = isEqual &&
                  (memcmp(params1->dstAddr, params2->dstAddr, sizeof(params1->dstAddr)) == 0);
        isEqual = isEqual &&
                  (memcmp(params1->srcAddr, params2->srcAddr, sizeof(params1->srcAddr)) == 0);
    }

    return isEqual;
}

static int32_t CpswMacPort_findInterVLANRouteId(const CpswMacPort_Handle hPort,
                                                const CpswMacPort_InterVlanRoutingParams *routeMatchParams,
                                                CpswMacPort_InterVLANRouteId *pRouteId)
{
    int32_t retVal = CPSW_MACPORT_E_ROUTENOTFOUND;
    CpswMacPort_InterVlanRoutingParams routeParams;
    bool matchFound;
    CpswMacPort_InterVLANRouteId i;

    matchFound = FALSE;
    for (i = CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST;
         i <= CPSW_MACPORT_INTERVLAN_ROUTEID_LAST;
         i = (CpswMacPort_InterVLANRouteId)(i + 1))
    {
        if (CpswMacPort_getInterVLANParams(hPort, i, &routeParams) == CPSW_SOK)
        {
            if (CpswMacPort_compareRoutes(routeMatchParams, &routeParams) == TRUE)
            {
                matchFound = TRUE;
                break;
            }
        }
    }

    if (matchFound)
    {
        Cpsw_assert(i <= CPSW_MACPORT_INTERVLAN_ROUTEID_LAST);
        *pRouteId = i;
        retVal = CPSW_SOK;
    }
    else
    {
        CPSWTRACE_DEBUG("MAC %u: no interVLAN found with matching params\n", hPort->portNum);
        retVal = CPSW_MACPORT_E_ROUTENOTFOUND;
    }

    return retVal;
}

static int32_t CpswMacPort_delInterVLANRoute(const CpswMacPort_Handle hPort,
                                             const CpswMacPort_InterVlanRoutingParams *interVLANConfig,
                                             CpswMacPort_InterVLANRouteId *pRouteId)
{
    int32_t retVal;

    retVal = CpswMacPort_findInterVLANRouteId(hPort, interVLANConfig, pRouteId);
    if (CPSW_SOK == retVal)
    {
        /* Entry already present. No need to add entry */
        retVal = CpswMacPort_validateInterVLANRouteId(hPort, *pRouteId);
        CPSWTRACE_ERROR_IF((retVal != CPSW_SOK),
                           "MAC %u: entry already present\n", hPort->portNum);
        Cpsw_assert(retVal == CPSW_SOK);

        CpswMacPort_clearInterVLANRouteId(hPort, *pRouteId);
    }
    else
    {
        CPSWTRACE_ERROR("MAC %u: failed to delete interVLAN route\n", hPort->portNum);
    }

    return retVal;
}

static int32_t CpswMacPort_allocSetInterVLANParams(const CpswMacPort_Handle hPort,
                                                   const CpswMacPort_InterVlanRoutingParams *interVLANConfig,
                                                   CpswMacPort_InterVLANRouteId *pRouteId)
{
    int32_t retVal;

    retVal = CpswMacPort_findInterVLANRouteId(hPort, interVLANConfig, pRouteId);
    if (CPSW_SOK == retVal)
    {
        /* Entry already present. No need to add entry */
        retVal = CpswMacPort_validateInterVLANRouteId(hPort, *pRouteId);
        CPSWTRACE_ERROR_IF((retVal != CPSW_SOK),
                           "MAC %u: entry already present\n", hPort->portNum);
        Cpsw_assert(retVal == CPSW_SOK);
    }
    else
    {
        CpswMacPort_InterVlanFreeRouteInfo freeRoutes;

        retVal = CpswMacPort_getInterVLANFreeRoutes(hPort, &freeRoutes);
        if (CPSW_SOK == retVal)
        {
            if (freeRoutes.numFreeRoutes > 0)
            {
                retVal = CpswMacPort_setInterVLANParams(hPort,
                                                        freeRoutes.freeRouteId[0],
                                                        interVLANConfig);
                if (CPSW_SOK == retVal)
                {
                    *pRouteId = freeRoutes.freeRouteId[0];
                }
                else
                {
                    CPSWTRACE_ERROR("MAC %u: failed to set interVLAN params: %d\n",
                                    hPort->portNum, retVal);
                }
            }
            else
            {
                CPSWTRACE_ERROR("MAC %u: no free routes available\n", hPort->portNum);
                retVal = CPSW_EFAIL;
            }
        }
        else
        {
            CPSWTRACE_ERROR("MAC %u: failed to get free interVLAN routes: %d\n",
                            hPort->portNum, retVal);
        }
    }

    return retVal;
}

uint32_t CpswMacPort_mapInterVLANRouteId2Pointer(CpswMacPort_InterVLANRouteId routeId)
{
    uint32_t interVlanPointer = CPSW_MACPORT_INTERVLAN_ROUTEID_BASE +
                                (routeId - CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST);

    return interVlanPointer;
}

static void CpswMacPort_clearAllInterVLANRoutes(const CpswMacPort_Handle hPort)
{
    int32_t status;
    CpswMacPort_InterVLANRouteId i;

    /* Clear interVlan routes only if supported by SoC */
    if (CpswMacPort_isInterVLANSupported(hPort->cpswType))
    {
        for (i = CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST;
             i <= CPSW_MACPORT_INTERVLAN_ROUTEID_LAST;
             i = (CpswMacPort_InterVLANRouteId)(i + 1))
        {
            if (!CpswMacPort_isInterVLANRouteIdFree(hPort, i))
            {
                status = CpswMacPort_clearInterVLANRouteId(hPort, i);
                CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                                   "MAC %u: failed to clear interVLAN route %u: %d\n",
                                   hPort->portNum, i, status);
                Cpsw_assert(status == CPSW_SOK);
            }
        }
    }
}

static bool CpswMacPort_isInterVLANSupported(Cpsw_Type cpswType)
{
    return (0U != (CPSW_FEATURE_INTERVLAN & CpswUtils_getIpFeatureBitMap(cpswType)));
}

static void CpswMacPort_printRegs(Cpsw_MacPort portNum,
                                  CSL_Xge_cpswEnetportRegs *cpswMacPortRegs)
{
    uint32_t *regAddr = (uint32_t *)((uintptr_t)cpswMacPortRegs + CPSW_MACPORT_START_REG_OFFSET);
    uint32_t regIdx = 0;

    while (((uintptr_t)regAddr) <= ((uintptr_t)cpswMacPortRegs + CPSW_MACPORT_END_REG_OFFSET))
    {
        if (*regAddr)
        {
            CpswUtils_printf("MACPORT %u: %u: 0x%08x\n", portNum, regIdx, *regAddr);
        }

        regAddr++;
        regIdx++;
    }
}

static int32_t CpswMacPort_setShortIPGCfg(const CpswMacPort_Handle hPort,
                                          const CpswMacPort_TxShortGapConfig *txShortIPGCfg)
{
    if (txShortIPGCfg->enableTxShortGapLimit)
    {
        CSL_CPGMAC_SL_enableTxShortGapLimit(hPort->cpswRegs, CPSW_NORMALIZE_MACPORT(hPort->portNum));
    }
    else
    {
        CSL_CPGMAC_SL_disableTxShortGapLimit(hPort->cpswRegs, CPSW_NORMALIZE_MACPORT(hPort->portNum));
    }

    if (txShortIPGCfg->enableTxShortGap)
    {
        CSL_CPGMAC_SL_enableTxShortGap(hPort->cpswRegs, CPSW_NORMALIZE_MACPORT(hPort->portNum));
    }
    else
    {
        CSL_CPGMAC_SL_disableTxShortGap(hPort->cpswRegs, CPSW_NORMALIZE_MACPORT(hPort->portNum));
    }

    return CPSW_SOK;
}

static int32_t CpswMacPort_getShortIPGCfg(const CpswMacPort_Handle hPort,
                                          CpswMacPort_TxShortGapConfig *txShortIPGCfg)
{
    txShortIPGCfg->enableTxShortGap = CSL_CPGMAC_SL_isTxShortGapEnabled(hPort->cpswRegs, CPSW_NORMALIZE_MACPORT(hPort->portNum));
    txShortIPGCfg->enableTxShortGapLimit = CSL_CPGMAC_SL_isTxShortGapLimitEnabled(hPort->cpswRegs, CPSW_NORMALIZE_MACPORT(hPort->portNum));
    return CPSW_SOK;
}

static int32_t CpswMacPort_enableCptsPortEvents(const CpswMacPort_Handle hPort,
                                                CpswMacPort_EnableTsEventInArgs *portEventCfg)
{
    int32_t status = CPSW_SOK;

    if ((hPort->portNum == portEventCfg->portNum) &&
        (portEventCfg->portCfg.seqIdOffset >= CPSW_MACPORT_PTP_SEQID_OFFSET_MIN_VAL))
    {
        CSL_CPSW_TSCONFIG timeSyncConfig;
        CpswMacPort_TsEventConfig *portCfg = &portEventCfg->portCfg;
        CpswMacPort_IpTsConfig *commonPortIPCfg = &portCfg->commonPortIpConfig;

        memset(&timeSyncConfig, 0U, sizeof(CSL_CPSW_TSCONFIG));
        timeSyncConfig.tsTxHostEnable   = portCfg->enableTxHostTs;
        timeSyncConfig.tsRxAnnexDEnable = portCfg->enableRxAnnexD;
        timeSyncConfig.tsRxAnnexEEnable = portCfg->enableRxAnnexE;
        timeSyncConfig.tsRxAnnexFEnable = portCfg->enableRxAnnexF;
        timeSyncConfig.tsTxAnnexDEnable = portCfg->enableTxAnnexD;
        timeSyncConfig.tsTxAnnexEEnable = portCfg->enableTxAnnexE;
        timeSyncConfig.tsTxAnnexFEnable = portCfg->enableTxAnnexF;
        timeSyncConfig.tsMsgTypeEnable  = portCfg->messageType;
        timeSyncConfig.tsSeqIdOffset    = portCfg->seqIdOffset;
        timeSyncConfig.tsDomainOffset   = portCfg->domainOffset;

        if ((portCfg->enableRxAnnexD == true)||
            (portCfg->enableRxAnnexE == true)||
            (portCfg->enableTxAnnexD == true)||
            (portCfg->enableTxAnnexE == true))
        {
            if ((portCfg->enableRxAnnexE == true)||
                (portCfg->enableTxAnnexE == true))
            {
                timeSyncConfig.tsMcastTypeEnable = portCfg->mcastType;
            }

            timeSyncConfig.tsTTLNonzeroEnable = commonPortIPCfg->enableTTLNonzero;
            timeSyncConfig.tsUniEnable        = commonPortIPCfg->enableUnicast;

            if (timeSyncConfig.tsUniEnable == false)
            {
                timeSyncConfig.ts107Enable = commonPortIPCfg->enableTsIp107;
                timeSyncConfig.ts129Enable = commonPortIPCfg->enableTsIp129;
                timeSyncConfig.ts130Enable = commonPortIPCfg->enableTsIp130;
                timeSyncConfig.ts131Enable = commonPortIPCfg->enableTsIp131;
                timeSyncConfig.ts132Enable = commonPortIPCfg->enableTsIp132;
                timeSyncConfig.ts319Enable = commonPortIPCfg->enableTsPort319;
                timeSyncConfig.ts320Enable = commonPortIPCfg->enableTsPort320;
            }
        }

        if ((portCfg->enableRxAnnexF == true)||
            (portCfg->enableTxAnnexF == true))
        {
            timeSyncConfig.tsLType2Enable  = portCfg->enableLType2;
            timeSyncConfig.tsLType1        = CPSW_ETHERTYPE_PTP;

            if (timeSyncConfig.tsLType2Enable == true)
            {
                timeSyncConfig.tsLType2    = CPSW_ETHERTYPE_PTP;
            }
        }

        if (portCfg->rxVlanType != CPSW_MACPORT_VLAN_TYPE_NONE)
        {
            timeSyncConfig.tsRxVlanLType1Enable = true;
            if (portCfg->rxVlanType == CPSW_MACPORT_VLAN_TYPE_STACKED_TAGS)
            {
                timeSyncConfig.tsRxVlanLType2Enable = true;
            }
        }

        if (portCfg->txVlanType != CPSW_MACPORT_VLAN_TYPE_NONE)
        {
            timeSyncConfig.tsTxVlanLType1Enable = true;
            if (portCfg->txVlanType == CPSW_MACPORT_VLAN_TYPE_STACKED_TAGS)
            {
                timeSyncConfig.tsTxVlanLType2Enable = true;
            }
        }

        if ((portCfg->rxVlanType != CPSW_MACPORT_VLAN_TYPE_NONE) ||
            (portCfg->txVlanType != CPSW_MACPORT_VLAN_TYPE_NONE))
        {
            timeSyncConfig.tsVlanLType1 = portCfg->vlanLType1;

            if ((portCfg->rxVlanType != CPSW_MACPORT_VLAN_TYPE_STACKED_TAGS)||
                (portCfg->txVlanType != CPSW_MACPORT_VLAN_TYPE_STACKED_TAGS))
            {
                timeSyncConfig.tsVlanLType2 = portCfg->vlanLType2;
            }
        }

        CSL_CPSW_setPortTimeSyncConfig(hPort->cpswRegs,
                                       (CPSW_NORMALIZE_MACPORT(portEventCfg->portNum) + 1U),
                                       &timeSyncConfig );
    }
    else
    {
        status = CPSW_EINVALID_PARAMS;
    }

    return status;
}

static int32_t CpswMacPort_disableCptsPortEvents(const CpswMacPort_Handle hPort,
                                                 Cpsw_MacPort portNum)
{
    int32_t status = CPSW_SOK;

    if (CPSW_NORMALIZE_MACPORT(portNum) <= CPSW_MAC_PORT_LAST)
    {
        CSL_CPSW_TSCONFIG timeSyncConfig;
        CSL_CPSW_getPortTimeSyncConfig(hPort->cpswRegs,
                                       (CPSW_NORMALIZE_MACPORT(portNum) + 1U),
                                       &timeSyncConfig);

        timeSyncConfig.tsTxHostEnable   = false;
        timeSyncConfig.tsRxAnnexDEnable = false;
        timeSyncConfig.tsRxAnnexEEnable = false;
        timeSyncConfig.tsRxAnnexFEnable = false;
        timeSyncConfig.tsTxAnnexDEnable = false;
        timeSyncConfig.tsTxAnnexEEnable = false;
        timeSyncConfig.tsTxAnnexFEnable = false;
        timeSyncConfig.tsMsgTypeEnable  = false;

        CSL_CPSW_setPortTimeSyncConfig(hPort->cpswRegs,
                                       (CPSW_NORMALIZE_MACPORT(portNum) + 1U),
                                       &timeSyncConfig );
    }
    else
    {
        status = CPSW_EINVALID_PARAMS;
    }

    return status;
}
