/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 *  \file cpsw_osal.c
 *
 *  \brief File containing the CPSW driver OSAL abstraction functions.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/include/cpsw_utils.h>
#include <ti/drv/cpsw/src/priv/cpsw_osal.h>
#include <ti/drv/cpsw/src/priv/cpsw_priv.h>

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

typedef struct CpswOsal_Obj_s
{
    CpswOsal_Prms osalPrms;
} CpswOsal_Obj;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/** \brief Variable to store the OSAL cache parameter across driver handle */
static CpswOsal_CachePrms gCpswOsalCachePrms =
{
    &CacheP_Inv, &CacheP_wb, &CacheP_wbInv,
};

/** \brief CPSW OSAL Object - we store OSAL config parameters passed by
 *        by application in driver object and directly use those to call the
 *        the OSAL functions.
 * \Note  NULL check before each call is skipped as it is done during OSAL init
 */
static CpswOsal_Obj gCpswOsalObj;

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

void CpswOsal_initPrms(CpswOsal_Prms *pOsalPrms)
{
    if (NULL != pOsalPrms)
    {
        pOsalPrms->disableAllIntr = &HwiP_disable;
        pOsalPrms->restoreAllIntr = &HwiP_restore;
        pOsalPrms->disableIntr    = &CpswUtils_disableIntr;
        pOsalPrms->restoreIntr    = &CpswUtils_enableIntr;
        pOsalPrms->createMutex    = &CpswUtils_mutexCreate;
        pOsalPrms->deleteMutex    = &CpswUtils_mutexDelete;
        pOsalPrms->lockMutex      = &CpswUtils_mutexLock;
        pOsalPrms->unlockMutex    = &CpswUtils_mutexUnlock;
        pOsalPrms->registerIntr   = &CpswUtils_registerIntr;
        pOsalPrms->unRegisterIntr = &CpswUtils_unRegisterIntr;
        pOsalPrms->timerRead      = &CpswUtils_timerRead;
    }

    return;
}

void CpswOsal_open(const CpswOsal_Prms *pOsalPrms)
{
    memset(&gCpswOsalObj, 0U, sizeof(gCpswOsalObj));

    CpswUtils_timerInit();

    Cpsw_assert(pOsalPrms->disableAllIntr != NULL);
    Cpsw_assert(pOsalPrms->restoreAllIntr != NULL);
    Cpsw_assert(pOsalPrms->disableIntr != NULL);
    Cpsw_assert(pOsalPrms->restoreIntr != NULL);
    Cpsw_assert(pOsalPrms->createMutex != NULL);
    Cpsw_assert(pOsalPrms->deleteMutex != NULL);
    Cpsw_assert(pOsalPrms->lockMutex != NULL);
    Cpsw_assert(pOsalPrms->unlockMutex != NULL);
    Cpsw_assert(pOsalPrms->registerIntr != NULL);
    Cpsw_assert(pOsalPrms->unRegisterIntr != NULL);

    gCpswOsalObj.osalPrms = *pOsalPrms;

    return;
}

void CpswOsal_close(void)
{
    CpswUtils_timerDeInit();

    memset(&gCpswOsalObj, 0U, sizeof(gCpswOsalObj));

    return;
}

void CpswOsal_initCachePrms(CpswOsal_CachePrms *cachePrms)
{
    if (NULL != cachePrms)
    {
        cachePrms->cacheInv = &CacheP_Inv;
        cachePrms->cacheWb  = &CacheP_wb;
    }

    return;
}

int32_t CpswOsal_setCachePrms(const CpswOsal_CachePrms *cachePrms)
{
    int32_t retVal = UDMA_SOK;

    if ((NULL == cachePrms) ||
        (NULL == cachePrms->cacheInv) ||
        (NULL == cachePrms->cacheWb))
    {
        retVal = UDMA_EBADARGS;
    }
    else
    {
        gCpswOsalCachePrms.cacheInv = cachePrms->cacheInv;
        gCpswOsalCachePrms.cacheWb  = cachePrms->cacheWb;
    }

    return(retVal);
}

void CpswOsal_cacheInv(const void *pVirtAddr,
                       int32_t size)
{
    bool isCacheCoherent = Udma_isCacheCoherent();

#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswOsalCachePrms.cacheInv != NULL);
#endif

    if (isCacheCoherent != true)
    {
        gCpswOsalCachePrms.cacheInv(pVirtAddr, size);
    }
}

void CpswOsal_cacheWb(const void *pVirtAddr,
                      int32_t size)
{
    bool isCacheCoherent = Udma_isCacheCoherent();

#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswOsalCachePrms.cacheWb != NULL);
#endif

    if (isCacheCoherent != true)
    {
        gCpswOsalCachePrms.cacheWb(pVirtAddr, size);
    }
}

void CpswOsal_cacheWbInv(const void *pVirtAddr,
                         int32_t size)
{
    bool isCacheCoherent = Udma_isCacheCoherent();

#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswOsalCachePrms.cacheWbInv != NULL);
#endif

    if (isCacheCoherent != true)
    {
        gCpswOsalCachePrms.cacheWbInv(pVirtAddr, size);
    }
}

void CpswOsal_restoreAllIntr(uintptr_t cookie)
{
#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswOsalObj.osalPrms.restoreAllIntr != NULL);
#endif

    gCpswOsalObj.osalPrms.restoreAllIntr(cookie);
}

uintptr_t CpswOsal_disableAllIntr(void)
{
#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswOsalObj.osalPrms.disableAllIntr != NULL);
#endif

    return gCpswOsalObj.osalPrms.disableAllIntr();
}

void *CpswOsal_registerIntr(CpswOsal_IsrFxn isrFxn,
                            uint32_t coreIntrNum,
                            uint32_t intrPriority,
                            void *arg)
{
#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswOsalObj.osalPrms.registerIntr != NULL);
#endif

    return gCpswOsalObj.osalPrms.registerIntr(isrFxn,
                                              coreIntrNum,
                                              intrPriority,
                                              arg);
}

void CpswOsal_unRegisterIntr(void *hwiHandle)
{
#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswOsalObj.osalPrms.unRegisterIntr != NULL);
#endif

    return gCpswOsalObj.osalPrms.unRegisterIntr(hwiHandle);
}

void CpswOsal_restoreIntr(uint32_t coreIntrNum)
{
#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswOsalObj.osalPrms.restoreIntr != NULL);
#endif

    gCpswOsalObj.osalPrms.restoreIntr(coreIntrNum);
}

void CpswOsal_disableIntr(uint32_t coreIntrNum)
{
#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswOsalObj.osalPrms.disableIntr != NULL);
#endif

    gCpswOsalObj.osalPrms.disableIntr(coreIntrNum);
}

/* We need only mutex in CPSW driver. Use semaphore in binary mode */
void *CpswOsal_mutexCreate(void)
{
#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswOsalObj.osalPrms.createMutex != NULL);
#endif

    return gCpswOsalObj.osalPrms.createMutex();
}

void CpswOsal_mutexDelete(void *mutexHandle)
{
#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswOsalObj.osalPrms.deleteMutex != NULL);
#endif

    gCpswOsalObj.osalPrms.deleteMutex(mutexHandle);
}

void CpswOsal_mutexLock(void *mutexHandle)
{
#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswOsalObj.osalPrms.lockMutex != NULL);
#endif

    gCpswOsalObj.osalPrms.lockMutex(mutexHandle);
}

void CpswOsal_mutexUnlock(void *mutexHandle)
{
#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswOsalObj.osalPrms.unlockMutex != NULL);
#endif

    gCpswOsalObj.osalPrms.unlockMutex(mutexHandle);
}

uint32_t CpswOsal_timerGetDiff(uint32_t startTime)
{
    uint32_t currTime;

#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswOsalObj.osalPrms.timerRead != NULL);
#endif

    currTime = gCpswOsalObj.osalPrms.timerRead();

    return(currTime - startTime);
}

uint32_t CpswOsal_timerRead(void)
{
#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(gCpswOsalObj.osalPrms.timerRead != NULL);
#endif

    return gCpswOsalObj.osalPrms.timerRead();
}

/* ========================================================================== */
/*                 Internal Function Declarations                             */
/* ========================================================================== */

/* end of file */
