/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file cpsw_phy.c
 *
 * \brief This file contains the ENETPHY functions.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <string.h>
#include <stdio.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/include/cpsw_phy.h>
#include <ti/drv/cpsw/include/cpsw_macport.h>
#include "priv/cpsw_phy_priv.h"
#include "priv/cpsw_mdio_priv.h"
#include "priv/cpsw_trace_priv.h"
#include "phy/generic_phy.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/*! \brief Capability string buffer length */
#define CPSW_PHY_CAPS_BUF_LEN                 (41U)

/*! \brief PHY mode string buffer length */
#define CPSW_PHY_MODE_BUF_LEN                 (25U)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* Miscellaneous capability-related auxiliary functions */

#if (CPSW_TRACE_CFG_TRACE_LEVEL >= CPSW_TRACE_CFG_LEVEL_WARN)
static const char *CpswPhy_getCapsString(uint32_t linkCaps);

#endif

#if (CPSW_TRACE_CFG_TRACE_LEVEL >= CPSW_TRACE_CFG_LEVEL_DEBUG)
static const char *CpswPhy_getModeString(Cpsw_Speed speed,
                                         Cpsw_Duplexity duplexity);

#endif

static void CpswPhy_showLinkPartnerCompat(CpswPhy_Handle hPhyDev,
                                          Cpsw_Speed speed,
                                          Cpsw_Duplexity duplexity);

static uint32_t CpswPhy_getManualCaps(Cpsw_Speed speed,
                                      Cpsw_Duplexity duplexity);

static uint32_t CpswPhy_findBestCap(uint32_t caps);

static void CpswPhy_capToMode(uint32_t caps,
                              Cpsw_Speed *speed,
                              Cpsw_Duplexity *duplexity);

static CpswPhy_MacMode CpswPhy_ifConfigToMacMode(CpswMacPort_Interface *interface);

/* MDIO wrapper functions */

static bool CpswPhy_isPhyAlive(CpswMdio_Handle hMdio,
                               uint8_t phyAddr);

static bool CpswPhy_isPhyLinked(CpswMdio_Handle hMdio,
                                uint8_t phyAddr);

/* PHY device related functions */

static void CpswPhy_setTimeoutConfig(CpswPhy_Handle hPhyDev,
                                     const CpswPhy_FsmTimeoutConfig *cfg);

static uint32_t CpswPhy_getVersion(CpswPhy_Handle hPhyDev,
                                   CpswPhy_Version *version);

static int32_t CpswPhy_bindDriver(CpswPhy_Handle hPhyDev);

static void CpswPhy_resetPhy(CpswPhy_Handle hPhyDev);

static int32_t CpswPhy_getLinkMode(CpswPhy_Handle hPhyDev,
                                   Cpsw_Speed *speed,
                                   Cpsw_Duplexity *duplexity);

static void CpswPhy_setPhyMode(CpswPhy_Handle hPhyDev,
                               uint32_t nwayCaps,
                               bool enableLoopback);

/* State machine related functions */

static int32_t CpswPhy_tick(CpswPhy_Handle hPhyDev,
                            uint32_t *mdioStatus);

static void CpswPhy_setNextState(CpswPhy_Handle hPhyDev,
                                 CpswPhy_FsmState state);

static bool CpswPhy_isNwayCapable(CpswPhy_Handle hPhyDev);

static uint32_t CpswPhy_getLocalCaps(CpswPhy_Handle hPhyDev);

static uint32_t CpswPhy_findCommonCaps(CpswPhy_Handle hPhyDev);

static uint32_t CpswPhy_findCommon1000Caps(CpswPhy_Handle hPhyDev);

static uint32_t CpswPhy_findCommonNwayCaps(CpswPhy_Handle hPhyDev);

static bool CpswPhy_mdixSupported(CpswPhy_Handle hPhyDev);

static void CpswPhy_mdixDelay(CpswPhy_Handle hPhyDev);

static void CpswPhy_phyTimeOut(CpswPhy_Handle hPhyDev);

static void CpswPhy_defaultState(CpswPhy_Handle hPhyDev);

static void CpswPhy_initState(CpswPhy_Handle hPhyDev);

static void CpswPhy_findingState(CpswPhy_Handle hPhyDev);

static void CpswPhy_foundState(CpswPhy_Handle hPhyDev);

static void CpswPhy_resetWaitState(CpswPhy_Handle hPhyDev);

static void CpswPhy_enableState(CpswPhy_Handle hPhyDev);

static void CpswPhy_setupNway(CpswPhy_Handle hPhyDev);

static void CpswPhy_setupManual(CpswPhy_Handle hPhyDev,
                                bool enableLoopback);

static void CpswPhy_loopbackState(CpswPhy_Handle hPhyDev);

static void CpswPhy_nwayStartState(CpswPhy_Handle hPhyDev);

static void CpswPhy_nwayWaitState(CpswPhy_Handle hPhyDev);

static void CpswPhy_linkWaitState(CpswPhy_Handle hPhyDev);

static void CpswPhy_linkedState(CpswPhy_Handle hPhyDev);

/* Wrappers for PHY-specific functionality */

static int32_t CpswPhy_runComplianceTest(CpswPhy_Handle hPhyDev,
                                         uint32_t testMode);

static int32_t CpswPhy_readExtReg(CpswPhy_Handle hPhyDev,
                                  uint32_t reg,
                                  uint16_t *val);

static int32_t CpswPhy_writeExtReg(CpswPhy_Handle hPhyDev,
                                   uint32_t reg,
                                   uint16_t val);

static void CpswPhy_printRegs(CpswPhy_Handle hPhyDev);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

extern CpswPhy_Drv gCpswPhyDrvGeneric;
extern CpswPhy_Drv gCpswPhyDrvDp83822;
extern CpswPhy_Drv gCpswPhyDrvDp83867;
extern CpswPhy_Drv gCpswPhyDrvVsc8514;

static CpswPhyDrv_Handle gCpswPhyDrvs[] =
{
    &gCpswPhyDrvVsc8514,          /* VSC8514 */
    &gCpswPhyDrvDp83822,          /* DP83822 */
    &gCpswPhyDrvDp83867,          /* DP83867 */
    &gCpswPhyDrvGeneric,          /* Generic PHY must be last */
};

/*! \brief CPSW MAC port objects */
static CpswPhy_Obj gCpswPhyObjs[CPSW_COUNT][CPSW_MAC_PORT_NUM];

#ifdef CPSW_PHY_MDIX_SUPPORT
static uint32_t gRandomSeed = 1U;
#endif

#if (CPSW_TRACE_CFG_TRACE_LEVEL >= CPSW_TRACE_CFG_LEVEL_DEBUG)
static const char *gCpswPhyStateNames[] =
{
    "INIT",
    "FINDING",
    "RESET_WAIT",
    "ENABLE",
    "FOUND",
    "NWAY_START",
    "NWAY_WAIT",
    "LINK_WAIT",
    "LINKED",
    "LOOPBACK",
};
#endif

#if (CPSW_TRACE_CFG_TRACE_LEVEL >= CPSW_TRACE_CFG_LEVEL_WARN)
static char gCpswPhyCapsBuf[CPSW_PHY_CAPS_BUF_LEN];
#endif

#if (CPSW_TRACE_CFG_TRACE_LEVEL >= CPSW_TRACE_CFG_LEVEL_DEBUG)
static char gCpswPhyModeBuf[CPSW_PHY_MODE_BUF_LEN];
#endif

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

void CpswPhy_initParams(CpswPhy_Config *phyConfig)
{
    memset(phyConfig, 0U, sizeof(*phyConfig));

    phyConfig->phyAddr                                    = CPSW_PHY_INVALID_PHYADDR;
    phyConfig->nwayCaps                                   = 0U; /* Defaults to all capabilities */
    phyConfig->mdixEnable                                 = true;
    phyConfig->isStrapped                                 = false;
    phyConfig->enableLoopback                             = false;
    phyConfig->masterMode                                 = false;
    phyConfig->extClkSource                               = false;
    phyConfig->skipExtendedConfig                         = false;
    phyConfig->timeoutConfig.findingStateTicks            = 0U;
    phyConfig->timeoutConfig.resetWaitStateTicks          = 0U;
    phyConfig->timeoutConfig.resetWaitStateResidenceTicks = 0U;
    phyConfig->timeoutConfig.nwayStartStateTicks          = 0U;
    phyConfig->timeoutConfig.nwayWaitStateTicks           = 0U;
    phyConfig->timeoutConfig.linkWaitStateTicks           = 0U;
    phyConfig->timeoutConfig.mdixTicks                    = 0U;
}

void CpswPhy_setExtendedParams(CpswPhy_Config *phyConfig,
                               const void *extendedConfig,
                               uint32_t extendedConfigSize)
{
    Cpsw_assert(extendedConfigSize <= CPSW_PHY_EXTENDED_CFG_SIZE_MAX);

    if ((extendedConfigSize > 0U) &&
        (extendedConfig != NULL))
    {
        memcpy(phyConfig->extendedConfig,
               extendedConfig,
               extendedConfigSize);
    }

    phyConfig->extendedConfigSize = extendedConfigSize;
}

CpswPhy_Handle CpswPhy_open(Cpsw_Type cpswType,
                            CpswMdio_Handle hMdio,
                            Cpsw_MacPort portNum,
                            CpswMacPort_Interface *interface,
                            const CpswPhy_Config *phyConfig,
                            const CpswMacPort_LinkConfig *linkConfig)
{
    CpswPhy_Handle hPhyDev = &gCpswPhyObjs[cpswType][portNum];
    int32_t status         = CPSW_SOK;
    bool manualMode        = false;
    bool alive;

    memset(hPhyDev, 0U, sizeof(*hPhyDev));

    hPhyDev->cpswType   = cpswType;
    hPhyDev->hDrv       = NULL;
    hPhyDev->hMdio      = hMdio;
    hPhyDev->phyConfig  = *phyConfig;
    hPhyDev->linkConfig = *linkConfig;
    hPhyDev->portNum    = portNum;
    hPhyDev->addr       = phyConfig->phyAddr;
    hPhyDev->macMode    = CpswPhy_ifConfigToMacMode(interface);

    /* State's linkCaps will be a mask for NWAY path and
     * single bit for manual settings */
    if ((linkConfig->speed == CPSW_SPEED_AUTO) ||
        (linkConfig->duplexity == CPSW_DUPLEX_AUTO))
    {
        hPhyDev->reqLinkCaps = phyConfig->nwayCaps;
        if (hPhyDev->reqLinkCaps == 0U)
        {
            hPhyDev->reqLinkCaps = CPSW_LINK_CAP_ALL;
        }
    }
    else
    {
        hPhyDev->reqLinkCaps = CpswPhy_getManualCaps(linkConfig->speed,
                                                     linkConfig->duplexity);
        manualMode = true;
    }

    hPhyDev->state.enableLoopback = phyConfig->enableLoopback;

    CpswPhy_setTimeoutConfig(hPhyDev, &phyConfig->timeoutConfig);

    /* Take a shortcut to FOUND state if PHY is already alive, will save some
     * hundreds of msecs.  Take a shorter shortcut to LINK_WAIT state if PHY
     * is strap configured */
    alive = CpswPhy_isPhyAlive(hPhyDev->hMdio, hPhyDev->addr);
    if (alive)
    {
        CpswPhy_initState(hPhyDev);
        if (hPhyDev->phyConfig.isStrapped)
        {
            CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_LINK_WAIT);
        }
        else
        {
            CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_FOUND);
        }

        status = CpswPhy_bindDriver(hPhyDev);
    }
    else
    {
        CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_INIT);
    }

    CPSWTRACE_DEBUG_IF((status == CPSW_SOK), "PHY %u: open\n", hPhyDev->addr);

    /* Show compatible link partner configuration when local PHY is in manual mode */
    if (manualMode)
    {
        CpswPhy_showLinkPartnerCompat(hPhyDev,
                                      linkConfig->speed,
                                      linkConfig->duplexity);
    }

    return (status == CPSW_SOK) ? hPhyDev : NULL;
}

void CpswPhy_close(CpswPhy_Handle hPhyDev)
{
    /* Power down and isolate the PHY if not strapped */
    if (!hPhyDev->phyConfig.isStrapped)
    {
        CPSWTRACE_DEBUG("PHY %u: disable\n", hPhyDev->addr);
        CpswPhy_rmwReg(hPhyDev, PHY_BMCR,
                       BMCR_ISOLATE | BMCR_PWRDOWN,
                       BMCR_ISOLATE | BMCR_PWRDOWN);
    }
}

int32_t CpswPhy_ioctl(CpswPhy_Handle hPhyDev,
                      uint32_t cmd,
                      Cpsw_IoctlPrms *prms)
{
    int32_t status = CPSW_SOK;

    switch (cmd)
    {
        case CPSW_PHY_IOCTL_PHY_TICK:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(CpswPhy_GenericInArgs),
                                         sizeof(uint32_t));
            if (status == CPSW_SOK)
            {
                CpswPhy_GenericInArgs *inArgs = (CpswPhy_GenericInArgs *)prms->inArgs;
                uint32_t *mdioStatus          = (uint32_t *)prms->outArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                CpswPhy_tick(hPhyDev, mdioStatus);
            }
        }
        break;

        case CPSW_PHY_IOCTL_PHY_GET_ID:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(CpswPhy_GenericInArgs),
                                         sizeof(CpswPhy_Version));
            if (status == CPSW_SOK)
            {
                CpswPhy_GenericInArgs *inArgs = (CpswPhy_GenericInArgs *)prms->inArgs;
                CpswPhy_Version *version      = (CpswPhy_Version *)prms->outArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                status = CpswPhy_getVersion(hPhyDev, version);
            }
        }
        break;

        case CPSW_PHY_IOCTL_PHY_GET_SUPPORTED_MODES:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(CpswPhy_GenericInArgs),
                                         sizeof(uint32_t));
            if (status == CPSW_SOK)
            {
                CpswPhy_GenericInArgs *inArgs = (CpswPhy_GenericInArgs *)prms->inArgs;
                uint32_t *modes               = (uint32_t *)prms->outArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                *modes = CpswPhy_getLocalCaps(hPhyDev);
            }
        }
        break;

        case CPSW_PHY_IOCTL_PHY_GET_LOOPBACK_STATE:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(CpswPhy_GenericInArgs),
                                         sizeof(bool));
            if (status == CPSW_SOK)
            {
                CpswPhy_GenericInArgs *inArgs = (CpswPhy_GenericInArgs *)prms->inArgs;
                bool *loopback                = (bool *)prms->outArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                /* Determine if the LOOPBACK state has been reached */
                *loopback = (hPhyDev->state.fsmState == CPSW_PHY_FSM_STATE_LOOPBACK);
            }
        }
        break;

        case CPSW_PHY_IOCTL_PHY_GET_LINK_STATE:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(CpswPhy_GenericInArgs),
                                         sizeof(bool));
            if (status == CPSW_SOK)
            {
                CpswPhy_GenericInArgs *inArgs = (CpswPhy_GenericInArgs *)prms->inArgs;
                CpswPhy_FsmState linkupFsmState;
                bool *linked = (bool *)prms->outArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                /* Target state for LINKUP/LINKDOWN events can be LOOPBACK or LINKED */
                linkupFsmState = hPhyDev->state.enableLoopback ?
                    CPSW_PHY_FSM_STATE_LOOPBACK : CPSW_PHY_FSM_STATE_LINKED;

                /* Determine if the target state has been reached */
                *linked = (hPhyDev->state.fsmState == linkupFsmState);
            }
        }
        break;

        case CPSW_PHY_IOCTL_PHY_GET_LINK_MODE:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(CpswPhy_GenericInArgs),
                                         sizeof(CpswMacPort_LinkConfig));
            if (status == CPSW_SOK)
            {
                CpswPhy_GenericInArgs *inArgs   = (CpswPhy_GenericInArgs *)prms->inArgs;
                CpswMacPort_LinkConfig *outArgs = (CpswMacPort_LinkConfig *)prms->outArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                status = CpswPhy_getLinkMode(hPhyDev,
                                             &outArgs->speed,
                                             &outArgs->duplexity);
            }
        }
        break;

        case CPSW_PHY_IOCTL_PHY_SET_MODE:
        {
            status = Cpsw_checkInArgs(prms,
                                      sizeof(CpswPhy_SetModeInArgs));
            if (status == CPSW_SOK)
            {
                CpswPhy_SetModeInArgs *inArgs = (CpswPhy_SetModeInArgs *)prms->inArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                if (!hPhyDev->phyConfig.isStrapped)
                {
                    CpswPhy_setPhyMode(hPhyDev, inArgs->mode, inArgs->loopbackEn);
                }
                else
                {
                    status = CPSW_EPERM;
                }
            }
        }
        break;

        case CPSW_PHY_IOCTL_PHY_RESET:
        {
            status = Cpsw_checkInArgs(prms,
                                      sizeof(CpswPhy_GenericInArgs));
            if (status == CPSW_SOK)
            {
                CpswPhy_GenericInArgs *inArgs = (CpswPhy_GenericInArgs *)prms->inArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                CpswPhy_resetPhy(hPhyDev);
            }
        }
        break;

        case CPSW_PHY_IOCTL_RUN_COMPLIANCE_TEST:
        {
            status = Cpsw_checkInArgs(prms,
                                      sizeof(CpswPhy_ComplianceTestInArgs));
            if (status == CPSW_SOK)
            {
                CpswPhy_ComplianceTestInArgs *inArgs = (CpswPhy_ComplianceTestInArgs *)prms->inArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                status = CpswPhy_runComplianceTest(hPhyDev, inArgs->testMode);
            }
        }
        break;

        case CPSW_PHY_IOCTL_PHY_IS_ALIVE:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(CpswPhy_GenericInArgs),
                                         sizeof(bool));
            if (status == CPSW_SOK)
            {
                CpswPhy_GenericInArgs *inArgs = (CpswPhy_GenericInArgs *)prms->inArgs;
                bool *alive                   = (bool *)prms->outArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                *alive = CpswPhy_isPhyAlive(hPhyDev->hMdio, hPhyDev->addr);
            }
        }
        break;

        case CPSW_PHY_IOCTL_PHY_IS_LINKED:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(CpswPhy_GenericInArgs),
                                         sizeof(bool));
            if (status == CPSW_SOK)
            {
                CpswPhy_GenericInArgs *inArgs = (CpswPhy_GenericInArgs *)prms->inArgs;
                bool *linked                  = (bool *)prms->outArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                *linked = CpswPhy_isPhyLinked(hPhyDev->hMdio, hPhyDev->addr);
            }
        }
        break;

        case CPSW_PHY_IOCTL_READ_REG:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(CpswPhy_ReadRegInArgs),
                                         sizeof(uint16_t));
            if (status == CPSW_SOK)
            {
                CpswPhy_ReadRegInArgs *inArgs = (CpswPhy_ReadRegInArgs *)prms->inArgs;
                uint16_t *val                 = (uint16_t *)prms->outArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                status = CpswPhy_readReg(hPhyDev, inArgs->reg, val);
            }
        }
        break;

        case CPSW_PHY_IOCTL_WRITE_REG:
        {
            status = Cpsw_checkInArgs(prms,
                                      sizeof(CpswPhy_WriteRegInArgs));
            if (status == CPSW_SOK)
            {
                CpswPhy_WriteRegInArgs *inArgs = (CpswPhy_WriteRegInArgs *)prms->inArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                status = CpswPhy_writeReg(hPhyDev, inArgs->reg, inArgs->val);
            }
        }
        break;

        case CPSW_PHY_IOCTL_READ_REG_EXT:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(CpswPhy_ReadRegInArgs),
                                         sizeof(uint16_t));
            if (status == CPSW_SOK)
            {
                CpswPhy_ReadRegInArgs *inArgs = (CpswPhy_ReadRegInArgs *)prms->inArgs;
                uint16_t *val                 = (uint16_t *)prms->outArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                status = CpswPhy_readExtReg(hPhyDev, inArgs->reg, val);
            }
        }
        break;

        case CPSW_PHY_IOCTL_WRITE_REG_EXT:
        {
            status = Cpsw_checkInArgs(prms,
                                      sizeof(CpswPhy_WriteRegInArgs));
            if (status == CPSW_SOK)
            {
                CpswPhy_WriteRegInArgs *inArgs = (CpswPhy_WriteRegInArgs *)prms->inArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                status = CpswPhy_writeExtReg(hPhyDev, inArgs->reg, inArgs->val);
            }
        }
        break;

        case CPSW_PHY_IOCTL_READ_CLAUSE45_REG:
        {
            status = Cpsw_checkInOutArgs(prms,
                                         sizeof(CpswPhy_Clause45ReadRegInArgs),
                                         sizeof(uint16_t));
            if (status == CPSW_SOK)
            {
                CpswPhy_Clause45ReadRegInArgs *inArgs = (CpswPhy_Clause45ReadRegInArgs *)prms->inArgs;
                uint16_t *val = (uint16_t *)prms->outArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                status = CpswPhy_readC45Reg(hPhyDev, inArgs->mmd, inArgs->reg, val);
            }
        }
        break;

        case CPSW_PHY_IOCTL_WRITE_CLAUSE45_REG:
        {
            status = Cpsw_checkInArgs(prms,
                                      sizeof(CpswPhy_Clause45WriteRegInArgs));
            if (status == CPSW_SOK)
            {
                CpswPhy_Clause45WriteRegInArgs *inArgs = (CpswPhy_Clause45WriteRegInArgs *)prms->inArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                status = CpswPhy_writeC45Reg(hPhyDev, inArgs->mmd, inArgs->reg, inArgs->val);
            }
        }
        break;

        case CPSW_PHY_IOCTL_PRINT_REGS:
        {
            status = Cpsw_checkInArgs(prms,
                                      sizeof(CpswPhy_GenericInArgs));
            if (status == CPSW_SOK)
            {
                CpswPhy_GenericInArgs *inArgs = (CpswPhy_GenericInArgs *)prms->inArgs;

                Cpsw_assert(hPhyDev->portNum == inArgs->portNum);

                CpswPhy_printRegs(hPhyDev);
            }
        }
        break;

        default:
        {
            status = CPSW_EINVALID_PARAMS;
        }
        break;
    }

    return status;
}

#if (CPSW_TRACE_CFG_TRACE_LEVEL >= CPSW_TRACE_CFG_LEVEL_WARN)
static const char *CpswPhy_getCapsString(uint32_t linkCaps)
{
    snprintf(gCpswPhyCapsBuf, CPSW_PHY_CAPS_BUF_LEN, "%s%s%s%s%s%s%s",
             ((linkCaps & CPSW_LINK_CAP_FD1000) != 0U) ? "FD1000 " : "",
             ((linkCaps & CPSW_LINK_CAP_HD1000) != 0U) ? "HD1000 " : "",
             ((linkCaps & CPSW_LINK_CAP_FD100) != 0U) ? "FD100 " : "",
             ((linkCaps & CPSW_LINK_CAP_HD100) != 0U) ? "HD100 " : "",
             ((linkCaps & CPSW_LINK_CAP_FD10) != 0U) ? "FD10 " : "",
             ((linkCaps & CPSW_LINK_CAP_HD10) != 0U) ? "HD10 " : "",
             ((linkCaps & CPSW_LINK_CAP_ALL) == 0U) ? "None" : "");

    return gCpswPhyCapsBuf;
}

#endif

#if (CPSW_TRACE_CFG_TRACE_LEVEL >= CPSW_TRACE_CFG_LEVEL_DEBUG)
static const char *CpswPhy_getModeString(Cpsw_Speed speed,
                                         Cpsw_Duplexity duplexity)
{
    snprintf(gCpswPhyModeBuf, CPSW_PHY_MODE_BUF_LEN, "%sbps %s-duplex",
             (speed == CPSW_SPEED_1GBIT) ? "1 G" :
             (speed == CPSW_SPEED_100MBIT) ? "100 M" : "10 M",
             (duplexity == CPSW_DUPLEX_FULL) ? "full" : "half");

    return gCpswPhyModeBuf;
}

#endif

static void CpswPhy_showLinkPartnerCompat(CpswPhy_Handle hPhyDev,
                                          Cpsw_Speed speed,
                                          Cpsw_Duplexity duplexity)
{
    switch (speed)
    {
        case CPSW_SPEED_10MBIT:
        case CPSW_SPEED_100MBIT:
            if (duplexity == CPSW_DUPLEX_HALF)
            {
                CPSWTRACE_DEBUG("PHY %u: recommended link partner config: "
                                "manual mode at %s or auto-negotiation\n",
                                hPhyDev->addr,
                                CpswPhy_getModeString(speed, duplexity));
            }
            else
            {
                CPSWTRACE_DEBUG("PHY %u: recommended link partner config: "
                                "manual mode at %s\n",
                                hPhyDev->addr,
                                CpswPhy_getModeString(speed, duplexity));
            }

            break;

        case CPSW_SPEED_1GBIT:
            CPSWTRACE_DEBUG("PHY %u: recommended link partner config: "
                            "manual mode at %s\n",
                            hPhyDev->addr,
                            CpswPhy_getModeString(speed, duplexity));
            break;

        default:
            break;
    }
}

static uint32_t CpswPhy_getManualCaps(Cpsw_Speed speed,
                                      Cpsw_Duplexity duplexity)
{
    uint32_t caps = 0U;

    Cpsw_assert((speed != CPSW_SPEED_AUTO) &&
                (duplexity != CPSW_DUPLEX_AUTO));

    switch (speed)
    {
        case CPSW_SPEED_10MBIT:
            caps = (duplexity == CPSW_DUPLEX_FULL) ?
                   CPSW_LINK_CAP_FD10 :
                   CPSW_LINK_CAP_HD10;
            break;

        case CPSW_SPEED_100MBIT:
            caps = (duplexity == CPSW_DUPLEX_FULL) ?
                   CPSW_LINK_CAP_FD100 :
                   CPSW_LINK_CAP_HD100;
            break;

        case CPSW_SPEED_1GBIT:
            caps = (duplexity == CPSW_DUPLEX_FULL) ?
                   CPSW_LINK_CAP_FD1000 :
                   CPSW_LINK_CAP_HD1000;
            break;

        default:
            break;
    }

    return caps;
}

static uint32_t CpswPhy_findBestCap(uint32_t caps)
{
    uint32_t i;

    for (i = CPSW_LINK_CAP_FD1000; i >= CPSW_LINK_CAP_HD10; i >>= 1U)
    {
        if ((caps & i) != 0U)
        {
            break;
        }
    }

    return i;
}

static void CpswPhy_capToMode(uint32_t caps,
                              Cpsw_Speed *speed,
                              Cpsw_Duplexity *duplexity)
{
    if (caps & CPSW_LINK_CAP_FD1000)
    {
        *speed     = CPSW_SPEED_1GBIT;
        *duplexity = CPSW_DUPLEX_FULL;
    }
    else if (caps & CPSW_LINK_CAP_HD1000)
    {
        *speed     = CPSW_SPEED_1GBIT;
        *duplexity = CPSW_DUPLEX_HALF;
    }
    else if (caps & CPSW_LINK_CAP_FD100)
    {
        *speed     = CPSW_SPEED_100MBIT;
        *duplexity = CPSW_DUPLEX_FULL;
    }
    else if (caps & CPSW_LINK_CAP_HD100)
    {
        *speed     = CPSW_SPEED_100MBIT;
        *duplexity = CPSW_DUPLEX_HALF;
    }
    else if (caps & CPSW_LINK_CAP_FD10)
    {
        *speed     = CPSW_SPEED_10MBIT;
        *duplexity = CPSW_DUPLEX_FULL;
    }
    else
    {
        *speed     = CPSW_SPEED_10MBIT;
        *duplexity = CPSW_DUPLEX_HALF;
    }
}

CpswPhy_MacMode CpswPhy_ifConfigToMacMode(CpswMacPort_Interface *interface)
{
    CpswPhy_MacMode macMode;

    if ((interface->layerType == CPSW_MAC_LAYER_MII) &&
        (interface->sublayerType == CPSW_MAC_SUBLAYER_REDUCED))
    {
        macMode = CPSW_PHY_MAC_MODE_RMII;
    }
    else if ((interface->layerType == CPSW_MAC_LAYER_GMII) &&
             (interface->sublayerType == CPSW_MAC_SUBLAYER_STANDARD))
    {
        macMode = CPSW_PHY_MAC_MODE_GMII;
    }
    else if ((interface->layerType == CPSW_MAC_LAYER_GMII) &&
             (interface->sublayerType == CPSW_MAC_SUBLAYER_REDUCED))
    {
        macMode = CPSW_PHY_MAC_MODE_RGMII;
    }
    else if ((interface->layerType == CPSW_MAC_LAYER_GMII) &&
             (interface->sublayerType == CPSW_MAC_SUBLAYER_SERIAL))
    {
        macMode = CPSW_PHY_MAC_MODE_SGMII;
    }
    else if ( (interface->layerType == CPSW_MAC_LAYER_GMII) &&
              ((interface->sublayerType == CPSW_MAC_SUBLAYER_QUAD_SERIAL_MAIN) ||
               (interface->sublayerType == CPSW_MAC_SUBLAYER_QUAD_SERIAL_SUB)) )
    {
        macMode = CPSW_PHY_MAC_MODE_QSGMII;
    }
    else
    {
        Cpsw_assert(FALSE);

        /* Compiler "'macMode' may be used uninitialized" warning fix */
        macMode = CPSW_PHY_MAC_MODE_RGMII;
    }

    return macMode;
}

static bool CpswPhy_isPhyAlive(CpswMdio_Handle hMdio,
                               uint8_t phyAddr)
{
    Cpsw_IoctlPrms prms;
    bool isAlive = false;
    int32_t status;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &phyAddr, &isAlive);
    status = CpswMdio_ioctl(hMdio, CPSW_MDIO_IOCTL_GET_ALIVE_STATUS, &prms);
    if (status != CPSW_SOK)
    {
        CPSWTRACE_ERROR("PHY %u: failed to get alive status: %d\n", phyAddr, status);
    }

    return isAlive;
}

static bool CpswPhy_isPhyLinked(CpswMdio_Handle hMdio,
                                uint8_t phyAddr)
{
    Cpsw_IoctlPrms prms;
    bool isLinked = false;
    int32_t status;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &phyAddr, &isLinked);
    status = CpswMdio_ioctl(hMdio, CPSW_MDIO_IOCTL_GET_LINK_STATUS, &prms);
    if (status != CPSW_SOK)
    {
        CPSWTRACE_ERROR("PHY %u: failed to get link status: %d\n", phyAddr, status);
    }

    return isLinked;
}

static void CpswPhy_setTimeoutConfig(CpswPhy_Handle hPhyDev,
                                     const CpswPhy_FsmTimeoutConfig *cfg)
{
    CpswPhy_FsmTimeoutConfig *timeoutCfg = &hPhyDev->timeoutConfig;

    *timeoutCfg = *cfg;

    if (timeoutCfg->findingStateTicks == 0U)
    {
        timeoutCfg->findingStateTicks = CPSW_PHY_TIMEOUT_FINDING;
    }

    if (timeoutCfg->resetWaitStateTicks == 0U)
    {
        timeoutCfg->resetWaitStateTicks = CPSW_PHY_TIMEOUT_RESET_WAIT;
    }

    if (timeoutCfg->resetWaitStateResidenceTicks == 0U)
    {
        timeoutCfg->resetWaitStateResidenceTicks = CPSW_PHY_RESIDENCE_RESET_WAIT;
    }

    if (timeoutCfg->nwayStartStateTicks == 0U)
    {
        timeoutCfg->nwayStartStateTicks = CPSW_PHY_TIMEOUT_NWAY_START;
    }

    if (timeoutCfg->nwayWaitStateTicks == 0U)
    {
        timeoutCfg->nwayWaitStateTicks = CPSW_PHY_TIMEOUT_NWAY_WAIT;
    }

    if (timeoutCfg->linkWaitStateTicks == 0U)
    {
        timeoutCfg->linkWaitStateTicks = CPSW_PHY_TIMEOUT_LINK_WAIT;
    }

    if (timeoutCfg->mdixTicks == 0U)
    {
        timeoutCfg->mdixTicks = CPSW_PHY_TIMEOUT_MDIX;
    }
}

static uint32_t CpswPhy_getVersion(CpswPhy_Handle hPhyDev,
                                   CpswPhy_Version *version)
{
    uint32_t phyAddr = hPhyDev->addr;
    uint16_t id1, id2;
    int32_t status = CPSW_SOK;
    bool alive;

    alive = CpswPhy_isPhyAlive(hPhyDev->hMdio, phyAddr);
    if (alive == false)
    {
        status = CPSW_EFAIL;
    }

    if (status == CPSW_SOK)
    {
        status = CpswPhy_readReg(hPhyDev, PHY_PHYIDR1, &id1);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "PHY %u: failed to read ID1 register: %d\n",
                           phyAddr, status);
    }

    if (status == CPSW_SOK)
    {
        status = CpswPhy_readReg(hPhyDev, PHY_PHYIDR2, &id2);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "PHY %u: failed to read ID2 register: %d\n",
                           phyAddr, status);
    }

    if (status == CPSW_SOK)
    {
        version->oui      = ((uint32_t)id1 << PHYIDR1_OUI_OFFSET);
        version->oui     |= ((id2 & PHYIDR2_OUI_MASK) >> PHYIDR2_OUI_OFFSET);
        version->model    = (id2 & PHYIDR2_VMDL_MASK) >> PHYIDR2_VMDL_OFFSET;
        version->revision = (id2 & PHYIDR2_VREV_MASK) >> PHYIDR2_VREV_OFFSET;
    }

    return status;
}

static int32_t CpswPhy_bindDriver(CpswPhy_Handle hPhyDev)
{
    CpswPhyDrv_Handle hDrv;
    CpswPhy_Version version;
    bool match;
    bool macSupported;
    uint32_t i;
    int32_t status;

    hPhyDev->hDrv = NULL;

    status = CpswPhy_getVersion(hPhyDev, &version);
    if (status == CPSW_SOK)
    {
        for (i = 0U; i < CPSW_UTILS_ARRAYSIZE(gCpswPhyDrvs); i++)
        {
            hDrv = gCpswPhyDrvs[i];

            Cpsw_assert(hDrv != NULL);
            Cpsw_assert(hDrv->isPhyDevSupported != NULL);

            /* Skip if this driver doesn't support detected PHY device */
            match = hDrv->isPhyDevSupported(hPhyDev, &version);
            if (!match)
            {
                CPSWTRACE_DEBUG("PHY %u: OUI:%06x Model:%02x Ver:%02x <-> '%s'\n",
                                hPhyDev->addr,
                                version.oui, version.model, version.revision,
                                hDrv->name);
                continue;
            }

            CPSWTRACE_INFO("PHY %u: OUI:%06x Model:%02x Ver:%02x <-> '%s' : OK\n",
                           hPhyDev->addr,
                           version.oui, version.model, version.revision,
                           hDrv->name);

            /* Skip if this driver doesn't support the MAC mode */
            macSupported = hDrv->isMacModeSupported(hPhyDev, hPhyDev->macMode);
            if (!macSupported)
            {
                CPSWTRACE_WARN("PHY %u: '%s' doesn't support MAC mode %d\n",
                               hPhyDev->addr, hDrv->name, hPhyDev->macMode);
            }

            /* Bind the device <-> driver now */
            hPhyDev->hDrv = hDrv;
            break;
        }
    }

    return (hPhyDev->hDrv != NULL) ? CPSW_SOK : CPSW_EFAIL;
}

static void CpswPhy_resetPhy(CpswPhy_Handle hPhyDev)
{
    CpswPhy_State *state = &hPhyDev->state;

    CPSWTRACE_WARN_IF((hPhyDev->hDrv == NULL),
                      "PHY %u: not bound to a driver, can't be reset\n",
                      hPhyDev->addr);

    /* PHY specific reset */
    if ((hPhyDev->hDrv != NULL) &&
        (hPhyDev->hDrv->reset != NULL))
    {
        hPhyDev->hDrv->reset(hPhyDev);

        if (hPhyDev->phyConfig.isStrapped)
        {
            /* Strapped PHY doesn't need either Auto-Negotiation or manual
             * configuration */
            state->needsManualConfig = false;
            state->needsNwayConfig   = false;
        }
        else
        {
            /* Typically manual/auto-negotiation should be reconfigured after
             * PHY reset.  Certain reset/restart implementations only reset the
             * logic not registers, but it's safer to assume reconfiguration
             * is needed */
            state->needsManualConfig = true;
            state->needsNwayConfig   = true;
        }
    }
}

static int32_t CpswPhy_getLinkMode(CpswPhy_Handle hPhyDev,
                                   Cpsw_Speed *speed,
                                   Cpsw_Duplexity *duplexity)
{
    CpswPhy_State *state = &hPhyDev->state;
    CpswPhy_FsmState linkupFsmState;
    int32_t status = CPSW_SOK;

    /* Target state for LINKUP/LINKDOWN events can be LOOPBACK or LINKED */
    linkupFsmState = state->enableLoopback ?
        CPSW_PHY_FSM_STATE_LOOPBACK : CPSW_PHY_FSM_STATE_LINKED;

    if (state->fsmState == linkupFsmState)
    {
        *speed = state->speed;
        *duplexity = state->duplexity;
    }
    else
    {
        CPSWTRACE_WARN("PHY %u: PHY is not linked, can't get link mode",
                       hPhyDev->addr);
        status = CPSW_EPERM;
    }

    return status;
}

static void CpswPhy_setPhyMode(CpswPhy_Handle hPhyDev,
                               uint32_t nwayCaps,
                               bool enableLoopback)
{
    CpswPhy_State *state = &hPhyDev->state;

    state->linkCaps = nwayCaps;
    if (state->linkCaps == 0U)
    {
        state->linkCaps = CPSW_LINK_CAP_ALL;
    }

    state->enableLoopback = enableLoopback;

    CPSWTRACE_DEBUG("PHY %u: NWAY capabilities: %s\n",
                    hPhyDev->addr, CpswPhy_getCapsString(nwayCaps));
    CPSWTRACE_DEBUG("PHY %u: loopback: %s\n",
                    hPhyDev->addr, enableLoopback ? "enabled" : "disabled");

    switch (state->fsmState)
    {
        case CPSW_PHY_FSM_STATE_NWAY_START:
        case CPSW_PHY_FSM_STATE_NWAY_WAIT:
        case CPSW_PHY_FSM_STATE_LINK_WAIT:
        case CPSW_PHY_FSM_STATE_LINKED:
        case CPSW_PHY_FSM_STATE_LOOPBACK:
            state->needsManualConfig = true;
            state->needsNwayConfig   = true;
            CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_FOUND);
            break;

        default:
            break;
    }
}

/* TODO: Tick function should not return true/false (currently returning true if
 * state has changed) */
static int32_t CpswPhy_tick(CpswPhy_Handle hPhyDev,
                            uint32_t *mdioStatus)
{
    CpswPhy_State *state = &hPhyDev->state;
    CpswPhy_FsmState prevFsmState = state->fsmState;
    CpswPhy_FsmState currFsmState;
    CpswPhy_FsmState linkupFsmState;
    int32_t ret = 0U;

    *mdioStatus = CPSW_PHY_EVENT_NOCHANGE;

    /* TODO: Check if pending state transitions needs to be applied,
     * i.e. from CpswPhy_setPhyMode() */

    /* Act on current state of the PHY */
    switch (state->fsmState)
    {
        case CPSW_PHY_FSM_STATE_INIT:
            CPSWTRACE_VERBOSE("PHY %u: INIT\n", hPhyDev->addr);
            CpswPhy_initState(hPhyDev);
            break;

        case CPSW_PHY_FSM_STATE_FINDING:
            CPSWTRACE_VERBOSE("PHY %u: FINDING (rem ticks %u)\n",
                              hPhyDev->addr, state->timeout);
            CpswPhy_findingState(hPhyDev);
            break;

        case CPSW_PHY_FSM_STATE_FOUND:
            CPSWTRACE_VERBOSE("PHY %u: FOUND\n", hPhyDev->addr);
            CpswPhy_foundState(hPhyDev);
            break;

        case CPSW_PHY_FSM_STATE_RESET_WAIT:
            CPSWTRACE_VERBOSE("PHY %u: RESET_WAIT\n", hPhyDev->addr);
            CpswPhy_resetWaitState(hPhyDev);
            break;

        case CPSW_PHY_FSM_STATE_ENABLE:
            CPSWTRACE_VERBOSE("PHY %u: ENABLE\n", hPhyDev->addr);
            CpswPhy_enableState(hPhyDev);
            break;

        case CPSW_PHY_FSM_STATE_NWAY_START:
            CPSWTRACE_VERBOSE("PHY %u: NWAY_START (rem ticks %u)\n",
                              hPhyDev->addr, state->timeout);
            CpswPhy_nwayStartState(hPhyDev);
            break;

        case CPSW_PHY_FSM_STATE_NWAY_WAIT:
            CPSWTRACE_VERBOSE("PHY %u: NWAY_WAIT (rem ticks %u)\n",
                              hPhyDev->addr, state->timeout);
            CpswPhy_nwayWaitState(hPhyDev);
            break;

        case CPSW_PHY_FSM_STATE_LINK_WAIT:
            CPSWTRACE_VERBOSE("PHY %u: LINK_WAIT (rem ticks %u)\n",
                              hPhyDev->addr, state->timeout);
            CpswPhy_linkWaitState(hPhyDev);
            break;

        case CPSW_PHY_FSM_STATE_LINKED:
            CPSWTRACE_VERBOSE("PHY %u: LINKED\n", hPhyDev->addr);
            CpswPhy_linkedState(hPhyDev);
            break;

        case CPSW_PHY_FSM_STATE_LOOPBACK:
            CPSWTRACE_VERBOSE("PHY %u: LOOPBACK\n", hPhyDev->addr);
            CpswPhy_loopbackState(hPhyDev);
            break;

        default:
            CPSWTRACE_VERBOSE("PHY %u: DEFAULT\n", hPhyDev->addr);
            CpswPhy_defaultState(hPhyDev);
            break;
    }

    /* Check if MDI/MDIX mode switch is needed */
    if (state->needsMdixSwitch)
    {
        /* TODO: Confirm we dont do into infinite loop */
        uint32_t mdix;

        state->needsMdixSwitch = false;

        mdix = state->enableMdix ? 1U : 0U;

        /* REVIEW: This is an indication to the application that MDI/MDIX has changed */
        ret = (int32_t)(CPSW_PHY_MIIMDIO_MDIXFLIP | mdix);
    }
    else
    {
        /* Return state change to user */

        /* CHECK: report link events as CPSW_PHY_EVENT_NOCHANGE,
         * CPSW_PHY_EVENT_LINKDOWN, CPSW_PHY_EVENT_PHYERROR,
         * CPSW_PHY_EVENT_LINKUP.
         * Currently ENETPHY_CHNG_MASK set for any state transition */
        if (state->fsmStateChanged)
        {
            currFsmState = state->fsmState;

            /* Target state for LINKUP/LINKDOWN events can be LOOPBACK or LINKED */
            linkupFsmState = state->enableLoopback ?
                CPSW_PHY_FSM_STATE_LOOPBACK : CPSW_PHY_FSM_STATE_LINKED;

            /* LINKUP if FSM entered target state */
            if (currFsmState == linkupFsmState)
            {
                *mdioStatus = CPSW_PHY_EVENT_LINKUP;
            }

            /* LINKDOWN if FSM left target state */
            if (prevFsmState == linkupFsmState)
            {
                *mdioStatus = CPSW_PHY_EVENT_LINKDOWN;
            }

            state->fsmStateChanged = false;
            ret = (int32_t)TRUE;
        }
        else
        {
            ret = (int32_t)FALSE;
        }
    }

    return ret;
}

/* TODO: Remove if not needed */
void CpswPhy_linkChange(CpswPhy_Handle hPhyDev)
{
    CpswPhy_State *state = &hPhyDev->state;
    uint16_t status      = 0U;

    if (state->fsmState == CPSW_PHY_FSM_STATE_LINKED)
    {
        CpswPhy_readReg(hPhyDev, PHY_BMSR, &status);
        if ((status & BMSR_LINKSTS) == 0U)
        {
            if (state->enableNway)
            {
                CpswPhy_rmwReg(hPhyDev, PHY_BMCR,
                               BMCR_ANEN | BMCR_ANRESTART,
                               BMCR_ANEN | BMCR_ANRESTART);

                CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_NWAY_START);
            }
            else
            {
                CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_LINK_WAIT);
            }
        }
    }
}

static void CpswPhy_setNextState(CpswPhy_Handle hPhyDev,
                                 CpswPhy_FsmState state)
{
    CpswPhy_FsmTimeoutConfig *timeoutCfg = &hPhyDev->timeoutConfig;
    CpswPhy_FsmState prevFsmState = hPhyDev->state.fsmState;
    uint32_t timeout;
    uint32_t residenceTime = 0U;

    switch (state)
    {
        case CPSW_PHY_FSM_STATE_INIT:
            timeout = 0U;
            break;

        case CPSW_PHY_FSM_STATE_FINDING:
            timeout = timeoutCfg->findingStateTicks;
            break;

        case CPSW_PHY_FSM_STATE_FOUND:
            timeout = 0U;
            break;

        case CPSW_PHY_FSM_STATE_RESET_WAIT:
            timeout       = timeoutCfg->resetWaitStateTicks;
            residenceTime = timeoutCfg->resetWaitStateResidenceTicks;
            break;

        case CPSW_PHY_FSM_STATE_ENABLE:
            timeout = 0U;
            break;

        case CPSW_PHY_FSM_STATE_NWAY_START:
            timeout = timeoutCfg->nwayStartStateTicks;
            break;

        case CPSW_PHY_FSM_STATE_NWAY_WAIT:
            timeout = timeoutCfg->nwayWaitStateTicks;
            break;

        case CPSW_PHY_FSM_STATE_LINK_WAIT:
            timeout = timeoutCfg->linkWaitStateTicks;
            break;

        case CPSW_PHY_FSM_STATE_LINKED:
            timeout = 0U;
            break;

        case CPSW_PHY_FSM_STATE_LOOPBACK:
            timeout = 0U;
            break;

        default:
            timeout = 0U;
            break;
    }

    CPSWTRACE_DEBUG("PHY %u: %s -> %s (%u ticks)\n",
                    hPhyDev->addr,
                    gCpswPhyStateNames[hPhyDev->state.fsmState],
                    gCpswPhyStateNames[state],
                    timeout);

    hPhyDev->state.fsmState        = state;
    hPhyDev->state.timeout         = timeout;
    hPhyDev->state.residenceTime   = residenceTime;
    hPhyDev->state.fsmStateChanged = (state != prevFsmState);
}

static bool CpswPhy_isNwayCapable(CpswPhy_Handle hPhyDev)
{
    uint16_t val = 0U;

    /* Get the PHY Status */
    CpswPhy_readReg(hPhyDev, PHY_BMSR, &val);

    return((val & BMSR_ANCAPABLE) != 0U);
}

static uint32_t CpswPhy_getLocalCaps(CpswPhy_Handle hPhyDev)
{
    uint32_t caps = 0U;
    uint16_t val  = 0U;

    /* Get 10/100 Mbps capabilities */
    CpswPhy_readReg(hPhyDev, PHY_BMSR, &val);
    if ((val & BMSR_100FD) != 0U)
    {
        caps |= CPSW_LINK_CAP_FD100;
    }

    if ((val & BMSR_100HD) != 0U)
    {
        caps |= CPSW_LINK_CAP_HD100;
    }

    if ((val & BMSR_10FD) != 0U)
    {
        caps |= CPSW_LINK_CAP_FD10;
    }

    if ((val & BMSR_10HD) != 0U)
    {
        caps |= CPSW_LINK_CAP_HD10;
    }

    /* Get extended (1 Gbps) capabilities if supported */
    if ((val & BMSR_GIGEXTSTS) != 0U)
    {
        CpswPhy_readReg(hPhyDev, PHY_GIGESR, &val);

        if ((val & GIGESR_1000FD) != 0U)
        {
            caps |= CPSW_LINK_CAP_FD1000;
        }

        if ((val & GIGESR_1000HD) != 0U)
        {
            caps |= CPSW_LINK_CAP_HD1000;
        }
    }

    return caps;
}

static uint32_t CpswPhy_findCommonCaps(CpswPhy_Handle hPhyDev)
{
    uint32_t localCaps   = 0U;
    uint32_t partnerCaps = 0U;
    uint16_t val         = 0U;

    /* Get local device capabilities */
    CpswPhy_readReg(hPhyDev, PHY_ANAR, &val);

    if ((val & ANAR_100FD) != 0U)
    {
        localCaps |= CPSW_LINK_CAP_FD100;
    }

    if ((val & ANAR_100HD) != 0U)
    {
        localCaps |= CPSW_LINK_CAP_HD100;
    }

    if ((val & ANAR_10FD) != 0U)
    {
        localCaps |= CPSW_LINK_CAP_FD10;
    }

    if ((val & ANAR_10HD) != 0U)
    {
        localCaps |= CPSW_LINK_CAP_HD10;
    }

    /* Get link partner capabilities */
    val = 0U;
    CpswPhy_readReg(hPhyDev, PHY_ANLPAR, &val);

    if ((val & ANLPAR_100FD) != 0U)
    {
        partnerCaps |= CPSW_LINK_CAP_FD100;
    }

    if ((val & ANLPAR_100HD) != 0U)
    {
        partnerCaps |= CPSW_LINK_CAP_HD100;
    }

    if ((val & ANLPAR_10FD) != 0U)
    {
        partnerCaps |= CPSW_LINK_CAP_FD10;
    }

    if ((val & ANLPAR_10HD) != 0U)
    {
        partnerCaps |= CPSW_LINK_CAP_HD10;
    }

    CPSWTRACE_DEBUG("PHY %u: local caps: %s\n",
                    hPhyDev->addr, CpswPhy_getCapsString(localCaps));
    CPSWTRACE_DEBUG("PHY %u: partner caps: %s\n",
                    hPhyDev->addr, CpswPhy_getCapsString(partnerCaps));
    CPSWTRACE_DEBUG("PHY %u: common caps: %s\n",
                    hPhyDev->addr, CpswPhy_getCapsString(localCaps & partnerCaps));

    return(localCaps & partnerCaps);
}

static uint32_t CpswPhy_findCommon1000Caps(CpswPhy_Handle hPhyDev)
{
    uint32_t localCaps   = 0U;
    uint32_t partnerCaps = 0U;
    uint16_t val;

    /* Get local device capabilities */
    CpswPhy_readReg(hPhyDev, PHY_GIGCR, &val);

    if ((val & GIGCR_1000FD) != 0U)
    {
        localCaps |= CPSW_LINK_CAP_FD1000;
    }

    if ((val & GIGCR_1000HD) != 0U)
    {
        localCaps |= CPSW_LINK_CAP_HD1000;
    }

    /* Get link partner capabilities */
    CpswPhy_readReg(hPhyDev, PHY_GIGSR, &val);

    if ((val & GIGSR_1000FD) != 0U)
    {
        partnerCaps |= CPSW_LINK_CAP_FD1000;
    }

    if ((val & GIGSR_1000FD) != 0U)
    {
        partnerCaps |= CPSW_LINK_CAP_HD1000;
    }

    CPSWTRACE_DEBUG("PHY %u: local caps: %s\n",
                    hPhyDev->addr, CpswPhy_getCapsString(localCaps));
    CPSWTRACE_DEBUG("PHY %u: partner caps: %s\n",
                    hPhyDev->addr, CpswPhy_getCapsString(partnerCaps));
    CPSWTRACE_DEBUG("PHY %u: common caps: %s\n",
                    hPhyDev->addr, CpswPhy_getCapsString(localCaps & partnerCaps));

    return(localCaps & partnerCaps);
}

static uint32_t CpswPhy_findCommonNwayCaps(CpswPhy_Handle hPhyDev)
{
    CpswPhy_State *state = &hPhyDev->state;
    uint32_t nwayCaps;

    nwayCaps = CpswPhy_findCommonCaps(hPhyDev);

    /* Find common gigabit capabilities if supported by PHY */
    if ((state->phyLinkCaps & CPSW_LINK_CAP_1000) != 0U)
    {
        nwayCaps |= CpswPhy_findCommon1000Caps(hPhyDev);
    }

    CPSWTRACE_DEBUG("PHY %u: common caps: %s\n",
                    hPhyDev->addr, CpswPhy_getCapsString(nwayCaps));

    if (nwayCaps != 0U)
    {
        /* Find the highest performance protocol from the common
         * capabilities of local device and link partner */
        nwayCaps = CpswPhy_findBestCap(nwayCaps);
    }
    else
    {
        /* Non-compiant PHY? Flag the negotiation error */
        CPSWTRACE_ERROR("PHY %u: no common caps found\n", hPhyDev->addr);
    }

    return nwayCaps;
}

static void CpswPhy_phyTimeOut(CpswPhy_Handle hPhyDev)
{
    CpswPhy_State *state = &hPhyDev->state;

    CPSWTRACE_DEBUG("PHY %u: timeout has occurred\n", hPhyDev->addr);

    /* TODO: Confirm this does not go into infinite loop of MDIX/MDI swap */
    if (CpswPhy_mdixSupported(hPhyDev))
    {
        /* Indicate MDI/MDIX mode switch is needed */
        state->needsMdixSwitch = true;

        /* Toggle the MDIX mode indicator */
        state->enableMdix = !state->enableMdix;
    }

    /* Reset state machine to FOUND for non-strapped PHY, or to LINK_WAIT for
     * state for strapped PHY as most FSM states are bypassed */
    if (hPhyDev->phyConfig.isStrapped)
    {
        CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_LINK_WAIT);
    }
    else
    {
        CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_FOUND);
    }
}

static void CpswPhy_initState(CpswPhy_Handle hPhyDev)
{
    CpswPhy_State *state = &hPhyDev->state;

    state->speed       = CPSW_SPEED_10MBIT;
    state->duplexity   = CPSW_DUPLEX_HALF;
    state->phyLinkCaps = 0U;

    if (hPhyDev->phyConfig.isStrapped)
    {
        state->needsManualConfig = false;
        state->needsNwayConfig   = false;
    }
    else
    {
        state->needsManualConfig = true;
        state->needsNwayConfig   = true;
    }

    CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_FINDING);
}

static void CpswPhy_defaultState(CpswPhy_Handle hPhyDev)
{
    /* Awaiting a CpswPhy_init call */
    hPhyDev->state.fsmStateChanged = true;
}

static void CpswPhy_findingState(CpswPhy_Handle hPhyDev)
{
    CpswPhy_State *state = &hPhyDev->state;
    uint32_t phyAddr     = hPhyDev->addr;
    int32_t status;
    bool alive;

    alive = CpswPhy_isPhyAlive(hPhyDev->hMdio, phyAddr);
    if (alive)
    {
        /* PHY found */
        CPSWTRACE_DEBUG("PHY %u: PHY has been found\n", phyAddr);

        status = CpswPhy_bindDriver(hPhyDev);
        if (status == CPSW_SOK)
        {
            if (hPhyDev->phyConfig.isStrapped)
            {
                CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_LINK_WAIT);
            }
            else
            {
                CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_FOUND);
            }
        }
        else
        {
            /* TODO: Go to a 'disable' state */
            CPSWTRACE_ERROR("PHY %u: couldn't be bound to any driver\n",
                            phyAddr);
        }
    }
    else
    {
        if (state->timeout != 0U)
        {
            state->timeout--;
        }
        else
        {
            CPSWTRACE_DEBUG("PHY %u: timed out\n", phyAddr);
            state->timeout = CPSW_PHY_TIMEOUT_FINDING;
        }
    }
}

static void CpswPhy_foundState(CpswPhy_Handle hPhyDev)
{
    Cpsw_assert(!hPhyDev->phyConfig.isStrapped);

    CpswPhy_resetPhy(hPhyDev);

    CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_RESET_WAIT);
}

static void CpswPhy_resetWaitState(CpswPhy_Handle hPhyDev)
{
    CpswPhy_State *state = &hPhyDev->state;
    bool complete;

    Cpsw_assert(hPhyDev->hDrv->isResetComplete != NULL);
    Cpsw_assert(!hPhyDev->phyConfig.isStrapped);

    /* Wait for PHY reset to complete */
    complete = hPhyDev->hDrv->isResetComplete(hPhyDev);
    if (complete)
    {
        if (state->residenceTime != 0U)
        {
            state->residenceTime--;
        }
        else
        {
            CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_ENABLE);
        }
    }
    else
    {
        if (state->timeout != 0U)
        {
            state->timeout--;
        }
        else
        {
            CpswPhy_phyTimeOut(hPhyDev);
        }
    }
}

static void CpswPhy_enableState(CpswPhy_Handle hPhyDev)
{
    CpswPhy_State *state = &hPhyDev->state;
    uint32_t socLinkCaps;
    uint32_t commonLinkCaps;

    Cpsw_assert(!hPhyDev->phyConfig.isStrapped);

    /* Enable PHY (resetting it if applicable) can be done only when
     * entering the ENABLE state for the first time */
    if (state->needsNwayConfig ||
        state->needsManualConfig)
    {
        /* Set PHY in normal mode */
        CPSWTRACE_DEBUG("PHY %u: enable\n", hPhyDev->addr);
        CpswPhy_rmwReg(hPhyDev, PHY_BMCR, BMCR_ISOLATE | BMCR_PWRDOWN, 0U);

        /* PHY-specific 'extended' configuration */
        if ((hPhyDev->hDrv->config != NULL) &&
            !hPhyDev->phyConfig.skipExtendedConfig)
        {
            hPhyDev->hDrv->config(hPhyDev,
                                  &hPhyDev->phyConfig,
                                  hPhyDev->macMode);
        }
    }

    /* Find common capabilities among app's, PHY's and SoC's */
    CPSWTRACE_DEBUG("PHY %u: req caps: %s\n",
                    hPhyDev->addr, CpswPhy_getCapsString(hPhyDev->reqLinkCaps));

    state->phyLinkCaps = CpswPhy_getLocalCaps(hPhyDev);
    CPSWTRACE_DEBUG("PHY %u: PHY caps: %s\n",
                    hPhyDev->addr, CpswPhy_getCapsString(state->phyLinkCaps));

    socLinkCaps = CpswSoc_getMacCaps(hPhyDev->cpswType, hPhyDev->portNum);
    CPSWTRACE_DEBUG("PHY %u: MAC %u caps: %s\n",
                    hPhyDev->addr, hPhyDev->portNum,
                    CpswPhy_getCapsString(socLinkCaps));

    commonLinkCaps = hPhyDev->reqLinkCaps &
                     state->phyLinkCaps &
                     socLinkCaps;
    CPSWTRACE_DEBUG("PHY %u: refined caps: %s\n",
                    hPhyDev->addr, CpswPhy_getCapsString(commonLinkCaps));

    if (commonLinkCaps != 0U)
    {
        state->linkCaps = commonLinkCaps;

        /* Enable NWAY if requested by app (via auto-speed or auto-duplex) and
         * supported by PHY */
        state->isNwayCapable = CpswPhy_isNwayCapable(hPhyDev);
        CPSWTRACE_DEBUG("PHY %u: PHY is %sNWAY-capable\n",
                        hPhyDev->addr, state->isNwayCapable ? "" : "not");

        state->enableNway = (hPhyDev->linkConfig.speed == CPSW_SPEED_AUTO) ||
                            (hPhyDev->linkConfig.duplexity == CPSW_DUPLEX_AUTO);
        if (state->enableNway && !state->isNwayCapable)
        {
            /* If PHY is not capable of auto negotiation, fallback to manual mode
             * with the highest refined capability */
            CPSWTRACE_WARN("PHY %u: falling back to manual mode\n",
                           hPhyDev->addr);
            state->enableNway = false;

            commonLinkCaps = CpswPhy_findBestCap(commonLinkCaps);
            CPSWTRACE_WARN("PHY %u: new link caps: %s\n",
                           hPhyDev->addr, CpswPhy_getCapsString(commonLinkCaps));
            state->linkCaps = commonLinkCaps;
        }
    }

    if (commonLinkCaps == 0U)
    {
        /* TODO: Should it go to a new "disabled" state where state machine doesn't run */
        CPSWTRACE_ERROR("PHY %u: no supported caps found\n",
                        hPhyDev->addr);
    }
    else if (state->enableLoopback)
    {
        CPSWTRACE_DEBUG("PHY %u: setup loopback\n", hPhyDev->addr);
        CpswPhy_setupManual(hPhyDev, true);
        CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_LOOPBACK);
    }
    else if (state->isNwayCapable && state->enableNway)
    {
        CPSWTRACE_DEBUG("PHY %u: setup NWAY\n", hPhyDev->addr);
        CpswPhy_setupNway(hPhyDev);
        CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_NWAY_START);
    }
    else
    {
        CPSWTRACE_DEBUG("PHY %u: manual setup\n", hPhyDev->addr);
        CpswPhy_setupManual(hPhyDev, false);
        CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_LINK_WAIT);
    }

    /* If AutoMdix add delay */
    CpswPhy_mdixDelay(hPhyDev);
}

static void CpswPhy_setupNway(CpswPhy_Handle hPhyDev)
{
    CpswPhy_State *state = &hPhyDev->state;
    uint16_t nwayAdvertise;
    uint16_t nway1000Advertise = 0U;

    CPSWTRACE_DEBUG("PHY %u: NWAY advertising: %s\n",
                    hPhyDev->addr, CpswPhy_getCapsString(state->linkCaps));

    CPSWTRACE_DEBUG("PHY %u: config is %sneeded\n",
                    hPhyDev->addr, state->needsNwayConfig ? "" : "not ");

    if (state->needsNwayConfig)
    {
        nwayAdvertise = ANAR_802P3;
        if ((state->linkCaps & CPSW_LINK_CAP_FD100) != 0U)
        {
            nwayAdvertise |= ANAR_100FD;
        }

        if ((state->linkCaps & CPSW_LINK_CAP_HD100) != 0U)
        {
            nwayAdvertise |= ANAR_100HD;
        }

        if ((state->linkCaps & CPSW_LINK_CAP_FD10) != 0U)
        {
            nwayAdvertise |= ANAR_10FD;
        }

        if ((state->linkCaps & CPSW_LINK_CAP_HD10) != 0U)
        {
            nwayAdvertise |= ANAR_10HD;
        }

        if ((state->linkCaps & CPSW_LINK_CAP_FD1000) != 0U)
        {
            nway1000Advertise |= GIGCR_1000FD;
        }

        if ((state->linkCaps & CPSW_LINK_CAP_HD1000) != 0U)
        {
            nway1000Advertise |= GIGCR_1000HD;
        }

        CpswPhy_rmwReg(hPhyDev, PHY_ANAR, ANAR_100 | ANAR_10, nwayAdvertise);

        if ((state->phyLinkCaps & CPSW_LINK_CAP_1000) != 0U)
        {
            CpswPhy_rmwReg(hPhyDev, PHY_GIGCR, GIGCR_1000, nway1000Advertise);
        }

        state->needsNwayConfig = false;
    }

    /* Restart auto-negotiation */
    CPSWTRACE_DEBUG("PHY %u: restart autonegotiation\n",
                    hPhyDev->addr);
    CpswPhy_rmwReg(hPhyDev, PHY_BMCR, BMCR_ANEN, BMCR_ANEN);

    /* TODO: is MII_ENETPHY_FD needed for auto-negotiation? */
    CpswPhy_rmwReg(hPhyDev, PHY_BMCR,
                   BMCR_ANRESTART | BMCR_FD,
                   BMCR_ANRESTART | BMCR_FD);
}

static void CpswPhy_setupManual(CpswPhy_Handle hPhyDev,
                                bool enableLoopback)
{
    CpswPhy_State *state = &hPhyDev->state;
    uint16_t val;

    /* Manual config is needed if entered the FOUND state for the first time
     * or if an explicit mode change was requested via CpswPhy_setPhyMode() */
    if (state->needsManualConfig)
    {
        /* Disable auto-negotiation if supported by PHY */
        if (state->isNwayCapable)
        {
            CPSWTRACE_DEBUG("PHY %u: disable NWAY\n", hPhyDev->addr);
            CpswPhy_rmwReg(hPhyDev, PHY_BMCR, BMCR_ANEN, 0U);
        }

        switch (state->linkCaps)
        {
            case CPSW_LINK_CAP_FD1000:
                val = BMCR_SPEED1000 | BMCR_FD;
                break;

            case CPSW_LINK_CAP_HD1000:
                val = BMCR_SPEED1000;
                break;

            case CPSW_LINK_CAP_FD100:
                val = BMCR_SPEED100 | BMCR_FD;
                break;

            case CPSW_LINK_CAP_HD100:
                val = BMCR_SPEED100;
                break;

            case CPSW_LINK_CAP_FD10:
                val = BMCR_FD;
                break;

            case CPSW_LINK_CAP_HD10:
            default:
                val = 0U;
                break;
        }

        if (enableLoopback)
        {
            val |= BMCR_LOOPBACK;
        }

        CpswPhy_rmwReg(hPhyDev, PHY_BMCR,
                       BMCR_LOOPBACK | BMCR_SPEED1000 | BMCR_SPEED100 | BMCR_FD,
                       val);

        /* Clear advertised capabilities values if supported */
        if (state->isNwayCapable)
        {
            CpswPhy_rmwReg(hPhyDev, PHY_ANAR, ANAR_100 | ANAR_10, 0U);

            if ((state->phyLinkCaps & CPSW_LINK_CAP_1000) != 0U)
            {
                CpswPhy_rmwReg(hPhyDev, PHY_GIGCR, GIGCR_1000, 0U);
            }
        }

        state->needsManualConfig = false;
    }

    CpswPhy_capToMode(state->linkCaps,
                      &state->speed,
                      &state->duplexity);

    CPSWTRACE_DEBUG("PHY %u: requested mode: %s\n",
                    hPhyDev->addr,
                    CpswPhy_getModeString(state->speed,
                                          state->duplexity));
}

static void CpswPhy_loopbackState(CpswPhy_Handle hPhyDev)
{
    bool loopback = false;
    int32_t status;
    uint16_t val;

    status = CpswPhy_readReg(hPhyDev, PHY_BMCR, &val);
    if (status == CPSW_SOK)
    {
        loopback = ((val & BMCR_LOOPBACK) != 0U);
    }
    else
    {
        CPSWTRACE_ERROR("PHY %u: failed to read BMCR register: %d\n",
                        hPhyDev->addr, status);
    }

    if (!loopback)
    {
        CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_FOUND);
    }
}

static void CpswPhy_nwayStartState(CpswPhy_Handle hPhyDev)
{
    CpswPhy_State *state = &hPhyDev->state;
    uint16_t mode;
    uint16_t val;

    Cpsw_assert(!hPhyDev->phyConfig.isStrapped);

    /* Wait for NWAY to start */
    CpswPhy_readReg(hPhyDev, PHY_BMCR, &mode);

    if ((mode & BMCR_ANRESTART) == 0U)
    {
        /* Flush pending latch bits */
        CpswPhy_readReg(hPhyDev, PHY_BMSR, &val);
        CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_NWAY_WAIT);

        /* If AutoMdix add delay */
        CpswPhy_mdixDelay(hPhyDev);
    }
    else
    {
        if (state->timeout != 0U)
        {
            state->timeout--;
        }
        else
        {
            CpswPhy_phyTimeOut(hPhyDev);
        }
    }
}

static void CpswPhy_nwayWaitState(CpswPhy_Handle hPhyDev)
{
    CpswPhy_State *state = &hPhyDev->state;
    uint32_t nwayCaps;
    uint16_t status;

    Cpsw_assert(!hPhyDev->phyConfig.isStrapped);

    CpswPhy_readReg(hPhyDev, PHY_BMSR, &status);
    if (0U != (status & BMSR_ANCOMPLETE))
    {
        nwayCaps = CpswPhy_findCommonNwayCaps(hPhyDev);
        if (nwayCaps != 0U)
        {
            CpswPhy_capToMode(nwayCaps,
                              &state->speed,
                              &state->duplexity);

            CPSWTRACE_DEBUG("PHY %u: negotiated mode: %s\n",
                            hPhyDev->addr,
                            CpswPhy_getModeString(state->speed,
                                                  state->duplexity));

            if ((status & BMSR_LINKSTS) != 0U)
            {
                CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_LINKED);
            }
            else
            {
                CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_LINK_WAIT);
            }
        }
    }
    else
    {
        if (state->timeout != 0U)
        {
            state->timeout--;
        }
        else
        {
            CpswPhy_phyTimeOut(hPhyDev);
        }
    }
}

static void CpswPhy_linkWaitState(CpswPhy_Handle hPhyDev)
{
    CpswPhy_State *state = &hPhyDev->state;
    uint32_t nwayCaps;
    uint16_t status;

    CpswPhy_readReg(hPhyDev, PHY_BMSR, &status);

    if ((status & BMSR_LINKSTS) != 0U)
    {
        /* Populate FSM state now as most FSM states are bypassed for
         * strapped PHYs */
        if (hPhyDev->phyConfig.isStrapped)
        {
            state->phyLinkCaps = CpswPhy_getLocalCaps(hPhyDev);
            nwayCaps           = CpswPhy_findCommonNwayCaps(hPhyDev);
            if (nwayCaps != 0U)
            {
                CpswPhy_capToMode(nwayCaps,
                                  &state->speed,
                                  &state->duplexity);

                CPSWTRACE_DEBUG("PHY %u: negotiated mode: %s\n",
                                hPhyDev->addr,
                                CpswPhy_getModeString(state->speed,
                                                      state->duplexity));
            }
        }

        CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_LINKED);
    }
    else
    {
        if (state->timeout != 0U)
        {
            state->timeout--;
        }
        else
        {
            CpswPhy_phyTimeOut(hPhyDev);
        }
    }
}

static void CpswPhy_linkedState(CpswPhy_Handle hPhyDev)
{
    bool linked;

    /* Alternatively, BMSR[2] Link Status bit can be checked */
    linked = CpswPhy_isPhyLinked(hPhyDev->hMdio, hPhyDev->addr);

    /* Not linked */
    if (!linked)
    {
        if (hPhyDev->phyConfig.isStrapped)
        {
            CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_LINK_WAIT);
        }
        else
        {
            CpswPhy_setNextState(hPhyDev, CPSW_PHY_FSM_STATE_FOUND);
        }

        /* If AutoMdix add delay */
        CpswPhy_mdixDelay(hPhyDev);
    }
}

static int32_t CpswPhy_runComplianceTest(CpswPhy_Handle hPhyDev,
                                         uint32_t testMode)
{
    int32_t status;

    if (hPhyDev->hDrv != NULL)
    {
        if (hPhyDev->hDrv->runComplianceTest != NULL)
        {
            status = hPhyDev->hDrv->runComplianceTest(hPhyDev, testMode);
        }
        else
        {
            status = CPSW_ENOTSUPPORTED;
        }
    }
    else
    {
        status = CPSW_EPERM;
    }

    return status;
}

int32_t CpswPhy_readReg(CpswPhy_Handle hPhyDev,
                        uint32_t reg,
                        uint16_t *val)
{
    Cpsw_IoctlPrms prms;
    CpswMdio_Clause22ReadRegInArgs inArgs;
    uint32_t phyAddr = hPhyDev->addr;
    int32_t status;

    inArgs.userCh  = CPSW_MDIO_USER_CHANNEL_0; /* TODO: unused */
    inArgs.phyAddr = phyAddr;
    inArgs.reg     = reg;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, val);
    status = CpswMdio_ioctl(hPhyDev->hMdio, CPSW_MDIO_IOCTL_CLAUSE22_READ_REG, &prms);
    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "PHY %u: failed to read reg %u: %d\n",
                       phyAddr, reg, status);
    CPSWTRACE_VERBOSE_IF((status == CPSW_SOK),
                         "PHY %u: reg %u val 0x%04x %s\n",
                         phyAddr, reg, *val);

    return status;
}

int32_t CpswPhy_writeReg(CpswPhy_Handle hPhyDev,
                         uint32_t reg,
                         uint16_t val)
{
    Cpsw_IoctlPrms prms;
    CpswMdio_Clause22WriteRegInArgs inArgs;
    uint32_t phyAddr = hPhyDev->addr;
    uint32_t status;

    inArgs.userCh  = CPSW_MDIO_USER_CHANNEL_0; /* TODO: unused */
    inArgs.phyAddr = phyAddr;
    inArgs.reg     = reg;
    inArgs.val     = val;

    CPSW_IOCTL_SET_IN_ARGS(&prms, &inArgs);
    status = CpswMdio_ioctl(hPhyDev->hMdio, CPSW_MDIO_IOCTL_CLAUSE22_WRITE_REG, &prms);
    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "PHY %u: failed to write reg %u: %d\n",
                       phyAddr, reg, status);
    CPSWTRACE_VERBOSE_IF((status == CPSW_SOK),
                         "PHY %u: reg %u val 0x%04x\n",
                         phyAddr, reg, val);

    return status;
}

int32_t CpswPhy_rmwReg(CpswPhy_Handle hPhyDev,
                       uint32_t reg,
                       uint16_t mask,
                       uint16_t val)
{
    Cpsw_IoctlPrms prms;
    CpswMdio_Clause22ReadRegInArgs readInArgs;
    CpswMdio_Clause22WriteRegInArgs writeInArgs;
    uint32_t phyAddr = hPhyDev->addr;
    uint32_t status;
    uint16_t data = 0U;

    readInArgs.userCh  = CPSW_MDIO_USER_CHANNEL_0; /* TODO: unused */
    readInArgs.phyAddr = phyAddr;
    readInArgs.reg     = reg;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &readInArgs, &data);
    status = CpswMdio_ioctl(hPhyDev->hMdio, CPSW_MDIO_IOCTL_CLAUSE22_READ_REG, &prms);
    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "PHY %u: failed to read reg %u: %d\n",
                       phyAddr, reg, status);
    CPSWTRACE_VERBOSE_IF((status == CPSW_SOK),
                         "PHY %u: read reg %u val 0x%04x\n",
                         phyAddr, reg, data);

    if (status == CPSW_SOK)
    {
        data                = (data & ~mask) | (val & mask);
        writeInArgs.userCh  = CPSW_MDIO_USER_CHANNEL_0; /* TODO: unused */
        writeInArgs.phyAddr = phyAddr;
        writeInArgs.reg     = reg;
        writeInArgs.val     = data;

        CPSW_IOCTL_SET_IN_ARGS(&prms, &writeInArgs);
        status = CpswMdio_ioctl(hPhyDev->hMdio, CPSW_MDIO_IOCTL_CLAUSE22_WRITE_REG, &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "PHY %u: failed to write reg %u: %d\n",
                           phyAddr, reg, status);
        CPSWTRACE_VERBOSE_IF((status == CPSW_SOK),
                             "PHY %u: write reg %u val 0x%04x\n",
                             phyAddr, reg, data);
    }

    return status;
}

static void CpswPhy_printRegs(CpswPhy_Handle hPhyDev)
{
    if ((hPhyDev->hDrv != NULL) &&
        (hPhyDev->hDrv->printRegs != NULL))
    {
        hPhyDev->hDrv->printRegs(hPhyDev);
    }
}

static int32_t CpswPhy_readExtReg(CpswPhy_Handle hPhyDev,
                                  uint32_t reg,
                                  uint16_t *val)
{
    int32_t status = CPSW_ENOTSUPPORTED;

    if ((hPhyDev->hDrv != NULL) &&
        (hPhyDev->hDrv->readExtReg != NULL))
    {
        status = hPhyDev->hDrv->readExtReg(hPhyDev, reg, val);
    }

    return status;
}

static int32_t CpswPhy_writeExtReg(CpswPhy_Handle hPhyDev,
                                   uint32_t reg,
                                   uint16_t val)
{
    int32_t status = CPSW_ENOTSUPPORTED;

    if ((hPhyDev->hDrv != NULL) &&
        (hPhyDev->hDrv->writeExtReg != NULL))
    {
        status = hPhyDev->hDrv->writeExtReg(hPhyDev, reg, val);
    }

    return status;
}

int32_t CpswPhy_readC45Reg(CpswPhy_Handle hPhyDev,
                           CpswMdio_Clause45Mmd mmd,
                           uint32_t reg,
                           uint16_t *val)
{
    Cpsw_IoctlPrms prms;
    CpswMdio_Clause45ReadRegInArgs inArgs;
    uint32_t phyAddr = hPhyDev->addr;
    int32_t status;

    inArgs.userCh = CPSW_MDIO_USER_CHANNEL_0; /* TODO: unused */
    inArgs.phyAddr = phyAddr;
    inArgs.mmd = mmd;
    inArgs.reg = reg;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &inArgs, val);
    status = CpswMdio_ioctl(hPhyDev->hMdio, CPSW_MDIO_IOCTL_CLAUSE45_READ_REG, &prms);
    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "PHY %u: failed to read MMD %u reg %u: %d\n",
                       phyAddr, mmd, reg, status);
    CPSWTRACE_VERBOSE_IF((status == CPSW_SOK),
                         "PHY %u: MMD %u reg %u val 0x%04x %s\n",
                         phyAddr, mmd, reg, *val);

    return status;
}

int32_t CpswPhy_writeC45Reg(CpswPhy_Handle hPhyDev,
                            CpswMdio_Clause45Mmd mmd,
                            uint32_t reg,
                            uint16_t val)
{
    Cpsw_IoctlPrms prms;
    CpswMdio_Clause45WriteRegInArgs inArgs;
    uint32_t phyAddr = hPhyDev->addr;
    uint32_t status;

    inArgs.userCh  = CPSW_MDIO_USER_CHANNEL_0; /* TODO: unused */
    inArgs.phyAddr = phyAddr;
    inArgs.mmd     = mmd;
    inArgs.reg     = reg;
    inArgs.val     = val;

    CPSW_IOCTL_SET_IN_ARGS(&prms, &inArgs);
    status = CpswMdio_ioctl(hPhyDev->hMdio, CPSW_MDIO_IOCTL_CLAUSE45_WRITE_REG, &prms);
    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "PHY %u: failed to write MMD %u reg %u: %d\n",
                       phyAddr, mmd, reg, status);
    CPSWTRACE_VERBOSE_IF((status == CPSW_SOK),
                         "PHY %u: MMD %u reg %u val 0x%04x\n",
                         phyAddr, mmd, reg, val);

    return status;
}

int32_t CpswPhy_rmwC45Reg(CpswPhy_Handle hPhyDev,
                          CpswMdio_Clause45Mmd mmd,
                          uint32_t reg,
                          uint16_t mask,
                          uint16_t val)
{
    Cpsw_IoctlPrms prms;
    CpswMdio_Clause45ReadRegInArgs readInArgs;
    CpswMdio_Clause45WriteRegInArgs writeInArgs;
    uint32_t phyAddr = hPhyDev->addr;
    uint32_t status;
    uint16_t data = 0U;

    readInArgs.userCh  = CPSW_MDIO_USER_CHANNEL_0; /* TODO: unused */
    readInArgs.phyAddr = phyAddr;
    readInArgs.mmd     = mmd;
    readInArgs.reg     = reg;

    CPSW_IOCTL_SET_INOUT_ARGS(&prms, &readInArgs, &data);
    status = CpswMdio_ioctl(hPhyDev->hMdio, CPSW_MDIO_IOCTL_CLAUSE45_READ_REG, &prms);
    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "PHY %u: failed to read MMD %u reg %u: %d\n",
                       phyAddr, mmd, reg, status);
    CPSWTRACE_VERBOSE_IF((status == CPSW_SOK),
                         "PHY %u: read MMD %u reg %u val 0x%04x\n",
                         phyAddr, mmd, reg, data);

    if (status == CPSW_SOK)
    {
        data = (data & ~mask) | (val & mask);
        writeInArgs.userCh  = CPSW_MDIO_USER_CHANNEL_0; /* TODO: unused */
        writeInArgs.phyAddr = phyAddr;
        writeInArgs.mmd     = mmd;
        writeInArgs.reg     = reg;
        writeInArgs.val     = data;

        CPSW_IOCTL_SET_IN_ARGS(&prms, &writeInArgs);
        status = CpswMdio_ioctl(hPhyDev->hMdio, CPSW_MDIO_IOCTL_CLAUSE45_WRITE_REG, &prms);
        CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                           "PHY %u: failed to write MMD %u reg %u: %d\n",
                           phyAddr, mmd, reg, status);
        CPSWTRACE_VERBOSE_IF((status == CPSW_SOK),
                             "PHY %u: write MMD %u reg %u val 0x%04x\n",
                             phyAddr, mmd, reg, data);
    }

    return status;
}

/* ========================================================================== */
/*                          Static Function Definitions                       */
/* ========================================================================== */

#ifdef CPSW_PHY_MDIX_SUPPORT
static uint32_t CpswPhy_getRandom(void)
{
    gCpswPhyRandomSeed = (gCpswPhyRandomSeed * (1103515245U)) + (12345U);

    return((uint32_t)(gCpswPhyRandomSeed / (65536U)) %
           (CPSW_PHY_COMMON_RANDOM_MAX + 1U));
}

static uint32_t CpswPhy_getRandomInRange(uint32_t min,
                                         uint32_t max)
{
    uint32_t tmp;

    tmp  = CpswPhy_getRandom();
    tmp %= ((max - min) + 1U);
    tmp += min;

    return tmp;
}

#endif

/* returns 0 if current PHY has AutoMdix support, otherwise 0 */
static bool CpswPhy_mdixSupported(CpswPhy_Handle hPhyDev)
{
#ifdef CPSW_PHY_MDIX_SUPPORT
    bool supported = true;

    if (hPhyDev->phyConfig.mdixEnable == FALSE)
    {
        /* PHY does not support AutoMdix */
        supported = false;
    }

    return supported;
#else
    return false;
#endif
}

/* If current PHY has AutoMdix support add Mdix delay to the timer state value */
static void CpswPhy_mdixDelay(CpswPhy_Handle hPhyDev)
{
#ifdef CPSW_PHY_MDIX_SUPPORT
    uint32_t delay;

    if (CpswPhy_mdixSupported(hPhyDev))
    {
        /* Get the delay value in milli-seconds and convert to ten-milli second value */
        delay = CpswPhy_getRandomInRange(CPSW_PHY_AUTOMDIX_DELAY_MIN,
                                         CPSW_PHY_AUTOMDIX_DELAY_MAX);
        delay /= 10U;

        /* Add auto-MDIX link delay and random switch delay */
        hPhyDev->state.timeout = CPSW_PHY_TIMEOUT_MDIX + delay;

        /* TODO How do we ensure current state is MDIX */
    }
#endif
}
