/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 *  \file cpsw_queue.c
 *
 *  \brief File containing the Queue utility functions.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/drv/cpsw/include/cpsw_queue.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/src/priv/cpsw_osal.h>
#include <ti/drv/cpsw/src/priv/cpsw_priv.h>

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define CPSW_Q_INIT_DONE (0xABCDABCD)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

void CpswUtils_initQ(queue_t *pQueue)
{
    pQueue->count = 0;
    pQueue->pHead = NULL;
    pQueue->pTail = NULL;
    pQueue->magic = CPSW_Q_INIT_DONE;
}

void CpswUtils_copyQ(queue_t *pDstQueue,
                     const queue_t *pSrcQueue)
{
    uintptr_t key = CpswOsal_disableAllIntr();

#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(pSrcQueue->magic == CPSW_Q_INIT_DONE);
#endif

    pDstQueue->magic = CPSW_Q_INIT_DONE;
    pDstQueue->pHead = pSrcQueue->pHead;
    pDstQueue->pTail = pSrcQueue->pTail;
    pDstQueue->count = pSrcQueue->count;

    CpswOsal_restoreAllIntr(key);
}

void CpswUtils_enQ(queue_t *pQueue,
                   node_t *pNode)
{
    uintptr_t key = CpswOsal_disableAllIntr();

#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(pQueue->magic == CPSW_Q_INIT_DONE);
#endif
    pNode->pNext = NULL;

    if (0U == pQueue->count)
    {
        /* Queue is empty - Initialize it with this one node */
        pQueue->pHead = pNode;
    }
    else
    {
        /* Queue is not empty - Push onto END */
        pQueue->pTail->pNext = pNode;
    }

    /* Make tail of queue point to new */
    pQueue->pTail = pNode;
    pQueue->count++;

    CpswOsal_restoreAllIntr(key);
}

void CpswUtils_enQHead(queue_t *pQueue,
                       node_t *pNode)
{
    uintptr_t key = CpswOsal_disableAllIntr();

#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(pQueue->magic == CPSW_Q_INIT_DONE);
#endif
    if (0U == pQueue->count)
    {
        pQueue->pTail = pNode;
    }

    pNode->pNext  = pQueue->pHead;
    pQueue->pHead = pNode;
    pQueue->count++;

    CpswOsal_restoreAllIntr(key);
}

node_t *CpswUtils_deQ(queue_t *pQueue)
{
    node_t *pNode;
    uintptr_t key = CpswOsal_disableAllIntr();

#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(pQueue->magic == CPSW_Q_INIT_DONE);
#endif
    pNode = pQueue->pHead;

    if (NULL != pNode)
    {
        pQueue->pHead = pNode->pNext;
        if (NULL == pQueue->pHead)
        {
            pQueue->pTail = NULL;
        }

        pQueue->count--;
        pNode->pNext = NULL;
    }

    CpswOsal_restoreAllIntr(key);

    return pNode;
}

void CpswUtils_appendQ(queue_t *pDstQueue,
                       queue_t *pSrcQueue)
{
    uintptr_t key = CpswOsal_disableAllIntr();

#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(pDstQueue->magic == CPSW_Q_INIT_DONE);
    Cpsw_assert(pSrcQueue->magic == CPSW_Q_INIT_DONE);
#endif
    if (0U != pSrcQueue->count)
    {
        if (NULL == pDstQueue->pHead)
        {
            /* queue is empty - Initialize it with the source queue */
            CpswUtils_copyQ(pDstQueue, pSrcQueue);
        }
        else
        {
            /* queue is not empty - Add free queue contents onto end */
            pDstQueue->pTail->pNext = pSrcQueue->pHead;
            pDstQueue->pTail        = pSrcQueue->pTail;
            pDstQueue->count       += pSrcQueue->count;
        }
    }

    CpswOsal_restoreAllIntr(key);
}

uint32_t CpswUtils_getQCount(queue_t *pQueue)
{
#if (1U == CPSW_CFG_DEV_ERROR)
    Cpsw_assert(pQueue->magic == CPSW_Q_INIT_DONE);
#endif
    return pQueue->count;
}

void CpswUtils_verifyQCount(queue_t *pQueue)
{
#if (1U == CPSW_CFG_DEV_ERROR)
    node_t *pNode;
    uint32_t qCount = 0U;
    uintptr_t key = CpswOsal_disableAllIntr();

    Cpsw_assert(pQueue->magic == CPSW_Q_INIT_DONE);
    pNode = pQueue->pHead;

    while (NULL != pNode)
    {
        qCount++;
        pNode = pNode->pNext;
    }

    Cpsw_assert(pQueue->count == qCount);
    CpswOsal_restoreAllIntr(key);
#endif
}

/* end of file */
