/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file cpsw_stats.c
 *
 * \brief This file contains statistics related functionality.
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/csl/csl_cpswitch.h>
#include <ti/drv/cpsw/cpsw.h>
#include <ti/drv/cpsw/include/cpsw_types.h>
#include <ti/drv/cpsw/include/cpsw_stats.h>
#include "priv/cpsw_stats_priv.h"
#include "priv/cpsw_priv.h"
#include "priv/cpsw_trace_priv.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static CpswStats_Handle CpswStats_getObject(Cpsw_Type cpswType);

static void CpswStats_resetHostStats(CpswStats_Handle hStats);

static void CpswStats_resetMacStats(CpswStats_Handle hStats,
                                    Cpsw_MacPort macPort);

static void CpswStats_readHostStats(CpswStats_Handle hStats);

static void CpswStats_readMacStats(CpswStats_Handle hStats,
                                   Cpsw_MacPort macPort);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

/*! \brief CPSW statistics objects */
CpswStats_Obj statsObjs[CPSW_COUNT];

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

CpswStats_Handle CpswStats_open(Cpsw_Type cpswType)
{
    CpswStats_Handle hStats = CpswStats_getObject(cpswType);
    CSL_CPSW_PORTSTAT portStat;
    uint32_t i;

    memset(hStats, 0U, sizeof(*hStats));

    hStats->cpswRegs = Cpsw_getCslRegs(cpswType);
    Cpsw_assert(hStats->cpswRegs != NULL);

    memset(&portStat, 0U, sizeof(CSL_CPSW_PORTSTAT));

    /* Enable statistics on all applicable ports */
    portStat.p0StatEnable = true;
    portStat.p1StatEnable = true;
    if (cpswType == CPSW_9G)
    {
        portStat.p2StatEnable = true;
        portStat.p3StatEnable = true;
        portStat.p4StatEnable = true;
        portStat.p5StatEnable = true;
        portStat.p6StatEnable = true;
        portStat.p7StatEnable = true;
        portStat.p8StatEnable = true;
    }
    else if (cpswType == CPSW_5G)
    {
        portStat.p2StatEnable = true;
        portStat.p3StatEnable = true;
        portStat.p4StatEnable = true;
    }
    else
    {
        /* won't come here */
    }

    CSL_CPSW_setPortStatsEnableReg(hStats->cpswRegs, &portStat);

    /* Clear all statistics counters */
    CpswStats_resetHostStats(hStats);
    for (i = 0U; i < Cpsw_getMacPortMax(cpswType); i++)
    {
        CpswStats_resetMacStats(hStats, (Cpsw_MacPort)i);
    }

    hStats->magic    = CPSW_STATS_MAGIC;
    hStats->cpswType = cpswType;

    return hStats;
}

CpswStats_Handle CpswStats_reopen(Cpsw_Type cpswType)
{
    CpswStats_Handle hStats = CpswStats_getObject(cpswType);

    memset(hStats, 0U, sizeof(*hStats));
    hStats->cpswRegs = Cpsw_getCslRegs(cpswType);
    Cpsw_assert(hStats->cpswRegs != NULL);

    /* Clear host statistics counters */
    CpswStats_resetHostStats(hStats);

    hStats->magic    = CPSW_STATS_MAGIC;
    hStats->cpswType = cpswType;

    return hStats;
}

void CpswStats_close(CpswStats_Handle hStats)
{
    if (hStats->magic == CPSW_STATS_MAGIC)
    {
        hStats->magic    = CPSW_NO_MAGIC;
        hStats->cpswRegs = NULL;
    }
    else
    {
        CPSWTRACE_ERROR("Stats is not open (handle=%p)\n", hStats);
    }
}

int32_t CpswStats_ioctl(CpswStats_Handle hStats,
                        uint32_t cmd,
                        Cpsw_IoctlPrms *prms)
{
    int32_t status = CPSW_SOK;

    if (hStats->magic != CPSW_STATS_MAGIC)
    {
        CPSWTRACE_ERROR("Stats is not open (handle=%p)\n", hStats);
        status = CPSW_EINVALID_PARAMS;
    }

    if (status == CPSW_SOK)
    {
        switch (cmd)
        {
            case CPSW_STATS_IOCTL_GET_HOSTPORT_STATS:
            {
                status = Cpsw_checkOutArgs(prms, sizeof(CpswStats_PortStats));
                if (status == CPSW_SOK)
                {
                    CpswStats_PortStats *portStats =
                        (CpswStats_PortStats *)prms->outArgs;

                    CpswStats_readHostStats(hStats);

                    memcpy(portStats,
                           &hStats->hostPortStats,
                           sizeof(CpswStats_PortStats));
                }
                else
                {
                    CPSWTRACE_ERROR("invalid GET_HOSTPORT_STATS params\n");
                }
            }
            break;

            case CPSW_STATS_IOCTL_GET_MACPORT_STATS:
            {
                status = Cpsw_checkInOutArgs(prms,
                                             sizeof(CpswStats_GenericMacPortInArgs),
                                             sizeof(CpswStats_PortStats));
                if (status == CPSW_SOK)
                {
                    CpswStats_GenericMacPortInArgs *inArgs =
                        (CpswStats_GenericMacPortInArgs *)prms->inArgs;
                    CpswStats_PortStats *portStats =
                        (CpswStats_PortStats *)prms->outArgs;

                    CpswStats_readMacStats(hStats, inArgs->portNum);

                    memcpy(portStats,
                           &hStats->macPortStats[inArgs->portNum],
                           sizeof(CpswStats_PortStats));
                }
                else
                {
                    CPSWTRACE_ERROR("invalid GET_MACPORT_STATS params\n");
                }
            }
            break;

            case CPSW_STATS_IOCTL_RESET_HOSTPORT_STATS:
            {
                CpswStats_resetHostStats(hStats);
            }
            break;

            case CPSW_STATS_IOCTL_RESET_MACPORT_STATS:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswStats_GenericMacPortInArgs));
                if (status == CPSW_SOK)
                {
                    CpswStats_GenericMacPortInArgs *inArgs =
                        (CpswStats_GenericMacPortInArgs *)prms->inArgs;

                    CpswStats_resetMacStats(hStats, inArgs->portNum);
                }
                else
                {
                    CPSWTRACE_ERROR("invalid RESET_MACPORT_STATS params\n");
                }
            }
            break;

            case CPSW_STATS_IOCTL_RESTORE_HOSTPORT_STATS:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswStats_PortStats));
                if (status == CPSW_SOK)
                {
                    CpswStats_PortStats *portStats =
                        (CpswStats_PortStats *)prms->inArgs;

                    memcpy(&hStats->hostPortStats,
                           portStats,
                           sizeof(CpswStats_PortStats));
                }
                else
                {
                    CPSWTRACE_ERROR("invalid RESTORE_HOSTPORT_STAT params\n");
                }
            }
            break;

            case CPSW_STATS_IOCTL_RESTORE_MACPORT_STATS:
            {
                status = Cpsw_checkInArgs(prms, sizeof(CpswStats_RestoreMacPortStatsInArgs));
                if (status == CPSW_SOK)
                {
                    CpswStats_RestoreMacPortStatsInArgs *inArgs =
                        (CpswStats_RestoreMacPortStatsInArgs *)prms->inArgs;

                    memcpy(&hStats->macPortStats[inArgs->portNum],
                           &inArgs->portStats,
                           sizeof(CpswStats_PortStats));
                }
                else
                {
                    CPSWTRACE_ERROR("invalid RESTORE_MACPORT_STATS params\n");
                }
            }
            break;

            case CPSW_STATS_IOCTL_SYNC:
            {
                uint32_t i;

                CpswStats_readHostStats(hStats);
                for (i = 0U; i < Cpsw_getMacPortMax(hStats->cpswType); i++)
                {
                    CpswStats_readMacStats(hStats, (Cpsw_MacPort)i);
                }
            }
            break;

            case CPSW_STATS_IOCTL_PRINT_REGS:
            {
            }
            break;

            default:
            {
                CPSWTRACE_ERROR("invalid IOCTL cmd: %u\n", cmd);
                status = CPSW_EINVALID_PARAMS;
            }
            break;
        }
    }

    return status;
}

static CpswStats_Handle CpswStats_getObject(Cpsw_Type cpswType)
{
    CpswStats_Handle hStats = NULL;

    switch (cpswType)
    {
        case CPSW_2G:
        case CPSW_5G:
        case CPSW_9G:
            hStats = &statsObjs[cpswType];
            break;

        default:
            Cpsw_assert(false);
    }

    return hStats;
}

static void CpswStats_resetHostStats(CpswStats_Handle hStats)
{
    CSL_CPSW_STATS portStats;

    /* Clear local stats and hardware stats through a dummy read */
    memset(&hStats->hostPortStats, 0U, sizeof(CpswStats_PortStats));
    memset(&portStats, 0U, sizeof(CSL_CPSW_STATS));
    CSL_CPSW_getPortStats(hStats->cpswRegs, 0U, &portStats);
}

static void CpswStats_resetMacStats(CpswStats_Handle hStats,
                                    Cpsw_MacPort macPort)
{
    CSL_CPSW_STATS portStats;
    uint32_t portNum = (uint32_t)macPort;

    /* Clear local stats and hardware stats through a dummy read */
    memset(&hStats->macPortStats[portNum], 0U, sizeof(CpswStats_PortStats));
    memset(&portStats, 0U, sizeof(CSL_CPSW_STATS));
    CSL_CPSW_getPortStats(hStats->cpswRegs, portNum + 1U, &portStats);
}

static void CpswStats_readHostStats(CpswStats_Handle hStats)
{
    CSL_CPSW_STATS portStats;
    uint64_t *stats64 = hStats->hostPortStats.val;
    uint32_t *stats32 = (uint32_t *)&portStats;
    uint32_t i;

    /* CSL blindly reads all registers in the statistics block regardless
     * of whether they are applicable or not to a CPSW instance type*/
    memset(&portStats, 0U, sizeof(CSL_CPSW_STATS));
    CSL_CPSW_getPortStats(hStats->cpswRegs, 0U, &portStats);

    for (i = 0U; i < CPSW_STATS_BLOCK_ELEM_NUM; i++)
    {
        stats64[i] += stats32[i];
    }

    /* Clear reserved fields */
    switch (hStats->cpswType)
    {
        case CPSW_2G:
        {
            CpswStats_HostPort_2g *hostStats2G = NULL;
            hostStats2G = (CpswStats_HostPort_2g *)stats64;

            hostStats2G->reserved4 = 0U;
            hostStats2G->reserved6 = 0U;
            hostStats2G->reserved8 = 0U;
            memset(hostStats2G->reserved17to25, 0U, sizeof(hostStats2G->reserved17to25));
            memset(hostStats2G->reserved52to95, 0U, sizeof(hostStats2G->reserved52to95));
            memset(hostStats2G->reserved97to128, 0U, sizeof(hostStats2G->reserved97to128));
        }
        break;

        case CPSW_5G:
        case CPSW_9G:
        {
            CpswStats_HostPort_9g *hostStats9G = NULL;
            hostStats9G = (CpswStats_HostPort_9g *)stats64;

            hostStats9G->reserved4 = 0U;
            hostStats9G->reserved6 = 0U;
            hostStats9G->reserved8 = 0U;
            memset(hostStats9G->reserved17to25, 0U, sizeof(hostStats9G->reserved17to25));
            memset(hostStats9G->reserved57to80, 0U, sizeof(hostStats9G->reserved57to80));
            memset(hostStats9G->reserved87to95, 0U, sizeof(hostStats9G->reserved87to95));
        }
        break;
        default:
            Cpsw_assert(false);
    }
}

static void CpswStats_readMacStats(CpswStats_Handle hStats,
                                   Cpsw_MacPort macPort)
{
    CSL_CPSW_STATS portStats;
    uint32_t portNum  = (uint32_t)macPort;
    uint64_t *stats64 = hStats->macPortStats[portNum].val;
    uint32_t *stats32 = (uint32_t *)&portStats;
    uint32_t i;

    /* CSL blindly reads all registers in the statistics block regardless
     * of whether they are applicable or not to a CPSW instance type*/
    memset(&portStats, 0U, sizeof(CSL_CPSW_STATS));
    CSL_CPSW_getPortStats(hStats->cpswRegs, portNum + 1U, &portStats);

    for (i = 0U; i < CPSW_STATS_BLOCK_ELEM_NUM; i++)
    {
        stats64[i] += stats32[i];
    }

    /* Clear reserved fields */
    switch (hStats->cpswType)
    {
        case CPSW_2G:
        {
            CpswStats_MacPort_2g *macStats2G = NULL;
            macStats2G = (CpswStats_MacPort_2g *)stats64;

            memset(macStats2G->reserved52to95, 0U, sizeof(macStats2G->reserved52to95));
        }
        break;

        case CPSW_5G:
        case CPSW_9G:
        {
            CpswStats_MacPort_9g *macStats9G = NULL;
            macStats9G = (CpswStats_MacPort_9g *)stats64;

            memset(macStats9G->reserved57to80, 0U, sizeof(macStats9G->reserved57to80));
            memset(macStats9G->reserved87to95, 0U, sizeof(macStats9G->reserved87to95));
        }
        break;

        default:
            Cpsw_assert(false);
    }
}
