/*
 *  Copyright (c) Texas Instruments Incorporated 2020
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file dp83822.c
 *
 * \brief DP83822 PHY specific implementation
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/drv/cpsw/include/cpsw_phy.h>
#include <ti/drv/cpsw/include/cpsw_utils.h>
#include <ti/drv/cpsw/src/priv/cpsw_phy_priv.h>
#include <ti/drv/cpsw/src/priv/cpsw_trace_priv.h>
#include <ti/drv/cpsw/include/dp83822.h>
#include "generic_phy.h"
#include "dp83822_priv.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define DP83822_OUI                           (0x080028U)
#define DP83822_MODEL                         (0x24U)
#define DP83822_REV                           (0x00U)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static bool Dp83822_isPhyDevSupported(CpswPhy_Handle hPhyDev,
                                      const CpswPhy_Version *version);

static bool Dp83822_isMacModeSupported(CpswPhy_Handle hPhyDev,
                                       CpswPhy_MacMode mode);

static int32_t Dp83822_config(CpswPhy_Handle hPhyDev,
                              const CpswPhy_Config *cfg,
                              CpswPhy_MacMode mode);

static void Dp83822_enableAutoMdix(CpswPhy_Handle hPhyDev,
                                   bool enable);

static void Dp83822_printRegs(CpswPhy_Handle hPhyDev);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

CpswPhy_Drv gCpswPhyDrvDp83822 =
{
    .name               = "dp83822",
    .isPhyDevSupported  = Dp83822_isPhyDevSupported,
    .isMacModeSupported = Dp83822_isMacModeSupported,
    .config             = Dp83822_config,
    .reset              = GenericPhy_reset,
    .isResetComplete    = GenericPhy_isResetComplete,
    .runComplianceTest  = NULL,
    .readExtReg         = GenericPhy_readExtReg,
    .writeExtReg        = GenericPhy_writeExtReg,
    .printRegs          = Dp83822_printRegs,
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

void Dp83822_initParams(Dp83822_Config *config)
{
}

static bool Dp83822_isPhyDevSupported(CpswPhy_Handle hPhyDev,
                                      const CpswPhy_Version *version)
{
    bool supported = false;

    if ((version->oui == DP83822_OUI) &&
        (version->model == DP83822_MODEL) &&
        (version->revision == DP83822_REV))
    {
        supported = true;
    }

    return supported;
}

static bool Dp83822_isMacModeSupported(CpswPhy_Handle hPhyDev,
                                       CpswPhy_MacMode mode)
{
    bool supported;

    switch (mode)
    {
        case CPSW_PHY_MAC_MODE_RMII:
            supported = true;
            break;

        /* This driver doesn't support MII and RGMII interfaces,
         * but the DP83822 PHY does support them */
        case CPSW_PHY_MAC_MODE_MII:
        case CPSW_PHY_MAC_MODE_RGMII:
        default:
            supported = false;
            break;
    }

    return supported;
}

static int32_t Dp83822_config(CpswPhy_Handle hPhyDev,
                              const CpswPhy_Config *cfg,
                              CpswPhy_MacMode mode)
{
    const Dp83822_Config *extendedCfg = (const Dp83822_Config *)cfg->extendedConfig;
    uint32_t extendedCfgSize = cfg->extendedConfigSize;
    bool enableAutoMdix;
    int32_t status = CPSW_SOK;

    if ((extendedCfg == NULL) ||
        (extendedCfgSize != sizeof(*extendedCfg)))
    {
        CPSWTRACE_ERROR("PHY %u: invalid config params (cfg=%p, size=%u)\n",
                        hPhyDev->addr, extendedCfg, extendedCfgSize);
        status = CPSW_EINVALID_PARAMS;
    }

    /* Auto-MDIX should be disabled in near-end loopback modes */
    enableAutoMdix = !cfg->enableLoopback;

    /* Enable Auto-MDIX and Robust Auto-MDIX */
    if (status == CPSW_SOK)
    {
        Dp83822_enableAutoMdix(hPhyDev, enableAutoMdix);
    }

    return status;
}

static void Dp83822_enableAutoMdix(CpswPhy_Handle hPhyDev,
                                   bool enable)
{
    CPSWTRACE_DEBUG("PHY %u: %s automatic cross-over %u\n",
                    hPhyDev->addr, enable ? "enable" : "disable");
    CpswPhy_rmwReg(hPhyDev, DP83822_PHYCR,
                   PHYCR_AUTOMDIX_ENABLE,
                   enable ? PHYCR_AUTOMDIX_ENABLE : 0);

    if (enable)
    {
        CPSWTRACE_DEBUG("PHY %u: enable Robust Auto-MDIX\n", hPhyDev->addr);
        CpswPhy_rmwReg(hPhyDev, DP83822_CR1,
                       CR1_ROBUSTAUTOMDIX,
                       CR1_ROBUSTAUTOMDIX);
    }
    else
    {
        CpswPhy_rmwReg(hPhyDev, DP83822_PHYCR,
                       PHYCR_FORCEMDIX_MASK,
                       PHYCR_FORCEMDIX_MDI);
    }
}

static void Dp83822_printRegs(CpswPhy_Handle hPhyDev)
{
    uint32_t phyAddr = hPhyDev->addr;
    uint16_t val;

    CpswPhy_readReg(hPhyDev, PHY_BMCR, &val);
    CpswUtils_printf("PHY %u: BMCR    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_BMSR, &val);
    CpswUtils_printf("PHY %u: BMSR    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_PHYIDR1, &val);
    CpswUtils_printf("PHY %u: PHYIDR1 = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_PHYIDR2, &val);
    CpswUtils_printf("PHY %u: PHYIDR2 = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_ANAR, &val);
    CpswUtils_printf("PHY %u: ANAR    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_ANLPAR, &val);
    CpswUtils_printf("PHY %u: ANLPAR  = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_ANER, &val);
    CpswUtils_printf("PHY %u: ANER    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_ANNPTR, &val);
    CpswUtils_printf("PHY %u: ANNPTR  = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_ANNPRR, &val);
    CpswUtils_printf("PHY %u: ANNPRR  = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, DP83822_CR1, &val);
    CpswUtils_printf("PHY %u: CR1     = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_GIGSR, &val);
    CpswUtils_printf("PHY %u: STS1    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_GIGESR, &val);
    CpswUtils_printf("PHY %u: 1KSCR   = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, DP83822_PHYCR, &val);
    CpswUtils_printf("PHY %u: PHYCR   = 0x%04x\n", phyAddr, val);
}
