/*
 *  Copyright (c) Texas Instruments Incorporated 2019
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file dp83867.c
 *
 * \brief DP83867 PHY specific implementation
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/drv/cpsw/include/cpsw_phy.h>
#include <ti/drv/cpsw/include/cpsw_utils.h>
#include <ti/drv/cpsw/src/priv/cpsw_phy_priv.h>
#include <ti/drv/cpsw/src/priv/cpsw_trace_priv.h>
#include <ti/drv/cpsw/include/dp83867.h>
#include "generic_phy.h"
#include "dp83867_priv.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

#define DP83867_OUI                           (0x080028U)
#define DP83867_MODEL                         (0x23U)
#define DP83867_REV                           (0x01U)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static bool Dp83867_isPhyDevSupported(CpswPhy_Handle hPhyDev,
                                      const CpswPhy_Version *version);

static bool Dp83867_isMacModeSupported(CpswPhy_Handle hPhyDev,
                                       CpswPhy_MacMode mode);

static int32_t Dp83867_config(CpswPhy_Handle hPhyDev,
                              const CpswPhy_Config *cfg,
                              CpswPhy_MacMode mode);

static void Dp83867_setLoopbackConfig(CpswPhy_Handle hPhyDev,
                                      bool enable);

static void Dp83867_enableAutoMdix(CpswPhy_Handle hPhyDev,
                                   bool enable);

static void Dp83867_setClkShift(CpswPhy_Handle hPhyDev,
                                bool txShiftEnable,
                                bool rxShiftEnable);

static int32_t Dp83867_setTxFifoDepth(CpswPhy_Handle hPhyDev,
                                      uint8_t depth);

static int32_t Dp83867_setClkDelay(CpswPhy_Handle hPhyDev,
                                   uint32_t txDelay,
                                   uint32_t rxDelay);

static int32_t Dp83867_setOutputImpedance(CpswPhy_Handle hPhyDev,
                                          uint32_t impedance);

static void Dp83867_setGpioMux(CpswPhy_Handle hPhyDev,
                               Dp83867_Gpio0Mode gpio0Mode,
                               Dp83867_Gpio1Mode gpio1Mode);

static void Dp83867_setLedMode(CpswPhy_Handle hPhyDev,
                               const Dp83867_LedMode *ledMode);

static void Dp83867_reset(CpswPhy_Handle hPhyDev);

static bool Dp83867_isResetComplete(CpswPhy_Handle hPhyDev);

static void Dp83867_rmwExtReg(CpswPhy_Handle hPhyDev,
                              uint32_t reg,
                              uint16_t mask,
                              uint16_t val);

static void Dp83867_printRegs(CpswPhy_Handle hPhyDev);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

CpswPhy_Drv gCpswPhyDrvDp83867 =
{
    .name               = "dp83867",
    .isPhyDevSupported  = Dp83867_isPhyDevSupported,
    .isMacModeSupported = Dp83867_isMacModeSupported,
    .config             = Dp83867_config,
    .reset              = Dp83867_reset,
    .isResetComplete    = Dp83867_isResetComplete,
    .runComplianceTest  = NULL,
    .readExtReg         = GenericPhy_readExtReg,
    .writeExtReg        = GenericPhy_writeExtReg,
    .printRegs          = Dp83867_printRegs,
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

void Dp83867_initParams(Dp83867_Config *config)
{
    config->txClkShiftEnable     = false;
    config->rxClkShiftEnable     = false;
    config->txDelayInPs          = 2000U;  /* 2.00 ns*/
    config->rxDelayInPs          = 2000U;  /* 2.00 ns */
    config->txFifoDepth          = 4U;     /* 4 bytes/nibbles */
    config->impedanceInMilliOhms = 50000U; /* 50 ohms */
    config->gpio0Mode            = DP83867_GPIO0_RXERR;
    config->gpio1Mode            = DP83867_GPIO1_COL;
    config->ledMode[0]           = DP83867_LED_LINKED;
    config->ledMode[1]           = DP83867_LED_LINKED_1000BT;
    config->ledMode[2]           = DP83867_LED_RXTXACT;
    config->ledMode[3]           = DP83867_LED_LINKED_100BTX;
}

static bool Dp83867_isPhyDevSupported(CpswPhy_Handle hPhyDev,
                                      const CpswPhy_Version *version)
{
    bool supported = false;

    if ((version->oui == DP83867_OUI) &&
        (version->model == DP83867_MODEL) &&
        (version->revision == DP83867_REV))
    {
        supported = true;
    }

    return supported;
}

static bool Dp83867_isMacModeSupported(CpswPhy_Handle hPhyDev,
                                       CpswPhy_MacMode mode)
{
    bool supported;

    switch (mode)
    {
        case CPSW_PHY_MAC_MODE_MII:
        case CPSW_PHY_MAC_MODE_GMII:
        case CPSW_PHY_MAC_MODE_RGMII:
            supported = true;
            break;

        default:
            supported = false;
            break;
    }

    return supported;
}

static int32_t Dp83867_config(CpswPhy_Handle hPhyDev,
                              const CpswPhy_Config *cfg,
                              CpswPhy_MacMode mode)
{
    const Dp83867_Config *extendedCfg = (const Dp83867_Config *)cfg->extendedConfig;
    uint32_t extendedCfgSize = cfg->extendedConfigSize;
    bool enableAutoMdix = true;
    int32_t status = CPSW_SOK;

    if ((extendedCfg == NULL) ||
        (extendedCfgSize != sizeof(*extendedCfg)))
    {
        CPSWTRACE_ERROR("PHY %u: invalid config params (cfg=%p, size=%u)\n",
                        hPhyDev->addr, extendedCfg, extendedCfgSize);
        status = CPSW_EINVALID_PARAMS;
    }

    /* Set loopback configuration: enable or disable */
    if (status == CPSW_SOK)
    {
        /* To maintain the desired operating mode, Auto-Negotiation should be
         * disabled before enabling the near-end loopback mode (MII loopback is
         * the only mode supported by the driver).  That's taken care by the
         * PHY framework when setting PHY's manual mode parameters */
        Dp83867_setLoopbackConfig(hPhyDev, cfg->enableLoopback);

        /* Auto-MDIX should be disabled before selecting the Near-End Loopback
         * mode. MDI or MDIX configuration should be manually configured */
        enableAutoMdix = !cfg->enableLoopback;
    }

    /* Enable Auto-MDIX and Robust Auto-MDIX */
    if (status == CPSW_SOK)
    {
        Dp83867_enableAutoMdix(hPhyDev, enableAutoMdix);
    }

    /* Config RGMII TX and RX clock shift and clock delay */
    if ((status == CPSW_SOK) &&
        (mode == CPSW_PHY_MAC_MODE_RGMII))
    {
        Dp83867_setClkShift(hPhyDev,
                            extendedCfg->txClkShiftEnable,
                            extendedCfg->rxClkShiftEnable);

        status = Dp83867_setClkDelay(hPhyDev,
                                     extendedCfg->txDelayInPs,
                                     extendedCfg->rxDelayInPs);
    }

    /* Config TX FIFO depth */
    if (status == CPSW_SOK)
    {
        status = Dp83867_setTxFifoDepth(hPhyDev,
                                        extendedCfg->txFifoDepth);
    }

    /* Set output impedance */
    if (status == CPSW_SOK)
    {
        status = Dp83867_setOutputImpedance(hPhyDev,
                                            extendedCfg->impedanceInMilliOhms);
    }

    /* Set GPIO mux control (RGZ devices only) */
    if (status == CPSW_SOK)
    {
        Dp83867_setGpioMux(hPhyDev,
                           extendedCfg->gpio0Mode,
                           extendedCfg->gpio1Mode);
    }

    /* Set LED configuration */
    if (status == CPSW_SOK)
    {
        Dp83867_setLedMode(hPhyDev, extendedCfg->ledMode);
    }

    return status;
}

static void Dp83867_setLoopbackConfig(CpswPhy_Handle hPhyDev,
                                      bool enable)
{
    bool complete;
    uint16_t val;

    CPSWTRACE_DEBUG("PHY %u: %s loopback\n",
                    hPhyDev->addr, enable ? "enable" : "disable");

    if (enable)
    {
        val = LOOPCR_CFG_LOOPBACK;
    }
    else
    {
        val = LOOPCR_CFG_NORMAL;
    }

    /* Specific predefined loopback configuration values are required for
     * normal mode or loopback mode */
    GenericPhy_writeExtReg(hPhyDev, DP83867_LOOPCR, val);

    /* Software restart is required after changing LOOPCR register */
    Dp83867_reset(hPhyDev);

    do
    {
        complete = Dp83867_isResetComplete(hPhyDev);
    }
    while (!complete);
}

static void Dp83867_enableAutoMdix(CpswPhy_Handle hPhyDev,
                                   bool enable)
{
    uint16_t val;

    if (enable)
    {
        val = PHYCR_MDICROSSOVER_AUTO;
    }
    else
    {
        val = PHYCR_MDICROSSOVER_MDI;
    }

    CPSWTRACE_DEBUG("PHY %u: %s automatic cross-over\n",
                    hPhyDev->addr, enable ? "enable" : "disable");
    CpswPhy_rmwReg(hPhyDev, DP83867_PHYCR,
                   PHYCR_MDICROSSOVER_MASK,
                   val);

    if (enable)
    {
        CPSWTRACE_DEBUG("PHY %u: enable Robust Auto-MDIX\n", hPhyDev->addr);
        CpswPhy_rmwReg(hPhyDev, DP83867_CFG3,
                       CFG3_ROBUSTAUTOMDIX,
                       CFG3_ROBUSTAUTOMDIX);
    }
}

static void Dp83867_setClkShift(CpswPhy_Handle hPhyDev,
                                bool txShiftEnable,
                                bool rxShiftEnable)
{
    uint16_t val;

    CPSWTRACE_DEBUG("PHY %u: clock shift TX:%s RX:%s\n",
                    hPhyDev->addr,
                    txShiftEnable ? "enable" : "disable",
                    rxShiftEnable ? "enable" : "disable");

    val  = RGMIICTL_RGMIIEN;
    val |= (txShiftEnable == true) ? RGMIICTL_TXCLKDLY : 0U;
    val |= (rxShiftEnable == true) ? RGMIICTL_RXCLKDLY : 0U;
    Dp83867_rmwExtReg(hPhyDev, DP83867_RGMIICTL,
                      RGMIICTL_RGMIIEN | RGMIICTL_TXCLKDLY | RGMIICTL_RXCLKDLY,
                      val);
}

static int32_t Dp83867_setTxFifoDepth(CpswPhy_Handle hPhyDev,
                                      uint8_t depth)
{
    uint16_t val   = 0U;
    int32_t status = CPSW_SOK;

    switch (depth)
    {
        case 3U:
            val = PHYCR_TXFIFODEPTH_3B;
            break;

        case 4U:
            val = PHYCR_TXFIFODEPTH_4B;
            break;

        case 6U:
            val = PHYCR_TXFIFODEPTH_6B;
            break;

        case 8U:
            val = PHYCR_TXFIFODEPTH_8B;
            break;

        default:
            status = CPSW_EINVALID_PARAMS;
            break;
    }

    if (status == CPSW_SOK)
    {
        CPSWTRACE_DEBUG("PHY %u: set FIFO depth %u\n", hPhyDev->addr, depth);
        CpswPhy_rmwReg(hPhyDev, DP83867_PHYCR, PHYCR_TXFIFODEPTH_MASK, val);
    }
    else
    {
        CPSWTRACE_ERROR("PHY %u: invalid FIFO depth: %u\n", hPhyDev->addr, depth);
    }

    return status;
}

static int32_t Dp83867_setClkDelay(CpswPhy_Handle hPhyDev,
                                   uint32_t txDelay,
                                   uint32_t rxDelay)
{
    uint16_t val;
    uint32_t delay;
    uint32_t delayCtrl;
    int32_t status = CPSW_SOK;

    if ((txDelay <= RGMIIDCTL_DELAY_MAX) &&
        (rxDelay <= RGMIIDCTL_DELAY_MAX))
    {
        CPSWTRACE_DEBUG("PHY %u: set delay %u ps TX, %u ps RX\n",
                        hPhyDev->addr, txDelay, rxDelay);

        /* Avoids wrong value of delayCtrl if txDelay is 0 */
        delay     = (txDelay > 0U) ? txDelay : 1U;
        delayCtrl = CPSW_UTILS_DIV_ROUNDUP(delay, RGMIIDCTL_DELAY_STEP) - 1U;
        val       = (uint16_t)((delayCtrl << RGMIIDCTL_TXDLYCTRL_OFFSET) & RGMIIDCTL_TXDLYCTRL_MASK);

        /* Avoids wrong value of delayCtrl if rxDelay is 0 */
        delay     = (rxDelay > 0U) ? rxDelay : 1U;
        delayCtrl = CPSW_UTILS_DIV_ROUNDUP(delay, RGMIIDCTL_DELAY_STEP) - 1U;
        val      |= (uint16_t)((delayCtrl << RGMIIDCTL_RXDLYCTRL_OFFSET) & RGMIIDCTL_RXDLYCTRL_MASK);

        GenericPhy_writeExtReg(hPhyDev, DP83867_RGMIIDCTL, val);
    }
    else
    {
        CPSWTRACE_ERROR("PHY %u: invalid delay (TX=%u RX=%u)\n",
                        hPhyDev->addr, txDelay, rxDelay);
        status = CPSW_EINVALID_PARAMS;
    }

    return status;
}

static int32_t Dp83867_setOutputImpedance(CpswPhy_Handle hPhyDev,
                                          uint32_t impedance)
{
    int32_t status = CPSW_SOK;
    uint32_t val;

    if ((impedance >= IOMUXCFG_IOIMPEDANCE_MIN) &&
        (impedance <= IOMUXCFG_IOIMPEDANCE_MAX))
    {
        CPSWTRACE_DEBUG("PHY %u: set output impedance to %u milli-ohms\n",
                        hPhyDev->addr, impedance);

        val = (IOMUXCFG_IOIMPEDANCE_MAX - impedance) * IOMUXCFG_IOIMPEDANCE_MASK;
        val = (val + IOMUXCFG_IOIMPEDANCE_RANGE / 2) / IOMUXCFG_IOIMPEDANCE_RANGE;

        Dp83867_rmwExtReg(hPhyDev, DP83867_IOMUXCFG,
                          IOMUXCFG_IOIMPEDANCE_MASK,
                          val);
    }
    else
    {
        CPSWTRACE_ERROR("PHY %u: out-of-range impedance: %u\n",
                        hPhyDev->addr, impedance);
        status = CPSW_EINVALID_PARAMS;
    }

    return status;
}

static void Dp83867_setGpioMux(CpswPhy_Handle hPhyDev,
                               Dp83867_Gpio0Mode gpio0Mode,
                               Dp83867_Gpio1Mode gpio1Mode)
{
    int16_t gpio0;
    int16_t gpio1;

    CPSWTRACE_DEBUG("PHY %u: set gpio0 = mode%u, gpio1 = mode%u\n",
                    hPhyDev->addr, gpio0Mode, gpio1Mode);

    gpio0  = (uint16_t)gpio0Mode << GPIOMUXCTRL_GPIO0_OFFSET;
    gpio0 &= GPIOMUXCTRL_GPIO0_MASK;

    gpio1  = (uint16_t)gpio1Mode << GPIOMUXCTRL_GPIO1_OFFSET;
    gpio1 &= GPIOMUXCTRL_GPIO1_MASK;

    Dp83867_rmwExtReg(hPhyDev, DP83867_GPIOMUXCTRL,
                      GPIOMUXCTRL_GPIO1_MASK | GPIOMUXCTRL_GPIO0_MASK,
                      gpio1 | gpio0);
}

static void Dp83867_setLedMode(CpswPhy_Handle hPhyDev,
                               const Dp83867_LedMode *ledMode)
{
    uint16_t val;

    CPSWTRACE_DEBUG("PHY %u: set LED0 = mode%u, LED1 = mode%u, LED2 = mode%u, LED3 = mode%u\n",
                    hPhyDev->addr, ledMode[0], ledMode[1], ledMode[2], ledMode[3]);

    val  = ((uint16_t)ledMode[0] << LEDCR1_LED0SEL_OFFSET) & LEDCR1_LED0SEL_MASK;
    val |= ((uint16_t)ledMode[1] << LEDCR1_LED1SEL_OFFSET) & LEDCR1_LED1SEL_MASK;
    val |= ((uint16_t)ledMode[2] << LEDCR1_LED2SEL_OFFSET) & LEDCR1_LED2SEL_MASK;
    val |= ((uint16_t)ledMode[3] << LEDCR1_LED3SEL_OFFSET) & LEDCR1_LED3SEL_MASK;

    CpswPhy_writeReg(hPhyDev, DP83867_LEDCR1, val);
}

static void Dp83867_reset(CpswPhy_Handle hPhyDev)
{
    /* Global software reset: all PHY internal circuits including IEEE-defined
     * registers and all extended registers are reset */
    CPSWTRACE_DEBUG("PHY %u: global soft-reset\n", hPhyDev->addr);
    CpswPhy_rmwReg(hPhyDev, DP83867_CTRL, CTRL_SWRESET, CTRL_SWRESET);
}

static bool Dp83867_isResetComplete(CpswPhy_Handle hPhyDev)
{
    int32_t status;
    uint16_t val;
    bool complete = false;

    /* Reset is complete when RESET bit has self-cleared */
    status = CpswPhy_readReg(hPhyDev, DP83867_CTRL, &val);
    if (status == CPSW_SOK)
    {
        complete = ((val & CTRL_SWRESET) == 0U);
    }

    CPSWTRACE_DEBUG("PHY %u: global soft-reset is %scomplete\n",
                    hPhyDev->addr, complete ? "" : "not");

    return complete;
}

static void Dp83867_rmwExtReg(CpswPhy_Handle hPhyDev,
                              uint32_t reg,
                              uint16_t mask,
                              uint16_t val)
{
    uint16_t devad = MMD_CR_DEVADDR;
    uint16_t data;
    int32_t status;

    CPSWTRACE_VERBOSE("PHY %u: write reg %u mask 0x%04x val 0x%04x\n",
                      hPhyDev->addr, reg, mask, val);

    CpswPhy_writeReg(hPhyDev, PHY_MMD_CR, devad | MMD_CR_ADDR);
    CpswPhy_writeReg(hPhyDev, PHY_MMD_DR, reg);
    CpswPhy_writeReg(hPhyDev, PHY_MMD_CR, devad | MMD_CR_DATA_NOPOSTINC);
    status = CpswPhy_readReg(hPhyDev, PHY_MMD_DR, &data);

    if (status == CPSW_SOK)
    {
        data = (data & ~mask) | (val & mask);
        CpswPhy_writeReg(hPhyDev, PHY_MMD_CR, devad | MMD_CR_DATA_NOPOSTINC);
        CpswPhy_writeReg(hPhyDev, PHY_MMD_DR, data);
    }
}

static void Dp83867_printRegs(CpswPhy_Handle hPhyDev)
{
    uint32_t phyAddr = hPhyDev->addr;
    uint16_t val;

    CpswPhy_readReg(hPhyDev, PHY_BMCR, &val);
    CpswUtils_printf("PHY %u: BMCR    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_BMSR, &val);
    CpswUtils_printf("PHY %u: BMSR    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_PHYIDR1, &val);
    CpswUtils_printf("PHY %u: PHYIDR1 = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_PHYIDR2, &val);
    CpswUtils_printf("PHY %u: PHYIDR2 = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_ANAR, &val);
    CpswUtils_printf("PHY %u: ANAR    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_ANLPAR, &val);
    CpswUtils_printf("PHY %u: ANLPAR  = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_ANER, &val);
    CpswUtils_printf("PHY %u: ANER    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_ANNPTR, &val);
    CpswUtils_printf("PHY %u: ANNPTR  = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_ANNPRR, &val);
    CpswUtils_printf("PHY %u: ANNPRR  = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_GIGCR, &val);
    CpswUtils_printf("PHY %u: CFG1    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_GIGSR, &val);
    CpswUtils_printf("PHY %u: STS1    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, PHY_GIGESR, &val);
    CpswUtils_printf("PHY %u: 1KSCR   = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, DP83867_PHYCR, &val);
    CpswUtils_printf("PHY %u: PHYCR   = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, DP83867_PHYSTS, &val);
    CpswUtils_printf("PHY %u: PHYSTS  = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, DP83867_MICR, &val);
    CpswUtils_printf("PHY %u: MICR    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, DP83867_ISR, &val);
    CpswUtils_printf("PHY %u: ISR     = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, DP83867_CFG2, &val);
    CpswUtils_printf("PHY %u: CFG2    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, DP83867_RECR, &val);
    CpswUtils_printf("PHY %u: RECR    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, DP83867_BISCR, &val);
    CpswUtils_printf("PHY %u: BISCR   = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, DP83867_STS2, &val);
    CpswUtils_printf("PHY %u: STS2    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, DP83867_LEDCR1, &val);
    CpswUtils_printf("PHY %u: LEDCR1  = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, DP83867_LEDCR2, &val);
    CpswUtils_printf("PHY %u: LEDCR2  = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, DP83867_LEDCR3, &val);
    CpswUtils_printf("PHY %u: LEDCR3  = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, DP83867_CFG3, &val);
    CpswUtils_printf("PHY %u: CFG3    = 0x%04x\n", phyAddr, val);
    CpswPhy_readReg(hPhyDev, DP83867_CTRL, &val);
    CpswUtils_printf("PHY %u: CTRL    = 0x%04x\n", phyAddr, val);
}
