/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file generic_phy.c
 *
 * \brief PHY specific implementation for an IEEE generic PHY
 */

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/drv/cpsw/include/cpsw_phy.h>
#include <ti/drv/cpsw/include/cpsw_utils.h>
#include <ti/drv/cpsw/src/priv/cpsw_phy_priv.h>
#include <ti/drv/cpsw/src/priv/cpsw_trace_priv.h>
#include "generic_phy.h"

/* ========================================================================== */
/*                           Macros & Typedefs                                */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

static bool GenericPhy_isPhyDevSupported(CpswPhy_Handle hPhyDev,
                                         const CpswPhy_Version *version);

static bool GenericPhy_isMacModeSupported(CpswPhy_Handle hPhyDev,
                                          CpswPhy_MacMode mode);

/* ========================================================================== */
/*                            Global Variables                                */
/* ========================================================================== */

CpswPhy_Drv gCpswPhyDrvGeneric =
{
    .name               = "generic",
    .isPhyDevSupported  = GenericPhy_isPhyDevSupported,
    .isMacModeSupported = GenericPhy_isMacModeSupported,
    .config             = NULL,
    .reset              = GenericPhy_reset,
    .isResetComplete    = GenericPhy_isResetComplete,
    .runComplianceTest  = NULL,
    .readExtReg         = GenericPhy_readExtReg,
    .writeExtReg        = GenericPhy_writeExtReg,
    .printRegs          = GenericPhy_printRegs,
};

/* ========================================================================== */
/*                          Function Definitions                              */
/* ========================================================================== */

static bool GenericPhy_isPhyDevSupported(CpswPhy_Handle hPhyDev,
                                         const CpswPhy_Version *version)
{
    /* All IEEE-standard PHY models are supported */
    return true;
}

static bool GenericPhy_isMacModeSupported(CpswPhy_Handle hPhyDev,
                                          CpswPhy_MacMode mode)
{
    /* All MAC modes are supported */
    return true;
}

void GenericPhy_reset(CpswPhy_Handle hPhyDev)
{
    CPSWTRACE_DEBUG("PHY %u: reset\n", hPhyDev->addr);

    /* Reset the PHY */
    CpswPhy_rmwReg(hPhyDev, PHY_BMCR, BMCR_RESET, BMCR_RESET);
}

bool GenericPhy_isResetComplete(CpswPhy_Handle hPhyDev)
{
    int32_t status;
    uint16_t val;
    bool complete = false;

    /* Reset is complete when RESET bit has self-cleared */
    status = CpswPhy_readReg(hPhyDev, PHY_BMCR, &val);
    if (status == CPSW_SOK)
    {
        complete = ((val & BMCR_RESET) == 0U);
    }

    CPSWTRACE_DEBUG("PHY %u: reset is %scomplete\n",
                    hPhyDev->addr, complete ? "" : "not");

    return complete;
}

int32_t GenericPhy_readExtReg(CpswPhy_Handle hPhyDev,
                              uint32_t reg,
                              uint16_t *val)
{
    uint16_t devad = MMD_CR_DEVADDR;
    int32_t status;

    status = CpswPhy_writeReg(hPhyDev, PHY_MMD_CR, devad | MMD_CR_ADDR);

    if (status == CPSW_SOK)
    {
        status = CpswPhy_writeReg(hPhyDev, PHY_MMD_DR, reg);
    }

    if (status == CPSW_SOK)
    {
        CpswPhy_writeReg(hPhyDev, PHY_MMD_CR, devad | MMD_CR_DATA_NOPOSTINC);
    }

    if (status == CPSW_SOK)
    {
        status = CpswPhy_readReg(hPhyDev, PHY_MMD_DR, val);
    }

    CPSWTRACE_VERBOSE_IF((status == CPSW_SOK),
                         "PHY %u: failed to read reg %u\n",
                         hPhyDev->addr, reg);
    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "PHY %u: read reg %u val 0x%04x\n",
                       hPhyDev->addr, reg, *val);

    return status;
}

int32_t GenericPhy_writeExtReg(CpswPhy_Handle hPhyDev,
                               uint32_t reg,
                               uint16_t val)
{
    uint16_t devad = MMD_CR_DEVADDR;
    int32_t status;

    CPSWTRACE_VERBOSE("PHY %u: write %u val 0x%04x\n",
                      hPhyDev->addr, reg, val);

    status = CpswPhy_writeReg(hPhyDev, PHY_MMD_CR, devad | MMD_CR_ADDR);
    if (status == CPSW_SOK)
    {
        CpswPhy_writeReg(hPhyDev, PHY_MMD_DR, reg);
    }

    if (status == CPSW_SOK)
    {
        CpswPhy_writeReg(hPhyDev, PHY_MMD_CR, devad | MMD_CR_DATA_NOPOSTINC);
    }

    if (status == CPSW_SOK)
    {
        CpswPhy_writeReg(hPhyDev, PHY_MMD_DR, val);
    }

    CPSWTRACE_ERROR_IF((status != CPSW_SOK),
                       "PHY %u: failed to write reg %u val 0x%04x\n",
                       hPhyDev->addr, reg, val);

    return status;
}

void GenericPhy_printRegs(CpswPhy_Handle hPhyDev)
{
    uint32_t i;
    uint16_t val;

    for (i = PHY_BMCR; i <= PHY_GIGESR; i++)
    {
        CpswPhy_readReg(hPhyDev, i, &val);
        CpswUtils_printf("PHY %u: reg 0x%02x = 0x%04x\n", hPhyDev->addr, i, val);
    }
}
