/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file cpsw_cpts_priv.h
 *
 * \brief This file contains the private definitions of the CPTS submodule.
 */

/*!
 * \addtogroup CPSW_CPTS_API
 * @{
 */

#ifndef CPSW_CPTS_PRIV_H_
#define CPSW_CPTS_PRIV_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/csl/csl_cpswitch.h>
#include <ti/drv/cpsw/include/cpsw_cpts.h>
#include <ti/drv/cpsw/include/cpsw_utils.h>

#include "cpsw_priv.h"

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

#define CPSW_IOCTL_CPTS_PRIV_BASE         (CPSW_IOCTL_CPTS_BASE | \
                                           CPSW_IOCTL_TYPE_PRIVATE)
#define CPSW_CPTS_PPM                     (1000000U)
#define CPSW_CPTS_PPM_GIGAHERTZ_VAL       (1000000000U)
#define CPSW_CPTS_SECONDS_PER_HOUR        (3600U)
#define CPSW_CPTS_POLL_INTERVAL           (10000U)
#define CPSW_CPTS_NUDGE_MIN_VAL           (-128)
#define CPSW_CPTS_NUDGE_MAX_VAL           (127)
#define CPSW_CPTS_PPM_MIN_VAL             (0x400U)
#define CPSW_CPTS_GENFN_LENGTH_MIN_VAL    (5U)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 * \brief CPTS private IOCTL commands
 */
typedef enum CpswCpts_PrivateIoctls_e
{
    /*!
     * \brief Handle the interrupt and notify the event's registered Callback
     *
     * IOCTL parameters:
     *   inArgs: None
     *  outArgs: None
     */
    CPSW_CPTS_IOCTL_HANDLE_INTR  = CPSW_IOCTL_CPTS_PRIV_BASE | CPSW_IOCTL_MIN(0U),

    /*!
     * \brief Enable CPTS interrupt
     *
     * IOCTL parameters:
     *   inArgs: None
     *  outArgs: None
     */
    CPSW_CPTS_IOCTL_ENABLE_INTR  = CPSW_IOCTL_CPTS_PRIV_BASE | CPSW_IOCTL_MIN(1U),

    /*!
     * \brief Disable CPTS interrupt
     *
     * IOCTL parameters:
     *   inArgs: None
     *  outArgs: None
     */
    CPSW_CPTS_IOCTL_DISABLE_INTR = CPSW_IOCTL_CPTS_PRIV_BASE | CPSW_IOCTL_MIN(2U),
} CpswCpts_PrivateIoctls;

/*!
 * \brief CPTS event memory object
 */
typedef struct CpswCpts_EventMemObj_s
{
    /*! Event memory pool*/
    CpswCpts_Event eventMemPool[CPSW_CFG_CPTS_EVENTS_POOL_SIZE];

    /* Index where the event is to be written */
    uint32_t index;
} CpswCpts_EventMemObj;

/*!
 * \brief CPTS submodule object
 */
typedef struct CpswCpts_Obj_s
{
    /*! Magic value to identify whether object handle is valid or not */
    Cpsw_Magic magic;

    /*! CPSW instance type */
    Cpsw_Type cpswType;

    /*! Flag to indicate that a SW push event is already active */
    volatile bool tsPushInFifo;

    /*! Current time stamp value maintained by driver, gets updated at every TS_PUSH_EVENT */
    uint64_t tsVal;

    /* CPTS register overlay */
    CSL_cptsRegs *cptsRegs;

    /*! Notify callback function for registered stack */
    CpswCpts_EventNotifyCb pEventNotifyCb;

    /*! Notify callback function argument for registered stack */
    void *hEventNotifyCbArg;

    /*! Maximum hardware push instances supported in hardware */
    uint32_t hwPushCount;

    /*! Hardware push event notify callback function pointers */
    CpswCpts_HwPushNotifyCb pHwPushNotifyCb[CPSW_CPTS_HW_PUSH_COUNT_MAX];

    /*! Hardware push event notify callback function arguments */
    void *pHwPushNotifyCbArg[CPSW_CPTS_HW_PUSH_COUNT_MAX];

    /*! Event memory object for ethernet Tx event */
    CpswCpts_EventMemObj ethTxEventPool;

    /*! Event memory object for ethernet Rx event */
    CpswCpts_EventMemObj ethRxEventPool;

    /*! Event memory object for hardware Ts Push event */
    CpswCpts_EventMemObj hwPushEventPool;

    /*! Event memory object for compare event */
    CpswCpts_EventMemObj cmpEventPool;

    /*! Event memory object for host Tx event */
    CpswCpts_EventMemObj hostTxEventPool;

    /*! Event statistics */
    CpswCpts_EventStats eventStats;
} CpswCpts_Obj;

/* ========================================================================== */
/*                         Global Variables Declarations                      */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

void CpswCpts_initParams(CpswCpts_Config *cfg);

CpswCpts_Handle CpswCpts_open(Cpsw_Type cpswType,
                              const CpswCpts_Config *cfg);

CpswCpts_Handle CpswCpts_reopen(Cpsw_Type cpswType);

void CpswCpts_close(CpswCpts_Handle hCpts);

int32_t CpswCpts_ioctl(CpswCpts_Handle hCpts,
                       uint32_t cmd,
                       Cpsw_IoctlPrms *prms);

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_CPTS_PRIV_H_ */

/* @} */
