/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_macport_priv.h
 *
 * \brief    This file contains the private definitions of the MAC port
 *           submodule.
 */

/*!
 * \addtogroup CPSW_MACPORT_API
 * @{
 */

#ifndef CPSW_MACPORT_PRIV_H_
#define CPSW_MACPORT_PRIV_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/drv/cpsw/include/cpsw_types.h>
#include <ti/drv/cpsw/include/cpsw_macport.h>
#include "cpsw_priv.h"

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

#define CPSW_MACPORT_PRIVATE_IOCTL(x)         (CPSW_IOCTL_MACPORT_BASE | \
                                               CPSW_IOCTL_TYPE_PRIVATE | \
                                               CPSW_IOCTL_MIN(x))

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 * \brief MAC port private IOCTL commands
 */
typedef enum CpswMacPort_PrivateIoctls_e
{
    /*!
     * \brief Enable MAC port
     *
     * IOCTL parameters:
     *   inArgs: CpswMacPort_LinkConfig
     *  outArgs: None
     */
    CPSW_MACPORT_IOCTL_ENABLE                       = CPSW_MACPORT_PRIVATE_IOCTL(0U),

    /*!
     * \brief Disable MAC port
     *
     * IOCTL parameters:
     *   inArgs: None
     *  outArgs: None
     */
    CPSW_MACPORT_IOCTL_DISABLE                      = CPSW_MACPORT_PRIVATE_IOCTL(1U),

    /*!
     * \brief Alloc free route and set interVLAN config for egress MAC port
     *
     * IOCTL parameters:
     *   inArgs: CpswMacPort_InterVlanRoutingParams
     *  outArgs: CpswMacPort_InterVlanRouteIdArgs
     */
    CPSW_MACPORT_IOCTL_SET_INTERVLAN_ROUTE          = CPSW_MACPORT_PRIVATE_IOCTL(2U),

    /*!
     * \brief Set interVLAN config for egress MAC port at given route index
     *
     * IOCTL parameters:
     *   inArgs: CpswMacPort_SetSpecificInterVlanRoutingInArgs
     *  outArgs: None
     */
    CPSW_MACPORT_IOCTL_SET_SPECIFIC_INTERVLAN_ROUTE = CPSW_MACPORT_PRIVATE_IOCTL(3U),

    /*!
     * \brief Get interVLAN config for egress MAC port
     *
     * IOCTL parameters:
     *   inArgs: CpswMacPort_InterVlanRouteIdArgs
     *  outArgs: CpswMacPort_InterVlanRoutingParams
     */
    CPSW_MACPORT_IOCTL_GET_INTERVLAN_ROUTE          = CPSW_MACPORT_PRIVATE_IOCTL(4U),

    /*!
     * \brief Delete the specified interVlan route
     *
     * IOCTL parameters:
     *   inArgs: CpswMacPort_InterVlanRoutingParams
     *  outArgs: CpswMacPort_InterVlanRouteIdArgs
     */
    CPSW_MACPORT_IOCTL_DELETE_INTERVLAN_ROUTE       = CPSW_MACPORT_PRIVATE_IOCTL(5U),

    /*!
     * \brief Get interVLAN free routes available for the given egress port
     *
     * IOCTL parameters:
     *   inArgs: NONE
     *  outArgs: CpswMacPort_InterVlanFreeRouteInfo
     */
    CPSW_MACPORT_IOCTL_GET_INTERVLAN_FREEROUTES     = CPSW_MACPORT_PRIVATE_IOCTL(6U),

    /*!
     * \brief Find interVLAN route id matching the given route params
     *
     * If unable to find the route returns CPSW_MACPORT_EROUTENOTFOUND
     * IOCTL parameters:
     *   inArgs: CpswMacPort_InterVlanRoutingParams
     *  outArgs: CpswMacPort_InterVlanRouteIdArgs
     */
    CPSW_MACPORT_IOCTL_FIND_INTERVLAN_ROUTE         = CPSW_MACPORT_PRIVATE_IOCTL(7U),

    /*!
     * \brief Check if given interVLAN route is free
     *
     * IOCTL parameters:
     *   inArgs: CpswMacPort_InterVlanRouteIdArgs
     *  outArgs: bool
     */
    CPSW_MACPORT_IOCTL_IS_INTERVLAN_ROUTE_FREE      = CPSW_MACPORT_PRIVATE_IOCTL(8U),

    /*!
     * \brief Set transmit short gap configuration for MAC Port
     *
     * IOCTL parameters:
     *   inArgs: CpswMacPort_TxShortGapConfigInArgs
     *  outArgs: NONE
     */
    CPSW_MACPORT_IOCTL_SET_SHORT_IPG                = CPSW_MACPORT_PRIVATE_IOCTL(9U),

    /*!
     * \brief Get transmit short gap configuration for MAC Port
     *
     * IOCTL parameters:
     *   inArgs: CpswMacPort_GenericInArgs
     *  outArgs: CpswMacPort_TxShortGapConfig
     */
    CPSW_MACPORT_IOCTL_GET_SHORT_IPG                = CPSW_MACPORT_PRIVATE_IOCTL(10U),

    /*!
     * \brief Get SGMII auto negotiate link status from the SGMII status register
     *
     * IOCTL parameters:
     *   inArgs: CpswMacPort_GenericInArgs
     *  outArgs: bool
     */
    CPSW_MACPORT_IOCTL_GET_SGMII_AN_LINK_STATUS     = CPSW_MACPORT_PRIVATE_IOCTL(11U),

    /*!
     * \brief Get SGMII link status from the SGMII status register
     *
     * IOCTL parameters:
     *   inArgs: CpswMacPort_GenericInArgs
     *  outArgs: bool
     */
    CPSW_MACPORT_IOCTL_GET_SGMII_LINK_STATUS     = CPSW_MACPORT_PRIVATE_IOCTL(12U),
} CpswMacPort_PrivateIoctls;

/*!
 * \brief Maximum number of interVLAN routes available per egress port
 */
#define CPSW_MACPORT_INTERVLAN_MAX_ROUTES ((CPSW_MACPORT_INTERVLAN_ROUTEID_LAST - \
                                            CPSW_MACPORT_INTERVLAN_ROUTEID_FIRST) + 1)

/**
 *  \name CPSW MACPORT internal failure codes
 *
 *  Error codes returned by CPSW MACPORT internal functions
 *  @{
 */
#define CPSW_MACPORT_E_ROUTENOTFOUND                     (-1)
/* @} */

/** @brief
 *
 *  Input args for CPSW_MACPORT_IOCTL_SET_SPECIFIC_INTERVLAN_ROUTE command
 */
typedef struct CpswMacPort_SetSpecificInterVlanRouteInArgs_s
{
    /** InterVLAN egress port route index . Can be value from 1 - 4*/
    CpswMacPort_InterVLANRouteId routeId;

    /** Egress route interVLAN configuration to be set for the route */
    CpswMacPort_InterVlanRoutingParams routeCfg;
} CpswMacPort_SetSpecificInterVlanRouteInArgs;

/** @brief
 *
 *  Input args for CPSW_MACPORT_IOCTL_GET_INTERVLAN_PARAMS command
 */
typedef struct CpswMacPort_InterVlanRouteIdArgs_s
{
    /** InterVLAN egress port route index . Can be value from 1 - 4*/
    CpswMacPort_InterVLANRouteId routeId;
} CpswMacPort_InterVlanRouteIdArgs;

/** @brief
 *
 *  Information on free interVLAN routes available for an egress port
 */
typedef struct CpswMacPort_InterVlanFreeRouteInfo_s
{
    /** Number of free interVLAN routes available */
    uint32_t numFreeRoutes;
    /** Array of free route indexes */
    CpswMacPort_InterVLANRouteId freeRouteId[CPSW_MACPORT_INTERVLAN_MAX_ROUTES];
} CpswMacPort_InterVlanFreeRouteInfo;

/*!
 * \brief MAC port submodule object
 */
typedef struct CpswMacPort_Obj_s
{
    /*! Magic value to identify whether object handle is valid or not */
    Cpsw_Magic magic;

    /*! CPSW instance type */
    Cpsw_Type cpswType;

    /*! MAC port number */
    Cpsw_MacPort portNum;

    /*! CPSW register overlay */
    CSL_Xge_cpswRegs *cpswRegs;

    /*! CPSW register overlay */
    CSL_CpsgmiiRegs *sgmiiRegs;

} CpswMacPort_Obj;

/* ========================================================================== */
/*                         Global Variables Declarations                      */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

void CpswMacPort_initParams(CpswMacPort_Config *cfg);

CpswMacPort_Handle CpswMacPort_open(Cpsw_Type cpswType,
                                    Cpsw_MacPort portNum,
                                    const CpswMacPort_Config *cfg,
                                    const CpswMacPort_Interface *interface,
                                    CpswMacPort_LinkConfig *linkConfig);

CpswMacPort_Handle CpswMacPort_reopen(Cpsw_Type cpswType,
                                      Cpsw_MacPort portNum);

void CpswMacPort_close(CpswMacPort_Handle hPort);

int32_t CpswMacPort_ioctl(CpswMacPort_Handle hPort,
                          uint32_t cmd,
                          Cpsw_IoctlPrms *prms);

uint32_t CpswMacPort_mapInterVLANRouteId2Pointer(CpswMacPort_InterVLANRouteId routeId);

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_MACPORT_PRIV_H_ */

/* @} */
