/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_mdio_priv.h
 *
 * \brief    This file contains the private definitions of the MDIO submodule.
 */

/*!
 * \addtogroup CPSW_MDIO_API
 * @{
 */

#ifndef CPSW_MDIO_PRIV_H_
#define CPSW_MDIO_PRIV_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/csl/csl_cpswitch.h>
#include "cpsw_priv.h"

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

#define CPSW_MDIO_PRIVATE_IOCTL(x)            (CPSW_IOCTL_MDIO_BASE |    \
                                               CPSW_IOCTL_TYPE_PRIVATE | \
                                               CPSW_IOCTL_MIN(x))

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 * \brief MDIO private IOCTL commands
 */
typedef enum CpswMdio_PrivateIoctls_e
{
    /*!
     * \brief Handle MDIO_INTR which combines MDIO_LINKINT and MDIO_USERINT
     * events.
     *
     * IOCTL parameters:
     *   inArgs: CpswMdio_Callbacks
     *  outArgs: None
     */
    CPSW_MDIO_IOCTL_HANDLE_INTR = CPSW_MDIO_PRIVATE_IOCTL(0U),
} CpswMdio_PrivateIoctls;

/*!
 * \brief CPSW MDIO link state change callback function
 *
 * Callback for the MDIO link state change interrupt (MDIO_LINKINT).
 * This callback is invoked from interrupt context.
 */
typedef void (*CpswMdio_LinkStateCallback)(CpswMdio_UserChannel userCh,
                                           CpswMdio_PhyStatus *phyStatus,
                                           void *appData);

/*!
 * \brief CPSW MDIO user access completion callback function
 */
typedef void (*CpswMdio_UserAccessCallback)(CpswMdio_UserChannel userCh,
                                            uint8_t phyAddr,
                                            void *appData);

/*!
 * \brief MDIO callback functions
 *
 * Optional callback functions that can be used to get notification
 * about link state change and user access completion events.
 */
typedef struct CpswMdio_Callbacks_s
{
    /*! Link state change interrupt callback function */
    CpswMdio_LinkStateCallback linkStateCb;

    /*! User access (read/write) completion callback function */
    CpswMdio_UserAccessCallback userAccessCb;

    /*! Application specific data to be passed to the callbacks */
    void *appData;
} CpswMdio_Callbacks;

/*!
 * \brief MDIO submodule object
 *
 * MDIO submodule internal/private driver structure.
 */
typedef struct CpswMdio_Obj_s
{
    /*! Magic value to identify whether object handle is valid or not */
    Cpsw_Magic magic;

    /*! MDIO register overlay */
    CSL_mdioHandle mdioRegs;

    /*! MDIO user channel */
    CpswMdio_UserChannel userCh;
} CpswMdio_Obj;

/* ========================================================================== */
/*                         Global Variables Declarations                      */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

void CpswMdio_initParams(CpswMdio_Config *cfg);

CpswMdio_Handle CpswMdio_open(Cpsw_Type cpswType,
                              const CpswMdio_Config *cfg);

CpswMdio_Handle CpswMdio_reopen(Cpsw_Type cpswType);

void CpswMdio_close(CpswMdio_Handle hMdio);

int32_t CpswMdio_ioctl(CpswMdio_Handle hMdio,
                       uint32_t cmd,
                       Cpsw_IoctlPrms *prms);

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_MDIO_PRIV_H_ */

/* @} */
