/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_phy_priv.h
 *
 * \brief    This file contains the internal structure definitions and function
 *           prototypes of the CPSW PHY driver.
 */

#ifndef CPSW_PHY_PRIV_H_
#define CPSW_PHY_PRIV_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/drv/cpsw/include/cpsw_phy.h>
#include <ti/drv/cpsw/include/cpsw_mdio.h>

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

/*! \brief Auto MDIX minimum delay in ms */
#define CPSW_PHY_AUTOMDIX_DELAY_MIN           (80U)

/*! \brief Auto MDIX maximum delay in ms */
#define CPSW_PHY_AUTOMDIX_DELAY_MAX           (200U)

/*! \brief Default FINDING state timeout (in ticks) */
#define CPSW_PHY_TIMEOUT_FINDING              (20U)

/*! \brief Default RESET_WAIT state timeout (in ticks) */
#define CPSW_PHY_TIMEOUT_RESET_WAIT           (10U)

/*! \brief Default RESET_WAIT state residence time (in ticks) */
#define CPSW_PHY_RESIDENCE_RESET_WAIT         (0U)

/*! \brief Default LINK_WAIT state timeout (in ticks) */
#define CPSW_PHY_TIMEOUT_LINK_WAIT            (50U)

/*! \brief Default NWAY_START state timeout (in ticks) */
#define CPSW_PHY_TIMEOUT_NWAY_START           (50U)

/*! \brief Default NWAY_WAIT state timeout (in ticks) */
#define CPSW_PHY_TIMEOUT_NWAY_WAIT            (80U)

/*! \brief Default timeout if MDIX is enabled (in ticks) */
#define CPSW_PHY_TIMEOUT_MDIX                 (27U)

/*! \brief Random number support */
#define CPSW_PHY_COMMON_RANDOM_MAX            (0xFFFFU)

/* ========================================================================== */
/*                         Structure Declarations                             */
/* ========================================================================== */

/*!
 * \brief MAC interface layer type
 */
typedef enum CpswPhy_MacMode_e
{
    /*! \brief ENET_CTRL_MODE value for MII mode */
    CPSW_PHY_MAC_MODE_MII        = 0U,

    /*! \brief ENET_CTRL_MODE value for RMII mode */
    CPSW_PHY_MAC_MODE_RMII       = 1U,

    /*! \brief ENET_CTRL_MODE value for GMII mode */
    CPSW_PHY_MAC_MODE_GMII       = 2U,

    /*! \brief ENET_CTRL_MODE value for RGMII mode */
    CPSW_PHY_MAC_MODE_RGMII      = 3U,

    /*! \brief ENET_CTRL_MODE value for SGMII mode */
    CPSW_PHY_MAC_MODE_SGMII      = 4U,

    /*! \brief ENET_CTRL_MODE value for QSGMII mode */
    CPSW_PHY_MAC_MODE_QSGMII     = 5U,

} CpswPhy_MacMode;

/*!
 * \brief PHY driver state-machine states
 */
typedef enum CpswPhy_FsmState_e
{
    /*! \brief INIT state */
    CPSW_PHY_FSM_STATE_INIT       = 0U,

    /*! \brief FINDING state */
    CPSW_PHY_FSM_STATE_FINDING    = 1U,

    /*! \brief RESET_WAIT state */
    CPSW_PHY_FSM_STATE_RESET_WAIT = 2U,

    /*! \brief ENABLE state */
    CPSW_PHY_FSM_STATE_ENABLE     = 3U,

    /*! \brief FOUND state */
    CPSW_PHY_FSM_STATE_FOUND      = 4U,

    /*! \brief NWAY_START state (auto-negotiation path) */
    CPSW_PHY_FSM_STATE_NWAY_START = 5U,

    /*! \brief NWAY_WAIT state (auto-negotiation path) */
    CPSW_PHY_FSM_STATE_NWAY_WAIT  = 6U,

    /*! \brief LINK_WAIT state */
    CPSW_PHY_FSM_STATE_LINK_WAIT  = 7U,

    /*! \brief LINKED state */
    CPSW_PHY_FSM_STATE_LINKED     = 8U,

    /*! \brief LOOPBACK state */
    CPSW_PHY_FSM_STATE_LOOPBACK   = 9U
} CpswPhy_FsmState;

typedef struct CpswPhy_State_s
{
    /*! PHY state-machine state */
    CpswPhy_FsmState fsmState;

    /*! Whether the PHY state-machine state has changed */
    bool fsmStateChanged;

    /*! PHY speed (auto-negotiated or manually set) */
    Cpsw_Speed speed;

    /*! PHY duplexity (auto-negotiated or manually set) */
    Cpsw_Duplexity duplexity;

    /*! Timeout (ticks) */
    uint32_t timeout;

    /*! Residence time (ticks) */
    uint32_t residenceTime;

    /*! Whether PHY is auto-negotiation capable */
    bool isNwayCapable;

    /*! Whether auto-negotiation is to be used or not */
    bool enableNway;

    /*! Whether manual mode needs to be configured */
    bool needsManualConfig;

    /*! Whether auto-negotiation advertisement needs to be configured */
    bool needsNwayConfig;

    /* Refined link capability mask (app, SoC, PHY) */
    uint32_t linkCaps;

    /* PHY link capability mask */
    uint32_t phyLinkCaps;

    /*! Whether PHY loopback is enabled or not */
    bool enableLoopback;

    /*! Whether MDIX switch is needed or not */
    bool needsMdixSwitch;

    /*! Whether MDIX is enabled or not (MDI) */
    bool enableMdix;
} CpswPhy_State;

/*!
 * \brief PHY specific driver
 *
 * PHY specific driver implementation. This driver code must not modify registers
 * and/or bitfields that are being modified by the main PHY driver. Otherwise,
 * PHY driver state machine may not work as the underlying state of the PHY would
 * be inconsistent.
 */
typedef struct CpswPhy_Drv_s
{
    /*! \brief Driver name */
    const char *name;

    /*! \brief Whether driver supports a given PHY version */
    bool (*isPhyDevSupported)(CpswPhy_Handle hPhyDev,
                              const CpswPhy_Version *version);

    /*! \brief Whether MAC mode is supported by the PHY */
    bool (*isMacModeSupported)(CpswPhy_Handle hPhyDev,
                               CpswPhy_MacMode mode);

    /*! \brief PHY specific configuration */
    int32_t (*config)(CpswPhy_Handle hPhyDev,
                      const CpswPhy_Config *cfg,
                      CpswPhy_MacMode mode);

    /*! \brief PHY specific soft reset */
    void (*reset)(CpswPhy_Handle hPhyDev);

    /*! \brief PHY specific soft reset status */
    bool (*isResetComplete)(CpswPhy_Handle hPhyDev);

    /*! \brief Run compliance test */
    int32_t (*runComplianceTest)(CpswPhy_Handle hPhyDev,
                                 uint32_t testMode);

    /*! \brief Read from PHY extended register */
    int32_t (*readExtReg)(CpswPhy_Handle hPhyDev,
                          uint32_t reg,
                          uint16_t *val);

    /*! \brief Write to PHY extended register */
    int32_t (*writeExtReg)(CpswPhy_Handle hPhyDev,
                           uint32_t reg,
                           uint16_t val);

    /*! \brief Print PHY registers */
    void (*printRegs)(CpswPhy_Handle hPhyDev);
} CpswPhy_Drv;

/*!
 * \brief PHY specific driver handle
 */
typedef CpswPhy_Drv *CpswPhyDrv_Handle;

/*!
 * \brief PHY driver object
 */
typedef struct CpswPhy_Obj_s
{
    /*! CPSW instance type */
    Cpsw_Type cpswType;

    /*! MDIO handle used to access PHY registers */
    CpswMdio_Handle hMdio;

    /*! MAC port number the PHY is connected to */
    Cpsw_MacPort portNum;

    /*! PHY configuration params */
    CpswPhy_Config phyConfig;

    /*! Port Link configuration (speed, duplexity) */
    CpswMacPort_LinkConfig linkConfig;

    /*! State-machine timeout configuration */
    CpswPhy_FsmTimeoutConfig timeoutConfig;

    /*! State-machine state */
    CpswPhy_State state;

    /*! PHY device address */
    uint32_t addr;

    /*! Requested link capability mask */
    uint32_t reqLinkCaps;

    /*! MAC mode */
    CpswPhy_MacMode macMode;

    /*! PHY driver */
    CpswPhyDrv_Handle hDrv;
} CpswPhy_Obj;

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/*!
 * \brief Initialize PHY config params
 *
 * Initialize PHY driver configuration parameters.
 *
 * \param phyConfig  PHY configuration params
 */
void CpswPhy_initParams(CpswPhy_Config *phyConfig);

/*!
 * \brief Set PHY extended parameters
 *
 * Sets the PHY-specific extended parameters to the PHY config structure.
 *
 * \param phyConfig            Pointer to the PHY config
 * \param extendedConfig       Pointer to the PHY extended config
 * \param extendedConfigSize   Size of the PHY extended config
 */
void CpswPhy_setExtendedParams(CpswPhy_Config *phyConfig,
                               const void *extendedConfig,
                               uint32_t extendedConfigSize);

/*!
 * \brief Open the PHY driver
 *
 * Open the CPSW PHY driver for the given MAC port number. The PHY driver
 * takes PHY specific configuration parameters, the MAC port type connection
 * and the desired link configuration (auto or manual).
 *
 * \param type       CPSW instance type (CPSW_2G, CPSW_9G)
 * \param hMdio      MDIO module handle
 * \param interface  MAC interface type
 * \param phyConfig  PHY configuration params
 * \param linkConfig Link configuration (auto or manual)
 *
 * \return CpswPhy_Handle if success, NULL otherwise
 */
CpswPhy_Handle CpswPhy_open(Cpsw_Type cpswType,
                            CpswMdio_Handle hMdio,
                            Cpsw_MacPort portNum,
                            CpswMacPort_Interface *interface,
                            const CpswPhy_Config *phyConfig,
                            const CpswMacPort_LinkConfig *linkConfig);

/*!
 * \brief Close the PHY driver
 *
 * Close the CPSW PHY driver.
 *
 * \param hPhyDev    PHY device handle
 */
void CpswPhy_close(CpswPhy_Handle hPhyDev);

/*!
 * \brief PHY driver IOCTL
 *
 * IOCTL implementation of the PHY driver
 *
 * \param hPhyDev    PHY device handle
 * \param cmd        IOCTL command
 * \param prms       IOCTL parameters
 *
 * \return \ref Cpsw_ErrorCodes
 */
int32_t CpswPhy_ioctl(CpswPhy_Handle hPhyDev,
                      uint32_t cmd,
                      Cpsw_IoctlPrms *prms);

/*!
 * \brief Read PHY register
 *
 * Read a PHY register. It's not meant for extended registers.
 *
 * \param hPhyDev    PHY device handle
 * \param reg        Register number
 * \param val        Pointer to the read value
 *
 * \return \ref Cpsw_ErrorCodes
 */
int32_t CpswPhy_readReg(CpswPhy_Handle hPhyDev,
                        uint32_t reg,
                        uint16_t *val);

/*!
 * \brief Write PHY register
 *
 * Write a PHY register. It's not meant for extended registers.
 *
 * \param hPhyDev    PHY device handle
 * \param reg        Register number
 * \param val        Value to be written
 *
 * \return \ref Cpsw_ErrorCodes
 */
int32_t CpswPhy_writeReg(CpswPhy_Handle hPhyDev,
                         uint32_t reg,
                         uint16_t val);

/*!
 * \brief Read-modify-write PHY register
 *
 * Read-modify-write a PHY register. It's not meant for extended registers.
 *
 * \param hPhyDev    PHY device handle
 * \param reg        Register number
 * \param mask       Bitmask to be applied on read value and value to be
 *                   written
 * \param val        Value to be written
 *
 * \return \ref Cpsw_ErrorCodes
 */
int32_t CpswPhy_rmwReg(CpswPhy_Handle hPhyDev,
                       uint32_t reg,
                       uint16_t mask,
                       uint16_t val);

/*!
 * \brief Read PHY register using Clause-45 frame
 *
 * Read a PHY register using Clause-45 frame.
 *
 * \param hPhyDev    PHY device handle
 * \param mmd        MMD
 * \param reg        Register number
 * \param val        Pointer to the read value
 *
 * \return \ref Cpsw_ErrorCodes
 */
int32_t CpswPhy_readC45Reg(CpswPhy_Handle hPhyDev,
                           CpswMdio_Clause45Mmd mmd,
                           uint32_t reg,
                           uint16_t *val);

/*!
 * \brief Write PHY register using Clause-45 frame
 *
 * Write a PHY register using Clause-45 frame.
 *
 * \param hPhyDev    PHY device handle
 * \param mmd        MMD
 * \param reg        Register number
 * \param val        Value to be written
 *
 * \return \ref Cpsw_ErrorCodes
 */
int32_t CpswPhy_writeC45Reg(CpswPhy_Handle hPhyDev,
                            CpswMdio_Clause45Mmd mmd,
                            uint32_t reg,
                            uint16_t val);

/*!
 * \brief Read-modify-write PHY register using Clause-45 frame
 *
 * Read-modify-write a PHY register using Clause-45 frame.
 *
 * \param hPhyDev    PHY device handle
 * \param mmd        MMD
 * \param reg        Register number
 * \param mask       Bitmask to be applied on read value and value to be
 *                   written
 * \param val        Value to be written
 *
 * \return \ref Cpsw_ErrorCodes
 */
int32_t CpswPhy_rmwC45Reg(CpswPhy_Handle hPhyDev,
                          CpswMdio_Clause45Mmd mmd,
                          uint32_t reg,
                          uint16_t mask,
                          uint16_t val);

/* ========================================================================== */
/*                       Static Function Definitions                          */
/* ========================================================================== */

/* None */

#ifdef __cplusplus
}
#endif

#endif /* CPSW_PHY_PRIV_H_ */
