/*
 *  Copyright (c) Texas Instruments Incorporated 2018
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!
 * \file     cpsw_priv.h
 *
 * \brief    This file contains the private generic definitions and structure
 *           definitions of the CPSW driver.
 */

/*!
 * \addtogroup CPSW_MAIN_API
 * @{
 */

#ifndef CPSW_PRIV_H_
#define CPSW_PRIV_H_

/* ========================================================================== */
/*                             Include Files                                  */
/* ========================================================================== */

#include <ti/csl/csl_cpswitch.h>
#include <ti/drv/cpsw/soc/cpsw_soc.h>

/* needed for CpswUtils_printf function */
#include <ti/drv/cpsw/include/cpsw_utils.h>

#if defined(__KLOCWORK__)
#include <stdlib.h>
#endif

#ifdef QNX_OS
#include <stdio.h>
#include <sys/mman.h>
#include <unistd.h>
#include <hw/inout.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* ========================================================================== */
/*                                 Macros                                     */
/* ========================================================================== */

#if defined(CPSW_CFG_ASSERT_ENABLE)
#if defined(__KLOCWORK__)
#define Cpsw_assert(cond)               do { if (!(cond)) abort(); } while (0)
#else
#define Cpsw_assert(cond)                                 \
    (Cpsw_assertLocal((bool)(cond), (const char *)# cond, \
                      (const char *)__FILE__, (int32_t)__LINE__))
#endif /* __KLOCWORK__ */
#else
#define Cpsw_assert(cond) (void)(cond)
#endif

/*!
 * \brief Priority Escalation value for switch scheduler
 *
 * When a port is in escalate priority, this is the number of higher priority
 * packets sent before the next lower priority is allowed to send a packet.
 * Escalate priority allows lower priority packets to be sent at a fixed rate
 * relative to the next higher priority.  The min value of esc_pri_ld_val = 2
 */
#define CPSW_CFG_ESC_PRI_LD_VAL                                           (2)

#define CPSW_COREKEY_CONVERT_MAGIC_NUM  (0x38ACB7E6U)
#define CPSW_COREID_2_COREKEY(coreId)   ((coreId * 100U) + CPSW_COREKEY_CONVERT_MAGIC_NUM)
#define CPSW_COREKEY_2_COREID(coreKey)  ((coreKey - CPSW_COREKEY_CONVERT_MAGIC_NUM) / 100U)

/* ========================================================================== */
/*                         Structures and Enums                               */
/* ========================================================================== */

/*!
 * \brief CPSW port
 */
typedef enum Cpsw_Port_e
{
    /*! Host port (CPPI port) */
    CPSW_PORT_HOST  = 0U,

    /*! MAC port 1 */
    CPSW_PORT_MAC_1 = 1U,

    /*! MAC port 2 */
    CPSW_PORT_MAC_2 = 2U,

    /*! MAC port 3 */
    CPSW_PORT_MAC_3 = 3U,

    /*! MAC port 4 */
    CPSW_PORT_MAC_4 = 4U,

    /*! MAC port 5 */
    CPSW_PORT_MAC_5 = 5U,

    /*! MAC port 6 */
    CPSW_PORT_MAC_6 = 6U,

    /*! MAC port 7 */
    CPSW_PORT_MAC_7 = 7U,

    /*! MAC port 8  */
    CPSW_PORT_MAC_8 = 8U,

    /*! Total number of ports. Not usable by application */
    CPSW_PORT_LAST  = CPSW_PORT_MAC_8,
}
Cpsw_Port;

/*! Number of CPSW ports */
#define CPSW_PORT_NUM (CPSW_PORT_LAST + 1U)

typedef enum Cpsw_Magic_e
{
    /*! CPSW magic number */
    CPSW_DRV_MAGIC      = 0x213D090FU,

    /*! ALE submodule magic number */
    CPSW_ALE_MAGIC      = 0x73CA4108U,

    /*! CPTS submodule magic number */
    CPSW_CPTS_MAGIC     = 0x5E6252B9U,

    /*! DMA magic number */
    CPSW_DMA_MAGIC      = 0x33D47140U,

    /*! Host port submodule magic number */
    CPSW_HOSTPORT_MAGIC = 0x26A4E4A7U,

    /*! MAC port submodule magic number */
    CPSW_MACPORT_MAGIC  = 0x102E5D0EU,

    /*! MDIO submodule magic number */
    CPSW_MDIO_MAGIC     = 0x7A94C29EU,

    /*! Statistics submodule magic number */
    CPSW_STATS_MAGIC    = 0x66172D9DU,

    /*! RM submodule magic number */
    CPSW_RM_MAGIC       = 0x5A108FECU,

    /*! Uninitialized magic */
    CPSW_NO_MAGIC       = 0x00000000U,
} Cpsw_Magic;

/* ========================================================================== */
/*                         Global Variables Declarations                      */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                          Function Declarations                             */
/* ========================================================================== */

/*!
 * \brief Check IOCTL input args for commands that expect an input
 *
 * Helper function to check the input arguments passed to an IOCTL for
 * commands that expect an input argument.
 *
 * \param prms         IOCTL parameters pointer
 * \param inArgsSize   Expected size of the input argument
 *
 * \return CPSW_SOK if inArgs pointer is not NULL and inArgsSize matches the
 *         expected size. Otherwise, CPSW_EMALFORMED_IOCTL.
 */
static inline uint32_t Cpsw_checkInArgs(const Cpsw_IoctlPrms *prms,
                                        uint32_t inArgsSize);

/*!
 * \brief Check IOCTL input args for commands that don't expect an input
 *
 * Helper function to check the input arguments passed to an IOCTL for
 * commands that don't expect an input.
 *
 * \param prms         IOCTL parameters pointer
 *
 * \return CPSW_SOK if inArgs pointer is NULL and inArgsSize is 0.
 *         Otherwise, CPSW_EMALFORMED_IOCTL.
 */
static inline uint32_t Cpsw_checkNoInArgs(const Cpsw_IoctlPrms *prms);

/*!
 * \brief Check IOCTL output args for commands that expect an output
 *
 * Helper function to check the output arguments passed to an IOCTL for
 * commands that expect an output argument.
 *
 * \param prms         IOCTL parameters pointer
 * \param outArgsSize  Expected size of the output argument
 *
 * \return CPSW_SOK if outArgs pointer is not NULL and outArgsSize matches the
 *         expected size. Otherwise, CPSW_EMALFORMED_IOCTL.
 */
static inline uint32_t Cpsw_checkOutArgs(const Cpsw_IoctlPrms *prms,
                                         uint32_t outArgsSize);

/*!
 * \brief Check IOCTL output args for commands that don't have output
 *
 * Helper function to check the output arguments passed to an IOCTL for
 * commands that don't have an output argument.
 *
 * \param prms         IOCTL parameters pointer
 *
 * \return CPSW_SOK if outArgs pointer is NULL and outArgsSize is 0.
 *         Otherwise, CPSW_EMAILFORMED_IOCTL.
 */
static inline uint32_t Cpsw_checkNoOutArgs(const Cpsw_IoctlPrms *prms);

/*!
 * \brief Check IOCTL input and output args
 *
 * Helper function to check the input and output arguments passed to an IOCTL
 * for commands that expect input and output arguments.
 *
 * \param prms         IOCTL parameters pointer
 * \param inArgsSize   Expected size of the input argument
 * \param outArgsSize  Expected size of the output argument
 *
 * \return CPSW_SOK if outArgs pointer is NULL and outArgsSize is 0.
 *         Otherwise, CPSW_EMAILFORMED_IOCTL.
 */
static inline uint32_t Cpsw_checkInOutArgs(const Cpsw_IoctlPrms *prms,
                                           uint32_t inArgsSize,
                                           uint32_t outArgsSize);

#if defined(CPSW_CFG_ASSERT_ENABLE) && !defined(__KLOCWORK__)
static inline void Cpsw_assertLocal(bool condition,
                                    const char *str,
                                    const char *fileName,
                                    int32_t lineNum);
#endif

/* ========================================================================== */
/*                        Deprecated Function Declarations                    */
/* ========================================================================== */

/* None */

/* ========================================================================== */
/*                       Static Function Definitions                          */
/* ========================================================================== */

static inline CSL_Xge_cpsw_ss_sRegs *Cpsw_getCslSsRegs(Cpsw_Type cpswType)
{
    CSL_Xge_cpsw_ss_sRegs *regs = NULL;
    const CpswSoc_CpswIpInfo *ipInfo = CpswUtils_getIpInfo(cpswType);

    Cpsw_assert((ipInfo != NULL) && (ipInfo->cpswType == cpswType));
    if (ipInfo != NULL)
    {
        regs = (CSL_Xge_cpsw_ss_sRegs *)ipInfo->cpswSSInfo.regBaseAddr;
    }

#ifdef QNX_OS
    uint32_t length = sizeof(CSL_Xge_cpsw_ss_sRegs);
    regs = (CSL_Xge_cpsw_ss_sRegs*) mmap_device_memory(0,
        length, PROT_READ|PROT_WRITE|PROT_NOCACHE, 0, (uint64_t)regs);
#endif

    return regs;
}

static inline CSL_CpsgmiiRegs* Cpsw_getCslSgmiiRegs(Cpsw_Type cpswType)
{
    CSL_CpsgmiiRegs* regs = NULL;
    const CpswSoc_CpswIpInfo *ipInfo = CpswUtils_getIpInfo(cpswType);

    Cpsw_assert((ipInfo != NULL) && (ipInfo->cpswType == cpswType));
    if (ipInfo != NULL)
    {
        regs = (CSL_CpsgmiiRegs *)ipInfo->sgmiiInfo.regBaseAddr;
    }
#ifdef QNX_OS
    uint32_t length = sizeof(CSL_CpsgmiiRegs);
    regs = (CSL_CpsgmiiRegs*) mmap_device_memory(0,
        length, PROT_READ|PROT_WRITE|PROT_NOCACHE, 0, (uint64_t)regs);
#endif

    return regs;
}

static inline CSL_Xge_cpswRegs *Cpsw_getCslRegs(Cpsw_Type cpswType)
{
    CSL_Xge_cpswRegs *regs = NULL;
    const CpswSoc_CpswIpInfo *ipInfo = CpswUtils_getIpInfo(cpswType);

    Cpsw_assert((ipInfo != NULL) && (ipInfo->cpswType == cpswType));
    if (ipInfo != NULL)
    {
        regs = (CSL_Xge_cpswRegs *)ipInfo->cpswInfo.regBaseAddr;
    }
#ifdef QNX_OS
    uint32_t length = sizeof(CSL_Xge_cpswRegs);
    regs = (CSL_Xge_cpswRegs*) mmap_device_memory(0,
        length, PROT_READ|PROT_WRITE|PROT_NOCACHE, 0, (uint64_t)regs);
#endif
    return regs;
}

static inline CSL_cptsRegs *Cpsw_getCslCptsRegs(Cpsw_Type cpswType)
{
    CSL_cptsRegs *regs = NULL;
    const CpswSoc_CpswIpInfo *ipInfo = CpswUtils_getIpInfo(cpswType);

    Cpsw_assert((ipInfo != NULL) && (ipInfo->cpswType == cpswType));
    if (ipInfo != NULL)
    {
        regs = (CSL_cptsRegs *)ipInfo->cptsInfo.regBaseAddr;
    }
#ifdef QNX_OS
    uint32_t length = sizeof(CSL_cptsRegs);
    regs = (CSL_cptsRegs *) mmap_device_memory(0,
        length, PROT_READ|PROT_WRITE|PROT_NOCACHE, 0, (uint64_t)regs);
#endif
    return regs;
}

static inline Cpsw_Type Cpsw_getType(CSL_Xge_cpswRegs *regs)
{
    Cpsw_Type cpswType = CPSW_9G;
    const CpswSoc_CpswIpInfoTbl *ipInfoTbl = CpswSoc_getIPInfoTbl();
    uint32_t i;

    for (i = 0; i < ipInfoTbl->numCpswIPs; i++)
    {
        const CpswSoc_CpswIpInfo *ipInfo = &ipInfoTbl->info[i];

        if (ipInfo->cpswInfo.regBaseAddr == (uintptr_t)regs)
        {
            break;
        }
    }
    Cpsw_assert (i < ipInfoTbl->numCpswIPs);
    if (i < ipInfoTbl->numCpswIPs)
    {
        cpswType = ipInfoTbl->info[i].cpswType;
    }
    return cpswType;
}

static inline uint32_t Cpsw_checkInArgs(const Cpsw_IoctlPrms *prms,
                                        uint32_t inArgsSize)
{
    return ((prms != NULL) &&
            (prms->inArgs != NULL) &&
            (prms->inArgsSize == inArgsSize)) ?
           CPSW_SOK : CPSW_EMALFORMED_IOCTL;
}

static inline uint32_t Cpsw_checkNoInArgs(const Cpsw_IoctlPrms *prms)
{
    return ((prms != NULL) &&
            (prms->inArgs == NULL) &&
            (prms->inArgsSize == 0U)) ?
           CPSW_SOK : CPSW_EMALFORMED_IOCTL;
}

static inline uint32_t Cpsw_checkOutArgs(const Cpsw_IoctlPrms *prms,
                                         uint32_t outArgsSize)
{
    return ((prms != NULL) &&
            (prms->outArgs != NULL) &&
            (prms->outArgsSize == outArgsSize)) ?
           CPSW_SOK : CPSW_EMALFORMED_IOCTL;
}

static inline uint32_t Cpsw_checkNoOutArgs(const Cpsw_IoctlPrms *prms)
{
    return ((prms != NULL) &&
            (prms->outArgs == NULL) &&
            (prms->outArgsSize == 0U)) ?
           CPSW_SOK : CPSW_EMALFORMED_IOCTL;
}

static inline uint32_t Cpsw_checkInOutArgs(const Cpsw_IoctlPrms *prms,
                                           uint32_t inArgsSize,
                                           uint32_t outArgsSize)
{
    return ((prms != NULL) &&
            (prms->inArgs != NULL) &&
            (prms->inArgsSize == inArgsSize) &&
            (prms->outArgs != NULL) &&
            (prms->outArgsSize == outArgsSize)) ?
           CPSW_SOK : CPSW_EMALFORMED_IOCTL;
}

#if defined(CPSW_CFG_ASSERT_ENABLE) && !defined(__KLOCWORK__)
#ifdef QNX_OS
extern void error_log(const char *str, const char *fileName, int32_t lineNum);
#endif
static inline void Cpsw_assertLocal(bool condition,
                                    const char *str,
                                    const char *fileName,
                                    int32_t lineNum)
{
#ifdef QNX_OS
    if (!(condition))
    {
        error_log(str, fileName, lineNum);
    }
#else
    volatile static bool gCpswAssertWaitInLoop = TRUE;

    if (!(condition))
    {
        CpswUtils_printf("Assertion @ Line: %d in %s: %s : failed !!!\n",
                         lineNum, fileName, str);
        while (gCpswAssertWaitInLoop)
        {
        }

        ;
    }
#endif

    return;
}
#endif

#ifdef __cplusplus
}
#endif

#endif /* CPSW_PRIV_H_ */

/* @} */
